import glob
import logging, logging.handlers
import math
import os
import sys
import tempfile
import time
import json

from errormsgs import PivotError
from field import Field
from calculation import Calculation, Lookup, Rex, Eval, StatsField, GeoIP
from stats import Stats

SPLUNK_HOME = os.environ['SPLUNK_HOME']

INTERNAL_OBJECT_NAMES = ['Base', 'BaseEvent', 'BaseSearch', 'BaseTransaction', 'BaseInterface', 'BaseImplements', 'BaseAggregate']
DEBUG = 0

def isSplunkField(fieldname): #FIXME: move to a util of sorts
  if fieldname in ['_time', 'host', 'source', 'sourcetype']:
    return True
  return False		

class Base:
    def __init__(self):
        self.stats=[]
        self.top=[]
        self.splitbyfields=[]
        self.calculations=[]
        self.filters = []
        self.postfilters = []
	self.searchtimefilters = []
        self.name = 'Base'
        self.parentname = ''
        self.lineage = ''
        self.fields = []
        self.changedinheritedfields = []
        self.displayName = ''
        self.childObjects = []
	self.tstatsNamespace = ''
	self.objectType=''
	self.comment=''
	self.replaceBoolLabels=''
	self.reformattime=''
	self.sortsearch=''
	self.colOther=True
	self.rowSummary=False
	self.colSummary='none' # true = 'show', false = 'none', calculate but don't show = 'hide'
	self.colLimit=0

    def setColLimit(self, colLimit):
	self.colLimit=colLimit

    def addComment(self, comment):
	self.comment=comment

    def getComment(self):
	return self.comment

    def setTsidxNamespace(self, namespace):
	if namespace == '':	
	  raise PivotError('setTsidxNamespace', 'must specify tsidx namespace')
	self.tstatsNamespace = namespace
	
    def unsetTsidxNamespace(self):
	self.tstatsNamespace = ''

    def getTsidxNamespace(self):
	return self.tstatsNamespace

    def getName(self):
	return self.name

    def setName(self, name):
        self.name = name

    def getParentName(self):
        return self.parentname

    def getLineage(self):
        return self.lineage

    def getObjectType(self):
        return self.objectType

    def setParentName(self, name):
        self.parentname = name

    def setLineage(self):
        if self.getLineage()!='':
	  self.lineage += '.' 
	self.lineage += self.getName()

    def getDisplayName(self):
        if self.displayName == '':
            self.displayName = self.name
        return self.displayName

    def setDisplayName(self, name):
        self.displayName = name

    def getChildNames(self):
	children = []
	for i in self.childObjects:
	  children.append(i.getName())
        return children

    def getChildObjects(self):
	return self.childObjects

    # add calculations field that counts by each child object type
    def addChildObject(self, child):
	outfield = {}
	outfield["fieldName"]= "is_" + child.getName()
	outfield["owner"]= self.getLineage()
	outfield["type"]= "childCount"
	outfield["required"]= False
	outfield["hidden"]= False
	valueSearch = 'if(searchmatch("'+child.getConstraintsWithoutParents().replace('"','\\\"')+'"),1,0)'
	self.evalField(self.getLineage(), outfield, valueSearch, editable=False) # FIXME: type=bool

	outfield["fieldName"]= "is_not_" + child.getName()
	valueSearch = 'if(searchmatch("NOT( '+child.getConstraintsWithoutParents().replace('"','\\\"')+')"),1,0)'
	self.evalField(self.getLineage(), outfield, valueSearch, editable=False) # FIXME: type=bool

        self.childObjects.append(child)

    def getChildObject(self, childName):
	for child in self.childObjects:
	  if child.getName() == childName:
	    return child
	return None 

    def addField(self, owner, fieldName,fieldType, required, multivalued, hidden=False, constraintSearch='', displayName='', comment='', editable=True):
	if displayName=='': displayName = fieldName
	if fieldType=="objectCount":
	  objectTypeField =  self.getFieldType("objectCount")
	  if objectTypeField != None: 
	  #remove parents' objectType fields
	    self.removeField(objectTypeField) 

	if self.hasField(fieldName): #updating hidden/required. should never have two fields with the same name
	  inheritedField = self.getField(fieldName).JSONdesc()
	  if inheritedField['hidden']!=hidden or inheritedField['required']!=required:
	    self.changedinheritedfields.append(fieldName)   
	  self.removeField(self.getField(fieldName))

        self.fields.append(Field(fieldName, constraintSearch, fieldType, owner, required, multivalued, hidden, displayName, comment=comment, editable=editable))


    def removeField(self, myfield):
        self.fields.remove(myfield)

    def supressField(self, myfield):
	# this should work for fields as well as calculations since both of them have a setHidden function
        # set hidden bit to True
	# FIXME: check if this object is the owner...if not, need to locally supess the field.
	myfield.setHidden(True)
    
    
    def sparkline(self, statsfunc, fieldName='', label=''):	
	if fieldName=='':
          fnfieldlist = {"function": 'sparkline',"field": statsfunc, "rename":label}
	else: 
          fnfieldlist = {"function": 'sparkline',"field": statsfunc+'('+fieldName+')', "rename":label}
        self.stats.append(fnfieldlist)

    def count(self, field='', label=''):
        fnfieldlist = {"function": 'count',"field": field, "rename":label}
        self.stats.append(fnfieldlist)
 
 
    def countif(self, field, value, label):
	#count(eval(host="fool01.splunk.com"))
	expression = 'eval(' + field + '="' + value +'")'
        fnfieldlist = {"function": 'count',"field": expression, "rename":label}
        self.stats.append(fnfieldlist)

    def getConstraintsWithoutParents(self):
	retval = ''
        for field in self.fields:
 	  if field.getOwner()==self.lineage: #don't want inherrited fields/constraints
	    if field.search()!='':
	      if retval == '': 
	        retval += field.search()
              else:
                retval += " " + field.search()

 	for filter in self.filters:
            if filter["owner"]==self.lineage and filter["search"].startswith('regex') == False:
              retval += " (" + filter["search"] + ") "
            #need to check if filter is "regex"...if it is, need to put it in separately with | 
        for filter in self.filters:
            if filter["owner"]==self.lineage and filter["search"].startswith('regex'):
              retval += ' | ' + filter["search"] + " "

	return retval

    def getConstraints(self):
	retval = ''
        for field in self.fields:
	  if field.search()!='':
	    if retval == '': 
	      retval += field.search()
            else:
              retval += " " + field.search()

 	for filter in self.filters:
            if filter["search"].startswith('regex') == False:
              retval += " (" + filter["search"] + ") "
            #need to check if filter is "regex"...if it is, need to put it in separately with | 
        for filter in self.filters:
            if filter["search"].startswith('regex'):
              retval += ' | ' + filter["search"] + " "

	return retval

    def getCalculationsWithoutParents(self):
	retval = '' 
	calcSearches=''
        for calc in self.calculations:
	  if calc.getOwner()==self.getLineage():
	    if retval != '': retval += " | "
            retval += calc.search() 
	    calcSearch = ''
	    for f in calc.getFields():
	      calcSearch += f.search()
	    if calcSearch != '':
	      calcSearches += calcSearch	
	if calcSearches != '':
	  retval += ' | search ' + calcSearches
	return retval


    # FIXME: Make this only usable with pivot UI	
    def filterLimit(self, rawObjectSearch, fieldname, attributename, statsfn, limittype='', limitamount=''):
        fnfieldlist = [{"function": statsfn, "field": attributename}]

	sortsearch = 'sort limit=' + str(limitamount) + ' '
	if limittype == 'default': 
	  sortsearch += fieldname
	else:
	  if limittype == 'highest': 
	    sortsearch += '-'
	  sortsearch += '"' + statsfn
	  if attributename !='': 
	    sortsearch += '(' + attributename + ')"' + ' | fields - "' + statsfn + '(' + attributename + ')"'
	  else:
	    sortsearch += '" | fields - ' + statsfn

	splitby = [{"field": fieldname, "roworcol": 'row'}]	
	# this should resemble the subsearch in the where clause in:
	# | tstats sum(Amount) from foobar where [| tstats sum(Amount) from foobar groupby Opportunity\ Owner | sort limit=10 -"sum(Amount)" | fields - "sum(Amount)"]
	if self.tstatsNamespace == '': 
 	  filtersearch = ' [' + rawObjectSearch + ' | ' + self.renameFieldsWithoutLineage() + Stats(fnfieldlist, splitby).getStatsString() + ' | ' + sortsearch  + ']'
        else: 
	  nodenamefilters=[]
	  for f in self.searchtimefilters:
	    if f["search"].startswith('nodename=')==True: 
	      nodenamefilters.append(f) # add nodename filters for model-wide acceleration
	  filtersearch = ' [' + Stats(fnfieldlist, splitby).tstatsSearch(self.tstatsNamespace, nodenamefilters) + ' | ' + sortsearch  + ']'
	# add filter searchstring to self.filters and self.searchtimefilters
	self.filters = self.filters + [{"search": filtersearch, "owner": self.lineage}]
	self.searchtimefilters = self.searchtimefilters + [{"search": filtersearch, "owner": self.lineage}]


    def setColOther(self, colOther):
	self.colOther=colOther

    def setRowSummary(self, rowSum):
	self.rowSummary=rowSum

    def setColSummary(self, colSum):
	# true = 'show', false = 'none', calculate but don't show = 'hide'
	self.colSummary=colSum

    def sortLimit(self, fieldname='', attributename='', statsfn='', limittype='default', limitamount=0):
	#FIXME: check for invalid limittype
	# FIXME: check - if fieldname='' and limittype==default --> error
	# FIXME: if limittype != default and (attributename=='' or statsfn='') --> error
	sortsearch = '| sort limit=' + str(limitamount) + ' '

	if limittype == 'default': 
	  sortsearch += '"' + fieldname.strip('"').strip("'") + '"'

	else:
	  # FIXME: check that statsfn and limitamount are nonempty	
	  if limittype == 'descending': 
	    sortsearch += '-'

	  statsFnRename=''; colSplitAll=''
	  for field in self.splitbyfields:
	    if 'roworcol' in field and field['roworcol']=='col':
	      colSplitAll='~~~'
	      break

	  # use rename from self.stats[]
	  for funcfield in self.stats:
            #iterate through and assign list of outputfields for the stats calculation
            if funcfield["function"]==statsfn and funcfield["field"].strip("'").strip('"')==attributename and funcfield["rename"]!="":
              statsFnRename = funcfield["rename"]
	      break;

	  if statsFnRename == '':
	    statsFnRename = statsfn
	    if attributename !='': 
	      statsFnRename += '(' + attributename + ')'
	  if len(self.stats) == 1 and colSplitAll!='':
	    statsFnRename = ''	

	  sortsearch += '"' + colSplitAll 
	  if statsFnRename!='' and colSplitAll!='':
	    sortsearch += ':::' 
	  sortsearch += statsFnRename + '"'
	  if self.colSummary=='hide' and colSplitAll != '':
 	    sortsearch += ' | fields - "' + colSplitAll 
 	    if len(self.stats)>1:
 	      sortsearch += ":::*"
 	    sortsearch +='"'


	if fieldname.strip("'").strip('"')!='_time' or limittype!='default':
	  sortsearch += ' | fields - _span ' 
	
	self.sortsearch = sortsearch

    # FIXME: Make this only usable with pivot UI	
    def statsTable(self, statsfunc, fieldname, label=''):
	if (self.tstatsNamespace!='' and self.hasFieldHierarchical(fieldname)==False) or (self.tstatsNamespace=='' and not (self.hasField(fieldname.strip("'")) or self.hasCalculatedField(fieldname.strip("'")))):
	  raise PivotError("statsTable", "fieldname " + fieldname + " does not exist in object " + self.getName()) 
	if statsfunc not in ["count", "sum", "dc", "list", "values", "first", "last", "min", "max", "avg", "stdev", "duration", "range", "earliest", "latest"]:
	  raise PivotError("statsTable", "unrecognized stats function " + statsfunc) 
 	for split in self.splitbyfields:
          if label != '' and label == split["label"]:
            raise PivotError("statsTable", "cannot have duplicate label  " + label)
 	for stat in self.stats:
          if label != '' and label == stat["rename"]:
            raise PivotError("statsTable", "cannot have duplicate label  " + label)

        fnfieldlist = {"function": statsfunc,"field": fieldname, "rename":label}
        self.stats.append(fnfieldlist)


    # FIXME: Make this only usable with pivot UI	
    def splitby(self, field, label='', roworcol='', span='', bins='', start='', end=''):
	if field=="":
	  raise PivotError("splitby", "must specify non-empty field name in splitby ") 
	if (self.tstatsNamespace!='' and self.hasFieldHierarchical(field)==False) or (self.tstatsNamespace=='' and not (self.hasField(field.strip("'")) or self.hasCalculatedField(field.strip("'")))):
	  raise PivotError("splitby", "fieldname " + field + " does not exist in object " + self.getName()) 
	for split in self.splitbyfields:
	  if field == split["field"] and roworcol == split["roworcol"]:
	    raise PivotError("splitby", "repeated " + roworcol + " splitby field  " + field) 
	  if label != '' and label == split["label"]:
	    raise PivotError("splitby", "cannot have duplicate label  " + label) 
 	for stat in self.stats:
          if label != '' and label == stat["rename"]:
            raise PivotError("splitby", "cannot have duplicate label  " + label)
	# some splitby fields can specify bucketing
	self.splitbyfields = self.splitbyfields + [{"field": field, "label": label, "roworcol": roworcol, "span": span, "bins": bins, "start": start, "end": end}]

 
    def topcmd(self, fields=[]):
        if len(self.top) == 0:
            self.top = fields
        else:
            self.top.append([fields])
 
    def hasField(self, fieldName):
        for field in self.fields:
            if field.getName() == fieldName:
                return True
        return False

    def hasDisplayName(self, fieldDisplayName):
        for field in self.fields:
            if field.getDisplayName() == fieldDisplayName:
                return True
        return False

    def hasFieldHierarchical(self, fieldName):
        for field in self.fields + self.getCalculatedFields():
	    if field.getOwner()[0:4]!="Base" and not isSplunkField(field.getName()):
	      hierarchicalFieldName = "'"+field.getOwner()+'.'+field.getName().strip('"').strip("'")+"'"
	    else: 
	      hierarchicalFieldName = "'"+field.getName().strip('"').strip("'")+"'"
            if hierarchicalFieldName == fieldName:
                return True
        return False

    def getFieldType(self, fieldType):
        for field in self.fields:
            if field.getType() == fieldType:
                return field

    def getField(self, fieldName):
        for f in self.fields:
            if f.getName() == fieldName:
                return f

    def getFields(self):
	return self.fields

    def getHierarchicalFieldNames(self, includeInherited=True):
	fieldnames = "" 
	for f in self.fields:
	  if f.getType() != "objectCount" and f.getType() != "childCount":
	    fname = f.getName()
	    fname = fname.strip('"')
	    fname = fname.strip("'")
	    if f.getOwner()[0:4]!="Base" and not isSplunkField(fname): #FIXME: special case these fields in separate function
	      fname = f.getOwner()+'.'+fname
	    if includeInherited or (self.getLineage()==f.getOwner()):
	      if fieldnames != "" and fname != "": fieldnames += ", "
	      fieldnames += '"'+fname+'"'
	return fieldnames

    def renameFieldsWithLineage(self):
	fieldnames = "" 
	for f in self.fields+self.getCalculatedFields():
	  if f.getType() != "objectCount" and f.getOwner()[0:4]!="Base" and not isSplunkField(f.getName()):
	    fname = f.getName()
	    fname = fname.strip('"')
	    fname = fname.strip("'")
	    if fieldnames != "" and fname != "": fieldnames += ", "
	    fieldnames += '"'+fname+'"' + ' AS "' + f.getOwner()+"."+fname+'"'
	if fieldnames != '': return "rename " + fieldnames
	else: return ''

    def renameFieldsWithoutLineage(self):
	fieldnames = "" 
	for f in self.fields+self.getCalculatedFields():
	  if f.getType() != "objectCount" and f.getOwner()[0:4]!="Base" and not isSplunkField(f.getName()):
	    fname = f.getName()
	    fname = fname.strip('"')
	    fname = fname.strip("'")
	    if fieldnames != "" and fname != "": fieldnames += ", "
	    fieldnames += '"'+f.getOwner()+"."+fname+'"'+ ' AS "' + fname+'"'
	if fieldnames != '': return "rename " + fieldnames
	else: return ''

    def hasCalculatedField(self, fieldName):
        for field in self.getCalculatedFields():
            if field.getName() == fieldName:
                return True
        return False

    def hasCalculatedDisplayName(self, fieldDisplayName):
        for field in self.getCalculatedFields():
            if field.getDisplayName() == fieldDisplayName:
                return True
        return False

    def getCalculatedFields(self):
	calcfields = [] 
	for calculation in self.calculations:
	  calcfields += calculation.getFields()
	return calcfields

    def getCalculatedField(self, fieldName):
      for field in self.getCalculatedFields():
        if field.getName() == fieldName:
          return field

    def getHierarchicalCalculatedFieldNames(self, includeInherited=True):
	fieldnames=''
	for field in self.getCalculatedFields():
          fname = field.getName()
	  fname = fname.strip('"')
	  fname = fname.strip("'")
	  if field.getOwner()[0:4]!="Base" and not isSplunkField(fname):
	      fname = field.getOwner()+'.'+fname
	  if includeInherited or (self.getLineage()==field.getOwner()):
            if fieldnames != "": fieldnames += ", "
            fieldnames += '"'+fname+'"'
        return fieldnames


    def filter(self, owner, f):
	if {"search": f, "owner": owner} not in self.filters and f != "":
          self.filters.append({"search": f, "owner": owner})

    # FIXME: Make this only usable with pivot UI	
    def filterwithname(self, fieldname, operation, RHS, fieldowner=''):
	#construct filtering search
	if operation == 'regex':
	  f = operation + " " + fieldname + '="' +RHS +'"'
	else:
	  f = fieldname + operation + RHS
	#check if fieldname is a required field or a calculated field 
	#if calculated field, need to do a postfilter
	if self.hasField(fieldname):
	  owner = self.getField(fieldname).getOwner()
	  if {"search": f, "owner": owner} not in self.filters and f != "":
            self.filters.append({"search": f, "owner": owner})
	else:
	  owner = self.lineage 
	  if {"search": f, "owner": owner} not in self.postfilters and f != "":
            self.postfilters.append({"search": f, "owner": owner})
	# tracking separate list of filters coming in at search time to make it easy for tstats
	# adding lineage to fieldname in this case since this is used only in tstats case
	if operation == '=' or operation == '!=' or operation == '<' or operation == '<=' or operation == '>' or operation == '>=':
	  if fieldowner!='' and fieldowner[0:4]!="Base" and not isSplunkField(fieldname): # should not be '' if coming from pivot. if base field, no lineage
            self.searchtimefilters = self.searchtimefilters + [{"search": '"'+fieldowner+'.'+fieldname+'"'+operation+RHS, "owner": self.lineage}]
	  else: 
            self.searchtimefilters = self.searchtimefilters + [{"search": f, "owner": self.lineage}]
	

    def getFilters(self):
	retval = []
	for field in self.fields:
            retval.append({"search": field.search(), "owner": field.getOwner()})
	for f in self.filters:
	    retval.append(f) 
        return retval


    def hasCalc(self, calculation):
	found=False
	for calc in self.calculations:
	  if calc.getCalculationID()!='':
 	    if calc.getCalculationID()==calculation.getCalculationID(): found=True
	  elif calculation.JSONdesc() == calc.JSONdesc(): found=True #no calc id assigned
	  # FIXME: compare everything except calculationID? that's a duplicate.. we may not care but it affects evals
	return found

    def reformatTimeValues(self, timefield):
	 self.reformattime+= ' | eval \"' + timefield + '\"= strftime(\'' + timefield + '\',  \"%FT%H:%M:%S.%3N%z\")'

    def replaceBoolValues(self, fieldname, trueLabel='', falseLabel=''):
	fieldname = '"'+fieldname.strip("'").strip('"')+'"'
	if trueLabel!='':
	  self.replaceBoolLabels+=' | replace \"true\" with \"' + trueLabel + '\" in ' + fieldname
	if falseLabel!='':
	  self.replaceBoolLabels+=' | replace \"false\" with \"' + falseLabel + '\" in ' + fieldname
	
	
    def evalField(self, myowner, outputfield, valueSearch, comment='', calcID='', editable=True):
        myeval = Eval(myowner, outputfield,valueSearch, comment, calcID, editable)
	if not self.hasCalc(myeval): self.calculations.append(myeval)

    def lookupFields(self, myowner, lookupname, inputfield, lookupfield='', outputFields=[], comment='', calcID='', editable=True):
        mylookup = Lookup(myowner, lookupname, inputfield, lookupfield, outputFields, comment=comment, calculationID=calcID, editable=editable)
	if not self.hasCalc(mylookup): self.calculations.append(mylookup)

    def geoipFields(self, myowner, inputfield, outputFields=[], comment='', calcID='', editable=True):
        mygeoip = GeoIP(myowner, inputfield, outputFields, comment, calcID, editable)
	if not self.hasCalc(mygeoip): self.calculations.append(mygeoip)

    def rexField(self, myowner, fieldname, outputFields, regex, maxmatch=1, comment='', calcID='', editable=True):
 	myrex = Rex(myowner, fieldname, outputFields, regex, maxmatch, comment, calcID, editable)
	if not self.hasCalc(myrex): self.calculations.append(myrex)

    def statsField(self, myowner, funcfieldnamelist, pivotby, comment='', calcID='', editable=True):
        mystats = StatsField(myowner, funcfieldnamelist, pivotby, comment, calcID, editable)
	if not self.hasCalc(mystats): self.calculations.append(mystats)

    def JSONdesc(self, conciseMode):
        #write out class name
        desc = {}
        desc["objectName"] = self.getName()
        desc["displayName"] = self.getDisplayName()
        #write parent name
        desc["parentName"] = self.getParentName()
	
	desc["comment"]=self.getComment()

        if conciseMode:
            desc["fields"] = []
	    for f in self.fields:
		if f.getOwner() == self.getLineage() and f.getType()!="objectCount":
	    	  desc["fields"].append(f.JSONdesc(conciseMode=True))
		elif f.getOwner()!=self.getLineage() and f.getName() in self.changedinheritedfields:
	    	  # if this object changed an inherited field's hidden/required bits or displayName, show in concise model but include owner
              	  desc["fields"].append(f.JSONdesc(conciseMode=True, inclOwner=True))
            desc["calculations"] = [c.JSONdesc(conciseMode=True) for c in self.calculations if c.getOwner() == self.getLineage() and c.JSONdesc()["outputFields"][0]["type"]!="childCount"]
            desc["constraints"] = [{"search":c["search"]} for c in self.filters+self.postfilters if not c == None and (not "owner" in c or c["owner"] == self.getLineage())]
        else:
            desc["lineage"] = self.getLineage()
	    desc["tsidxNamespace"] = self.getTsidxNamespace()
            desc["fields"] = map(Field.JSONdesc, self.fields)
            desc["calculations"] = map(Calculation.JSONdesc, self.calculations)
            desc["constraints"] = self.filters + self.postfilters
            desc["autoextractSearch"] = self.search(mode="autoextract") #this one's for Jason to use in the editor
            desc["previewSearch"] = self.search(mode="preview") #this one's for Jason to use in the editor
            desc["objectSearchNoFields"] = self.search() #this one's for Jason to use in the editor
	    # adding | fields here for simon to use in tscollect
	    desc["objectSearch"]=desc["objectSearchNoFields"]
	    if self.getHierarchicalFieldNames()!='':
              desc["objectSearch"] += ' | fields ' + self.getHierarchicalFieldNames() 
	      if self.getHierarchicalCalculatedFieldNames()!='':
                desc["objectSearch"] += ", " + self.getHierarchicalCalculatedFieldNames()
	    elif self.getHierarchicalCalculatedFieldNames()!='':
              desc["objectSearch"] += " | fields " + self.getHierarchicalCalculatedFieldNames()
            desc["children"] = self.getChildNames()
        return desc


	
    def search(self, mode='full'):

      if mode not in ["simple", "full", "preview", "autoextract", "drilldown"]:
	mode = "full" #default to full search. 

      startingsearch=''
      if mode not in ['preview', 'autoextract', "drilldown"] and self.tstatsNamespace != '' and self.stats !=[]: #produce tstats search
	statsObj = Stats(self.stats, self.splitbyfields, self.rowSummary, self.colSummary, self.colLimit, self.colOther, self.replaceBoolLabels, self.sortsearch)
	retval = statsObj.tstatsSearch(self.tstatsNamespace, self.searchtimefilters)
	retval += self.reformattime

      else: # not tstats...just regular stats
	lineagelist = []
	curlineage = self.getLineage()
	# get sorted list of objects by lineage
	while curlineage!='':	
	  lineagelist = [curlineage]+lineagelist
          curlineage = '.'.join(curlineage.split('.')[:-1])
	lineagelist = [self.getObjectType()]+lineagelist
	
	retval = ''

	prevobjhaspipe=0
	if mode=='full' and self.renameFieldsWithLineage()!='':
	  # rename raw data fields to hierarchical fieldnames (to avoid collisions with other objects' fields)
          retval += '| ' + self.renameFieldsWithLineage() 
	  prevobjhaspipe=1
	  # rename hierarchical fieldnames to fieldnames (since filters/calculations will be on regular fieldnames)
	  if self.renameFieldsWithLineage() != '':
            retval += " | " + self.renameFieldsWithoutLineage() 

	childCountCalcs = ''
	for myobj in lineagelist:	
	  retvalmyobj=''
  	  # get field searches, basically constraints on each field
	  fieldsearchesandfilters = ''
          for field in self.fields:
	    if field.getOwner()==myobj and field.search()!=' ':
              fieldsearchesandfilters += field.search()

          for filter in self.filters:
            if filter["search"].startswith('regex') == False:
	        if filter["owner"][0:4]=="Base" or (filter["owner"] == self.getLineage().split('.')[0] and myobj==self.getLineage().split('.')[0]):
		  startingsearch += " (" + filter["search"] + ") " 
		elif filter["owner"]==myobj:
                  fieldsearchesandfilters += " (" + filter["search"] + ") "


	  if fieldsearchesandfilters != '': 
	    if prevobjhaspipe==1 or (mode=="preview" and startingsearch=='') or self.getObjectType()=="BaseSearch":
		retvalmyobj += " | search " + fieldsearchesandfilters
	    else: #append to constraints for parent object since nothing new was calculated/extracted
		retvalmyobj += " "+fieldsearchesandfilters
	    prevobjhaspipe=0

          #need to check if filter is "regex"...if it is, need to put it in separately with | 
          for filter in self.filters:
            if filter["owner"]==myobj and filter["search"].startswith('regex'):
              retvalmyobj += " |" + filter["search"] + " "
	      prevobjhaspipe=1

	  calcSearches = ''
          for calc in self.calculations:
	    if calc.getOwner()==myobj: 
	      if calc.getFields()[0].getType()!='childCount':
                retvalmyobj += ' | ' + calc.search()
	        prevobjhaspipe=1
	        calcSearch = ''
	        for f in calc.getFields():
	          calcSearch += f.search()
		if calcSearch != '':
		  calcSearches += calcSearch	
	      else: 
		childCountCalcs += ' | ' + calc.search()
		# do evals for childCount objects after all other calculations, allowing post calculation constraints in children
	  retval += retvalmyobj
	  if calcSearches!='':
	    retval += " | search " + calcSearches

	if mode == 'autoextract': 
	  return startingsearch

	if mode!='preview':
	  retval += childCountCalcs

	# move root object constraint before renames  
 	if startingsearch!='':
	    retval = ' | search ' + startingsearch + retval

        if self.postfilters!=[]:
            retval += ' | search '
            for filter in self.postfilters:
                if filter["search"].startswith('regex') == False:
                   retval += " (" + filter["search"] + ") "
            #need to check if filter is "regex"...if it is, need to put it in separately with | 
            for filter in self.postfilters:
                if filter["search"].startswith('regex'):
                   retval += " |" + filter["search"] + " "
	
	if mode=="preview": return retval
	
 	# rename raw data fields to hierarchical fieldnames...pivot uses hierarchical naming
	if (self.tstatsNamespace !='' or mode=='full') and self.renameFieldsWithLineage()!='':
	  retval += ' | ' + self.renameFieldsWithLineage()

	if mode=="drilldown":
	  tempsplitbyfields=[]
	  for split in self.splitbyfields:
	    split['roworcol']='row'
	    tempsplitbyfields.append(split)
	  self.splitbyfields = tempsplitbyfields
	  self.rowSummary = False

        if self.stats != []:
	    if (self.tstatsNamespace =='' and mode=='full') and self.renameFieldsWithLineage()!='':
	      retval += ' | ' + self.renameFieldsWithoutLineage()
            retval += Stats(self.stats, self.splitbyfields, self.rowSummary, self.colSummary, self.colLimit, self.colOther, self.replaceBoolLabels, self.sortsearch).search()
	    retval += self.reformattime
	
      return retval


    def getSearchComponents(self):
      # this function returns the different pieces of the search string separately so they can be assembled as necessary for things like transactions
	fieldsSearch = ''; calculationSearch = ''; prefilterSearch = ''; postfilterSearch = ''; statsSearch = ''; renameWithLineage = ''; renameWithoutLineage = ''
        for field in self.fields:
            fieldsSearch += " " + field.search()
        for filter in self.filters:
            if filter["search"].startswith('regex') == False:
              prefilterSearch += " (" + filter["search"] + ") "
            #need to check if filter is "regex"...if it is, need to put it in separately with | 
        for filter in self.filters:
            if filter["search"].startswith('regex'):
              prefilterSearch += ' | ' + filter["search"] + " "

	calcSearches = ''
        for calc in self.calculations:
            calculationSearch += calc.search() + ' | '
	    calcSearch = ''
	    for f in calc.getFields():
	      calcSearch += f.search()
	    if calcSearch != '':
	      calcSearches += calcSearch	
	if calcSearches!='':
	  calculationSearch += " | search " + calcSearches
	    

        if self.postfilters!=[]:
            postfilterSearch += ' | search '
            for filter in self.postfilters:
                if filter["search"].startswith('regex') == False:
                   postfilterSearch += " (" + filter["search"] + ") "
            #need to check if filter is "regex"...if it is, need to put it in separately with | 
            for filter in self.postfilters:
                if filter["search"].startswith('regex'):
                   postfilterSearch += " |" + filter["search"] + " "

        renameWithLineage = self.renameFieldsWithLineage() 
	# rename hierarchical fieldnames to fieldnames (since filters/calculations will be on regular fieldnames)
	if self.tstatsNamespace != '' and self.renameFieldsWithLineage() != '':
          renameWithoutLineage = self.renameFieldsWithoutLineage() 

        if self.stats != []:
            statsSearch = Stats(self.stats, self.splitbyfields, self.rowSummary, self.colSummary, self.colLimit, self.colOther, self.replaceBoolLabels, self.sortsearch).search()

        retval = {"fields": fieldsSearch, "calculations": calculationSearch, "prefilter": prefilterSearch, "postfilter": postfilterSearch, "statsSearch": statsSearch, "renameWithLineage": renameWithLineage, "renameWithoutLineage": renameWithoutLineage}	

        return retval

	
  # This function returns a search string and a list of fields corresponding to the current object and its descendants
    def getSearchTree(self):
      fields=""
      if self.getParentName() in INTERNAL_OBJECT_NAMES:
        search = self.getConstraints() + ' | eval nodename = "'+self.getLineage()+'"'
      else:
        search =' | eval nodename = mvappend(nodename, if(searchmatch(" nodename='+ self.getLineage()[0:-len(self.getName())-1]+" " +self.getConstraintsWithoutParents().replace('"','\\\"')+'"),"'+ self.getLineage() + '",null()))'
      if self.getCalculationsWithoutParents()!='':
        search += ' | ' + self.getCalculationsWithoutParents()
      if self.renameFieldsWithLineage()!='':
        search += ' | ' + self.renameFieldsWithLineage()
        myfields = self.getHierarchicalFieldNames(includeInherited=False)
        mycalcfields = self.getHierarchicalCalculatedFieldNames(includeInherited=False)
        fields += myfields
        if myfields!='' and mycalcfields!='': fields += ', '+mycalcfields
        else: fields += mycalcfields
      for c in self.getChildObjects():
        if c.renameFieldsWithoutLineage()!='':
          search += ' | ' + c.renameFieldsWithoutLineage()
        searchTree=c.getSearchTree()
        search += searchTree["search"]
        if fields!='' and searchTree["fields"]!='':
          fields+=", "
        fields += searchTree["fields"]

      return {"search": search, "fields": fields}

