import json
import glob
import logging, logging.handlers
import math
import os
import sys
import tempfile
import time
import string
import re
import hashlib

from field import Field
from stats import Stats
from errormsgs import DMError


import splunk
import splunk.auth as auth
import splunk.search as search
import splunk.util as util


SPLUNK_HOME = os.environ['SPLUNK_HOME']

class Calculation():
    def __init__(self, calculationType='', calculationID='', inputfield='', owner='', outputfields=[], data='', hidden=False, comment='', editable=True):
	self.inputfield = inputfield
	self.type = calculationType
	self.owner = owner
	self.data = data
        self.isHidden = hidden
        self.isEditable = editable
	self.comment=comment
	self.calculationID=calculationID
	self.calculationType=calculationType
	self.outputfields = []
	for field in outputfields:
	    if "displayName" in field and field["displayName"]!='': dispName = field["displayName"]
	    else: dispName = field["fieldName"]	
	    if " " in field["fieldName"]: 
	      raise DMError("OBJECT_CALCULATION_SPACE_IN_FIELDNAME", {'fieldName': field["fieldName"], 'objectName': owner})
	    for f in self.outputfields:
	      if field["fieldName"] == f.getName():
	        raise DMError("OBJECT_CALCULATION_DUPLICATE_NAME", {'fieldName': field["fieldName"], 'objectName': owner})
	      if f.getDisplayName() == dispName:
	        raise DMError("OBJECT_CALCULATION_DUPLICATE_DISPLAYNAME", {'displayName': dispName, 'objectName': owner})
	    self.outputfields.append(Field(field["fieldName"], type=field["type"], owner=owner, hidden=field["hidden"], required=field["required"], displayName=dispName, editable=self.isEditable))
	if calculationID=='':
	  self.setCalculationID()	

    def getData(self):
	return self.data
	
    def getOwner(self):
	if self.owner!='':
	  return self.owner

    def getFields(self):
	return self.outputfields

    def setHidden(self,value):
        self.isHidden=value

    def getCalculationID(self):	
	return self.calculationID

    def getCalculationType(self):	
	return self.calculationType

    def getComment(self):	
	return self.comment

    def setCalculationID(self):	
 	desc = {"calculationType": self.getCalculationType(), "outputFields": map(Field.JSONdesc, self.getFields()), "owner": self.owner}
        desc.update(self.getData())
	self.calculationID = hashlib.sha1(json.dumps(desc, sort_keys=True)).hexdigest()

    def JSONdesc(self, conciseMode=False):
	if conciseMode==True:
	  desc = {"calculationID": self.calculationID, "calculationType": self.getCalculationType(), "outputFields": [f.JSONdesc(conciseMode=True) for f in self.getFields()], "comment": self.getComment()}
	else:
	  desc = {"calculationID": self.calculationID, "calculationType": self.__class__.__name__, "outputFields": map(Field.JSONdesc, self.getFields()), "owner": self.owner, "comment": self.getComment(), "editable": self.isEditable}
	desc.update(self.getData())
	return desc

class Lookup(Calculation):
    def __init__(self, myowner, lookupname, inputfield, lookupfield='', outputfields=[], comment='', calculationID='', editable=True):
	if lookupname=='':
	  raise DMError("OBJECT_CALCULATION_LOOKUP_EMPTY_LOOKUPNAME", {'objectName': myowner})
	if inputfield=='':
	  raise DMError("OBJECT_CALCULATION_LOOKUP_EMPTY_INPUTFIELD", {'objectName': myowner})
	if outputfields==[]: #this is a check for datamodels only..in general we don't have to specify outputfields for lookups
	  raise DMError("OBJECT_CALCULATION_LOOKUP_EMPTY_OUTPUTFIELDS", {'objectName': myowner})
	if lookupfield=='':
	  lookupfield = inputfield
        self.lookupname = lookupname
        self.lookupfield = lookupfield
	data = {"lookupName": self.lookupname, "inputField": inputfield, "lookupField": self.lookupfield}
	Calculation.__init__(self, calculationID=calculationID, calculationType='Lookup', owner=myowner, outputfields=[], inputfield=inputfield, comment=comment, data=data, editable=editable)
        for field in outputfields:
	    if field["fieldName"]=='':  
	      raise DMError("OBJECT_CALCULATION_LOOKUP_EMPTY_OUTPUTFIELDS", {'objectName': myowner})
            if "displayName" in field and field["displayName"]!='': dispName = field["displayName"]
            else: dispName = field["fieldName"]
	    for f in self.outputfields:
	      if field["fieldName"] == f.getName():
	        raise DMError("OBJECT_CALCULATION_DUPLICATE_NAME", {'fieldName': field["fieldName"], 'objectName': myowner})
	      if f.getDisplayName() == dispName:
	        raise DMError("OBJECT_CALCULATION_DUPLICATE_DISPLAYNAME", {'displayName': dispName, 'objectName': myowner})
	    if "lookupOutputFieldName" in field: 
              self.outputfields.append(Field(field["fieldName"], type=field["type"], owner=myowner, hidden=field["hidden"], required=field["required"], displayName=dispName, lookupOutputFieldName=field['lookupOutputFieldName'], editable=editable))
            else:
	      self.outputfields.append(Field(field["fieldName"], type=field["type"], owner=myowner, hidden=field["hidden"], required=field["required"], displayName=dispName, editable=editable))



    def search(self):
	lookupstring = "lookup " + self.lookupname + " \"" + self.lookupfield + "\" AS \""  + self.inputfield + "\" OUTPUT"
	for f in self.outputfields:
	  if 'lookupOutputFieldName' in f.JSONdesc():
	    lookupstring += ' \"' + f.JSONdesc()['lookupOutputFieldName'] + '\" AS ' + f.getName()
	  else: 
	    lookupstring += ' ' + f.getName() # NOTE: don't need renaming since we rename with lineage later anyway.
	# FIXME: might need renaming to display name if we allow multiple geoip calcs in the same object.
        return lookupstring

class GeoIP(Calculation):
    def __init__(self, myowner, inputfield, outputfields=[], comment='', calculationID='', editable=True):
	if inputfield=='':
	  raise DMError("OBJECT_CALCULATION_GEOIP_EMPTY_INPUTFIELD", {'objectName': myowner})
	data = {"inputField": inputfield}
	Calculation.__init__(self, calculationID=calculationID, calculationType='GeoIP', owner=myowner, outputfields=[], inputfield=inputfield, comment=comment, data=data, editable=editable)
	for field in outputfields:
            if "displayName" in field and field["displayName"]!='': dispName = field["displayName"]
            else: dispName = field["fieldName"]
	    for f in self.outputfields:
	      if field["fieldName"] == f.getName():
	        raise DMError("OBJECT_CALCULATION_DUPLICATE_NAME", {'fieldName': field["fieldName"], 'objectName': myowner})
	      if f.getDisplayName() == dispName:
	        raise DMError("OBJECT_CALCULATION_DUPLICATE_DISPLAYNAME", {'displayName': dispName, 'objectName': myowner})
            if "lookupOutputFieldName" in field:
	      if field["lookupOutputFieldName"] not in ["lat", "lon", "Country", "Region", "City"]:
	        raise DMError("OBJECT_CALCULATION_GEOIP_INVALID_OUTPUTFIELD", {'lookupOutputFieldName': field["lookupOutputFieldName"], 'objectName': myowner})
              self.outputfields.append(Field(field["fieldName"], type=field["type"], owner=myowner, hidden=field["hidden"], required=field["required"], displayName=dispName, lookupOutputFieldName=field['lookupOutputFieldName'], editable=editable))
            else:
              self.outputfields.append(Field(field["fieldName"], type=field["type"], owner=myowner, hidden=field["hidden"], required=field["required"], displayName=dispName, editable=editable))


    #City	Country	CountryCode	Region	clientip	lat	lon
    def search(self):
        return "iplocation \"" + self.inputfield + "\" prefix=\"" + self.inputfield+'_\"'

class Rex(Calculation):
    def __init__(self, myowner, inputfield, outputFields, expression, maxmatch=1, comment='', calculationID='', editable=True):
	if inputfield=='':
	  raise DMError("OBJECT_CALCULATION_REX_EMPTY_FIELD", {'objectName': myowner})
	if outputFields==[]:
	  raise DMError("OBJECT_CALCULATION_REX_EMPTY_OUTPUTFIELDS", {'objectName': myowner})
	if expression=='':
	  raise DMError("OBJECT_CALCULATION_REX_EMPTY_EXPRESSION", {'objectName': myowner})
	self.maxmatch = maxmatch
	self.expression = expression
        data = {"inputField": inputfield, "expression": self.expression}
        extractedOutputFields = re.findall('<[^>]*>', self.expression)	#match everything except > so we read the fieldname correctly
	# check that output fields specified match the fields extracted from expression.
	for field in outputFields:
	    if not '<'+field["fieldName"]+'>' in extractedOutputFields: #strip out < and >
	      raise DMError("OBJECT_CALCULATION_REX_INVALID_FIELD", {'fieldName': field['fieldName'], 'objectName': myowner})
	Calculation.__init__(self, calculationID=calculationID, calculationType="Rex", owner=myowner, outputfields=outputFields, inputfield=inputfield, comment=comment, data=data, editable=editable)


    def search(self):
        return "rex field=\"" + self.inputfield + "\" \"" + self.expression + "\" " + "max_match=" + str(self.maxmatch)


class StatsField(Calculation):
    def __init__(self, myowner, funcandfieldlist, pivotfields=[], hidden=False, comment='', calculationID='', editable=True):
	self.funcfieldlist = funcandfieldlist
	self.pivotfields = pivotfields
        data = {"functionFieldList": self.funcfieldlist, "pivotFields": self.pivotfields}
	outputfields = Stats(self.funcfieldlist, self.pivotfields).getOutputFields()
	# FIXME: need to check consistency between outputfields in funcandfieldlist and outputfields passed in
	outfields=[]
	for varname in outputfields:
	  outfields.append(Field(varname, type='string', owner=myowner, hidden=hidden, editable=editable).JSONdesc())
	Calculation.__init__(self, calculationID=calculationID, calculationType="StatsField", owner=myowner, comment=comment, hidden=hidden, outputfields=outfields, data=data, editable=editable)

    def search(self):
        return Stats(self.funcfieldlist, self.pivotfields).getStatsString() 


class Eval(Calculation):
    def __init__(self, myowner, outputfield, varvalue, comment='', calculationID='', editable=True):
	if outputfield==[] or outputfield["fieldName"]=='':
	  raise DMError("OBJECT_CALCULATION_EVAL_EMPTY_OUTPUTFIELD", {'objectName': myowner})
	if varvalue=='':
	  raise DMError("OBJECT_CALCULATION_EVAL_EMPTY_VARVALUE", {'objectName': myowner})
	self.varvalue = varvalue
        data = {"expression": self.varvalue}
	Calculation.__init__(self, calculationID=calculationID, calculationType="Eval", owner=myowner, outputfields=[outputfield], comment=comment, hidden=outputfield["hidden"], data=data, editable=editable)

    def search(self):
        return "eval \"" + self.outputfields[0].getName() + "\"=" + self.varvalue 
