import glob
import logging, logging.handlers
import math
import os
import sys
import tempfile
import time
import json
import re
import traceback
import logging


from errormsgs import DMError
from base import Base
from baseEvent import BaseEvent
from baseEventSet import BaseTransaction, BaseAggregate
from baseSearch import BaseSearch
from baseInterface import BaseInterface, BaseImplements
from field import Field
# FIXME - we should find a more robust way to do this. "import *" is not awesome.

logger = logging.getLogger("splunk")

# This class encapsulates the metadata for a data model, including available classes and (if applicable) mappings. 
# It should work the same way regardless of whether the model was generated from JSON or Python.
# It provides constructors for all available objects.
# An object is not guaranteed to be present in a model unless Model.hasObject() returns True.

class Model:
    #init from dictionary
    def __init__(self, raw, modelName, isAccelerated=False):
        self.childLists = {}
        self.implementsLists = {}

        if raw != None:
            #FIXME: use KeyError here to catch malformed models!
            self.rawData = raw # the dictionary we got from JSON, storing to avoid passing it around all over the place            

            self.displayName = raw["displayName"]
            self.description = raw["description"]            

            self.name = modelName
            
            #FIXME: the model is allowed to not have a modelName (the uniqueID).  Handle this.
           
            # FIXME: update this when marklar changes the tstats namespace syntax
	    if isAccelerated:
              self.tsidxNamespace = 'datamodel="'+modelName+'"'
 	    else: self.tsidxNamespace=''

            #allow for empty object list
            if not 'objects' in self.rawData:
                self.rawData['objects'] = []

            self.canAccelerate = False #will be updated if we load an acceleratable object
           	 
            self.loadObjectsFromList(self.rawData['objects'])

        else:
            raise DMError("MODEL_EMPTY", {'modelName': '-unknown-'})

    #method to fill in the object names and factories.  Separated from constructor to allow updating later.
    def loadObjectsFromList(self, objectList):
        self.childLists = {}
        self.implementsLists = {}
        self.objectFactories = {} # a dict of constructor for objects in the model
        self.objectNames = [] #list of valid objects in this model
            #iterate through objects in the raw data, add them to the model
	self.firstBaseEvent = ''

        for objectData in objectList:
	    try: 	
	      objectData['tsidxNamespace']=self.tsidxNamespace
              if self.hasObject(objectData["objectName"]):
	        logger.exception('Cannot add object with the same objectName as an existing object. Traceback: '+traceback.format_exc())
		raise DMError('OBJECT_DUPLICATE_OBJECTNAME', {'objectName': objectData['objectName'], 'modelName': self.name})

              self.loadObject(objectData)
	    except DMError as DMErr:
 	      logger.exception('loadObject failed in loadObjectsFromList. Traceback: '+traceback.format_exc())
	      params={'modelName': self.name}
              params.update(DMErr.__json__()['parameters'])
              raise DMError(DMErr.__json__()['errorCode'], params)
	    except:
	      logger.exception("unknown error loading object "+objectData["objectName"]+". Traceback: "+traceback.format_exc())
              raise DMError('MODEL_OBJECT_LOADING_ERROR', {'objectName': objectData['objectName'], 'modelName': self.name})


    def getName(self):
        return self.name

    def setName(self, name):
	pattern = r'[^_a-z0-9]'
    	if re.search(pattern, name):
	  raise DMError('OBJECT_INVALIDNAME', {}) # check that only understcore, alpha/numberic 
        self.name = name


    # this function returns a search string for all event-based objects in the entire datamodel
    def buildModelSearch(self):
        search = ''
        fields = ''
        rootObjects=[]
        for o in self.objectNames:
          curobj = self.getObject(o)
          if curobj.getParentName()=="BaseEvent":
            rootObjects.append(curobj)
        if len(rootObjects)>=1:
          searchtree = rootObjects[0].getSearchTree()
          search = '(index=* OR index=_*) ' + searchtree["search"]
          fields = searchtree["fields"]
          search+= " | fields nodename, _time, host, source, sourcetype"  # default fields to always include 
          if fields.strip(' ')!='': search+= ', '+fields
	if len(rootObjects)>1:
	  logger.exception("accelerating only the first BaseEvent object " + rootObjects[0].getName() + " in the model "+self.name + ". There are other BaseEvent objects that are not accelerated")
        return search

    # this function returns the first BaseEvent object in the model. this object and its decendants can be accelerated.
    def getFirstBaseEvent(self):
	return self.firstBaseEvent

    #returns true if an object has been loaded to the model
    def hasObject(self, name):
        return name in self.objectNames

    #internal method to tell whether or not an object is valid in this model (though it may not be loaded yet)
    def validObject(self, name):
        if self.hasObject(name):
            return True
        for objectData in self.rawData["objects"]:
            if 'objectName' in objectData and objectData["objectName"] == name:
                return True
        return False

    def objectDictForName(self, objectName):
        for o in self.rawData["objects"]:
            if 'objectName' in objectData and o["objectName"] == objectName:
                return o
        return None

    # Find a unique name for an object based on the displayName supplied.  Similar to code for                                                      
    # unique model names in controller.py. We could use hashes to minimize the change of collisions,
    # but this errs on the side of human-readable object names
    def uniqueObjectName(self, name):
        name = name.replace(' ', '')
        counter = 0
        isDuplicate = False
        for o in self.objectNames:
            if o == name:
                isDuplicate = True
                num = re.sub("[a-zA-Z]+", "", o)
                if num != '':
                    counter = max(int(num), counter)
        if isDuplicate:
            return name + str(counter + 1)
        else:
            return name

    #this method takes in the data for an object and adds a constructor for that object to the "objects" array
    def loadObject(self, objectData):
        #first make sure the names are filled out and unique
        if not 'objectName' in objectData:
            raise DMError("OBJECT_NONAME", {'modelName': self.name})

	if objectData['objectName'] in ['Base', 'BaseEvent', 'BaseSearch', 'BaseTransaction', 'BaseInterface', 'BaseImplements', 'BaseAggregate']:
            raise DMError("OBJECT_INTERNAL_OBJECT_NAME", {'objectName': objectData['objectName'], 'modelName': self.name})

        if ' ' in objectData['objectName']: #special snowflake cervelli error: see  SPL-71488
            raise DMError('OBJECT_INVALIDNAME', {'objectName': objectData['objectName'], 'modelName': self.name})

        if not 'displayName' in objectData:
            objectData['displayName'] = objectData['objectName']

        #only load the object if we haven't already done so
        if not self.hasObject(objectData["objectName"]):
            #create and store the constructor
            self.objectFactories[objectData["objectName"]] = self.objectFactory(objectData)
            if(self.objectFactories[objectData["objectName"]] == None):
                raise DMError("OBJECT_NULL_OBJECTFACTORY", {'objectName': objectData["objectName"], 'modelName': self.name})

            #register object with "valid objects" list
            self.objectNames.append(objectData["objectName"])

            #if there is a root event object, the model can be accelerated. We need to keep track of this.
            if objectData["parentName"] == "BaseEvent":
                self.canAccelerate = True

            try:
                if not objectData["parentName"] in self.childLists:
                    self.childLists[objectData["parentName"]] = []
                self.childLists[objectData["parentName"]].append(objectData["objectName"])
	   
            except:
                logger.exception("failed to add child to parent " + objectData["parentName"]+" Traceback: "+traceback.format_exc())
                raise DMError("OBJECT_CANNOT_ADD_CHILD", {'parentName': objectData["parentName"], 'objectName': objectData["objectName"], 'modelName': self.name})
	    
	    if objectData["parentName"] == "BaseImplements":
              try:
                if not objectData["implementsInterface"] in self.implementsLists:
                    self.implementsLists[objectData["implementsInterface"]] = []
                self.implementsLists[objectData["implementsInterface"]].append(objectData["objectName"])

              except:
                logger.exception("failed to add implements to interface " + objectData["implementsInterface"]+" Traceback: "+traceback.format_exc())
                raise DMError("OBJECT_CANNOT_ADD_IMPLEMENTS_TO_INTERFACE", {'interfaceObjectName': objectData["implementsInterface"], 'objectName': objectData["objectName"], 'modelName': self.name})
		
    
    #manufactures a class with the given characteristics
    def objectFactory(self, data):
	try:
          parentFactory = self.getParentFactory(data)
	except DMError as DMErr:
	  raise DMError(DMErr.__json__()['errorCode'], DMErr.__json__()['parameters']) 
        
        if parentFactory != None:
            #define a method that initializes this object
            def myFactory():
            #create parent object
                myInstance = parentFactory()
                if myInstance == None:
                    raise DMError("OBJECT_CANNOT_CREATE_PARENT", {'objectName': data["objectName"], 'modelName': self.name})

                #change the name and parent
                myInstance.setName(data["objectName"])
                myInstance.setDisplayName(data["displayName"])
                myInstance.setParentName(data["parentName"])
		if "comment" in data and data["comment"]!='':
		  myInstance.addComment(data["comment"])

		if myInstance.getObjectType() == "BaseEvent" and self.firstBaseEvent=='':
		  self.firstBaseEvent = myInstance.getName()

		if data['tsidxNamespace']!='' and myInstance.getObjectType() == "BaseEvent":
                  myInstance.setTsidxNamespace(namespace=data['tsidxNamespace'])

	
		# need to track lineage to identify where we are in the hierarchy
                myInstance.setLineage()
		data["lineage"]=myInstance.getLineage()

		if myInstance.getObjectType() == "BaseEvent" and data["lineage"].split(".")[0] != self.firstBaseEvent:
                  myInstance.unsetTsidxNamespace()

                #add/modify fields
                for f in data["fields"]:
		  if ("owner" not in f) or ("owner" in f and f["owner"]==""): f["owner"] = data["lineage"]
		  #may be objectName when reading a model from a file vs lineage when from editor UI
                  hidden = False
		  editable = True
		  comment=''
                  if 'editable' in f: editable = f['editable']
                  if 'hidden' in f: hidden = f['hidden']
                  if 'displayName' in f and f['displayName']!='': displayName = f['displayName']
                  else: displayName = f["fieldName"]
		  if 'comment' in f: comment=f['comment']
		  if myInstance.hasDisplayName(displayName) and not myInstance.hasField(f["fieldName"]):
          	    raise DMError("OBJECT_DUPLICATE_DISPLAYNAME", {'displayName': displayName, 'objectName': myInstance.getName(), 'modelName': self.name})
		  if f["owner"]==data["objectName"] or f["owner"]==data["lineage"]:
		    # addField takes care of updates if the field already exists. 
                    myInstance.addField(owner=data["lineage"], fieldName=f["fieldName"], fieldType=f["type"], required=f["required"], multivalued=f["multivalue"], hidden=hidden, displayName=displayName, comment=comment, editable=editable)
                  else: #handle changes to hidden/required bits in inherited fields 
		    myInstance.addField(owner=f["owner"], fieldName=f["fieldName"], fieldType=f["type"], required=f["required"], multivalued=f["multivalue"], hidden=hidden, displayName=displayName, comment=comment, editable=editable)



                #add/modify calculations
                for c in data["calculations"]:
		  if ("owner" not in c) or ("owner" in c and c["owner"]==""): c["owner"] = data["lineage"]
		  #may be objectName when reading a model from a file vs lineage when from editor UI
		  if c["owner"]==data["objectName"] or c["owner"]==data["lineage"]:
		    for d in c["outputFields"]:
		      if myInstance.hasField(d["fieldName"]) or myInstance.hasCalculatedField(d["fieldName"]):
          	        raise DMError("OBJECT_DUPLICATE_FIELDNAME", {'fieldName': d["fieldName"], 'objectName': myInstance.getName(), 'modelName': self.name})
		      if "displayName" in d and d["displayName"]!='' and (myInstance.hasDisplayName(d["displayName"]) or myInstance.hasCalculatedDisplayName(d["displayName"])):
          	        raise DMError("OBJECT_DUPLICATE_DISPLAYNAME", {'displayName': d["displayName"], 'objectName': myInstance.getName(), 'modelName': self.name})
                    cType = c["calculationType"]
		    calcID=''; comment=''; editable=True
		    if 'editable' in c: editable=c['editable']
		    if 'calculationID' in c: calcID=c['calculationID']
		    if 'comment' in c: comment=c['comment']
                    if cType == "StatsField":
                        myInstance.statsField(data["lineage"], c["functionFieldList"], c["pivotFields"], comment=comment, calcID=calcID)
                    elif cType == "Lookup":
			lookupfield=''
			if "lookupField" in c: lookupfield = c["lookupField"]
                        myInstance.lookupFields(data["lineage"], c["lookupName"], c["inputField"], lookupfield=lookupfield, outputFields=c["outputFields"], comment=comment, calcID=calcID, editable=editable) #FIXME: check data formats here
                    elif cType == "GeoIP":
                        myInstance.geoipFields(data["lineage"], c["inputField"], c["outputFields"], comment=comment, calcID=calcID, editable=editable) #FIXME: check data formats here
                    elif cType == "Eval": # note: only one outputfield is used for eval
		  	#don't want to add childCount fields since addChild handles it...this avoids having childCount
		  	#fields leftover for deleted childobjects
			if c["outputFields"][0]["type"]!="childCount":
                          myInstance.evalField(data["lineage"], c["outputFields"][0], c["expression"], comment=comment, calcID=calcID, editable=editable) #FIXME: check formats
                    elif cType == "Rex":
                        myInstance.rexField(data["lineage"], c["inputField"], c["outputFields"], c["expression"], comment=comment, calcID=calcID, editable=editable)#FIXME: check formats
                    else:
                        raise DMError("OBJECT_INVALID_CALCULATION_TYPE", {'objectName': data["objectName"], 'calculationType': cType, 'modelName': self.name})
    


                #add/modify constraints
                for c in data["constraints"]: 
		  # may have empty owner when creating a new object
		  if "owner" not in c or c["owner"] == "" or c["owner"]==data["objectName"] or c["owner"]==data["lineage"]:
                    if "|" in c["search"]: 
		      logger.exception("Object constraints cannot contain pipes. Constraint search: " + c["search"])
		      raise DMError("OBJECT_PIPE_IN_CONSTRAINT", {'objectName': data["objectName"], 'modelName': self.name})
                    if "[" in c["search"] and "]" in c["search"]: 
		      logger.exception("Object constraints cannot contain subsearches. Constraint search: " + c["search"])
		      raise DMError("OBJECT_SUBSEARCH_IN_CONSTRAINT", {'objectName': data["objectName"], 'modelName': self.name})
                    myInstance.filter(data["lineage"],c["search"])

                return myInstance

            return myFactory

        #Should not hit this if data model is well formed
        return None

    def JSONdesc(self, conciseMode):
        retval = self.shortJSONdesc()
        retval["objects"] = []

        #We're actually instantiating every object twice when we get the full model description.  This could definitely be changed to improve performance.
        #Might be worthwhile for a bubbles maintenence release? Wait and see if customers are hurting for performance
        #Alice noticed this on 9/23/13
        for o in self.objectNames:
            obj = self.getObject(o)
	    try:
              retval["objects"].append(obj.JSONdesc(conciseMode))
	    except DMError as DMErr:
	      params=DMErr.__json__()['parameters']
	      params.update({'objectName': obj.getName()})
              raise DMError(DMErr.__json__()['errorCode'], params)
	    except:
	      logger.exception("unknown error obtaining JSONdesc. Traceback: "+traceback.format_exc())	    
	      raise DMError("OBJECT_CANNOT_OBTAIN_JSON_DESCRIPTION", {'objectName': obj.getName(), 'modelName': self.name})	    
        return retval

    def shortJSONdesc(self):
	event = 0; transaction=0; search=0; interface=0; implements=0;
	for o in self.objectNames:
   	  obj = self.getObject(o)
	  if obj.getObjectType() == "BaseEvent": event+=1
	  elif obj.getObjectType() == "BaseTransaction": transaction+=1
	  elif obj.getObjectType() == "BaseSearch": search+=1
	  elif obj.getObjectType() == "BaseInterface": interface+=1
	  elif obj.getObjectType() == "BaseImplements": implements+=1
	
        return {"modelName": self.name, "displayName": self.displayName, "description": self.description, "objectNameList": self.objectNames, "objectSummary": {"Event-Based": event, "Transaction-Based": transaction, "Search-Based": search, "Interfaces": interface, "Interface Implementations": implements}}


    def objectDictForName(self, objectName):
        for o in self.rawData["objects"]:
            if o["objectName"] == objectName:
                return o
        return None


    #gets the factory for the parent of the object specified.
    #this factory may be parameterized by information contained in the current object's data (e.g. Transaction)
    def getParentFactory(self, data):
        name = data["parentName"]
        
        if ' ' in name: #special snowflake cervelli error: see  SPL-71488
            raise DMError('OBJECT_INVALIDNAME', {'objectName': objectData['objectName'], 'modelName': self.name})

        if name == "BaseEvent":
            def eventClosure():
                return BaseEvent()
            return eventClosure

        elif name == "BaseTransaction":
	    groupByFields=[]; transactionMaxTimeSpan=''; transactionMaxPause='';
            #look up groupByFields, transactionMaxTimeSpan, transactionMaxPause
            if "groupByFields" in data: groupByFields = data["groupByFields"]
            if "transactionMaxTimeSpan" in data: transactionMaxTimeSpan = data["transactionMaxTimeSpan"]
            if "transactionMaxPause" in data: transactionMaxPause = data["transactionMaxPause"]
	    if groupByFields==[] and transactionMaxTimeSpan=='' and transactionMaxPause=='':
		raise DMError("OBJECT_BASETRANSACTION_MISSING_GROUPBY_SPAN_PAUSE", {'objectName': data["objectName"], 'modelName': self.name})
            #look up base events
	    if "objectsToGroup" not in data or data["objectsToGroup"]==[]:
		raise DMError("OBJECT_BASETRANSACTION_MISSING_OBJECTSTOGROUP", {'objectName': data["objectName"], 'modelName': self.name})
            baseObjectNames = data["objectsToGroup"]
            objectsToGroup = []
	    groupByFieldsExist = [0 for i in range(len(groupByFields))]
            for name in baseObjectNames:
	      try: 
                obj = self.loadAndGetObject(name)
	      except DMError as DMErr:
	        params=DMErr.__json__()['parameters']
	        params.update({'objectName': data['objectName'], 'baseObjectName': name, 'modelName': self.name})
                raise DMError(DMErr.__json__()['errorCode'], params)
              if obj == None:
                logger.exception("Error loading transaction baseObject "+name + ". Traceback: "+traceback.format_exc())
		raise DMError("OBJECT_BASETRANSACTION_CANNOT_LOAD_BASEOBJECT", {'objectName': data["objectName"], 'baseObjectName': name, 'modelName': self.name})
              objectsToGroup.append(obj)
	      for field in groupByFields:
		if obj.hasField(field) or obj.hasCalculatedField(field): 
		  groupByFieldsExist[groupByFields.index(field)]=1
	    if 0 in groupByFieldsExist:
		raise DMError("OBJECT_BASETRANSACTION_INVALID_GROUPBYFIELD", {'objectName': data["objectName"], 'fieldName': groupByFields[groupByFieldsExist.index(0)], 'modelName': self.name})

            def transactionClosure():
                #return parameterized Transaction object
                return BaseTransaction(objectsToGroup, groupByFields, span=transactionMaxTimeSpan, pause=transactionMaxPause)

            return transactionClosure

        elif name == "BaseSearch":
	    if "baseSearch" not in data or data["baseSearch"]=="":
		raise DMError("OBJECT_BASESEARCH_MISSING_BASESEARCH", {'objectName': data["objectName"], 'modelName': self.name})
            baseSearch = data["baseSearch"]

            def searchClosure():
                return BaseSearch(baseSearch)

            return searchClosure

        elif name == "BaseInterface":
            def interfaceClosure():
                return BaseInterface()
            return interfaceClosure

        elif name == "BaseImplements":
	    if "implementsInterface" not in data or data["implementsInterface"]=="":
		raise DMError("OBJECT_BASEIMPLEMENTS_MISSING_IMPLEMENTSINTERFACE", {'objectName': data["objectName"], 'modelName': self.name})
	    implementsinterface = data["implementsInterface"]
	    try: 
                interfaceobj = self.loadAndGetObject(implementsinterface)
	    except DMError as DMErr:
	      params=DMErr.__json__()['parameters']
	      params.update({'objectName': data['objectName'], 'interfaceObjectName': implementsinterface})
              raise DMError(DMErr.__json__()['errorCode'], params)
            if interfaceobj==None:
                logger.exception("Error loading implementsInterface object "+ implementsinterface + ". Traceback: "+traceback.format_exc())
		raise DMError("OBJECT_BASEIMPLEMENTS_CANNOT_LOAD_IMPLEMENTSINTERFACE", {'interfaceObjectName': implementsinterface, 'objectName': data["objectName"], 'modelName': self.name})

            def implementsClosure():
                return BaseImplements(interfaceobj)

            return implementsClosure

        elif name == "BaseAggregate":
	    uniqueID=''; aggregates=[]
            if "uniqueID" in data: uniqueID = data["uniqueID"]
            if "aggregates" in data: aggregates = data["aggregates"]
	    if uniqueID=='':
		raise DMError("BaseAggregate requires a uniqueID field", {'objectName': data["objectName"], 'modelName': self.name})
            #look up base events
	    if "objectsToGroup" not in data or data["objectsToGroup"]==[]:
		raise DMError("BaseAggregate objects require at least one item in objectsToGroup", {'objectName': data["objectName"], 'modelName': self.name})
            baseObjectNames = data["objectsToGroup"]
            objectsToGroup = []
            for name in baseObjectNames:
	      try: 
                obj = self.loadAndGetObject(name)
	      except DMError as DMErr:
	        params=DMErr.__json__()['parameters']
	        params.update({'objectName': data['objectName'], 'baseObjectName': name, 'modelName': self.name})
                raise DMError(DMErr.__json__()['errorCode'], params)
              if obj==None:
                logger.exception("Error loading aggregate baseObject "+ name + ". Traceback: "+traceback.format_exc())
                raise DMError("Error loading aggregate baseObject ", {'objectName': data["objectName"], 'baseObjectName': name, 'modelName': self.name})
              objectsToGroup.append(obj)

            def aggregateClosure():
                #return parameterized Aggregate object
                return BaseAggregate(objectsToGroup, uniqueID, aggregates)

            return aggregateClosure

        elif self.validObject(name):
            self.loadObject(self.objectDictForName(name))
            return self.objectFactories[name]

        logger.exception("Cannot create parent object. Traceback: "+traceback.format_exc())
        raise DMError("OBJECT_CANNOT_CREATE_PARENT", {'objectName': name, 'modelName': self.name})


    #FIXME: consolidate these and make them return the child list as part of the object

    #returns an instance of the named object, if it exists in the model
    def loadAndGetObject(self, name):
        #check that object is valid
        if self.validObject(name):
            #make sure we've loaded it already
            self.loadObject(self.objectDictForName(name))
            return self.getObject(name)
        return None

    def getObject(self, name):
        if(name in self.objectNames):
            obj = self.objectFactories[name]()

	    if obj.getObjectType()=="BaseInterface":
              if name in self.implementsLists:
                for implements in self.implementsLists[name]:
		  implObj = self.getObject(implements)
    		  obj.addImplementsObject(implObj)
	      elif obj.getParentName()!="BaseInterface": # might be a descendant of an interface:
		rootInterfaceObj = obj.getLineage().split(".")[0]
                if rootInterfaceObj in self.implementsLists:
                  for implements in self.implementsLists[rootInterfaceObj]:
		    implObj = self.getObject(implements)
    		    obj.addImplementsObject(implObj)
		  self.implementsLists[name] = self.implementsLists[rootInterfaceObj]
	        else: 
		  raise DMError("Interface object doesn't have any implementations", {'objectName': name})

            if name in self.childLists:
                for child in self.childLists[name]:
                    childObj = self.getObject(child)
                    obj.addChildObject(childObj)

	    obj.addField(owner=obj.getLineage(), fieldName=obj.getName(), displayName=obj.getDisplayName(), fieldType="objectCount", required=False, multivalued=False, editable=False)

	else:
	    raise DMError("MODEL_NONEXISTENT_OBJECT", {'objectName': name})
        return obj

    def update(self, modelData):
        
       #check displayName
        if 'displayName' in modelData:
            self.displayName = modelData['displayName']

        #check description
        if 'description' in modelData:
            self.description = modelData['description']

        #check objects
        if 'objects' in modelData:
            self.rawData['objects'] = modelData['objects'] 
	    self.loadObjectsFromList(modelData['objects'])

	else: 
            self.rawData['objects'] = [] 
	    self.childLists = {}
            self.implementsLists = {}
            self.objectFactories = {} # a dict of constructor for objects in the model
            self.objectNames = [] 

'''      
class PythonModel(Model):

    def __init__(self, modelName, displayName, description, objects, tstats='', accelerate=False, accelerationWindow=''):
        self.name = modelName
        self.displayName = displayName
        self.description = description
        self.objectNames = []
        self.objectFactories = {}
        self.childLists = {}
        self.implementsLists = {}
	self.tsidxNamespace = tstats
	self.accelerate = accelerate
	self.accelerationWindow = accelerationWindow

        for o in objects:
            self.objectNames.append(o.__name__)
            self.objectFactories[o.__name__] = o
            obj = o()
            parent = obj.getParentName()
            if not parent in self.childLists:
                self.childLists[parent] = [obj.getName()]
            else:
                self.childLists[parent].append(obj.getName())
	    if parent == "BaseImplements":
	      implementsinterface = obj.getInterfaceName()
              if not implementsinterface in self.implementsLists:
                self.implementsLists[implementsinterface] = [obj.getName()]
              else:
                self.implementsLists[implementsinterface].append(obj.getName())
       
        
        
    def shortJSONdesc(self):
        retval = Model.shortJSONdesc(self)
        return retval
'''
