import glob
import logging, logging.handlers
import math
import os
import sys
import tempfile
import time
import string
import itertools

import splunk
import splunk.auth as auth
import splunk.search as search
import splunk.util as util

from errormsgs import PivotError

SPLUNK_HOME = os.environ['SPLUNK_HOME']
logger = logging.getLogger("splunk")

def valaggformat(): # FIXME: move to util
  # note: default for xyseries is format=$VAL$:$AGG$. default for timechart is $AGG$:$VAL$
  return 'format=$VAL$:::$AGG$'
  # FIXME: option to return 'format=$AGG$:::$VAL$'

#Object corresponding a stats command.  May be wrapped by a statsCalculation.
class Stats: 
    
    def __init__(self, funcandfieldlist, pivotfields=[], rowSummary=False, colSummary='none', colLimit=0, colOther=True, replaceBoolLabels='', sortsearch=''):
	if funcandfieldlist==[]:
	  raise PivotError("Stats::__init__", "must have at least one stats function ")
	self.funcandfieldlist = funcandfieldlist
	self.pivotfields = pivotfields
	self.outputfields = []
	self.fillnullfields = '' 
	self.rowSummary = rowSummary
	self.colSummary = colSummary  # true = 'show', false = 'none', calculate but don't show = 'hide'
	self.colLimit = colLimit
	self.colOther = colOther
	self.replaceBoolLabels = replaceBoolLabels
	self.sortSearch = sortsearch

        for funcfield in self.funcandfieldlist:
	  #iterate through and assign list of outputfields for the stats calculation
	    if funcfield["function"]=="":
	  	raise PivotError("Stats::__init__", "must specify a stats function ")
		
            if "rename" in funcfield and funcfield["rename"] != '':
                outputfield = '"'+funcfield["rename"]+'"'
            else:
		if funcfield["function"]=="count" and funcfield["field"]=='':
		  # the front end guards against this case but this is just backup 
		  outputfield = '"count"'
		else:
                  outputfield = '"'+funcfield["function"]+'('+funcfield["field"]+')"'
            self.outputfields = self.outputfields + [outputfield]
	    if funcfield["function"]=="count" or funcfield["function"]=="dc":
	      self.fillnullfields += ' ' + outputfield

    # This function returns the list of stats functions being returned
    def getOutputFields(self):
	return self.outputfields

    # This functino returns the list of row field names or col field names based on dimension specified
    def getRowColList(self, dimension):
	if dimension != 'col' and dimension != 'row': 
	  PivotError("Stats::getRowColList", "Must specify either row or column for a field's dimension")
	retval = [] 
	for x in self.pivotfields:
	  if 'roworcol' in x and x['roworcol']==dimension:
	    retval.append('"'+x['field'].strip('"').strip("'")+'"')
	return retval

    # JSON description for Stats 
    def JSONDesc(self):
        return {"functionFieldList": self.funcandfieldlist, "pivotFields": self.pivotfields}
	
    # This function returns the tstats search
    # | tstats <statsfn> from <namespace> groupby <pivotfields> where <filters> | stats/chart/timechart ...
    def tstatsSearch(self, tstatsNamespace, filters):

        searchstring = "| tstats " 
	searchstring += self.getStatsFuncString(tstatsMode=True)

	searchstring += ' from ' + tstatsNamespace

	# add 'WHERE' clause here with filters
	if filters != []:
	  searchstring += ' where '  
	  for filter in filters:
            if filter["search"].startswith('regex') == False:
              searchstring +=  " (" + filter["search"] + ") "


	if self.pivotfields!=[]:
	    searchstring += " groupby "
	searchstring += self.getRowColString()
        #only one span should be used...lowest common denominator
	if self.getMinSpan("_time")!='':
	  searchstring += 'span=' + self.getMinSpan("_time")
	      # FIXME:  need ability to add spans on fields other than _time, once tsidx supports it
	      # FIXME:  need to handle multiple spans on _time...not sure how this will happen
	      # FIXME: need to pass along _span so UI can use it...
	searchstring += ' prestats=true'


	# if only one splitby row and it's _time, use timechart
	if self.getRowColString(dimension='row', delim='+ " ::: " +', quotes="'").strip('"').strip("'")=="_time":
	  timefield = (item for item in self.pivotfields if item["field"].strip('"').strip("'") == "_time").next() 
	  if 'span' in timefield and timefield['span']!='': 
	    searchstring += ' span=' +timefield['span']  + ' '
	  else: 
	    searchstring += ' span=auto '

	searchstring += self.search(tstatsMode=True)
	return searchstring

    # This function returns the appropriately formatted string for stats functions
    # count AS <somelabel>, dc(somefield) etc
    def getStatsFuncString(self, tstatsMode=False):
	  statsstring = ''
	  for funcfield in self.funcandfieldlist:
            statsfunc = funcfield["function"]
            xfield = funcfield["field"].strip('"').strip("'")
	    if 'rename' in funcfield:
              rename = funcfield["rename"]
	    else: rename = ''

	    if statsfunc == "count" and xfield == '':
	        statsstring += statsfunc + " "
	    elif statsfunc == "count" and xfield[0:5] == 'eval(':
	        statsstring += statsfunc + "(" + xfield + ") " # no quotes around eval
	    elif statsfunc == "list" and tstatsMode==True:
                statsstring += '"'+xfield + '" '
	    else: 
	        statsstring += statsfunc + "("
                statsstring += '"'+xfield + '") '
	    if rename != '':
		statsstring += 'AS "' + rename + '" '
	  return statsstring


    # This function returns a string containing all row/col field names with optional delimeter/quotes
    def getRowColString(self, dimension='all', delim=' ', quotes='"'):
	retval = '' 
	for x in self.pivotfields:
	  if dimension=='all' or ('roworcol' in x and x['roworcol']==dimension):
	    if retval != '': retval += delim 
	    retval+=quotes+x['field'].strip('"').strip("'")+quotes
	return retval


    # This function returns the smallest discretization required for _time
    def getMinSpan(self, fieldname):
	spans = []
	for pivotfield in self.pivotfields:
	  if pivotfield["field"] == fieldname and 'span' in pivotfield:
	    spans.append(pivotfield['span'])
	if '1s' in spans: return '1s'	
	if '1m' in spans: return '1m'	
	if '1h' in spans: return '1h'	
	if '1d' in spans: return '1d'	
	if '7d' in spans: return '7d' 	
	if '1mon' in spans: return '1mon'	
	if '12mon' in spans: return '12mon'	
	# FIXME: need to add options for other spans...
	return ''

    # This function returns the appropriate bucketing string to discretize fields
    def getBucketString(self,pivotfield):
	bucketstring=''
  	if ('span' in pivotfield and pivotfield["span"]!=''):
  	    bucketstring +=' span=' +pivotfield['span']  + ' '
  	if ('bins' in pivotfield and pivotfield["bins"]!=''):
  	    bucketstring += ' bins=' +pivotfield['bins'] 
  	    if ('start' in pivotfield and pivotfield["start"]!=''):
  	      bucketstring += ' start=' +pivotfield['start'] 
  	    if ('end' in pivotfield and pivotfield["end"]!=''):
  	      bucketstring += ' end=' +pivotfield['end'] 
	
	return bucketstring

    # This function returns the appropriate bucketing string to discretize fields
    def getStartEndFilter(self,pivotfield):
	bucketstring=''
  	if ('bins' not in pivotfield or ('bins' in pivotfield and pivotfield["bins"]=='')):
  	    if ('start' in pivotfield and pivotfield["start"]!=''):
  	      bucketstring += ' "' + pivotfield['field'].strip("'").strip('"') + '">=' +pivotfield['start'] 
  	    if ('end' in pivotfield and pivotfield["end"]!=''):
  	      bucketstring += ' "' + pivotfield['field'].strip("'").strip('"') + '"<' +pivotfield['end'] 
	
	return bucketstring

    # This function returns the pivot search string using | stats
    def getStatsString(self):
	statsstring=''
	for pivotfield in self.pivotfields:
	  if self.getBucketString(pivotfield)!='':
	    statsstring += ' | bucket "' + pivotfield["field"].strip("'").strip('"') + '" ' + self.getBucketString(pivotfield)
	  if self.getStartEndFilter(pivotfield)!='':
	    statsstring += ' | search ' + self.getStartEndFilter(pivotfield)
	
	rowFields = self.getRowColList('row')	

	if self.rowSummary and rowFields!=[]:
    	  statsstring += ' | multireport ['
	else: statsstring += ' |'


	corestatsstring = ' stats '
	corestatsstring += self.getStatsFuncString()
	statsstring += corestatsstring
	if self.pivotfields!=[]:
	  statsstring += "by "
	statsstring += self.getRowColString() 

	statsstring += self.sortSearch
 
	if self.rowSummary and rowFields!=[]:
    	  statsstring += ']'
	  statsstring += ' [' + corestatsstring 
          for row in rowFields: # each field in row should have this set
              statsstring += ' | eval '+ row + ' = "~~~"'
          statsstring += "]"
	
	statsstring += " | fields " + ', '.join(rowFields) + ', ' + ', '.join(self.outputfields)
	return statsstring 

    # this function returns the pivot search string using | timechart
    def getTimechartString(self, tstatsMode=False):
	statsstring=''
	splitbycollabel=''
	timefield = (item for item in self.pivotfields if item["field"].strip('"').strip("'") == "_time").next() 
	for pivotfield in self.pivotfields:
	  if pivotfield["field"].strip('"').strip("'") != '_time' and self.getBucketString(pivotfield)!='':
	    statsstring += ' | bucket "' + pivotfield["field"].strip("'").strip('"') + '" ' + self.getBucketString(pivotfield)
	  if pivotfield["field"].strip('"').strip("'") != '_time' and self.getStartEndFilter(pivotfield)!='':
	    statsstring += ' | search ' + self.getStartEndFilter(pivotfield)
	  if pivotfield["roworcol"]=='col':
	    if splitbycollabel!='': splitbycollabel += ' ::: '
	    splitbycollabel += pivotfield["label"]
	splitbycols = self.getRowColString(dimension='col', delim='+ " ::: " +', quotes="'")
	if splitbycols != '': 
	    statsstring += ' | eval \"' + splitbycollabel +'\"=' + splitbycols
	if self.colSummary != 'none':
	    statsstring += ' | eval \"' + splitbycollabel + '\"=mvappend(\'' + splitbycollabel + '\',"~~~")'
	    if tstatsMode:
	      statsstring +=' | mvexpand "'+splitbycollabel + '"'
	if self.rowSummary:
    	  statsstring += ' | multireport ['
	else: statsstring += ' |'

        statsstring += " timechart " 
	if self.colLimit!=0: 
	  statsstring += "limit=" + str(self.colLimit) + ' '
	if splitbycols != '':
	  if self.colOther == True:
	    statsstring += "useother=t "
	  else: 
	    statsstring += "useother=f "
	statsstring += self.getBucketString(timefield) 
	
 	statsstring += self.getStatsFuncString()
	if splitbycols != '': 
	    statsstring+= ' by "' + splitbycollabel + '"'
	statsstring += ' ' + valaggformat()

	statsstring += self.sortSearch

	if self.rowSummary:  
    	  statsstring += ']'
	  statsstring += ' [ eval _time="~~~" | chart ' 
	  if self.colLimit!=0: 
	    statsstring += "limit=" + str(self.colLimit) + ' '
	  if self.colOther == True:
	    statsstring += "useother=t "
	  else: 
	    statsstring += "useother=f "
	  statsstring += self.getStatsFuncString() + ' by _time'
	  if splitbycols != '': 
	    statsstring+= ' "' + splitbycollabel + '"'
	  statsstring += ' ' + valaggformat()
          statsstring += "]"

	statsstring += " | fields _time *"

	return statsstring 

    # This function generates the "pivot" part of the search string. 
    # we choose between "stats", "timechart" and "chart" and format the fields
    # "chart" is where we handle row vs column splitting. 
    # --we create "newX" as an eval of all row splitbys
    # --we create  "newY" as an eval of all column splitbys. 
    # --then we | chart
    # field renaming is done here if there's a separate label/displayName
    def search(self, tstatsMode=False):

	statsstring = ''
	rowstring=''	
	columnstring = ''
	xLabelList = []
	renamestringRows = ''
	replaceRowAllGeneric = ''
	replaceRowAllSpecific = ''
	renameColumnAll = ''
	bucketString = ''
	newX = ''
	newY = ''
	
	for x in self.pivotfields:
	  if 'roworcol' in x:
	    x['field'] = x['field'].strip('"').strip("'")
  	    if x['roworcol']=='row':	
	      if 'label' not in x: 
		x['label']=x['field']
	      if x['field'] != x['label']: 
		renamestringRows+= ' "'+x['field']+'" AS "'+x['label']+'"'

  	      if rowstring != '': 
		newX += ' ::: '
		rowstring += '+" ::: "+'
  	      rowstring += "'"+x['field']+"'"
	      newX += x['label']

	      xLabelList.append(x['label'])

	      if self.rowSummary: # FIXME: move at the end?
	        replaceRowAllSpecific +=  '"'+x['label']+'" ' #using ~~~ for rowsummaries

  	    elif x['roworcol']=='col':
  	      if columnstring != '': 
		columnstring += '+" ::: "+'
		newY += ' ::: '
  	      columnstring += "'"+x['field']+"'"
  	      newY += x['label']
	      if self.colSummary == 'show': 
	        renameColumnAll += ' | rename *~~~* AS *ALL*' #using ~~~ for columnsummaries 

	  if self.getBucketString(x)!='':
	    bucketString += ' | bucket "' + x["field"].strip("'").strip('"') + '" ' + self.getBucketString(x)
	  if self.getStartEndFilter(x)!='':
	    statsstring += ' | search ' + self.getStartEndFilter(x)

	if rowstring != '':
	  rowstring = " | eval \"" + newX + "\"=" + rowstring

	if columnstring != '':
	  columnstring = " | eval \"" + newY + "\"=" + columnstring

	statsstring += self.replaceBoolLabels
	# timechart special case when only one splitby row and it is _time
	# check if _time is the only row and splitby all the columns for timechart
	if self.getRowColString(dimension='row', delim='+ " ::: " +', quotes="'").strip('"').strip("'")=="_time":
	    statsstring+= self.getTimechartString(tstatsMode)

	#regular stats if columnstring is empty
	elif columnstring=='':
	    statsstring+= self.getStatsString()
	    if self.fillnullfields != '':
	      statsstring += "| fillnull"+self.fillnullfields
	    if renamestringRows != '': 
	      statsstring += "| rename" + renamestringRows

	elif columnstring!='': 
	      if rowstring=='': 
	        newXFields = ' | fields - newX'  #remove dummy field at the end
	        rowstring = ' | eval newX = " "' #dummy newX so we can use chart
		newX = 'newX'
	      elif len(xLabelList)>1:
		fieldsOrder = ' | fields - "' + newX + '" | fields'
		newXFields = ' | makemv "' + newX + '" delim=" ::: "'; idx=0
		for x in xLabelList:
		  newXFields += ' | eval "' + x + '"= mvindex(\'' + newX + '\',' + str(idx) + ')'
		  idx = idx+1
		  fieldsOrder += ' "' +x +'"'
		newXFields += fieldsOrder + ' *'
	      else:
		newXFields = ' | fields "' + newX + '" *'

	      statsstring += bucketString + rowstring + columnstring 
	      if self.colSummary != 'none':
		statsstring += ' | eval "' + newY + '"=mvappend(\'' + newY + '\',"~~~")'
		if tstatsMode:
		  statsstring += ' | mvexpand "' + newY + '"'
	      if self.rowSummary and self.getRowColList('row')!=[]: 
	        statsstring += ' | multireport ['
	      else: statsstring += ' |'
	      corestatsstring = ' chart '
	      if self.colLimit!=0: 
	        corestatsstring += "limit=" + str(self.colLimit) + ' '
	      if self.colOther == True:
	  	corestatsstring += "useother=t "
	      else: 
	  	corestatsstring += "useother=f "
	      corestatsstring += self.getStatsFuncString()
	      corestatsstring += 'by "' + newX + '" "' + newY + '" ' + valaggformat()
	
	      statsstring += corestatsstring

	      statsstring += self.sortSearch

	      # Row Summaries
	      rowAllStr=''
              for row in self.getRowColList('row'): # each field in row should have this set
		if rowAllStr!='': 
		  rowAllStr += " ::: "
                rowAllStr += "~~~" 
	      if self.rowSummary and self.getRowColList('row')!=[]:  
	        statsstring += '] [eval ' + newX + ' = "' + rowAllStr + '" | ' 
	  	statsstring += corestatsstring 
            	statsstring += "]"

	      statsstring += newXFields

	if replaceRowAllSpecific!='': # FIXME: make this if self.rowSummary ...
          statsstring += "| replace *~~~* with *ALL* in "+replaceRowAllSpecific
              # internal fields like _time must be explicitly stated in 'replace'

	fieldsstring =''
	for f in xLabelList:
	  fieldsstring += ' "' + f + '"'

          
        if columnstring == '' and len(self.outputfields) > 1:
          fieldsstring += " " +' '.join(self.outputfields)

	if columnstring!='' or xLabelList!=[]:
	  fieldsstring += ' *' 
	if renameColumnAll!='':
	  if len(self.outputfields)>1:
	    fieldsstring += ' ' + ', '.join(['"ALL:::'+i.strip('"')+'"' for i in self.outputfields])
	  else:
	    fieldsstring += ' ALL'
	if fieldsstring!='':
	  fieldsstring = " | fields" + fieldsstring

	statsstring += renameColumnAll  + fieldsstring

	return statsstring


