import splunk.util
from panelElement import *
from row import Row
from dashboard import SimpleDashboard
import forminput
from base import *
import json
import logging

logger = logging.getLogger('splunk.models.view_escaping.fromdash')

def createDashboardFromXml(dashboardLxmlNode, viewName=None, digest=None, sourceApp=None):
    '''

    @param dashboardLxmlNode: lxml node representing a dashboard or form
    @param viewName: the name of the xml file
    @param digest: boolean
    @return:
    '''

    logger.info("Parsing view: simplexml=%s" % viewName)
    dashboard = SimpleDashboard(viewName=viewName, digest=digest, sourceApp=sourceApp)

    # Parse global options
    dashboard.label = dashboardLxmlNode.findtext('./label')
    dashboard.description = dashboardLxmlNode.findtext('./description')

    dashboard.customScript = dashboardLxmlNode.get('script', None)
    dashboard.customStylesheet = dashboardLxmlNode.get('stylesheet', None)
    dashboard.onUnloadCancelJobs = normalizeBoolean(dashboardLxmlNode.get('onunloadCancelJobs', dashboard.onUnloadCancelJobs))
    # master search template
    dashboard.searchTemplate = dashboardLxmlNode.findtext('./searchTemplate')
    dashboard.searchEarliestTime = dashboardLxmlNode.findtext('./earliestTime')
    dashboard.searchLatestTime = dashboardLxmlNode.findtext('./latestTime')
    dashboard.statusBuckets = 0

    dashboard.matchTagName = dashboardLxmlNode.tag
    if dashboard.matchTagName not in ['form', 'dashboard']:
        raise Exception(_('Invalid root element'))

    if dashboardLxmlNode.tag == 'form':
        # update the fieldset
        fieldsetNode = dashboardLxmlNode.find('./fieldset')
        if fieldsetNode is not None:
            dashboard.autoRun = splunk.util.normalizeBoolean(
                fieldsetNode.get('autoRun', False))
            dashboard.submitButton = splunk.util.normalizeBoolean(
                fieldsetNode.get('submitButton', True))

            for item in fieldsetNode:
                logger.debug("Found new element in fieldsetn: form=%s item=%s" % (viewName, item.tag))
                if item.tag == 'html':
                    panelInstance = createPanelElementFromXml(item)
                    dashboard.fieldset.append(panelInstance)
                elif item.tag == 'input':
                    inputInstance = forminput.createInput(
                        item.get('type'), dict(searchWhenChanged=not dashboard.submitButton))
                    inputInstance.fromXml(item, sourceApp)
                    dashboard.fieldset.append(inputInstance)

    # set core view attributes
    for k in dashboard.standardAttributeMap:
        v = dashboardLxmlNode.get(k)
        if v is not None:
            if k in dashboard.booleanAttributeKeys:
                v = splunk.util.normalizeBoolean(v)
            elif k in dashboard.integerAttributeKeys:
                try:
                    v = int(v)
                except:
                    msg = "Key, %s, should have an integer value. Value found, %s" % (k,v)
                    logger.error(msg)
                    raise Exception(msg)
            setattr(dashboard, k, v)

    for rowLxmlNode in dashboardLxmlNode.findall('row'):
        logger.debug("Appending new row to view=%s" % viewName)
        dashboard.rows.append(createRowFromXml(rowLxmlNode))

    for row in dashboard.rows:
        for panel in row.panels:
            for panelElement in panel.panelElements:
                if hasattr(panelElement, 'noSearch') and panelElement.noSearch and dashboard.searchTemplate:
                    panelElement.searchMode = None
                    panelElement.searchCommand = None
                    
                    # If this panel is an event, and it is tied to the global search template, then
                    # we need to tell that template to have status buckets      
                    if isinstance(panelElement, Event):
                        dashboard.statusBuckets = 300
                
                if isinstance(panelElement, Event):
                    panelElement.statusBuckets = 300
                    
                if hasattr(panelElement, 'searchMode') and panelElement.searchMode and panelElement.searchMode == POST_SEARCH_MODE:
                    dashboard.statusBuckets = 300
                        
                    

    normalizeIdentifiers(dashboard)
    return dashboard


def normalizeIdentifiers(dashboard):
    '''
    Ensure all dashboard elements have an unique id

    @param dashboard: Dashboard object
    @return: no return value
    '''
    seen = set()
    for el in dashboard.all_elements():
        if el.id is not None:
            if el.id in seen:
                raise Exception(_('Duplicate panel element ID "%s"') % el.id)
            seen.add(el.id)
    for el in dashboard.fieldset:
        if el.id is not None:
            if el.id in seen:
                raise Exception(_('Duplicate panel element ID "%s"') % el.id)
            seen.add(el.id)
    panelIndex = 1
    ctxId = 1
    fieldId = 1
    globalSearch = None
    if getattr(dashboard, 'searchTemplate', None) is not None:
        dashboard.context = globalSearch = "search%d" % ctxId
        ctxId += 1

    for el in dashboard.all_elements():
        if el.id is None:
            while True:
                el.id = 'element%d' % panelIndex
                panelIndex += 1
                if not el.id in seen:
                    break
        if el.context is None:
            if getattr(el,'searchMode', None):
                el.context = "search%d" % ctxId
                ctxId += 1
            elif globalSearch is not None:
                el.context = globalSearch
    for field in dashboard.fieldset:
        if field.id is None:
            while True:
                field.id = 'field%d' % fieldId
                fieldId += 1
                if not field.id in seen:
                    break
        if getattr(field, 'context', None) is None:
            if getattr(field, 'search', None) or getattr(field, 'savedSearch', None):
                field.context = "search%d" % ctxId
                ctxId += 1


def createRowFromXml(rowLxmlNode):
    '''
    Parses a row xml node
    @param rowLxmlNode: Lxml representing a form or dashboard element
    @return:
    '''
    logger.debug("parsing dashboard row node")
    if rowLxmlNode.get('grouping'):
        rowGroupings = map(
            int,
            rowLxmlNode.get('grouping').replace(' ', '').strip(',').split(','))
        logger.debug("Found row grouping=%s" % rowGroupings)
    else:
        rowGroupings = None

    row = Row(rowGroupings)

    for panelElementLxmlNode in rowLxmlNode:
        panel = createPanelElementFromXml(panelElementLxmlNode)
        if panel:
            row.appendPanelElement(panel)
    return row


def createPanelElementFromXml(panelElementLxmlNode):

    def createHtmlPanelElementFromXml(panelElementLxmlNode, panelInstance):
        '''
        Override default parser and just get the text
        '''
        logger.debug("parsing html panel element")
        src = panelElementLxmlNode.get('src')
        if src:
            panelInstance.options['serverSideInclude'] = src
            return

        flatString = et.tostring(panelElementLxmlNode)
        flatString = flatString.replace("<html>", "").replace("</html>", "").strip()
        panelInstance.options['rawcontent'] = flatString

    def createTablePanelElementFromXml(panelElementLxmlNode, panelInstance):
        """
        Add a format tag to provide cell formatting
        (primarily for sparklines at the current time)
        Each format tag may have some option sub-tags
        Each option may contain list or option tags generating
        lists or dictionaries respectively

        Each format tag may specify a field to filter on
        (defaults to all fields) and a type to apply (eg 'sparkline')
        """
        logger.debug("parsing table panel element")

        def parseOption(node):
            """Extract nested options/lists"""
            listNodes = node.findall('list')
            optionNodes = node.findall('option')

            if listNodes and optionNodes:
                raise ValueError("Tag cannot contain both list and option subtags")
            elif listNodes:
                result = []
                for listNode in listNodes:
                    result.append(parseOption(listNode))
                return result
            elif optionNodes:
                result = {}
                for optionNode in optionNodes:
                    result[optionNode.get('name')] = parseOption(optionNode)
                return result
            else:
                return node.text

        createDefaultPanelElementFromXml(panelElementLxmlNode, panelInstance)
        fieldFormats = {}
        for formatNode in panelElementLxmlNode.findall('format'):
            logger.debug("Parsing format view node")
            field = formatNode.get('field', '*')
            formatType = formatNode.get('type', 'text')
            options = {}
            for optionNode in formatNode.findall('option'):
                options[optionNode.get('name')] = parseOption(optionNode)
            fieldFormats.setdefault(field, []).append({
                'type': formatType,
                'options': options
            })
        panelInstance.fieldFormats = fieldFormats

    def createDefaultPanelElementFromXml(panelElementLxmlNode, panelInstance):
        logger.debug("Parsing default panel element options")
        # define common XML node -> object property mappings
        commonNodeMap = [
            ('title', 'title'),
            ('earliestTime', 'searchEarliestTime'),
            ('latestTime', 'searchLatestTime')
        ]

        # define search mode XML node -> object property mappings
        searchModeNodeMap = [
            ('searchString', TEMPLATE_SEARCH_MODE),
            ('searchName', SAVED_SEARCH_MODE),
            ('searchTemplate', TEMPLATE_SEARCH_MODE),
            ('searchPostProcess', POST_SEARCH_MODE)
        ]
        for nodeName, memberName in commonNodeMap:
            val = panelElementLxmlNode.findtext(nodeName)
            if val:
                setattr(panelInstance, memberName, val)

        optionTypeMap = {}
        if hasattr(panelInstance.__class__, 'optionTypeMap'):
            optionTypeMap = getattr(panelInstance.__class__, 'optionTypeMap')

        # option params get their own container
        for node in panelElementLxmlNode.findall('option'):
            optionName = node.get('name')
            optionValue = node.text
            if optionName in optionTypeMap:
                optionValue = optionTypeMap[optionName](optionValue)
            if isinstance(optionValue, str) and optionValue != None:
                optionValue = optionValue.strip()
            panelInstance.options[optionName] = optionValue

        # handle different search modes
        if getattr(panelInstance.__class__, 'hasSearch'):
            foundSearch = False
            for pair in searchModeNodeMap:
                if panelElementLxmlNode.find(pair[0]) is not None:
                    foundSearch = True
                    panelInstance.searchMode = pair[1]
                    panelInstance.searchCommand = (
                        panelElementLxmlNode.findtext(pair[0])).replace("\n", " ")
                    break
            if not foundSearch:
                panelInstance.searchMode = TEMPLATE_SEARCH_MODE
                panelInstance.searchCommand = ''
                panelInstance.noSearch = True

        # handle field lists
        if panelElementLxmlNode.find('fields') is not None:
            fields = panelElementLxmlNode.findtext('fields').strip()
            if len(fields) and fields[0] == '[' and fields[-1] == ']':
                panelInstance.searchFieldList = json.loads(fields)
            else:
                panelInstance.searchFieldList = splunk.util.stringToFieldList(fields)

        # extract simple XML drilldown params
        drilldownNode = panelElementLxmlNode.find('drilldown')
        panelInstance.hasDrilldownTag = (drilldownNode is not None)
        panelInstance.drilldownTarget = drilldownNode.get('target') if (drilldownNode is not None) else None
        for linkNode in panelElementLxmlNode.findall('drilldown/link'):
            if linkNode.text and len(linkNode.text) == 0:
                continue

            field = linkNode.attrib.get('field')
            series = linkNode.attrib.get('series')
            if not field and series:
                field = series
            if field:
                if len(field) == 0:
                    raise AttributeError(
                        'Unable to process drilldown because field attribute'
                        ' is missing: %s' % et.tostring(node))
            else:
                field = '*'

            logger.debug("Found drill down link field=%s link=%s" % (field, linkNode.text))
            panelInstance.simpleDrilldown[field] = linkNode.text.strip().replace("\n", " ")

        # extract the contents of all top-level comment nodes
        for node in panelElementLxmlNode.xpath('./comment()'):
            panelInstance.comments.append(node.text)

        # extract the comments from inside the drilldown tag
        for node in panelElementLxmlNode.xpath('./drilldown/comment()'):
            panelInstance.drilldownComments.append(node.text)

    def createPanel(name):
        '''
        Factory method for creating an appropriate panel object based upon the
        name.  Returns an instance of a BasePanel subclass, or throws a
        NotImplementedError if no suitable mapper is found.

        This method works by inspecting all objects that sublcass BasePanel and
        attempting to match their matchTagName class attribute.
        '''

        if not name:
            raise ValueError('Cannot create panel from nothing')

        for key, obj in globals().items():
            try:
                if issubclass(obj, BasePanel) and name == obj.matchTagName:
                    #  only Chart objects need to be instantiated
                    if obj is Chart:
                        return Chart()
                    else:
                        return obj()
            except:
                pass
        raise NotImplementedError(
            'Cannot find object mapper for panel type: %s' % name)

    if not isinstance(panelElementLxmlNode.tag, basestring):
        return False
    panelType = panelElementLxmlNode.tag
    try:
        panelInstance = createPanel(panelType)
    except NotImplementedError:
        return False

    id = panelElementLxmlNode.attrib.get('id')
    if id is not None:
        panelInstance.id = id

    logger.debug("found panel element type=%s" % panelType)
    if panelType == 'table':
        createTablePanelElementFromXml(panelElementLxmlNode, panelInstance)
    elif panelType == 'html':
        createHtmlPanelElementFromXml(
            panelElementLxmlNode, panelInstance)
    else:
        createDefaultPanelElementFromXml(panelElementLxmlNode, panelInstance)
    return panelInstance


if __name__ == '__main__':
    import unittest
    import lxml.etree as et

    def getRowXml(args='', panels=1):
        nodes = []
        nodes.append('<row %(args)s>')
        for i in range(0, panels):
            nodes.append('<single></single>')
        nodes.append('</row>')
        xml = ''.join(nodes)
        return xml % {'args': args}

    def getPanelElementXml(type="foo", options=None, args=None):
        options = options or (
            '<searchString> | metadata type="sources" | '
            'stats count</searchString>')
        args = args or ''
        xml = '''
            <%(type)s %(args)s>
                %(options)s
            </%(type)s>
        '''
        return xml % {'type': type, 'options': options, 'args': args}

    class CreatePanelElementTests(unittest.TestCase):

        def createPanel(self, type="foo", options=None, args=None):
            xml = getPanelElementXml(type, options, args)
            root = et.fromstring(xml)
            return createPanelElementFromXml(root)

        def testCreateUnknownPanel(self):
            d = self.createPanel('foo')
            self.assertFalse(d)

        def testCreateAllowedPanels(self):
            for type in ['single', 'chart', 'table',
                         'html', 'map', 'event', 'list']:
                d = self.createPanel(type)
                self.assertTrue(d.matchTagName == type)

        def testCreateHTMLPanel(self):
            d = self.createPanel(
                'html', args='src="/foo/bar"')
            self.assertTrue(d.options.get('serverSideInclude') == '/foo/bar')
            self.assertFalse(d.options.get('rawcontent'))
            d = self.createPanel(
                'html', options='<div>Test</div>', args='src="/foo/bar"')
            self.assertTrue(d.options.get('serverSideInclude') == '/foo/bar')
            self.assertFalse(d.options.get('rawcontent'))
            d = self.createPanel(
                'html', options='<div>Test</div>')
            self.assertFalse(d.options.get('serverSideInclude'))
            self.assertTrue(d.options.get('rawcontent') == '<div>Test</div>')

        def testCreateTablePanel(self):
            '''tables only have special format fields'''
            d = self.createPanel('table', options='''
                <format>
                    <option name="fff">
                        <list>foo</list>
                        <list>bop</list>
                    </option>
                    <option name="bippity">
                        <option name="bippity">bop</option>
                    </option>
                </format>
                <format field="foobar" type="num">
                    <option name="fff">
                        <list>foo</list>
                        <list>bop</list>
                    </option>
                    <option>bar</option>
                </format>
                ''')
            self.assertTrue(d.fieldFormats == {
                '*': [{'type': 'text',
                       'options': {'bippity': {'bippity': 'bop'},
                                   'fff': ['foo', 'bop']
                                   }
                       }],
                'foobar': [{'type': 'num',
                            'options': {None: 'bar',
                                        'fff': ['foo', 'bop']
                                        }
                            }]
            })
            #  formats can't mix options and lists.
            #  lists can contain options and options can contain list but
            #     neither can contain both.
            with self.assertRaises(ValueError):
                d = self.createPanel(
                    'table',
                    options='''
                        <format>
                            <option name="fff">
                                <list>foo</list>
                                <list>bop</list>
                                <option name="should">not work</option>
                            </option>
                            <option>bar</option>
                        </format>
                    ''')

        def testCreateDefaultPanel(self):
            '''All panels have some things in common'''
            #s test common nodes
            d = self.createPanel('single', options='''
                <title>Title1</title>
                <earliestTime>0</earliestTime>
                <latestTime>50</latestTime>
                <fields>foo bar baz</fields>
                ''')
            self.assertEqual(d.searchEarliestTime, '0')
            self.assertEqual(d.searchLatestTime, '50')
            self.assertEqual(d.title, 'Title1')
            self.assertEqual(d.searchFieldList, 'foo bar baz'.split())

            # should be able to accept one of the search modes
            d = self.createPanel('single', options='''
                <searchString>search string</searchString>
                ''')
            self.assertEqual(getattr(d, 'searchMode'), 'template')
            self.assertEqual(getattr(d, 'searchCommand'), 'search string')
            d = self.createPanel('single', options='''
                <searchName>search saved</searchName>
                <searchTemplate>search template</searchTemplate>
                <searchPostProcess>search postsearch</searchPostProcess>
                ''')
            self.assertEqual(getattr(d, 'searchMode'), 'saved')
            self.assertEqual(getattr(d, 'searchCommand'), 'search saved')
            d = self.createPanel('single', options='''
                <searchTemplate>search template</searchTemplate>
                ''')
            self.assertEqual(getattr(d, 'searchMode'), 'template')
            self.assertEqual(getattr(d, 'searchCommand'), 'search template')
            d = self.createPanel('single', options='''
                <searchPostProcess>search postsearch</searchPostProcess>
                ''')
            self.assertEqual(getattr(d, 'searchMode'), 'postsearch')
            self.assertEqual(getattr(d, 'searchCommand'), 'search postsearch')

            # any option show be saved into a dictionary
            optionsDict = {'f': ' foo', 'b': 'bar', 'z': 'zip'}
            options = ''
            for k, v in optionsDict.iteritems():
                options += '<option name="%s">%s</option>' % (k, v)
            d = self.createPanel('single', options=options)
            self.assertEqual(d.options['f'], optionsDict['f'].strip())
            self.assertEqual(d.options['b'], optionsDict['b'].strip())
            self.assertEqual(d.options['z'], optionsDict['z'].strip())

    class CreateRowTests(unittest.TestCase):
        def getRowLxml(self, args='', panels=1):
            xml = getRowXml(args, panels)
            root = et.fromstring(xml)
            return createRowFromXml(root)

        def testCreateRow(self):
            d = self.getRowLxml()
            self.assertTrue(d)
            self.assertEqual(len(d.panels), 1)
            self.assertEqual(
                d.panels[0].panelElements[0].matchTagName, 'single')

        def testRowGrouping(self):
            d = self.getRowLxml(args='grouping="2,1"', panels=3)
            self.assertTrue(d)
            self.assertEqual(len(d.panels), 2)
            self.assertEqual(len(d.panels[0].panelElements), 2)
            self.assertEqual(len(d.panels[1].panelElements), 1)

        def testCreateRowWith3Panels(self):
            d = self.getRowLxml(args='', panels=3)
            self.assertTrue(d)
            self.assertEqual(len(d.panels), 3)

    class CreateDashboardTests(unittest.TestCase):
        def getSimpleLxml(self, root='dashboard', rows=1, fieldset=''):
            nodes = []
            nodes.append('<%(root)s>')
            nodes.append(fieldset)
            for i in range(0, rows):
                nodes.append(getRowXml())
            nodes.append('</%(root)s>')
            xml = ''.join(nodes)
            xml = xml % {'root': root}
            root = et.fromstring(xml)
            return createDashboardFromXml(root)

        def testCreateDashboard(self):
            d = self.getSimpleLxml()
            self.assertTrue(d)
            self.assertEqual(d.matchTagName, 'dashboard')

        def testCreateForm(self):
            d = self.getSimpleLxml(root='form')
            self.assertTrue(d)
            self.assertEqual(d.matchTagName, 'form')

        def testCreateFormWithFieldset(self):
            fieldset = '''
            <fieldset>
                <input token="foo" searchWhenChanged="True"></input>
                <html></html>
                <shouldntShow></shouldntShow>
            </fieldset>
            '''
            d = self.getSimpleLxml(root='form', fieldset=fieldset)
            self.assertTrue(d)
            self.assertTrue(d.fieldset)
            self.assertEqual(len(d.fieldset), 2)
            self.assertEqual(d.fieldset[0].__class__.__name__, 'TextInput')
            self.assertEqual(d.fieldset[0].searchWhenChanged, True)
            self.assertEqual(d.fieldset[1].matchTagName, 'html')
            self.assertEqual(d.matchTagName, 'form')

        def testCreateUnsupportedRoot(self):
            with self.assertRaises(Exception):
                self.getSimpleLxml(root='notFormOrDashboard')

    loader = unittest.TestLoader()
    suites = []
    suites.append(loader.loadTestsFromTestCase(CreatePanelElementTests))
    suites.append(loader.loadTestsFromTestCase(CreateRowTests))
    suites.append(loader.loadTestsFromTestCase(CreateDashboardTests))
    unittest.TextTestRunner(verbosity=2).run(unittest.TestSuite(suites))
