define(function(require, exports, module) {
    var _ = require("underscore");
    var mvc = require('./mvc');
    var Backbone = require("backbone");
    var BaseSplunkView = require("./basesplunkview");
    var console = require('util/console');
    var TimeRangeView = require("./timerangeview");
    var InternalSearchBar = require("views/shared/searchbar/Master");
    var TimeRangeModel = require('models/TimeRange');
    var utils = require('./utils'), SplunkConfig = require('splunk.config');
    var sharedModels = require('./sharedmodels');

    require("css!../css/search-bar");
    
    var SearchBarView = BaseSplunkView.extend(
        // Instance
        {
            moduleId: module.id,
            
            className: "splunk-searchbar",
            
            options: {
                "default": undefined,
                managerid: null,
                timerange: true,
                value: undefined
            },
            
            initialize: function() {
                var that = this;
                
                this.configure();
                this.settings.enablePush("value");
                
                if (this.settings.has('timepicker')) {
                    console.warn(
                        'The "%s" setting of class "%s" is deprecated. Use "%s" instead.',
                        'timepicker', 'SearchBarView', 'timerange');
                    this.settings.set('timerange', this.settings.get('timepicker'));
                    this.settings.unset('timepicker');
                }
                
                // Initialize value with default, if provided
                this._onDefaultChange();
                
                this._state = new Backbone.Model({
                    'dispatch.earliest_time': this.settings.get("earliest_time"),
                    'dispatch.latest_time': this.settings.get("latest_time"),
                    'search': this.settings.get('value') || ""
                });
                
                // Get the shared models
                var appModel = sharedModels.get("app");
                var userModel = sharedModels.get("user");
                var appLocalModel = sharedModels.get("appLocal");
                var timesCollection = sharedModels.get("times");
                
                var timeRangeModel = new TimeRangeModel();
                timeRangeModel.save({
                    'earliest': this.settings.get("earliest_time"),
                    'latest': this.settings.get("latest_time")
                });
                
                var createTimeRange = function(settings, searchbar) {
                    var timeRangeOptions = settings.extractWithPrefix('timerange_');
                    var timePickerOptions = settings.extractWithPrefix('timepicker_');
                    if (!_.isEmpty(timePickerOptions)) {
                        console.warn(
                            'The "%s" settings of class "%s" are deprecated. Use "%s" instead.',
                            'timepicker_*', 'SearchBarView', 'timerange_*');
                    }
                    
                    var options = _.extend(
                        { managerid: settings.get("managerid") },
                        timeRangeOptions,
                        timePickerOptions);
                    
                    if (searchbar) {
                        options["timepicker"] = searchbar.children.timeRangePicker;
                        options["el"] = searchbar.children.timeRangePicker.el;
                    }
                    
                    return new TimeRangeView(options);
                };
                
                // We cannot create the searchbar/timerange until these internal models
                // have been fetched, and so we wait on them being done.
                this._dfd = $.when(timesCollection.dfd, userModel.dfd, appLocalModel.dfd).done(function() {
                    that.searchbar = new InternalSearchBar({
                        showTimeRangePicker: true,
                        collection: {
                            times: timesCollection
                        },
                        model: {
                            state: that._state,
                            timeRange: timeRangeModel,
                            user: userModel,
                            appLocal: appLocalModel,
                            application: appModel
                        }
                    });
                    that.searchbar.children.searchInput.options.useAssistant = false;
                    that.searchbar.children.searchInput.options.disableOnSubmit = false;
                    
                    // We can either have the timerange created or not by this 
                    // point (depending on whether the deferred was done before
                    // $.when was called). We create it if it is not already created,
                    // otherwise we set the appropriate properties on it.
                    if (that.timerange) {
                        // Set the appropriate things on the timerange now that it is
                        // created. Note that we are depending on the fact that this
                        // will execute before the deferred handler in the timerange.
                        that.timerange.setElement(that.searchbar.children.timeRangePicker.el);
                        that.timerange.settings.set("timepicker", that.searchbar.children.timeRangePicker);
                    }
                    else {
                        that.timerange = createTimeRange(that.settings, that.searchbar);
                        
                        // Permit deprecated access to the 'timepicker' field
                        that.timepicker = that.timerange;
                    }
                });

    
                if (!this.timerange) {
                    // We create the timerange wrapper ahead of time so that we can
                    // reference it (e.g. mySearchbar.timerange) before the deferreds
                    // are resolved.
                    this.timerange = createTimeRange(this.settings);
                    
                    // Permit deprecated access to the 'timepicker' field
                    this.timepicker = this.timerange;
                }
                
                // Update view if model changes
                this.settings.on("change:value", function(model, value, options) {
                    options = options || {};
                    var suppressValSet = options._self;
                    if (!suppressValSet) {
                        that.val(value || "");
                    }
                });
                
                this.bindToComponentSetting('managerid', this._onManagerChange, this);
                
                this._state.on("change:search", this._onSearchChange, this);
                this.settings.on("change:timerange", this._onDisplayTimeRangeChange, this);
                this.settings.on("change:default", this._onDefaultChange, this);
            },
            
            _onDefaultChange: function(model, value, options) {
                // Initialize value with default, if provided
                var oldDefaultValue = this.settings.previous("default");
                var defaultValue = this.settings.get("default");
                var currentValue = this.settings.get('value');
                
                if (defaultValue !== undefined &&
                    (currentValue === oldDefaultValue || currentValue === undefined))
                {
                    this.settings.set('value', defaultValue);
                }
            },
            
            _onDisplayTimeRangeChange: function() {
                // We cannot work with the searchbar/timerange until they
                // are done being created.
                var that = this;
                $.when(this._dfd).done(function() {
                    if (that.settings.get("timerange")) {
                        that.searchbar.children.timeRangePicker.$el.show();
                    }
                    else {
                        that.searchbar.children.timeRangePicker.$el.hide();
                    }
                });
            },
            
            _onManagerChange: function(managers, manager) {
                if (this.manager) {
                    this.manager.off(null, null, this);
                    this.manager = null;
                }

                this.manager = manager;
                
                // We defer setting the query to let the underlying search bar
                // have enough to finish setting up. Since we might get a 
                // a manager synchronously, it may have not finished setting up
                // (e.g. it is setting its own deferred actions).
                var that = this;
                _.defer(function() {
                    that._updateQuery();
                });
            },
            
            render: function() {
                // We cannot work with the searchbar/timerange until they
                // are done being created.
                var that = this;
                $.when(this._dfd).done(function() {
                    // Remove the internal searchbar
                    that.searchbar.$el.detach();
                    
                    // Clear ourselves
                    that.$el.html('');
                    
                    // Re-render and re-append
                    that.searchbar.render();
                    that.$el.append(that.searchbar.el);
                    
                    // Ensure we properly show/hide the timerange
                    that._onDisplayTimeRangeChange();
                });
                                
                return this;
            },
            
            val: function(value) {
                if (value !== undefined) {
                    this._setSearch(value);
                    
                    /* 
                     * Force firing of a new change event, even if the new
                     * value is the same as the old value. This provides
                     * the expected behavior if the user presses enter in
                     * a search box to refresh the search.
                     */
                    this.settings.unset("value", {_self: true});
                    this.settings.set("value", value, {_self: true});
                    
                    this.trigger("change", value, this);
                }
                else {
                    return this._getSearch();
                }
            },
            
            _onSearchChange: function(model, value, options) {
                options = options || {};
                var suppressValSet = options._self;
                if (!suppressValSet) {
                    this.val(value);
                }
            },
            
            _getSearch: function() {
                return this._state.get("search");
            },
            
            _setSearch: function(newSearch) {
                this._state.set("search", newSearch, {_self: true});
            },
            
            _updateQuery: function() {
                // If we have a previous search query set, display it
                if (this.manager) {
                    var currentSearch = this._state.get("search") || "";
                    var newSearch = (this.manager.query.resolve() || "").trim();
                    
                    if (!currentSearch && newSearch) {
                        this._setSearch(newSearch);
                    }
                }
            }
        }
    );
    
    return SearchBarView;
});
