define(function(require) {
    var _ = require('underscore'),
            Backbone = require('backbone'),
            mvc = require('splunkjs/mvc'),
            console = require('util/console');

    var MONITORED_PROPERTIES = ['dispatchState', 'scanCount', 'resultCount', 'eventCount',
        'eventAvailableCount', 'isDone', 'isFailed', 'isFinalized', 'doneProgress'];

    var JobTracker = function() {
        this.initialize.apply(this, arguments);
    };
    _.extend(JobTracker.prototype, Backbone.Events, {
        initialize: function() {
            this._active = {};
            this._inflight = null;
            this._timer = null;
            this._fresh = false;
            this.service = mvc.createService({ app: '-', owner: '-' });
            this._fetch = _.bind(this._fetch, this);
        },
        /**
         * Track the given search job
         * @param job the job to track
         * @param callbacks an object containing callback functions (ready, progress, done, failed, error)
         */
        track: function(job, callbacks) {
            this._active[job.sid] = this._initializeItem(job, callbacks || {});
            this._fresh = true;
            this._scheduleNextFetch();
        },
        /**
         * Stop tracking the job with the given SID
         * @param sid
         */
        stop: function(sid) {
            var item = this._active[sid];
            if (item) {
                item._stop = true;
                delete this._active[sid];
            }
        },
        _initializeItem: function(job, callbacks) {
            var item = {
                job: job,
                model: new Backbone.Model()
            };
            if (_.isFunction(callbacks.progress)) {
                item.model.on('change', function() {
                    callbacks.progress(job);
                });
            }
            if (_.isFunction(callbacks.ready)) {
                item.model.once('change:dispatchState', item.callbacks.ready);
            }
            if (_.isFunction(callbacks.done)) {
                item.model.on('done', callbacks.done);
            }
            if (_.isFunction(callbacks.failed)) {
                item.model.on('failed', callbacks.failed);
            }
            if (_.isFunction(callbacks.error)) {
                item.model.on('error', callbacks.error);
            }
            return item;
        },
        _scheduleNextFetch: function() {
            if (this._timer || this._inflight) {
                return;
            }
            // Fetch status faster if there have been new jobs added to the tracker recently
            var timeout = this._fresh ? 200 : 1000;
            this._timer = setTimeout(this._fetch, timeout);
        },
        _fetch: function() {
            this._timer = null;
            var requestedJobs = _.clone(this._active);
            if (_.isEmpty(requestedJobs)) {
                console.log('No more jobs to track');
                return;
            }
            // Create search filter for splunkd request
            var requestedSIDs = _(requestedJobs).keys();
            var filter = _(requestedSIDs).map(function(sid) { return ['sid', sid].join('='); }).join(' OR ');
            var tracker = this;
            this._inflight = this.service.jobs().get("", { search: filter, count: requestedSIDs.length },
                    function(err, response) {
                        tracker._inflight = null;
                        if (err) {
                            console.error('Error fetching search jobs: %o', err);
                            _(requestedJobs).chain().pluck('model').invoke('trigger', [ 'error', err ]);
                            this._scheduleNextFetch();
                            return;
                        }
                        var data = response.data;

                        var newActive = {}, oldActive = tracker._active;
                        // Extract items added after requesting the status
                        _(oldActive).each(function(item, sid) {
                            if (!requestedJobs.hasOwnProperty(sid)) {
                                newActive[sid] = item;
                            }
                        });
                        if (_.isEmpty(newActive)) {
                            tracker._fresh = false;
                        } else {
                            console.log('Have new jobs: %o', newActive);
                            tracker._fresh = true;
                        }

                        var returnedSIDs = _(data.entry).map(function(entry) {
                            var sid = entry.content.sid;
                            var item = requestedJobs[sid],
                                    job = item.job,
                                    model = item.model;

                            // Don't update the the content for cancelled jobs
                            if (item._stop) {
                                return sid;
                            }

                            // Update job content
                            job._load(entry);
                            model.set(_.pick(entry.content, MONITORED_PROPERTIES));

                            var props = job.properties();
                            var dispatchState = props.dispatchState;

                            if (dispatchState === 'DONE' && props.isDone) {
                                model.trigger('done', job);
                                model.off();
                            } else if (dispatchState === 'FAILED' && props.isFailed) {
                                model.trigger('failed', job);
                                model.off();
                            } else {
                                newActive[sid] = item;
                            }

                            return sid;
                        });

                        console.log('Splunkd responded with %o items', returnedSIDs.length);

                        // Collected all jobs that haven't been returned from splunkd and trigger an error
                        returnedSIDs = _(returnedSIDs);
                        _(requestedSIDs).each(function(sid) {
                            if (!returnedSIDs.contains(sid)) {
                                console.log('SID not returned: %o', sid);
                                var item = oldActive[sid];
                                if (item._stop) {
                                    return;
                                }
                                if (item.model.has('dispatchState') === undefined) {
                                    newActive[sid] = item;
                                    tracker._fresh = true;
                                } else {
                                    item.model.trigger('error', _("The job has not been found").t());
                                    tracker._itemDone(item);
                                }
                            }
                        });

                        tracker._active = newActive;
                        tracker._scheduleNextFetch();
                    });
        },
        _itemDone: function(item) {
            item.model.off();
        }
    });

    // Singleton instance
    return new JobTracker();

});