define(function(require, exports, module) {
    var _ = require("underscore");
    var mvc = require('./mvc');
    var Backbone = require("backbone");
    var BaseSplunkView = require("./basesplunkview");
    var InternalTimeRangePicker = require('views/shared/timerangepicker/Master');
    var TimeRangeModel = require('models/TimeRange');
    var utils = require('./utils'), SplunkConfig = require('splunk.config');
    var sharedModels = require('./sharedmodels');

    require("css!../css/timerange-picker");
    
    // DOC: Both the 'timepicker' field and setting are private.
    var TimeRangeView = BaseSplunkView.extend(
        // Instance
        {
            moduleId: module.id,
            
            className: "splunk-timerange",
            
            options: {
                'default': undefined,
                managerid: null,
                earliest_time: '',
                latest_time: '',
                preset: undefined,
                value: undefined
            },
            
            initialize: function() {
                var that = this;
                
                this.configure();
                this.settings.enablePush("value");
                
                // Initialize value with default, if provided
                this._onDefaultChange();
                
                var initialValue = this.settings.get("value");
                if (initialValue === undefined) {
                    initialValue = {
                        'earliest_time': this.settings.get("earliest_time"),
                        'latest_time': this.settings.get("latest_time")
                    };
                }
                
                this._state = new Backbone.Model({
                    'dispatch.earliest_time': initialValue['earliest_time'],
                    'dispatch.latest_time': initialValue['latest_time']
                });
                
                // Get the shared models
                var appModel = sharedModels.get("app");
                var userModel = sharedModels.get("user");
                var appLocalModel = sharedModels.get("appLocal");
                var timesCollection = this._timesCollection = sharedModels.get("times");
                
                // We cannot create the timepicker until these internal models
                // have been fetched, and so we wait on them being done.
                this._pickerDfd = $.when(timesCollection.dfd, userModel.dfd, appLocalModel.dfd).done(function() {
                    if (that.settings.get("timepicker")) {
                        that.timepicker = that.settings.get("timepicker");
                        that._state = that.timepicker.model.state;
                        
                        // Unset from settings, so we don't keep it around
                        that.settings.unset("timepicker");
                    } 
                    else {                  
                        var timeRangeModel = new TimeRangeModel();
                        timeRangeModel.save({
                            'earliest': that.settings.get("earliest_time"),
                            'latest': that.settings.get("latest_time")
                        });
                        
                        that.timepicker = new InternalTimeRangePicker({
                            className: "controls",
                            model: {
                                state: that._state,
                                timeRange: timeRangeModel,
                                user: userModel,
                                appLocal: appLocalModel,
                                application: appModel
                            },
                            collection: timesCollection
                        });
                    }
                
                    // We don't register the change handler on the internal 
                    // state until we're done creating the timepicker
                    that._state.on(
                        "change:dispatch.earliest_time change:dispatch.latest_time", 
                        _.debounce(that._onTimeRangeChange), 
                        that
                    );
                });
                    
                // Whenever the times collection changes and/or the preset changes,
                // we update the timepicker
                this._timesCollection.on("change reset", this._onTimePresetUpdate, this);
                this.settings.on("change:preset", this._onTimePresetUpdate, this);
                
                // Update view if model changes
                this.settings.on("change:value", function(model, value, options) {
                    that.val(value || {earliest_time: '', latest_time: ''});
                });
                
                // Update the default if it changes
                this.settings.on("change:default", this._onDefaultChange, this);
                
                // Update model if view changes
                this.on("change", function(value) {
                    that.settings.set("value", value);
                });
                
                this.bindToComponent(this.settings.get("managerid"), this._onManagerChange, this);
            },
            
            _onDefaultChange: function(model, value, options) {
                // Initialize value with default, if provided
                var oldDefaultValue = this.settings.previous("default");
                var defaultValue = this.settings.get("default");
                var currentValue = this.settings.get('value');
                
                if (defaultValue !== undefined &&
                    (_.isEqual(currentValue, oldDefaultValue) || currentValue === undefined))
                {
                    this.settings.set('value', defaultValue);
                }
            },
            
            _onTimePresetUpdate: function() {
                var preset = this.settings.get("preset");
                if (!preset) {
                    return;
                }
                
                // We can only look at the times collection when it has finished
                // fetching
                var that = this;
                $.when(this._pickerDfd).done(function() {
                    var timeModel = that._timesCollection.find(function(model) {
                        return model.entry.content.get("label") === preset;
                    });
                    
                    if (!timeModel) {
                        return;
                    }
                    
                    that.val({
                        earliest_time: timeModel.entry.content.get('earliest_time'),
                        latest_time: timeModel.entry.content.get('latest_time')
                    });
                });
            },
            
            _onManagerChange: function(managers, manager) {
                if (this.manager) {
                    this.manager.off(null, null, this);
                    this.manager = null;
                }

                this.manager = manager;
                if (manager.search) {
                    this.listenTo(
                        manager.search, 
                        "change:earliest_time change:latest_time", 
                        _.debounce(this._onManagerTimeRangeChange)
                    );
                    
                    this._onManagerTimeRangeChange();
                }
            },
            
            _onManagerTimeRangeChange: function() {
                // Get the current time range                
                var timeRange = this.val();
                
                // The manager's time range, if it has one, overrides the timepicker's
                // current range.
                timeRange.earliest_time = this.manager.search.get("earliest_time") || timeRange.earliest_time;
                timeRange.latest_time = this.manager.search.get("latest_time") || timeRange.latest_time;
                
                // Set it back
                this._setTimeRange(timeRange);
            },
            
            _onTimeRangeChange: function(model, value, options) {
                if (!options || (options && !options._self)) {
                    this.trigger("change", this._getTimeRange(), this);
                }
            },
            
            render: function() {
                var that = this;
                $.when(this._pickerDfd).done(function() {
                    that.timepicker.render();
                    that.$el.append(that.timepicker.el);
                    
                    // Ensure we have the right time preset set
                    that._onTimePresetUpdate();
                });
                                
                return this;
            },
            
            val: function(newValue) {
                if (newValue !== undefined) {
                    var oldValue = this.val();
                    
                    this._setTimeRange(newValue);
                    var realNewValue = this.val();
                    
                    // Trigger change event manually since programmatic DOM
                    // manipulations don't trigger events automatically.
                    if (!_.isEqual(realNewValue, oldValue)) {
                        this.trigger('change', realNewValue, this);
                    }
                }
                else {
                    return this._getTimeRange();
                }
            },
            
            _setTimeRange: function(value) {
                this._state.set({
                    "dispatch.earliest_time": value.earliest_time,
                    "dispatch.latest_time": value.latest_time 
                }, { _self: true });
            },
            
            _getTimeRange: function() {
                return {
                    "earliest_time": this._state.get("dispatch.earliest_time"),
                    "latest_time": this._state.get("dispatch.latest_time")
                };
            }
        }
    );
    
    return TimeRangeView;
});
