define(['underscore'], function(_) {

    /**
     * Expose a post-hook for the tooltip getAnchor method.
     *
     * @param Tooltip (Highcharts tooltip constructor)
     */

    var patchTooltipGetAnchor = function(Tooltip) {
        Tooltip.prototype.getAnchor = _(Tooltip.prototype.getAnchor).wrap(function(originalFn, points, mouseEvent) {
            var anchor = originalFn.call(this, points, mouseEvent),
                options = this.options;

            if(typeof options.getAnchorPostHook === 'function') {
                return options.getAnchorPostHook(points, mouseEvent, anchor);
            }
            return anchor;
        });
    };

    /**
     * Expose pre- and post-hooks for the rendering of the legend items.
     *
     * @param Legend (Highcharts legend constructor)
     */

    var patchLegendRenderItem = function(Legend) {
        Legend.prototype.renderItem = _(Legend.prototype.renderItem).wrap(function(originalFn, item) {
            var options = this.options,
                allItems = this.allItems;

            if(typeof options.renderItemsPreHook === 'function' && item === allItems[0]) {
                options.renderItemsPreHook(this);
                originalFn.call(this, item);
            }
            else if(typeof options.renderItemsPostHook === 'function' && item === _(allItems).last()) {
                originalFn.call(this, item);
                options.renderItemsPostHook(this);
            }
            else {
                originalFn.call(this, item);
            }
        });
    };

    /**
     * Expose a pre-hook to the series drawPoints routine
     *
     * @param Series (Highcharts series constructor)
     */

    var patchSeriesDrawPoints = function(Series) {
        Series.prototype.drawPoints = _(Series.prototype.drawPoints).wrap(function(originalFn) {
            var options = this.options;
            if(typeof options.drawPointsPreHook === 'function') {
                options.drawPointsPreHook(this);
            }
            originalFn.call(this);
        });
    };

    /**
     * Expose a post-hook to the series plotGroup routine
     *
     * @param Series (Highcharts series constructor)
     */

    var patchSeriesPlotGroup = function(Series) {
        Series.prototype.plotGroup = _(Series.prototype.plotGroup).wrap(function(originalFn, prop, name, visibility, zIndex, parent) {
            var group = originalFn.call(this, prop, name, visibility, zIndex, parent),
                options = this.options;

            if(typeof options.plotGroupPostHook === 'function') {
                options.plotGroupPostHook(this, group);
            }
            return group;
        });
    };

    /**
     * Expose a post-hook to the series render routine
     *
     * @param Series (Highcharts series constructor)
     */

    var patchSeriesRender = function(Series) {
        Series.prototype.render = _(Series.prototype.render).wrap(function(originalFn) {
            originalFn.call(this);
            var options = this.options;
            if(typeof options.renderPostHook === 'function') {
                options.renderPostHook(this);
            }
        });
    };

    /**
     * Expose a post-hook to the series afterAnimate routine
     *
     * @param Series (Highcharts series constructor)
     */

    var patchSeriesAfterAnimate = function(Series) {
        Series.prototype.afterAnimate = _(Series.prototype.afterAnimate).wrap(function(originalFn) {
            originalFn.call(this);
            var options = this.options;
            if(typeof options.afterAnimatePostHook === 'function') {
                options.afterAnimatePostHook(this);
            }
        });
    };

    /**
     * Expose a pre-hook to the series destroy routine
     *
     * @param Series (Highcharts series constructor)
     */

    var patchSeriesDestroy = function(Series) {
        Series.prototype.destroy = _(Series.prototype.destroy).wrap(function(originalFn) {
            var options = this.options;
            if(typeof options.destroyPreHook === 'function') {
                options.destroyPreHook(this);
            }
            originalFn.call(this);
        });
    };

    /**
     * Expose a pre-hook for the pie series translate routine
     *
     * @param PieSeries (Highcharts pie series constructor)
     */

    var patchPieSeriesTranslate = function(PieSeries) {
        PieSeries.prototype.translate = _(PieSeries.prototype.translate).wrap(function(originalFn) {
            var options = this.options;
            if(typeof options.translatePreHook === 'function') {
                options.translatePreHook(this);
            }
            originalFn.call(this);
        });
    };

    /**
     * Expose pre- and post-hooks for the pie chart data label drawing routing
     *
     * @param PieSeries (Highcharts pie series constructor)
     */

    var patchPieSeriesDrawDataLabels = function(PieSeries) {
        PieSeries.prototype.drawDataLabels = _(PieSeries.prototype.drawDataLabels).wrap(function(originalFn) {
            var options = this.options;
            options.dataLabels = options.dataLabels || {};

            if(typeof options.dataLabels.drawDataLabelsPreHook === 'function') {
                options.dataLabels.drawDataLabelsPreHook(this);
            }
            originalFn.call(this);
            if(typeof options.dataLabels.drawDataLabelsPostHook === 'function') {
                options.dataLabels.drawDataLabelsPostHook(this);
            }
        });
    };

    /**
     * Expose a pre-hook for the axis getOffset routine
     *
     * @param Axis (Highcharts axis constructor)
     */

    var patchAxisGetOffset = function(Axis) {
        Axis.prototype.getOffset = _(Axis.prototype.getOffset).wrap(function(originalFn) {
            var options = this.options;
            if(typeof options.getOffsetPreHook === 'function') {
                options.getOffsetPreHook(this);
            }
            originalFn.call(this);
        });
    };

    /**
     * Expose a post-hook for the axis getSeriesExtremes routine
     *
     * @param Axis (Highcharts axis constructor)
     */

    var patchAxisGetSeriesExtremes = function(Axis) {
        Axis.prototype.getSeriesExtremes = _(Axis.prototype.getSeriesExtremes).wrap(function(originalFn) {
            var options = this.options;
            originalFn.call(this);
            if(typeof options.getSeriesExtremesPostHook === 'function') {
                options.getSeriesExtremesPostHook(this);
            }
        });
    };

    /**
     * Expose a hook for the setting of the axis tick positions
     *
     * @param Axis (Highcharts axis constructor)
     */

    var patchAxisSetTickPositions = function(Axis) {
        Axis.prototype.setTickPositions = _(Axis.prototype.setTickPositions).wrap(function(originalFn, secondPass) {
            var options = this.options;
            if(typeof options.setTickPositionsPreHook === 'function') {
                options.setTickPositionsPreHook(this, secondPass);
            }
            originalFn.call(this, secondPass);
            if(typeof options.setTickPositionsPostHook === 'function') {
                options.setTickPositionsPostHook(this, secondPass);
            }
        });
    };

    /**
     * Expose a pre-hook for setting the size of the axis
     *
     * @param Axis (Highcharts axis constructor)
     */

    var patchAxisSetSize = function(Axis) {
        Axis.prototype.setAxisSize = _(Axis.prototype.setAxisSize).wrap(function(originalFn) {
            var options = this.options;
            if(typeof options.setSizePreHook === 'function') {
                options.setSizePreHook(this);
            }
            originalFn.call(this);
        });
    };

    /**
     * Expose a post-hook for the render routine of a tick.
     * Also, make sure the handleOverflow method is called for every render.
     *
     * @param Tick (Highcharts tick constructor)
     */

    var patchTickRender = function(Tick) {
        Tick.prototype.render = _(Tick.prototype.render).wrap(function(originalFn, index, old) {
            originalFn.call(this, index, old);
            var axisOptions = this.axis.options;
            if(typeof axisOptions.tickRenderPostHook === 'function') {
                axisOptions.tickRenderPostHook(this, index, old);
            }
        });
    };

    /**
     * Expose an override hook for the tick handleOverflow method
     *
     * @param Tick (Highcharts tick constructor)
     */

    var patchTickHandleOverflow = function(Tick) {
        Tick.prototype.handleOverflow = _(Tick.prototype.handleOverflow).wrap(function(originalFn, index, xy) {
            var axisOptions = this.axis.options;
            if(typeof axisOptions.tickHandleOverflowOverride === 'function') {
                return axisOptions.tickHandleOverflowOverride(this, index, xy);
            }
            return originalFn.call(this, index, xy);
        });
    };

    return ({

        applyPatches: function(Highcharts) {
            var Tooltip = Highcharts.Tooltip;
            patchTooltipGetAnchor(Tooltip);

            var Legend = Highcharts.Legend;
            patchLegendRenderItem(Legend);

            var Series = Highcharts.Series;
            patchSeriesDrawPoints(Series);
            patchSeriesPlotGroup(Series);
            patchSeriesRender(Series);
            patchSeriesAfterAnimate(Series);
            patchSeriesDestroy(Series);

            var PieSeries = Highcharts.seriesTypes.pie;
            patchPieSeriesTranslate(PieSeries);
            patchPieSeriesDrawDataLabels(PieSeries);

            var Axis = Highcharts.Axis;
            patchAxisGetOffset(Axis);
            patchAxisGetSeriesExtremes(Axis);
            patchAxisSetTickPositions(Axis);
            patchAxisSetSize(Axis);

            var Tick = Highcharts.Tick;
            patchTickRender(Tick);
            patchTickHandleOverflow(Tick);
        }

    });

});