// (c) 2012 Uzi Kilon, Splunk Inc.
// Backbone Poller may be freely distributed under the MIT license.
define([
    'underscore',
    'jquery',
    'backbone',
    'helpers/Session'
],
function(
    _,
    $,
    Backbone,
    Session
){
        // Private variables
        var pollerDefaults = {
            delay: 1000,
            stopOnError: true,
            condition: function(){return true;}
        };
        var eventTypes = ['start', 'stop', 'success', 'error', 'complete'];

        /**
         * Private Poller
         */
        var Poller = function Poller(model, options) {
            this.set(model, options);
            if (options.ui_inactivity) {
                Session.on('timeout', function() {
                    this.stop();
                    Session.off(null, null, this);
                }, this);
            }
        };
        _.extend(Poller.prototype, Backbone.Events, {
            set: function(model, options) {
                this.model = model;
                this.options = _.extend(_.clone(pollerDefaults), options || {});

                _.each(eventTypes, function(eventName){
                    var handler = this.options[eventName];
                    if(typeof handler === 'function') {
                        this.on(eventName, handler, this);
                    }
                }, this);

                if ( this.model instanceof Backbone.Model ) {
                    this.model.on('destroy', this.stop, this);
                }

                return this.stop({silent: true});
            },
            start: function(options){
                if(this.active() === true) {
                    return this;
                }
                options = options || {};
                if(!options.silent) {
                    this.trigger('start');
                }
                this.options.active = true;
                run(this);
                return this;
            },
            stop: function(options){
                options = options || {};
                if(!options.silent) {
                    this.trigger('stop');
                }
                this.options.active = false;
                clearTimeout(this.timeoutId);
                if(this.xhr && typeof this.xhr.abort === 'function') {
                    this.xhr.abort();
                }
                this.xhr = null;
                return this;
            },
            active: function(){
                return this.options.active === true;
            }
        });

        // private methods
        function run(poller) {
            if ( poller.active() !== true ) {
                window.clearTimeout(poller.timeoutId);
                return ;
            }

            var onSuccess = function(){
                poller.trigger('success');
                poller.model.off('sync', onSuccess);
            };
            
            var options = $.extend(true, {}, poller.options || {}, {
                success: function(model, response, options) {
                    poller.trigger('success');
                    if( poller.options.condition(poller.model) !== true ) {
                        poller.trigger('complete');
                        poller.stop({silent: true});
                    }
                    else {
                        poller.timeoutId = window.setTimeout(function(){ run(poller); }, poller.options.delay);
                    }
                },
                error: function(model, response, options){
                    poller.trigger('error');
                    if (poller.options.stopOnError) {
                        poller.stop({silent: true});
                    } else {
                        poller.timeoutId = window.setTimeout(function(){ run(poller); }, poller.options.delay);
                    }
                }
            });
            poller.xhr = poller.model.fetch(options);
        }

        /**
         * Polling Manager
         */
        var pollers = [];

        return {
            get: function(model) {
                return _.find(pollers, function(poller){
                    return poller.model === model;
                });
            },
            getPoller: function(model, options){
                var poller = this.get(model);
                if( poller ) {
                    poller.set(model, options);
                }
                else {
                    poller = new Poller(model, options);
                    pollers.push(poller);
                }
                return poller;
            },
            start: function(model, options) {
                return this.getPoller(model, options).start({silent: true});
            },
            stop: function(model) {
                var poller = this.get(model);
                if( poller ) {
                    poller.stop();
                    return true;
                }
                return false;
            },
            size: function(){
                return pollers.length;
            }
        };
    }
);
