define([
            'jquery',
            'underscore',
            './Axis',
            './CategoryAxis',
            '../../helpers/Formatter',
            '../../util/lang_utils',
            'util/time_utils',
            '../../util/time_utils',
            '../../util/dom_utils'
        ],
        function(
            $,
            _,
            Axis,
            CategoryAxis,
            Formatter,
            langUtils,
            splunkTimeUtils,
            timeUtils,
            domUtils
        ) {

    var TimeAxis = function(properties) {
        this.originalCategories = properties['axis.categories'];
        this.spanData = properties['axis.spanData'];
        var timeCategoryInfo = timeUtils.convertTimeToCategories(this.originalCategories, this.spanData, this.NUM_LABEL_CUTOFF);
        this.granularity = timeCategoryInfo.granularity;
        this.pointSpan = timeUtils.getPointSpan(this.originalCategories);
        properties['axis.categories'] = timeCategoryInfo.categories;
        CategoryAxis.call(this, properties);
    };
    langUtils.inherit(TimeAxis, CategoryAxis);

    $.extend(TimeAxis.prototype, {

        NUM_LABEL_CUTOFF: 6,

        getConfig: function() {
            var config = CategoryAxis.prototype.getConfig.call(this);
            $.extend(true, config, {
                showLastLabel: true,
                setTickPositionsPostHook: _(this.setTickPositionsPostHook).bind(this)
            });
            return config;
        },

        formatLabel: function(info) {
            return info;
        },

        formatValue: function(value) {
            return timeUtils.formatIsoStringAsTooltip(value, this.pointSpan) || 'Invalid timestamp'; // TODO: localize
        },

        /**
         * @author sfishel
         *
         * Before the getOffset routine runs, align the axis labels to the right of each tick
         */

        getOffsetPreHook: function(axis) {
            var options = axis.options,
                chart = axis.chart,
                categories = options.categories,
                tickLabelPadding = (this.isVertical) ? 2 : 3,
                axisLength = (this.isVertical) ? chart.plotHeight : chart.plotWidth,
                tickSpacing = (categories.length > 1) ? (axisLength / categories.length) : axisLength + (tickLabelPadding * 2);

            options.adjustedTickSpacing = tickSpacing;
            if(this.isVertical) {
                var labelFontSize = parseInt((options.labels.style.fontSize.split('px'))[0], 10);
                options.labels.y = (tickSpacing / 2) + labelFontSize + tickLabelPadding;
            }
            else {
                if(options.tickmarkPlacement === 'on') {
                    options.labels.align = 'left';
                    options.labels.x = tickLabelPadding;
                }
                else {
                    options.labels.align = 'left';
                    options.labels.x = (tickSpacing / 2) + tickLabelPadding;
                }
            }
        },

        /**
         * @author sfishel
         *
         * Make adjustments to the tick positions to label only the appropriate times
         */

        setTickPositionsPostHook: function(axis, secondPass) {
            var options = axis.options,
                categories = options.categories;

            axis.tickPositions = [];
            if(!options.originalCategories) {
                options.originalCategories = $.extend(true, [], categories);
            }
            var i,
                originalCategories = options.originalCategories;

            for(i = 0; i < originalCategories.length; i++) {
                if(originalCategories[i] && originalCategories[i] !== " ") {
                    if(options.tickmarkPlacement === 'on') {
                        axis.tickPositions.push(i);
                    }
                    else {
                        // if the tickmark placement is 'between', we shift everything back one
                        // interestingly, HighCharts will allow negatives here, and in fact that's what we need to label the first point
                        axis.tickPositions.push(i - 1);
                        categories[i - 1] = originalCategories[i];
                    }
                }
            }

            // adjust the axis label CSS so that soft-wrapping will not occur
            // for the VML renderer we also have to make sure the tick labels will accurately report their own widths
            options.labels.style.width = 'auto';
            if(!domUtils.hasSVG) {
                options.labels.style['white-space'] = 'nowrap';
            }
        },

        /**
         * @author sfishel
         *
         * Use the handleOverflow override hook to handle any collisions among the axis labels
         * All branches should return true, resolveCollisions will show/hide the correct ticks and labels
         *
         * NOTE: Highcharts renders the second tick first, and the first tick last
         */

        tickHandleOverflowOverride: function(tick, index, xy) {
            // FIXME: this is a little hacky...
            // use the second tick as an indicator that we're starting a new render routine and reset the collisionDetected flag
            // can't do the regular collision detection because the first tick isn't there yet
            if(index === 1) {
                this.collisionDetected = false;
                this.lastTickFits = true;
                return true;
            }
            // FIXME: again a little hacky...
            // use the first tick as an indicator that this is the last step in rendering
            // only now can we check if the first and second ticks overlap (if there is a second tick)
            if(tick.isFirst) {
                if(!tick.isLast) {
                    this.collisionDetected = this.collisionDetected || this.tickOverlapsNext(tick, index, xy);
                }
                this.resolveCollisionDetection(tick.axis, this.collisionDetected, this.lastTickFits);
                return true;
            }

            // if we haven't seen a collision yet, figure out if the current tick causes one
            this.collisionDetected = this.collisionDetected || this.tickOverlapsPrevious(tick, index, xy);

            // keep track of whether the last tick fits without overflow
            if(tick.isLast) {
                this.lastTickFits = CategoryAxis.prototype.tickHandleOverflowOverride.call(this, tick, index, xy);
            }
            return true;
        },

        tickOverlapsPrevious: function(tick, index, xy) {
            var axis = tick.axis,
                // assume this won't be called with the first tick
                previous = axis.ticks[axis.tickPositions[index - 1]],
                previousXY = previous.getPosition(axis.horiz, previous.pos, axis.tickmarkOffset);

            // check for the vertical axis case
            if(this.isVertical) {
                var previousBottom = previousXY.y + this.getTickLabelExtremesY(previous)[1];
                return (xy.y < previousBottom);
            }

            // otherwise handle the horizontal axis case
            var previousRight = previousXY.x + this.getTickLabelExtremesX(previous)[1];
            return (xy.x < previousRight);
        },

        tickOverlapsNext: function(tick, index, xy) {
            var axis = tick.axis,
                // assume this won't be called with the last tick
                next = axis.ticks[axis.tickPositions[index + 1]],
                nextXY = next.getPosition(axis.horiz, next.pos, axis.tickmarkOffset);

            if(!next) {
                return false;
            }

            // check for the vertical axis case
            if(this.isVertical) {
                var myBottom = xy.y + this.getTickLabelExtremesY(tick)[1];
                return (myBottom > nextXY.y);
            }

            // otherwise handle the horizontal case
            var myRight = xy.x + this.getTickLabelExtremesX(tick)[1];
            return (myRight > nextXY.x);
        },

        resolveCollisionDetection: function(axis, hasCollisions, lastLabelFits) {
            var formatter = new Formatter(axis.chart.renderer),
                tickPositions = axis.tickPositions,
                collisionTickPositions = tickPositions.slice(1),
                ticks = axis.ticks,
                rawLabels = this.originalCategories,
                labelGranularity = this.granularity,
                positionOffset = this.hasTickmarksBetween() ? 1 : 0;

            if(hasCollisions) {
                _(collisionTickPositions).each(function(pos, i) {
                    i++; // do this because we sliced out the first tick
                    var tick = ticks[pos];
                    if(i % 2 === 0) {
                        var bdTime = splunkTimeUtils.extractBdTime(rawLabels[tick.pos + positionOffset]),
                            prevTick = ticks[tickPositions[i - 2]],
                            prevBdTime = splunkTimeUtils.extractBdTime(rawLabels[prevTick.pos + positionOffset]),
                            newLabel = (timeUtils.formatBdTimeAsAxisLabel(bdTime, prevBdTime, labelGranularity) || ['']).join('<br/>');

                        tick.label.attr({ text: newLabel });
                    }
                    else {
                        tick.label.hide();
                        tick.mark.hide();
                    }
                });
            }
            else {
                _(collisionTickPositions).each(function(pos, i) {
                    i++; // do this because we sliced out the first tick
                    var tick = ticks[pos];
                    tick.label.show();
                    tick.mark.show();
                    if(i % 2 === 0) {
                        var bdTime = splunkTimeUtils.extractBdTime(rawLabels[pos + positionOffset]),
                            prevTick = ticks[tickPositions[i - 1]],
                            prevBdTime = splunkTimeUtils.extractBdTime(rawLabels[prevTick.pos + positionOffset]),
                            newLabel = (timeUtils.formatBdTimeAsAxisLabel(bdTime, prevBdTime, labelGranularity) || ['']).join('<br/>');

                        tick.label.attr({ text: newLabel });
                    }
                });
            }
            if(!lastLabelFits && (!hasCollisions || tickPositions.length % 2 !== 0)) {
                axis.ticks[_(tickPositions).last()].label.hide();
            }

            formatter.destroy();
        },

        // have to make some adjustments to get the correct answer when tickmarkPlacement = between
        getTickLabelExtremesX: function(tick) {
            var extremes = CategoryAxis.prototype.getTickLabelExtremesX.call(this, tick),
                axisOptions = tick.axis.options;

            if(axisOptions.tickmarkPlacement === 'between') {
                return _(extremes).map(function(extreme) { return extreme - (axisOptions.adjustedTickSpacing / 2); });
            }
            return extremes;
        },

        // inheritance gets a little weird here, the TimeAxis wants to go back to the base Axis behavior for this method
        getTickLabelExtremesY: function(tick) {
            return Axis.prototype.getTickLabelExtremesY.apply(this, arguments);
        }

    });

    return TimeAxis;

});