define([
            'jquery',
            'underscore',
            './ManyShapeSeries',
            '../util/lang_utils',
            '../util/parsing_utils',
            '../util/time_utils',
            'util/time_utils',
            'splunk.i18n'
        ], 
        function(
            $,
            _,
            ManyShapeSeries,
            langUtils,
            parsingUtils,
            timeUtils,
            splunkTimeUtils,
            i18n
        ) {

    var PieSeries = function(properties) {
        ManyShapeSeries.call(this, properties);
        this.collapseFieldName = this.properties.sliceCollapsingLabel || 'other';
        this.collapsePercent = !this.properties.hasOwnProperty('sliceCollapsingThreshold') ? 0.01 :
                                    parseFloat(this.properties.sliceCollapsingThreshold);
    };
    langUtils.inherit(PieSeries, ManyShapeSeries);

    $.extend(PieSeries.prototype, {

        CHARTING_PROPERTY_WHITELIST:_.union(
            ['sliceCollapsingThreshold', 'sliceCollapsingLabel', 'showPercent'],
            ManyShapeSeries.prototype.CHARTING_PROPERTY_WHITELIST
        ),

        type: 'pie',
        hasPrettyData: false,

        fieldList: [],

        getConfig: function() {
            return $.extend(ManyShapeSeries.prototype.getConfig.call(this), {
                translatePreHook: _(this.translatePreHook).bind(this)
            });
        },

        setData: function(inputData) {
            this.data = [];
            this.prettyData = [];
            var that = this,
                nameSeries = inputData.names,
                sizeSeries = inputData.sizes,
                spanSeries = inputData.spans,
                isTimeBased = inputData.isTimeBased,
                totalSize = _(sizeSeries).reduce(function(sum, value) { return (sum + value); }, 0),
                cardinality = sizeSeries.length,
                collapsedSize = 0,
                numCollapsed = 0,
                numLessThanThresh = 0,
                granularity = null,

                passesThreshold = function(value) {
                    return (value > 0 && (value / totalSize) > that.collapsePercent);
                };

            if(isTimeBased) {
                granularity = splunkTimeUtils.determineLabelGranularity(nameSeries);
                this.hasPrettyData = true;
            }

            this.fieldList = _(nameSeries).map(parsingUtils.escapeSVG, parsingUtils);
            _(sizeSeries).each(function(value, i) {
                if(!(passesThreshold(sizeSeries[i]))) {
                    numLessThanThresh++;
                }
            }, this);

            _(nameSeries).each(function(name, i) {
                var sizeValue = sizeSeries[i];
                if(passesThreshold(sizeValue) || numLessThanThresh === 1 || cardinality <=10) {                    
                    if(isTimeBased) {
                        var bdTime = splunkTimeUtils.extractBdTime(name),
                            humanizedName = timeUtils.formatBdTimeAsAxisLabel(bdTime, null, granularity).join(' '),
                            spanValue = spanSeries[i];
                        this.data.push([name, sizeValue, spanValue]);
                        this.prettyData.push([humanizedName, sizeValue, spanValue]);
                    }
                    else {
                        this.data.push([name, sizeValue]);
                    }
                }
                else {
                    collapsedSize += sizeValue;
                    numCollapsed++;
                    this.fieldList = _(this.fieldList).without(name);
                }
            }, this);

            if(numCollapsed > 0) {
                var collapsedName = this.collapseFieldName + ' (' + numCollapsed + ')';
                this.data.push([collapsedName, collapsedSize]);
                // Doesn't make sense to attach a span value to the collapsed section
                this.prettyData.push([collapsedName, collapsedSize, null]);
                this.fieldList.push('__other');
            }

            if(parsingUtils.normalizeBoolean(this.properties.showPercent)) {
                _([this.data, this.prettyData]).each(function(data) {
                    _(data).each(function(dataPoint) {
                        dataPoint[0] += ', ' + i18n.format_percent(dataPoint[1] / totalSize);
                    });
                });
            }

        },

        getFieldList: function() {
            return this.fieldList;
        },

        // returns the series data after any processing (like slice collapsing) has been applied
        getData: function() {
            return this.data;
        },

        getPrettyData: function() {
            return this.prettyData;
        },

        highlightPoint: function(hcPoint) {
            var pointColor = hcPoint.color;
            hcPoint.graphic.attr({
                'fill': pointColor,
                'stroke-width': 0,
                'stroke': pointColor
            });
        },

        getLeftColData: function(info) {
            return [info.sliceFieldName, info.seriesName, info.seriesName + "%"];
        },

        getRightColData: function(info) {
            return [info.sliceName, info.yValue, info.yPercent];
        },

        /**
         * @author sfishel
         *
         * Dynamically adjust the pie size based on the height and width of the container
         */

        translatePreHook: function(pieSeries) {
            var chart = pieSeries.chart;
            pieSeries.options.size = Math.min(chart.plotHeight * 0.75, chart.plotWidth / 3);
        },

        tooltipTemplate: '\
            <table class="highcharts-tooltip"\
                <% if(willWrap) { %>\
                    style="word-wrap: break-word; white-space: normal;"\
                <% } %>>\
                <tr>\
                    <td style="text-align: left; color: #cccccc; max-width: <%= scaledMaxLeftColWidth %>;"><%- sliceFieldName %>:&nbsp;&nbsp;</td>\
                    <td style="text-align: right; color: #ffffff; max-width: <%= scaledMaxRightColWidth %>;"><%- sliceName %></td>\
                </tr>\
                <tr>\
                    <td style="text-align: left; color: <%= sliceColor %>; max-width: <%= scaledMaxLeftColWidth %>;"><%- seriesName %>:&nbsp;&nbsp;</td>\
                    <td style="text-align: right; color: #ffffff; max-width: <%= scaledMaxRightColWidth %>;"><%- yValue %></td>\
                </tr>\
                <tr>\
                    <td style="text-align: left; color: <%= sliceColor %>; max-width: <%= scaledMaxLeftColWidth %>;"><%- seriesName %>%:&nbsp;&nbsp;</td>\
                    <td style="text-align: right; color: #ffffff; max-width: <%= scaledMaxRightColWidth %>;"><%- yPercent %></td>\
                </tr>\
            </table>\
        '

    });
    
    return PieSeries;
    
});