
define([
            'jquery',
            'underscore',
            'highcharts',
            '../Visualization',
            '../../components/ColorPalette',
            '../../components/axes/TimeAxis',
            '../../components/axes/CategoryAxis',
            '../../components/axes/NumericAxis',
            '../../components/Legend',
            '../../components/PanningScrollbar',
            '../../components/Tooltip',
            '../../helpers/HoverEventThrottler',
            '../../series/series_factory',
            'splunk.util',
            '../../util/lang_utils',
            '../../util/testing_utils',
            '../../util/parsing_utils',
            '../../util/color_utils',
            '../../util/time_utils',
            '../../util/dom_utils',
            '../../../helpers/user_agent',
            'util/console'
       ],
       function(
           $,
           _,
           Highcharts,
           Visualization,
           ColorPalette,
           TimeAxis,
           CategoryAxis,
           NumericAxis,
           Legend,
           PanningScrollbar,
           Tooltip,
           HoverEventThrottler,
           seriesFactory,
           splunkUtils,
           langUtils,
           testingUtils,
           parsingUtils,
           colorUtils,
           timeUtils,
           domUtils,
           userAgent,
           console
       ) {

    var Chart = function(container, properties) {
        Visualization.call(this, container, properties);
    };
    langUtils.inherit(Chart, Visualization);

    $.extend(Chart.prototype, {

        HOVER_DELAY: 160,
        EXPORT_WIDTH: 600,
        EXPORT_HEIGHT: 400,
        POINT_CSS_SELECTOR: domUtils.hasSVG ? '.highcharts-series > rect' : '.highcharts-series > shape',

        hasLegend: true,
        hasTooltip: true,
        hasXAxis: true,
        hasYAxis: true,

        requiresExternalColors: true,
        externalPaletteMapping: {},
        externalPaletteSize: 0,

        prepare: function(dataSet, properties) {
            // FIXME: should be smart enough here to update the chart in place when possible
            Visualization.prototype.prepare.call(this, dataSet, properties);
            this.benchmark('Prepare Started');
            this.initializeFields();
            var isEmpty = this.isEmpty();
            if(!isEmpty) {
                this.initializeColorPalette();
            }
            this.initializeSeriesList();
            this.axesAreInverted = _(this.seriesList).any(function(series) {
                return (series.getType() === 'bar');
            });
            if(this.hasXAxis) {
                this.initializeXAxisList();
            }
            if(this.hasYAxis) {
                this.initializeYAxisList();
            }
            if(!isEmpty) {
                if(this.hasLegend) {
                    this.initializeLegend();
                }
                if(this.hasTooltip) {
                    this.initializeTooltip();
                }
                this.setAllSeriesData();
                this.bindSeriesEvents();
                if(this.needsPanning){
                    this.initializePanning();
                    this.bindPanningEvents();
                }
            }
        },

        getFieldList: function() {
            return _(this.seriesList).chain().invoke('getFieldList').flatten(true).value();
        },

        setExternalColorPalette: function(fieldIndexMap, paletteTotalSize) {
            this.externalPaletteMapping = $.extend({}, fieldIndexMap);
            this.externalPaletteSize = paletteTotalSize;
        },

        // TODO [sff] the callback shouldn't be called with the Highcharts object
        draw: function(callback) {
            console.debug('drawing a chart with dimensions:', { width: this.width, height: this.height });
            console.debug('drawing a chart with properties:', this.properties);
            console.debug('drawing a chart with data:', this.dataSet.toJSON());
            this.benchmark('Draw Started');
            this.applyColorPalette();

            this.hcConfig = $.extend(true, {}, this.BASE_CONFIG, {
                chart: this.getChartConfig(),
                series: this.getSeriesConfigList(),
                xAxis: this.getXAxisConfig(),
                yAxis: this.getYAxisConfig(),
                legend: this.getLegendConfig(),
                tooltip: this.getTooltipConfig(),
                plotOptions: this.getPlotOptionsConfig()
            });

            var that = this;
            if(this.exportMode) {
                if(this.seriesIsTimeBased(this.xFields)) {
                    _(this.hcConfig.xAxis).each(function(xAxis) {
                        var xAxisMargin;
                        if(that.axesAreInverted) {
                            xAxisMargin = -50;
                        }
                        else {
                            var spanSeries = that.dataSet.getSeriesAsFloats('_span'),
                                span = (spanSeries && spanSeries.length > 0) ? parseInt(spanSeries[0], 10) : 1,
                                secsPerDay = 60 * 60 * 24,
                                secsPerYear = secsPerDay * 365;

                            if(span >= secsPerYear) {
                                xAxisMargin = 15;
                            }
                            else if(span >= secsPerDay) {
                                xAxisMargin = 25;
                            }
                            else {
                                xAxisMargin = 35;
                            }
                        }
                        xAxis.title.margin = xAxisMargin;
                    });
                }
                $.extend(true, this.hcConfig, {
                    plotOptions: {
                        series: {
                            enableMouseTracking: false,
                            shadow: false
                        }
                    }
                });

            }
            this.addEventHandlers();
            console.debug('config object to be sent to highcharts:', this.hcConfig);

            if(this.hcChart) {
                this.hcChart.destroy();
            }
            new Highcharts.Chart(this.hcConfig, function(chart) {
                that.hcChart = chart;
                if(that.testMode) {
                    testingUtils.initializeTestingMetaData(that, that.xFields, that.getClassName());
                    testingUtils.createGlobalReference(that, chart);
                }
                // this event is actually coming from the push state listener in js_charting/js_charting.js
                // we are just using the Highcharts object as a shared event bus
                $(Highcharts).on('baseUriChange.' + that.id, function() {
                    that.$container.find('[clip-path]').each(function() {
                        // just need to unset and reset the clip-path to force a refresh with the new base URI
                        var $this = $(this),
                            clipPath = $this.attr('clip-path');

                        $this.removeAttr('clip-path');
                        $this.attr('clip-path', clipPath);
                    });
                });

                if(callback) {
                    that.benchmark('Draw Finished');
                    callback(that, that.benchmarks);
                }
                // DEBUGGING
                // window.chart = that
            });
        },

        setSize: function(width, height) {
            if(!this.hcChart) {
                return;
            }
            this.hcChart.setSize(width, height, false);
        },

        destroy: function() {
            if(this.hcChart) {
                this.onChartDestroy();
                this.hcChart.destroy();
                this.hcChart = null;
            }
        },

        getSVG: function() {
            var chart = this.hcChart;
            if(this.hcConfig.legend.enabled) {
                if(this.exportMode && chart.type !== 'scatter') {
                    $(chart.series).each(function(i, loopSeries) {
                        if(!loopSeries.legendSymbol) {
                            return false;
                        }
                        loopSeries.legendSymbol.attr({
                            height: 8,
                            translateY: 4
                        });
                    });
                }
                if(chart.legend.nav) {
                    chart.legend.nav.destroy();
                }
            }

            var $svg = $('.highcharts-container').find('svg');
            $svg.siblings().remove();
            $svg.find('.highcharts-tracker').remove();

            // SPL-65745, remove the clip path that is being applied to the legend, or it will cause labels to be hidden
            $svg.find('.highcharts-legend g[clip-path]').each(function() {
                $(this).removeAttr('clip-path');
            });

            return $svg.parent().html();
        },

        /////////////////////////////////////////////////////////////////////////////////////////
        // [end of public interface]

        processProperties: function() {
            Visualization.prototype.processProperties.call(this);

            // handle enabling chart/legend clicks, there are an annoying number of different ways to specify this
            // the "drilldown" property trumps all others
            if(this.properties.hasOwnProperty('drilldown')) {
                this.chartClickEnabled = this.legendClickEnabled = this.properties['drilldown'] === 'all';
            }
            else {
                if(this.properties.hasOwnProperty('chart.clickEnabled')) {
                    this.chartClickEnabled = parsingUtils.normalizeBoolean(this.properties['chart.clickEnabled']);
                }
                else {
                    this.chartClickEnabled = parsingUtils.normalizeBoolean(this.properties['enableChartClick']);
                }
                if(this.properties.hasOwnProperty('chart.legend.clickEnabled')) {
                    this.legendClickEnabled = parsingUtils.normalizeBoolean(this.properties['chart.legend.clickEnabled']);
                }
                else {
                    this.legendClickEnabled = parsingUtils.normalizeBoolean(this.properties['enableLegendClick']);
                }
            }

            if(this.properties['legend.placement'] === 'none') {
                this.hasLegend = false;
            }

            //panning disabled until further notice
            if(false && this.properties['resultTruncationLimit'] > 0){
                this.truncationLimit = this.properties['resultTruncationLimit'];
                this.sliceDataForPanning = true;
            } else if(true || this.properties['resultTruncationLimit'] < 0) {
                this.sliceDataForPanning = false;
            } else {
                var actualPointsPerSeries = (this.dataSet.seriesList.length > 0) ? this.dataSet.seriesList[0].length : 0,
                    maxColumnPoints = ($.browser.msie &&  {"6.0": true, "7.0": true, "8.0": true}.hasOwnProperty($.browser.version)) ? 1000 : 1200,
                    maxLinePoints = 2000,
                    maxPoints = ({line: true, area: true}.hasOwnProperty(this.properties.chart) ? maxLinePoints : maxColumnPoints),
                    numSeries = this.dataSet.fields.length,
                    pointsAllowedPerSeries = Math.floor(maxPoints / numSeries);

                if(actualPointsPerSeries > pointsAllowedPerSeries){
                    this.truncationLimit = pointsAllowedPerSeries;
                    this.sliceDataForPanning = true;
                } else {
                    this.sliceDataForPanning = false;
                }
            }
            if(this.hasXAxis || this.hasYAxis) {
                this.axisColorScheme = {
                    'axis.foregroundColorSoft': this.foregroundColorSoft,
                    'axis.foregroundColorSofter': this.foregroundColorSofter,
                    'axis.fontColor': this.fontColor
                };
            }
            if(this.properties.hasOwnProperty('legend.masterLegend') &&
                    (!this.properties['legend.masterLegend'] || $.trim(this.properties['legend.masterLegend']) === 'null')) {
                this.requiresExternalColors = false;
            }
            this.stackMode = this.properties['chart.stackMode'] || 'default';
            this.legendLabels = parsingUtils.stringToArray(this.properties['legend.labels'] || '[]');
            this.showHideMode = this.properties['data.fieldListMode'] === 'show_hide';
            this.fieldHideList = _.union(
                this.properties['fieldHideList'] || [],
                parsingUtils.stringToArray(this.properties['data.fieldHideList']) || []
            );
            this.fieldShowList = parsingUtils.stringToArray(this.properties['data.fieldShowList']) || [];

            var seriesColorsSetting = this.properties['chart.seriesColors'] || this.properties['seriesColors'];
            this.seriesColors = parsingUtils.stringToHexArray(seriesColorsSetting) || null;
            var fieldColorsSetting = this.properties['chart.fieldColors'] || this.properties['fieldColors'];
            this.internalFieldColors = parsingUtils.stringToObject(fieldColorsSetting || '{}');
            // after the initial parse the field color will be a string representation of a hex number,
            // so loop through and make them true hex numbers (or zero if they can't be parsed)
            _(this.internalFieldColors).each(function(value, key) {
                var hexColor = parseInt(value, 16);
                this.internalFieldColors[key] = _(hexColor).isNaN() ? 0 : hexColor;
            }, this);

            // DEBUGGING: read in yAxisMapping property and yScaleMapping property for multiple y-axes
            this.yAxisMapping = parsingUtils.stringToObject(this.properties['yAxisMapping'] || "{}");
            this.yScaleMapping = parsingUtils.stringToArray(this.properties['yScaleMapping'] || "[]");

            // DEBUGGING: read in xAxisFields and xAxisMapping for multiple x-axes
            this.xAxisFields = parsingUtils.stringToArray(this.properties['xAxisFields'] || "[]");
            this.xAxisMapping = parsingUtils.stringToObject(this.properties['xAxisMapping'] || "{}");

            // DEBUGGING: read in map of field name to series types
            this.seriesTypeMapping = parsingUtils.stringToObject(this.properties['seriesTypeMapping'] || "{}");
        },

        //////////////////////////////////////////////////////////////////////////////////////////////
        // methods for initializing chart components

        initializeFields: function() {
            // TODO: this is where user settings could determine the x-axis field(s)

            var allDataFields = this.dataSet.allDataFields(),
                defaultXField = allDataFields[0];

            this.xFields = (this.xAxisFields && this.xAxisFields.length > 0) ? this.xAxisFields : [defaultXField];

            if(this.isRangeSeriesMode()) {
                var rangeConfig = this.getRangeSeriesConfig();
                allDataFields = _(allDataFields).without(rangeConfig.lower, rangeConfig.upper);
            }
            this.yFields = _(allDataFields).filter(function(field) {
                return (_(this.xFields).indexOf(field) === -1);
            }, this);
            var fieldWhiteList = $.extend([], this.fieldShowList),
                fieldBlackList = $.extend([], this.fieldHideList),
                intersection = _.intersection(fieldWhiteList, fieldBlackList);

            if(this.showHideMode) {
                fieldBlackList = _.difference(fieldBlackList, intersection);
            }
            else {
                fieldWhiteList = _.difference(fieldWhiteList, intersection);
            }

            this.yFields = _.difference(this.yFields, fieldBlackList);
            if(fieldWhiteList.length > 0) {
                this.yFields = _.intersection(this.yFields, fieldWhiteList);
            }
            // handle the user-specified legend labels
            if(this.legendLabels.length > 0) {
                this.yFields = _.union(this.legendLabels, this.yFields);
            }

        },

        isEmpty: function() {
            return (this.yFields.length === 0);
        },

        initializeColorPalette: function() {
            this.colorPalette = new ColorPalette(this.seriesColors);
        },

        initializeSeriesList: function() {
            if(this.isEmpty()) {
                this.seriesList = [seriesFactory.create({ type: this.type })];
                return;
            }

            var rangeFieldName,
                isRangeSeriesMode = this.isRangeSeriesMode();

            if(isRangeSeriesMode) {
                rangeFieldName = this.getRangeSeriesConfig().predicted;
            }
            this.seriesList = _(this.yFields).map(function(field) {
                // TODO: this is where user settings could determine series type and/or axis mappings
                var customType;
                if(rangeFieldName && rangeFieldName === field) {
                    customType = 'range';
                }
                else if(this.seriesTypeMapping.hasOwnProperty(field)) {
                    customType = this.seriesTypeMapping[field];
                }

                var properties = $.extend(true, {}, this.properties, {
                    type: customType || this.type,
                    name: field,
                    stacking: isRangeSeriesMode ? 'default' : this.stackMode,
                    // TODO [sff] should we just deal with this in the chart click handler?
                    clickEnabled: this.chartClickEnabled
                });

                if(customType === 'range') {
                    properties.names = this.getRangeSeriesConfig();
                }

                // DEBUGGING: allow series to be assigned to y-axes via the 'yAxisMapping' property
                if(this.yAxisMapping[field]) {
                    properties.yAxis = parseInt(this.yAxisMapping[field], 10);
                }
                // DEBUGGING: allow series to be assigned to x-axes via the 'xAxisMapping' property
                if(this.xAxisMapping[field]) {
                    properties.xAxis = parseInt(this.xAxisMapping[field], 10);
                }
                return seriesFactory.create(properties);
            }, this);
        },

        initializeXAxisList: function() {
            var isEmpty = this.isEmpty();

            // TODO: this is where user settings could specify multiple x-axes
            // TODO: this is where the x-axis type can be inferred from the series types attached to each axis
            this.xAxisList = _(this.xFields).map(function(field, i) {
                var tickmarksBetween = _(this.seriesList).any(function(series) {
                    return (series.getXAxisIndex() === i && { column: true, bar: true }.hasOwnProperty(series.getType()));
                });
                var axisProperties = $.extend(parsingUtils.getXAxisProperties(this.properties), this.axisColorScheme, {
                    'axis.orientation': this.axesAreInverted ? 'vertical' : 'horizontal',
                    'axisLabels.tickmarkPlacement': tickmarksBetween ? 'between' : 'on',
                    'isEmpty': isEmpty
                });
                if(!axisProperties['axisTitle.text']) {
                    axisProperties['axisTitle.text'] = field;
                }
                if(this.seriesIsTimeBased(field)) {
                    axisProperties['axis.spanData'] = this.dataSet.getSeriesAsFloats('_span');
                    axisProperties['axis.categories'] = this.dataSet.getSeriesAsTimestamps(field);
                    return new TimeAxis(axisProperties);
                }
                axisProperties['axis.categories'] = this.dataSet.getSeries(field);
                return new CategoryAxis(axisProperties);
            }, this);
        },

        initializeYAxisList: function() {
            // TODO: this is where user settings could specify multiple y-axes
            var that = this,
                isEmpty = this.isEmpty();

            this.yAxisList = [];

            // DEBUGGING: for now just make sure there are enough y-axes to satisfy the series
            var maxAxisIndex = _(this.seriesList).chain().invoke('getYAxisIndex').max().value();
            _(maxAxisIndex + 1).times(function(i) {
                var axisProperties = $.extend(parsingUtils.getYAxisProperties(that.properties), that.axisColorScheme, {
                    'axis.orientation': that.axesAreInverted ? 'horizontal' : 'vertical',
                    'isEmpty': isEmpty
                });

                // FIXME: we should do something more intelligent here
                // currently if there is only one series for an axis, use that series's name as the default title
                if(!axisProperties['axisTitle.text']) {
                    var axisSeries = _(that.seriesList).filter(function(series) { return series.getYAxisIndex() === i; });
                    if(axisSeries.length === 1) {
                        axisProperties['axisTitle.text'] = axisSeries[0].getName();
                    }
                }
                
                // log scale is not respected if the chart has stacking
                if(that.stackMode !== 'default') {
                    axisProperties['axis.scale'] = 'linear';
                }

                // DEBUGGING: use the yScaleMapping list to assign a different scale to each y-axis
                if(that.yScaleMapping && that.yScaleMapping.length > i) {
                    axisProperties['axis.scale'] = that.yScaleMapping[i];
                }

                that.yAxisList.push(new NumericAxis(axisProperties));
            });

        },

        initializeLegend: function() {
            var legendProps = parsingUtils.getLegendProperties(this.properties);
            if(_(legendProps['clickEnabled']).isUndefined()) {
                legendProps['clickEnabled'] = this.legendClickEnabled;
            }
            $.extend(legendProps, {
                fontColor: this.fontColor
            });
            this.legend = new Legend(legendProps);

            var that = this,
                properties = {
                    highlightDelay: 125,
                    unhighlightDelay: 50,
                    onMouseOver: function(fieldName) {
                        that.handleLegendMouseOver(fieldName);
                    },
                    onMouseOut: function(fieldName) {
                        that.handleLegendMouseOut(fieldName);
                    }
                },
                throttle = new HoverEventThrottler(properties);

            this.legend.on('mouseover', function(e, fieldName) {
                throttle.mouseOverHappened(fieldName);
            });
            this.legend.on('mouseout', function(e, fieldName) {
                throttle.mouseOutHappened(fieldName);
            });
            this.legend.on('click', function(e, fieldName) {
                that.handleLegendClick(e, fieldName);
            });
        },

        initializePanning: function(){
            var panningProperties = {
                truncationLimit: this.truncationLimit,
                numSeries: this.dataSet.seriesList.length - 2,
                seriesSize: this.dataSet.seriesList[0].length,
                container: this.$container
            };
            this.panning = new PanningScrollbar(panningProperties);
        },

        initializeTooltip: function() {
            var tooltipProps = {
                borderColor: this.foregroundColorSoft
            };
            this.tooltip = new Tooltip(tooltipProps);
        },

        setAllSeriesData: function() {
            var that = this,
                numberOfSeries = this.seriesList.length,
                totalDataPoints = numberOfSeries * this.dataSet.seriesList[0].length;

            _(this.seriesList).each(function(series) {
                if(series.getType() === 'range') {
                    this.setRangeSeriesData(series);
                } else if (that.sliceDataForPanning && (totalDataPoints > that.truncationLimit)) {
                    this.needsPanning = true;
                    this.setBasicSeriesData(series);
                } else {
                    this.setBasicSeriesData(series);
                }
            }, this);
        },

        setBasicSeriesData: function(series) {
            var xInfo = this.getSeriesXInfo(series),
                yInfo = this.getSeriesYInfo(series);

            if(xInfo.axis instanceof NumericAxis) {
                series.setData({
                    x: this.formatAxisData(xInfo.axis, xInfo.fieldName),
                    y: this.formatAxisData(yInfo.axis, yInfo.fieldName)
                });
            }
            else if(xInfo.axis instanceof TimeAxis) {
                // SPL-67612, handle the case where the last data point was a total value
                // the axis handlers will have removed it from the timestamps, so we just have to sync the array lengths
                var axisTimestamps = xInfo.axis.getCategories(),
                    rawData = this.formatAxisData(yInfo.axis, yInfo.fieldName);

                series.setData({
                    y: rawData.slice(0, axisTimestamps.length)
                });
            }
            else {
                series.setData({
                    y: this.formatAxisData(yInfo.axis, yInfo.fieldName)
                });
            }
            if(this.needsPanning){
                this.sliceSeriesForPanning(series, this.truncationLimit);
            }
        },

        setRangeSeriesData: function(series) {
            var xInfo = this.getSeriesXInfo(series),
                yInfo = this.getSeriesYInfo(series),
                rangeConfig = this.getRangeSeriesConfig(),
                rangeData = {
                    predicted: this.formatAxisData(yInfo.axis, rangeConfig.predicted),
                    lower: this.formatAxisData(yInfo.axis, rangeConfig.lower),
                    upper: this.formatAxisData(yInfo.axis, rangeConfig.upper)
                };

            if(xInfo.axis instanceof NumericAxis) {
                rangeData.x = this.formatAxisData(xInfo.axis, xInfo.fieldName);
            }
            series.setData(rangeData);
        },

        bindSeriesEvents: function() {
            var that = this,
                properties = {
                    highlightDelay: 125,
                    unhighlightDelay: 50,
                    onMouseOver: function(point, series) {
                        that.handlePointMouseOver(point, series);
                    },
                    onMouseOut: function(point, series) {
                        that.handlePointMouseOut(point, series);
                    }
                },
                throttle = new HoverEventThrottler(properties);

            _(this.seriesList).each(function(series) {
                series.on('mouseover', function(e, targetPoint, targetSeries) {
                    throttle.mouseOverHappened(targetPoint, targetSeries);
                });
                series.on('mouseout', function(e, targetPoint, targetSeries) {
                    throttle.mouseOutHappened(targetPoint, targetSeries);
                });
                series.on('click', function(e, targetPoint, targetSeries) {
                    that.handlePointClick(e, targetPoint, targetSeries);
                });
            });

        },

        bindPanningEvents: function() {
            var that = this;
            this.panning.on('scroll', function(e, windowToDisplay){
                that.panningScroll = true;
                that.windowToDisplay = windowToDisplay;
                that.setAllSeriesData();
                that.setPannedData();
                that.setPannedLabels();
                that.hcChart.redraw();
            });
        },

        setPannedData: function(){
            var that = this;
            _(this.hcChart.series).each(function(oldSeries, i){
                oldSeries.setData(that.seriesList[i].data, false);
            });
        },

        setPannedLabels: function() {
            var i,
                originalLabels,
                originalSpans,
                timeCategoryInfo;

            originalLabels = this.dataSet.getSeries('_time');
            originalSpans  = this.dataSet.getSeriesAsFloats('_span');
            originalLabels = originalLabels
                .slice(this.windowToDisplay,
                (this.windowToDisplay)
                    + this.pointsToPlotPerSeries);
            originalSpans = originalSpans
                .slice(this.windowToDisplay,
                (this.windowToDisplay)
                    + this.pointsToPlotPerSeries);
            timeCategoryInfo = timeUtils.convertTimeToCategories(originalLabels, originalSpans, 6);

            //FIXME: we need to handle the bizarre negative indexes when creating the lookup
            //for tooltip formatting (this carries over)
            this.categoryIndexLookup = timeCategoryInfo.categories;

            for(i = -1; i<timeCategoryInfo.categories.length; i++){
                if(i!==timeCategoryInfo.categories.length-1){
                    timeCategoryInfo.categories[i] = timeCategoryInfo.categories[i+1];
                }
            }

            _(this.hcChart.xAxis).each(function(xAxis) {
                xAxis.setCategories(timeCategoryInfo.categories, false);
            });

        },

        handlePointClick: function(event, point, series) {
            var rowContext = {},
                pointIndex = point.index,
                pointInfo = this.getSeriesPointInfo(series, point.index),
                pointClickEvent = {
                    type: 'pointClick',
                    modifierKey: event.modifierKey,
                    name: pointInfo.xAxisName,
                    // 'value' will be inserted later based on the x-axis type
                    name2: pointInfo.yAxisName,
                    value2: pointInfo.yValue
                };

            if(pointInfo.xAxisIsTime) {
                var isoTimeString = this.dataSet.getSeries(pointInfo.xAxisName)[pointIndex];
                pointClickEvent.value = splunkUtils.getEpochTimeFromISO(isoTimeString);
                pointClickEvent._span = this.dataSet.getSeriesAsFloats('_span')[pointIndex];
                rowContext['row.' + pointInfo.xAxisName] = pointClickEvent.value;
            }
            else {
                pointClickEvent.value = pointInfo.xValue;
                rowContext['row.' + pointInfo.xAxisName] = pointInfo.xValue;
            }

            _(this.yFields).each(function(fieldName) {
                rowContext['row.' + fieldName] = this.dataSet.getSeries(fieldName)[pointIndex];
            }, this);
            pointClickEvent.rowContext = rowContext;
            this.trigger(pointClickEvent);
        },

        handlePointMouseOver: function(targetPoint, targetSeries) {
            _(this.seriesList).each(function(series) {
                if(series.matchesName(targetSeries.getName())) {
                    series.handlePointMouseOver(targetPoint);
                }
                else {
                    series.unHighlight();
                }
            });
            if(this.legend) {
                this.legend.selectField(targetSeries.getLegendKey());
            }
        },

        handlePointMouseOut: function(targetPoint, targetSeries) {
            _(this.seriesList).each(function(series) {
                if(series.matchesName(targetSeries.getName())) {
                    series.handlePointMouseOut(targetPoint);
                }
                else {
                    series.highlight();
                }
            });
            if(this.legend) {
                this.legend.unSelectField(targetSeries.getLegendKey());
            }
        },

        handleLegendClick: function(event, fieldName) {
            var legendClickEvent = {
                type: 'legendClick',
                modifierKey: event.modifierKey,
                name2: fieldName
            };
            this.trigger(legendClickEvent);
        },

        handleLegendMouseOver: function(fieldName) {
            _(this.seriesList).each(function(series) {
                if(series.matchesName(fieldName)) {
                    series.handleLegendMouseOver(fieldName);
                }
                else {
                    series.unHighlight();
                }
            });
        },

        handleLegendMouseOut: function(fieldName) {
            _(this.seriesList).each(function(series) {
                if(series.matchesName(fieldName)) {
                    series.handleLegendMouseOut(fieldName);
                }
                else {
                    series.highlight();
                }
            });
        },

        applyColorPalette: function() {
            if(this.isEmpty()) {
                return;
            }
            var colorMapping = {};
            _(this.getFieldList()).each(function(field, i, fieldList) {
                colorMapping[field] = this.computeFieldColor(field, i, fieldList);
            }, this);
            _(this.seriesList).invoke('applyColorMapping', colorMapping);
        },

        //////////////////////////////////////////////////////////////////////////////////
        // methods for generating config objects from chart objects

        getSeriesConfigList: function() {
            return _(this.seriesList).chain().invoke('getConfig').flatten(true).value();
        },

        getXAxisConfig: function() {
            if(!this.hasXAxis) {
                return [];
            }
            return _(this.xAxisList).map(function(axis, i) {
                var config = axis.getConfig();
                if(i > 0) {
                    config.offset = 40;
                }
                return config;
            }, this);
        },

        getYAxisConfig: function() {
            if(!this.hasYAxis) {
                return [];
            }
            return _(this.yAxisList).map(function(axis, i) {
                var config = axis.getConfig();
                if(i % 2 !== 0) {
                    config.opposite = true;
                }
                return config;
            });
        },

        getLegendConfig: function() {
            if(!this.hasLegend || !this.legend) {
                return {};
            }
            return this.legend.getConfig();
        },

        getTooltipConfig: function() {
            if(!this.tooltip) {
                return {};
            }
            var that = this;
            return $.extend(this.tooltip.getConfig(), {
                formatter: function() {
                    // need to look up the instance of Splunk.JSCharting.BaseSeries, not the HighCharts series
                    var series = this.series.splSeries;
                    return that.formatTooltip(series, this.point);

                }
            });
        },

        formatTooltip: function(series, hcPoint) {
            var pointInfo;
            if(this.panningScroll){
                pointInfo = this.getSeriesPointInfo(series, this.categoryIndexLookup.indexOf(hcPoint.category) + this.windowToDisplay);
            } else {
                pointInfo = this.getSeriesPointInfo(series, hcPoint.index);
            }
            return series.getTooltipHtml(pointInfo, this.hcChart);
        },

        getChartConfig: function() {
            var config = {
                type: this.type,
                renderTo: this.container,
                backgroundColor: this.backgroundColor,
                borderColor: this.backgroundColor
            };
            // in export mode we need to set explicit width and height
            // we'll honor the width and height of the parent node, unless they are zero
            if(this.exportMode) {
                config.width = this.width || this.EXPORT_WIDTH;
                config.height = this.height || this.EXPORT_HEIGHT;
            }
            return config;
        },

        getPlotOptionsConfig: function() {
            return $.extend(true, {}, this.BASE_PLOT_OPTIONS_CONFIG, {
                series: {
                    cursor: this.chartClickEnabled ? 'pointer' : 'default',
                    enableMouseTracking: this.useMouseTracking()
                }
            });
        },

        ////////////////////////////////////////////////////////////////////////////////////////
        // methods for managing event handlers and effects

        addEventHandlers: function() {
            var that = this;
            this.hcConfig.chart.events = {
                load: function(e) {
                    var chart = this;
                    that.onChartLoad(chart, e);
                    that.onChartLoadOrResize(chart, e);
                },
                redraw: function(e) {
                    var chart = this;
                    that.onChartResize(chart, e);
                    that.onChartLoadOrResize(chart, e);
                }
            };
        },

        onChartLoad: function(chart) {
            if(this.legend) {
                this.legend.onChartLoad(chart);
            }
            if(this.panning){
                this.panning.onChartLoad(chart);
            }
            if(this.xAxisList) {
                _(this.xAxisList).each(function(axis) {
                    axis.onChartLoad(chart);
                });
            }
            if(this.yAxisList) {
                _(this.yAxisList).each(function(axis) {
                    axis.onChartLoad(chart);
                });
            }
            _(this.seriesList).each(function(series) {
                series.onChartLoad(chart);
            });
            if(!this.useMouseTracking()) {
                this.addPointListeners(chart);
            }
        },

        onChartResize: function() { },

        onChartLoadOrResize: function(chart) {
            if(this.panning){
                this.panning.onChartLoadOrResize(chart);
            }
            if(this.xAxisList) {
                _(this.xAxisList).each(function(axis) {
                    axis.onChartLoadOrResize(chart);
                });
            }
            if(this.yAxisList) {
                _(this.yAxisList).each(function(axis) {
                    axis.onChartLoadOrResize(chart);
                });
            }
        },

        onChartDestroy: function() {
            $(Highcharts).off('baseUriChange.' + this.id);
            if(this.legend) {
                this.legend.destroy();
            }
            if(this.xAxisList) {
                _(this.xAxisList).each(function(axis) {
                    axis.destroy();
                });
            }
            if(this.yAxisList) {
                _(this.yAxisList).each(function(axis) {
                    axis.destroy();
                });
            }
            _(this.seriesList).each(function(series) {
                series.destroy();
            });
            if(!this.useMouseTracking()) {
                // clean up any listeners created in addPointListeners
                $(this.hcChart.container).off('.splunk_jscharting');
            }
        },

        addPointListeners: function(chart) {
            var $chartContainer = $(chart.container),
                tooltipSelector = ".highcharts-tooltip *",
                hoveredPoint = null,
                // helper function to extract the corresponding point from a jQuery element
                extractPoint = function($el) {
                    return (chart.series[$el.attr('data-series-index')].data[$el.attr('data-point-index')]);
                };

            // with the VML renderer, have to explicitly destroy the tracker so it doesn't block mouse events
            if(!domUtils.hasSVG && chart.tracker) {
                chart.tracker.destroy();
            }
            // create a closure around the tooltip hide method so that we can make sure we always hide the selected series when it is called
            // this is a work-around for the situation when the mouse moves out of the chart container element without triggering a mouse event
            if(chart.tooltip) {
                var tooltipHide = _(chart.tooltip.hide).bind(chart.tooltip);
                chart.tooltip.hide = function(silent) {
                    tooltipHide();
                    if(!silent && hoveredPoint) {
                        hoveredPoint.firePointEvent('mouseOut');
                        hoveredPoint = null;
                    }
                };
            }

            // decorate each point element with the info we need to map it to its corresponding data object
            var that = this;
            $(chart.series).each(function(i, series) {
                // SPL-71093, with the VML renderer explicitly hide the marker group so it doesn't block mouse events
                if(!domUtils.hasSVG && series.markerGroup && chart.options.chart.type !== 'scatter') {
                    // it's ok to reference the element property because we know this will be VML mode
                    $(series.markerGroup.element).hide();
                }
                $(series.data).each(function(j, point) {
                    if(point.graphic && point.graphic.element) {
                        var $element = $(point.graphic.element);
                        $element.attr('data-series-index', i);
                        $element.attr('data-point-index', j);
                        if(that.chartClickEnabled) {
                            $element.css('cursor', 'pointer');
                        }
                    }
                });
            });
            // we are not using mouse trackers, so attach event handlers to the chart's container element
            $chartContainer.on('click.splunk_jscharting', function(event) {
                var $target = $(event.target);
                if($target.is(that.POINT_CSS_SELECTOR)) {
                    var point = extractPoint($target);
                    point.firePointEvent('click', event);
                }
            });
            // handle all mouseover events in the container here
            // if they are over the tooltip, ignore them (this avoids the dreaded tooltip flicker)
            // otherwise hide any point that is currently in a 'hover' state and 'hover' the target point as needed
            $chartContainer.on('mouseover.splunk_jscharting', function(event) {
                var $target = $(event.target);
                if($target.is(tooltipSelector)) {
                    return;
                }
                var point = null;
                if($target.is(that.POINT_CSS_SELECTOR)) {
                    point = extractPoint($target);
                }
                if(hoveredPoint && !(point && hoveredPoint === point)) {
                    hoveredPoint.firePointEvent('mouseOut');
                    chart.tooltip.hide(true);
                    hoveredPoint = null;
                }
                if(point) {
                    point.firePointEvent('mouseOver');
                    chart.tooltip.refresh(point);
                    hoveredPoint = point;
                }
            });

        },

        /////////////////////////////////////////////////////////////////////////////////////
        // re-usable helpers

        getSeriesXInfo: function(series) {
            var xIndex = series.getXAxisIndex();
            return ({
                axis: this.xAxisList[xIndex],
                fieldName: this.xFields[xIndex]
            });
        },

        getSeriesYInfo: function(series) {
            return ({
                axis: this.yAxisList[series.getYAxisIndex()],
                fieldName: series.getName()
            });
        },

        getSeriesPointInfo: function(series, pointIndex) {
            var xInfo = this.getSeriesXInfo(series),
                yInfo = this.getSeriesYInfo(series),
                xSeries = this.dataSet.getSeries(xInfo.fieldName),
                ySeries = this.dataSet.getSeries(yInfo.fieldName);

            return ({
                xAxisIsTime: (xInfo.axis instanceof TimeAxis),
                xAxisName: xInfo.fieldName,
                xValue: xSeries[pointIndex],
                xValueDisplay: xInfo.axis.formatValue(xSeries[pointIndex]),
                yAxisName: yInfo.fieldName,
                yValue: ySeries[pointIndex],
                yValueDisplay: yInfo.axis.formatValue(ySeries[pointIndex])
            });
        },

        isRangeSeriesMode: function() {
            return (this.dataSet.hasField('_predicted')
                && this.dataSet.hasField('_lower')
                && this.dataSet.hasField('_upper'));
        },

        getRangeSeriesConfig: function() {
            var predictedName = _(this.dataSet.getSeries('_predicted')).find(function(value) { return !!value; }),
                lowerName = _(this.dataSet.getSeries('_lower')).find(function(value) { return !!value; }),
                upperName = _(this.dataSet.getSeries('_upper')).find(function(value) { return !!value; });

            return ({
                predicted: predictedName,
                lower: lowerName,
                upper: upperName
            });
        },

        useMouseTracking: function() {
            // iOS Safari does not support event delegation bubbling up the DOM for non-links like every other
            // browser does. This was causing drilldown click events getting lost on iPhone/iPad (SPL-66456).
            // So if the user is on one of those devices, just use Highcharts' mouse tracking.
            if(userAgent.isSafariiPhone() || userAgent.isSafariiPad()) {
                return true;
            }
            return _(this.seriesList).any(function(series) {
                return series.getType() in { line: true, area: true, range: true };
            });
        },

        // by convention, we consider a series to be time-based if it is called _time, and there is also a _span series
        // with the exception that if there is only one data point _span does not need to be there
        seriesIsTimeBased: function(fieldName) {
            return (/^_time/).test(fieldName) && (this.dataSet.getSeries(fieldName).length <= 1 || this.dataSet.hasField('_span'));
        },

        sliceSeriesForPanning: function(series, truncationLimit) {
            var  numberOfSeries = this.seriesList.length;

            if(numberOfSeries > truncationLimit){
                this.pointsToPlotPerSeries = 1;
            } else {
                this.pointsToPlotPerSeries = Math.ceil(truncationLimit / numberOfSeries);
            }

            if(this.windowToDisplay){
                series.sliceData(series,
                    this.windowToDisplay,
                    (this.windowToDisplay) + this.pointsToPlotPerSeries);
            } else {
                series.sliceData(series, 0, this.pointsToPlotPerSeries);
            }

        },

        formatAxisData: function(axis, fieldName) {
            if(!this.dataSet.hasField(fieldName)) {
                return [];
            }
            return this.dataSet.getSeriesAsFloats(fieldName, {
                scale: axis.isLogScale() ? 'log' : 'linear',
                nullValueMode: this.properties['chart.nullValueMode']
            });
        },

        computeFieldColor: function(field, index, fieldList) {
            if(this.internalFieldColors.hasOwnProperty(field)) {
                return colorUtils.colorFromHex(this.internalFieldColors[field]);
            }
            var useExternalPalette = !_(this.externalPaletteMapping).isEmpty(),
                paletteIndex = useExternalPalette ? this.externalPaletteMapping[field] : index,
                paletteSize = useExternalPalette ? this.externalPaletteSize : fieldList.length;

            return this.colorPalette.getColorAsRgb(field, paletteIndex, paletteSize);
        },

        /////////////////////////////////////////////////////////////////////////////////////
        // templates and default settings

        BASE_CONFIG: {
            chart: {
                animation: false,
                showAxes: true,
                reflow: false
            },
            credits: {
                enabled: false
            },
            legend: {
                enabled: false
            },
            plotOptions: {
                series: {
                    animation: false,
                    events: {
                        legendItemClick: function() {
                            return false;
                        }
                    },
                    borderWidth: 0,
                    shadow: false,
                    turboThreshold: 0
                }
            },
            title: {
                text: null
            },
            tooltip: {
                enabled: false,
                useHTML: true
            }
        },

        BASE_PLOT_OPTIONS_CONFIG: {
            line: {
                stickyTracking: true,
                marker: {
                    states: {
                        hover: {
                            enabled: true,
                            radius: 6
                        }
                    },
                    symbol: 'square',
                    radius: 6
                }
            },
            area: {
                stickyTracking: true,
                lineWidth: 1,
                trackByArea: true,
                marker: {
                    states: {
                        hover: {
                            enabled: true,
                            radius: 6
                        }
                    },
                    symbol: 'square',
                    radius: 6,
                    enabled: false
                }
            },
            column: {
                markers: {
                    enabled: false
                },
                stickyTracking: false
            },
            bar: {
                markers: {
                    enabled: false
                },
                stickyTracking: false
            }
        }

    });

   return Chart;

});
