define(
    [
        'jquery',
        'underscore',
        'backbone',
        'models/services/search/Job',
        'helpers/polling_manager',
        'util/splunkd_utils',
        'util/time_utils',
        'util/console',
        'splunk.util',
        'helpers/Session',
        'util/Ticker'
    ],
    function($, _, Backbone, SearchJob, polling_manager, splunkd_utils, time_utils, console, splunkUtil, Session, Ticker) {
        
        // -------------- private helpers --------------- //
        
        var DEFAULT_POLLING_OPTIONS = {
                delay: SearchJob.DEFAULT_POLLING_INTERVAL,
                condition: function(model) {
                    return ((model.entry.content.get('isDone') !== true) && (model.entry.content.get('isFailed') !== true));
                }
            };
        
        // a helper object to encapsulate the behavior of a job-associated observable model or collection
        var ManagedChildObject = function(observable, fetchParams, parent) {
            this.observable = observable;
            this.fetchParams = fetchParams || {};
            var oldFetch = observable.fetch;

            observable.fetch = function(options) {
                options = options || {};
                options.sid = parent.id;
                return oldFetch.call(observable, options);
            };
            this.xhr = null;
        };
        
        ManagedChildObject.prototype = {
                
            fetch: function(baseParams) {
                this.abortXhr();
                var customParams = _(this.fetchParams).isFunction() ? this.fetchParams() : this.fetchParams;
                this.xhr = this.observable.fetch($.extend(true, {}, customParams, baseParams));
            },

            destroy: function() {
                this.abortXhr();
                if(this.observable instanceof Backbone.Model) {
                    this.observable.clear();
                }
                else if(this.observable instanceof Backbone.Collection) {
                    this.observable.reset();
                }
                this.observable.trigger('destroy');
            },

            getObservable: function() {
                return this.observable;
            },

            abortXhr: function() {
                if(this.xhr && _(this.xhr.abort).isFunction()) {
                    this.xhr.abort();
                }
                this.xhr = null;
            }

        };
        
        var JobModel = SearchJob.extend({
            initialize: function(attributes, options) {
                SearchJob.prototype.initialize.apply(this, arguments);
                options = options || {};
                
                this.enableSimplePoller = options.enableSimplePoller || false;
                
                //poller setup
                this.pollingOptions = $.extend(true, {}, DEFAULT_POLLING_OPTIONS, options);
                
                this.prepared = (this.entry.content.get('dispatchState') && !this.isPreparing());
                this.processKeepAlive = options.processKeepAlive || false;
                this.keepAlivePoller = null;
                this.keepAliveInterval = options.keepAliveInterval || SearchJob.DEFAULT_KEEP_ALIVE_INTERVAL;

                this.entry.content.on('change', function() {
                    var changedAttributes = this.entry.content.changedAttributes(),
                        previousAttributes = this.entry.content.previousAttributes();

                    this.handleJobProgress(changedAttributes, previousAttributes);
                }, this);
                
                this.entry.content.on("change:dispatchState", function() {
                    if (this.entry.content.get('dispatchState') && !this.isPreparing() && !this.prepared) {
                        this.prepared = true;
                        this.trigger("prepared");
                    }
                }, this);

                //new simplistic polling with safeFetch and a ticker
                if (this.enableSimplePoller) {
                    
                    //poll aggressively to begin with
                    this.ticker = new Ticker(100);
                    this.ticker.on('tick', function() {
                        this.safeFetch({
                            data: {
                                app: this.entry.acl.get('app'),
                                owner: this.entry.acl.get('owner'),
                                sharing: this.entry.acl.get('sharing')
                            }
                        });
                    }, this);
                    
                    this.entry.content.on('change:dispatchState', function() {                        
                        //stop polling when the job is done
                        if (this.entry.content.get('isDone')) {
                            this.ticker.stop();
                        }
                    }, this);
                    
                    this.on("prepared", function() {
                        var attemptCount = 0;
                        var maxAttempts = 30;
                       
                        var _private_progress = function() {
                            var count;
                            attemptCount++;
                            
                            if (this.entry.content.get('isDone')) {
                                this.off('sync', _private_progress, this);
                                return;
                            }
                            
                            if (this.isReportSearch()) {
                                if (this.entry.content.get('isPreviewEnabled')) {
                                    count = this.entry.content.get('resultPreviewCount');
                                } else {
                                    count = this.entry.content.get('resultCount');
                                }   
                            } else {
                                count = this.entry.content.get('eventAvailableCount');
                            }
                            
                            if ((count > 0) || (attemptCount === maxAttempts)) {
                                this.off('sync', _private_progress, this);
                                this.trigger('slowDownPoller');
                            }
                            
                        }.bind(this);
                        
                        this.on('slowDownPoller', function() {
                            this.ticker.reset({
                                delay: this.pollingOptions.delay, 
                                force: true
                            });
                        }, this);
                        
                        this.on('sync', _private_progress, this);
                    }, this);
                    
                //traditional complex and not intuitive 
                } else {
                    this.jobPoller = polling_manager.getPoller(this, this.pollingOptions);
                    this.childObjects = [];
                    this.preserveChildrenOnDestroy = options.preserveChildrenOnDestroy || false;
                    this.entry.acl.on('change', function() {
                        //update the job poller to be able to fetch from the correct location
                        $.extend(this.jobPoller.options, {
                            data: {
                                app: this.entry.acl.get("app"),
                                owner: this.entry.acl.get("owner"),
                                sharing: this.entry.acl.get("sharing")
                            }
                        });
                    }, this);
                }
                
                this.error.on("change", function(){
                    if (splunkd_utils.messagesContainsOneOfTypes(this.error.get("messages"), [splunkd_utils.NOT_FOUND])) {
                        this.handleJobDestroy();
                    }
                }, this);
                
                this.entry.content.on("change:isFailed", function(){
                    if (this.entry.content.get("isFailed") === true) {
                        this.handleJobDestroy();
                    }
                }, this);
                
                this.on('destroy', this.handleJobDestroy, this);
            },
            
            handleJobProgress: function(changed, previous) {
                if (this.isNew()) {
                    return false;
                }
                
                var jobIsRealTime = this.entry.content.get('isRealTimeSearch'),
                    jobIsReportSearch = this.entry.content.get('reportSearch'),
                    jobAdHocModeIsVerbose = (this.getAdhocSearchMode() === splunkd_utils.VERBOSE),
                    jobPreviewEnabled = this.entry.content.get('isPreviewEnabled'),
                    jobIsDoneInModel = this.entry.content.get('isDone'),
                    changeIsEmpty = _.isEmpty(changed);
                
                // if the job is not previewable and also not done, ignore the progress 
                if (
                    changeIsEmpty ||
                    (!jobIsDoneInModel && jobIsReportSearch && !jobPreviewEnabled && !jobAdHocModeIsVerbose)
                ) {
                    return false;
                }
                
                // examine changes to the job model and determine if the child objects should be updated
                var scanCountChanged = !_(changed.scanCount).isUndefined() && (changed.scanCount > previous.scanCount),
                    eventCountChanged = !_(changed.eventCount).isUndefined(),
                    resultCountChanged = !_(changed.resultCount).isUndefined(),
                    jobIsDone = !_(changed.isDone).isUndefined() && (changed.isDone === true),
                    jobIsUpdated = scanCountChanged || eventCountChanged || resultCountChanged || jobIsDone || jobIsRealTime;

                if (!jobIsUpdated) {
                    return false;
                }
                
                //we have determined that the job has been updated, so go and fetch the managed child object
                this.trigger("jobProgress");
                
                if (!this.enableSimplePoller) {
                    this.fetchManagedChildren();
                }
                
                if (jobIsDone && this.processKeepAlive) {
                    this.startKeepAlive();
                }
                
                return true;
            },
            
            handleJobDestroy: function() {
                if (!this.enableSimplePoller) {
                    if (!this.preserveChildrenOnDestroy) {
                        _(this.childObjects).each(function(observer) {
                            observer.destroy();
                        });
                    }
                }
                
                this.stopPolling();
                this.processKeepAlive = false;
                this.stopKeepAlive();
            },
            
            startKeepAlive: function() {
                //remove/add session observers to start/stop the keep alive poller based on UI session 
                Session.off('timeout', null, this);
                Session.off('start', null, this);
                Session.on('timeout', this.stopKeepAlive, this);
                Session.on('start', this.startKeepAlive, this);
                
                this.stopKeepAlive();//ensure you never create more than one keep alive poller
                this.keepAlivePoller = new Ticker(this.keepAliveInterval);
                this.keepAlivePoller.on('tick', function() {
                    this.control.save({}, {
                        data: {
                            action: 'touch' 
                        },
                        success: function(model, response, options) {
                            console.log('touched job:', model.id);
                        }.bind(this),
                        error: function(model, response, options) {
                            if (response.hasOwnProperty('status') && (response.status == 0 || response.status == 12029)) {
                                return;
                            }
                            console.log("error touching job (stopping keep alive):", model.id);
                            this.keepAlivePoller.stop();
                        }.bind(this)
                    });
                }, this);
                this.keepAlivePoller.start();
            },
            
            stopKeepAlive: function() {
                if (this.keepAlivePoller) {
                    this.keepAlivePoller.stop();
                    this.keepAlivePoller.off();
                }
            },
            
            startPolling: function(force) {
                //remove/add session observers to start/stop the job poller based on UI session
                Session.off('timeout', null, this);
                Session.off('start', null, this);
                Session.on('timeout', this.stopPolling, this);
                Session.on('start', function() {
                    this.startPolling(true);//force a new update from the server
                }, this);
                
                if (this.enableSimplePoller) {
                    if (force || (!this.entry.content.get('isDone') && !this.entry.content.get('isFailed'))) {
                        this.ticker.start(true);
                    }
                } else {
                    this.jobPoller.start();
                }
            },
            
            stopPolling: function() {
                if (this.enableSimplePoller) {
                    this.ticker.stop();
                } else {
                    this.jobPoller.stop();
                }
            },
            
            noParamControlUpdate: function(action, options) {
                if(options && options.data) {
                    delete options.data;
                }
                return this.control.save({}, $.extend(true, options, { data: { action: action } }));                
            },

            pause: function(options) {
                return this.noParamControlUpdate("pause", options);
            },

            unpause: function(options) {
                return this.noParamControlUpdate("unpause", options);
            },

            finalize: function(options) {
                return this.noParamControlUpdate("finalize", options);
            },
            
            cancel: function(options) {
                return this.noParamControlUpdate("cancel", options);
            },
            
            touch: function(options) {
                return this.noParamControlUpdate("touch", options);
            },
            
            setTTL: function(ttl, options) {
                if(options && options.data) {
                    delete options.data;
                }
                return this.control.save({}, $.extend(true, options, { data: { action: 'setttl', ttl: ttl} }));
            },
            
            setPriority: function(priority, options) {
                if(options && options.data) {
                    delete options.data;
                }
                return this.control.save({}, $.extend(true, options, { data: { action: 'setpriority', priority: priority} } ));
            },
            
            enablePreview: function(options) {
                return this.noParamControlUpdate("enablepreview", options);
            },
            
            disablePreview: function(options) {
                return this.noParamControlUpdate("disablepreview", options);
            },
            
            saveControlUpdate: function(action, options) {
                options = options || {};
                options.data = options.data || {};
                
                var data = {
                    action: action,
                    auto_cancel: options.data.auto_cancel,
                    auto_pause: options.data.auto_pause,
                    email_list: options.data.email_list,
                    email_subject: options.data.email_subject,
                    email_results: options.data.email_results,
                    ttl: options.data.ttl
                }; 
                
                if(options && options.data) {
                    delete options.data;
                }
                
                return this.control.save({}, $.extend(true, options, {data: data}));               
            },
            
            saveJob: function(options) {                
                return this.saveControlUpdate("save", options);
            },
            
            unsaveJob: function(options) {
                return this.saveControlUpdate("unsave", options);
            },
            
            makeWorldReadable: function(options) {
                var owner = this.entry.acl.get("owner"),
                    data = {
                        sharing: splunkd_utils.GLOBAL,
                        owner: this.entry.acl.get("owner"),
                        'perms.read': "*"                
                    };
                
                if (options && options.data) {
                    delete options.data;
                }
                
                return this.acl.save({}, $.extend(true, options, { data: data }));                
            },
            
            undoWorldReadable: function(options) {
                var owner = this.entry.acl.get("owner"),
                    data = {
                        sharing: splunkd_utils.GLOBAL,
                        owner: owner,
                        'perms.read': ""              
                    };
                
                if (options && options.data) {
                    delete options.data;
                }
                
                return this.acl.save({}, $.extend(true, options, { data: data }));                
            },
            
            isSharedAccordingToTTL: function(defaultSaveTTL) {
                var perms = this.entry.acl.permsToObj();
                if ((perms.read.indexOf("*") != -1) && (this.entry.content.get("ttl") === defaultSaveTTL)) {
                    return true;
                }
                return false;
            },
            
            saveIsBackground: function(options) {
                this.entry.content.custom.set("isBackground", "1");
                return this.save({}, options);
            },
            
            isBackground: function() {
                return splunkUtil.normalizeBoolean(this.entry.content.custom.get("isBackground"));
            }, 
            
            /**
             * Register an external model or collection to be refreshed whenever the job has progress.
             *
             * @param observable {Model | Collection} the model or collection to register
             * @param fetchParams {Object | Function} at fetch time the object or the result of calling the function will be merged with the default fetch parameters
             */

            register: function(observable, fetchParams) {
                var childObject = new ManagedChildObject(observable, fetchParams, this);
                this.childObjects.push(childObject);
                if(this.entry.content.get('isDone') === true) {
                    childObject.fetch({ sid: this.id });
                }
            },
            
            /**
             * Explicitly triggers a fetch of a previously-registered model or collection
             *
             * @param observable {Model | Collection} to be looked up in the instance-level childObjects array
             */

            fetchManagedChild: function(observable) {
                var childObject = _(this.childObjects).find(function(object) { return object.getObservable() === observable; });
                if(!childObject) {
                    console.error('Search Job Manager: can only refresh an object that has already been registered', observable);
                    return false;
                }
                childObject.fetch({ sid: this.id });
            },
            
            fetchManagedChildren: function() {
                _(this.childObjects).each(function(observer) {
                    observer.fetch({ sid: this.id });
                }.bind(this));               
            },
            
            resultCountSafe: function() {
                return (this.entry.content.get('isPreviewEnabled') && !this.entry.content.get('isDone')) ? this.entry.content.get('resultPreviewCount') : this.entry.content.get('resultCount');
            },
            
            eventAvailableCountSafe: function() {
                return (this.entry.content.get('statusBuckets') == 0) ? this.resultCountSafe() : this.entry.content.get('eventAvailableCount');
            },


            // a job can be dispatched without a latest time, in which case return the published time
            latestTimeSafe: function() {
                var entry = this.entry;
                return entry.content.get('latestTime') || entry.get('published');
            },
            
            isQueued: function() {
                return this.checkUppercaseValue('dispatchState', JobModel.QUEUED);
            },
            
            isParsing: function() {
                return this.checkUppercaseValue('dispatchState', JobModel.PARSING);
            },
            
            isFinalizing: function() {
                return this.checkUppercaseValue('dispatchState', JobModel.FINALIZING);
            },
            
            isPreparing: function() {
                return this.isQueued() || this.isParsing();
            },
            
            isRunning: function() {
                return !this.isNew() && !this.entry.content.get('isPaused') && !this.entry.content.get('isDone') && !this.isPreparing() && !this.isFinalizing();
            },
            
            isReportSearch: function() {
                return (this.entry.content.get('reportSearch') ? true : false);
            },

            // returns true if the job was dispatched over all time, returns false for all-time real-time
            isOverAllTime: function() {
                var dispatchEarliestTime = this.getDispatchEarliestTime();
                return ((!dispatchEarliestTime || dispatchEarliestTime === '0') && !this.getDispatchLatestTime());
            },

            isRealtime: function() {
                return this.entry.content.get('isRealTimeSearch');
            },
            
            checkUppercaseValue: function(key, uc_value) {
                var value = this.entry.content.get(key);
                if (!value) {
                    return false;
                }
                return (value.toUpperCase() === uc_value);
            },
            
            getDispatchEarliestTime: function() {
                var earliest = this.entry.content.request.get('earliest_time');
                if (earliest === void(0)) {
                    return this.entry.content.get('searchEarliestTime');
                }
                return earliest;
            },
            
            getDispatchEarliestTimeOrAllTime: function() {
                if (this.entry.content.get('delegate')==='scheduler' && this.entry.content.get('searchLatestTime') !== void(0)) {
                    // windowed rt or all time rt alerts
                    return this.entry.content.get('searchEarliestTime') || '0';
                } else {
                    return this.entry.content.request.get('earliest_time') || '0';
                }
            },
            
            getDispatchLatestTime: function() {
                var latest = this.entry.content.request.get('latest_time');
                if (latest === void(0)) {
                    return this.entry.content.get('searchLatestTime');
                }
                return latest;
            },
            
            getDispatchLatestTimeOrAllTime: function() {
                var searchLatestTime = this.entry.content.get('searchLatestTime');
                if (this.entry.content.get('delegate')==='scheduler' && searchLatestTime !== void(0)) {
                    return searchLatestTime;
                } else {
                    return this.entry.content.request.get('latest_time') || '';
                }
            },
            
            getStrippedEventSearch: function() {
                var search = this.entry.content.get('eventSearch');
                if (search) {
                    search = splunkUtil.stripLeadingSearchCommand(search);
                }
                return search;
            },
            
            getSearch: function() {
                var search = this.entry.get('name');
                if (!search) {
                    return this.entry.content.request.get('search');
                }
                return search;
            },
            
            getAdhocSearchMode: function() {
                var mode = this.entry.content.request.get('adhoc_search_level'),
                    statusBuckets = this.entry.content.get('statusBuckets'),
                    delegate;
                
                if (!mode) {
                    if ((statusBuckets !== void(0)) && (statusBuckets == 0)) {
                        return splunkd_utils.FAST;
                    }
                    
                    delegate = this.entry.content.get('delegate');
                    if (this.entry.content.get('isSavedSearch') && delegate === 'scheduler') {
                        return splunkd_utils.FAST;
                    }
                }
                
                return mode;
            },
            
            canBePausedOnRemove: function() {
                if (!this.isNew() &&
                        (!this.entry.content.get("isDone") && !this.get("cannotPauseOnRemove")) &&
                        !this.entry.content.get("isPaused") &&
                        !this.isBackground() && 
                        !this.entry.content.get("isSaved")) {
                    return true;
                }
                return false;
            },

            deepOff: function () {
                SearchJob.prototype.deepOff.apply(this, arguments);
                Session.off(null, null, this);
            }
            
        },
        {            
            createMetaDataSearch: function(search, deferred, applicationModel) {
                var job = new JobModel({}, {delay: SearchJob.DEFAULT_METADATA_POLLING_INTERVAL, enableSimplePoller: true});
                    
                job.save({}, {
                    data: {
                        app: applicationModel.get("app"),
                        owner: applicationModel.get("owner"),
                        search: search,
                        preview: "true",
                        earliest_time: "rt",
                        latest_time: "rt",
                        auto_cancel: SearchJob.DEFAULT_AUTO_CANCEL,
                        max_count: 100000
                    },
                    success: function(model, response) {
                        deferred.resolve();
                    },
                    error: function(model, response) {
                        deferred.resolve();
                    }
                });
                
                return job;
            }
        });
        
        return JobModel;
    }
);
