define(
    [
        'jquery',
        'underscore',
        'backbone',
        'models/Base',
        'models/fetch_data/ResultsFetchData',
        'collections/Base',
        'util/splunkd_utils',
        'util/time_utils',
        'splunk.i18n'
    ],
    function(
        $,
        _,
        Backbone,
        BaseModel,
        ResultsFetchData,
        BaseCollection,
        splunkd_utils,
        time_utils, 
        i18n
    ) 
    {
        var Result = Backbone.Model.extend({
            initialize: function(attributes, options) {
                //cache buster
                this.on('change', function() {
                    this._keys = void(0); 
                    this._strip = void(0);
                }, this);
            },
            idAttribute: '1ee63861-c188-11e2-8743-0017f209b4d8',
            systemFields: [
                 'host', 
                 'source', 
                 'sourcetype',
                 'punct'
            ],
            timeFields: [
                 '_time',
                 'date_zone',
                 'date_year',
                 'date_month',
                 'date_mday',
                 'date_wday',
                 'date_hour',
                 'date_minute',
                 'date_second',
                 'timeendpos',
                 'timestartpos'
            ],
            //cache keys
            keys: function() {
                if(!this._keys) {
                    this._keys = _.keys(this.toJSON());
                }
                return this._keys; 
            },
            strip: function() {
                if(!this._strip){
                    this._strip = _.intersection(this.keys(),  _.filter(this.keys(), function(key) {
                        return !(key.indexOf('_')===0 || key.indexOf('tag::')===0 || key.indexOf('_raw')===0);
                    }));
                }
                return this._strip;
            },
            system: function() {
                 return _.intersection(this.strip(), this.systemFields);
            },
            notSystemOrTime: function() {
                var fields = this.strip();
                return _.intersection(fields, _.difference(fields, _.union(this.time(), this.system())));
            },
            time: function() {
                return _.intersection(this.keys(), this.timeFields);
            },
            isTruncated: function() {
                return this.has('_fulllinecount'); 
            },
            getFieldsLength: function(fields) {
                return _(fields).reduce(function(m, f){ 
                    return m + this.get(f).length; 
                }, 0, this);
            },
            getTags: function(field) {
                var tag = this.get('tag::' + field);
                if(tag && !_.isArray(tag)){
                    return [tag];
                } else if(!tag){
                    return [];
                }
                return tag;
            },
            setTags: function(field, tags) {
                this.set('tag::'+field, tags);
            },
            rawToHTML: function(segment) {
                var raw    = this.get('_raw');
                if (!raw) {return '';}
                var tokens = raw.tokens,
                    stree  = raw.segment_tree,
                    types  = raw.types;
        
                function _rawToHtml (stree) {
                    var html = [], i = 0;
                    for (i=0; i<stree.length; i++) {
                        var leaf = stree[i];
                        if(typeof leaf === 'object' ) {
                            html.push('<span class="t'+((leaf.highlight)?' a':'')+'">', _rawToHtml(leaf.array), '</span>');
                        } else {
                            html.push(_.escape(tokens[leaf] || ''));
                        }
                    }
                    return html.join('');
                }
                return _rawToHtml(stree);
            },
            rawToText: function() {
                var raw = this.get('_raw');
                if (!raw) {return '';}
                return _.isArray(raw) ? raw[0]: raw.tokens.join('');
            },
            getRaw: function(){
                var raw = this.get('_raw');
                return _.isArray(raw) ? _.escape(raw[0]): this.rawToHTML();
            },
            formattedTime: function() {
                var time = this.get('_time');
                if (! time ) { return '';}
                return i18n.format_datetime_microseconds(
                    time_utils.jsDateToSplunkDateTimeWithMicroseconds(time_utils.isoToDateObject(time)), 'short', 'full'
                );    
            },
            replace: function() {
                return BaseModel.prototype.replace.apply(this, arguments);
            },
            deepOff: function() {
                this.off();
            }
        });
        
        var Model = BaseModel.extend({
            url: '',
            initialize: function(data, options) {
                options = options || {};
                options.fetchData = options.fetchData || new ResultsFetchData();
                BaseModel.prototype.initialize.call(this, data, options);
                this.initializeAssociated();
                if (options && options.splunkDPayload) {
                    this.setFromSplunkD(options.splunkDPayload, {silent: true});
                }
            },
            initializeAssociated: function() {
                // do a dynamic lookup of the current constructor so that this method is inheritance-friendly
                var RootClass = this.constructor;
                this.associated = this.associated || {};
                
                //instance level models
                this.results = this.results || new RootClass.Results();
                this.associated.results = this.results;
                this.messages = this.messages || new RootClass.Messages();
                this.associated.messages = this.messages;
                this.fields = this.fields || new RootClass.Fields();
                this.associated.fields = this.fields;
            },
            sync: function(method, model, options) {
                if (method!=='read') {
                    throw new Error('invalid method: ' + method);
                }
                options = options || {};
                var appOwner = {},
                    defaults = {
                        data: {output_mode: 'json'},
                        dataType: 'text'
                    },
                    url = _.isFunction(model.url) ? model.url() : model.url || model.id;
                    
                if(options.data){
                    appOwner = $.extend(appOwner, { //JQuery purges undefined
                        app: options.data.app || undefined,
                        owner: options.data.owner || undefined,
                        sharing: options.data.sharing || undefined
                    });
                }
                
                defaults.url = splunkd_utils.fullpath(url, appOwner);
                $.extend(true, defaults, options);
                
                delete defaults.data.app;
                delete defaults.data.owner;
                delete defaults.data.sharing;
                
                return Backbone.sync.call(this, method, model, defaults);
            },
            setFromSplunkD: function(payload, options) {
                options = options || {};
                if(!options.clone) {
                    this.responseText = JSON.stringify(payload);    
                }

                this.attributes = {};
                if (payload) {
                    if (payload.messages) {
                        this.messages.reset(payload.messages, options);
                        delete payload.messages;
                    }
                    if (payload.fields) {
                        this.fields.reset(payload.fields, options);
                        delete payload.fields;
                    }
                    this.set(payload, options);
                    
                    if (payload.results) {
                        _(payload.results).each(function(result, idx) {
                            _.each(result, function(v, k) {
                                result[k] = (typeof v === 'string') ? [v]: v;
                            },this);
                        },this);
                        this.results.reset(payload.results, options);
                        delete payload.results;
                    }
                }
            },
            parse: function(response, options) {
                this.initializeAssociated();
                this.responseText = response; //store string representation
                
                response = JSON.parse(response);
                
                if (options.sparseMode) {
                    this.results.reset(_(response.results).map(function() { return {}; }));
                    return {};
                }
                
                _(response.results).each(function(result, idx) {
                    _.each(result, function(v, k) {
                        result[k] = (typeof v === 'string') ? [v]: v;
                    },this);
                },this);
                
                this.messages.reset(response.messages);
                delete response.messages;
                
                this.fields.reset(response.fields);
                delete response.fields;
                
                var results = response.results;
                delete response.results;
                
                this.set(response);
                delete response;

                this.results.reset(results);
                return {};
            },
            endOffset: function() {
                return (this.get('init_offset') || 0) + this.results.length;
            },
            lineNumber: function(offset, isRealTime) {
                return isRealTime ? this.endOffset() - offset : (this.get('init_offset') || 0) + offset + 1;
            },
            offset: function(index) {
                return (this.get('init_offset') || 0) + index;
            },
            getTags: function(field, value) {
                var tags = this.get('tags'),
                    matched = [];
      
                if (tags) {
                    matched = tags[field];
                    if (matched) {
                        matched = matched[value] || [];
                    }
                }
                return matched || [];
            },
            setTags: function(field, value, tags) {
                if (!this.get('tags')) {
                   this.set('tags', {});
                }
                if (!this.get('tags')[field]) {
                    this.get('tags')[field] = {}; 
                }
                if (!this.get('tags')[field][value]) {
                    this.get('tags')[field][value] = [];
                }
                this.get('tags')[field][value] = tags;
                this.trigger('tags-updated'); //TODO: must fix this (make synthetic/namespaced)
            }
        },
        {
            Results: BaseCollection.extend(
                {
                    model: Result,
                    get: function(obj) {
                      if (obj == null) return void 0;
                      this._idAttr || (this._idAttr = this.model.prototype.idAttribute);
                      return this._byId[obj[this._idAttr]];
                    }
                },
                {
                    Result: Result
                }
            ),
            Messages: BaseCollection.extend({model: BaseModel}),
            Fields: BaseCollection.extend({model: BaseModel})
        });
    
        return Model;
    }
);
