define(function(require, exports, module) {
    var _ = require('underscore');
    var console = require('util/console');
    var BaseInputView = require("./baseinputview");
    var Messages = require("./messages");
    var Tooltip = require("bootstrap.tooltip");

    require("css!../css/choice.css");

    /**
     * Private abstract base class for form input views that 
     * present static and dynamic choices.  
     *
     * This class presents choices, which consist of a value 
     * and an optional label to display.  If the label is not
     * provided, the value will be displayed.
     *
     * Options:
     * 
     *     valueField: Field to use for option value (and optionally,
     *         option label)
     *
     *     labelField: Field to use for option label, will default to
     *         valueField if not provided.
     */
    var BaseChoiceView = BaseInputView.extend({
        options: {
            "choices": [],
            "selectFirstChoice": false
        },

        initialize: function() {
            var that = this;
            
            this._$messageEl = this.$(".splunk-choice-input-message span");

            this.options = _.extend({}, BaseInputView.prototype.options, this.options);
            BaseInputView.prototype.initialize.apply(this, arguments);

            this.manager = null;
            this.resultsModel = null;

            this.settings.on("change:value", this._onValueChange, this);
            this.settings.on("change:choices change:valueField change:labelField change:default",
                             _.debounce(this.render, 0), this);
        },

        updateDomVal: function(value) {
            // Given the value passed in, change the HTML of this
            // control to reflect the current value.
            throw new Error("Abstract method.  Must override");
        },
        
        _onValueChange: function(ctx, value, options) {
            this.updateDomVal(value);
            this.trigger('change', this.val(), this);
        },
        
        _displayMessage: function(messageName) {
            var info = messageName;
            if (_.isString(messageName)) {
                info = Messages.resolve(messageName);
            }
            
            // For the choice views, we have very limited space to render
            // messages, and so we render them to a specific message container
            // created in _updateView. We also replace the original message with
            // one that is more appropriate for the choice view.
            var message = "";
            var originalMessage = "";
            switch (messageName) {
                case "no-events":
                case "no-results":
                case "no-stats": {
                    message = _("Search produced no results.").t();
                    originalMessage = "";
                    
                    // We need to update the view with the empty search results,
                    // otherwise we may end up displaying stale data.
                    this._updateView(this._viz, []);
                    break;
                }
                case "waiting":
                case "waiting-queued":
                case "waiting-preparing": {
                    message = _("Populating...").t();
                    originalMessage = "";
                    break;
                }
                default: {
                    if (info.level === "error") {
                        message = _("Could not create search.").t();
                        originalMessage = info.message || "";
                    }
                    else {
                        message = "";
                        originalMessage = "";
                    }
                    break;
                }
            }
            
            // Put the message as the text, but also put the original message
            // as the tooltip.
            this._$messageEl.text(message);
            this._$messageEl.attr("title", originalMessage);
            this._$messageEl.tooltip('destroy');
            this._$messageEl.tooltip({animation: false});
        },

        convertDataToChoices: function(data) {
            // Given a new set of dynamic data, transforms all sources
            // of choices into a value/label pair suitable for DOM
            // rendering.  Merges static and dynamic data into a
            // single array.
            data = data || this._data;
            var valueField = this.settings.get("valueField") || 'value';
            var labelField = this.settings.get("labelField") || valueField;
            var choices = Array.prototype.slice.call(this.settings.get('choices') || []);

            choices = choices.concat(_.map((data || []), function(row) {
                return {
                    label: row[labelField],
                    value: row[valueField]
                };
            }));

            // De-duplicate values list, as HTML controls don't handle
            // them well.
            var originalChoicesLength = choices.length;
            choices = _.uniq(choices, false, function(i) { return i.value; });
            if (originalChoicesLength != choices.length) {
                console.log("Choice control received search result with duplicate values. Recommend dedupe of data source.");
            }
            return choices;
        },

        _updateView: function(viz, data) {
            data = this.convertDataToChoices(data);

            if (!this._viz) {
                this._createView(data); 
                if (!this._viz) {
                    return; 
                }
            }
            this.updateView(this._viz, data);
            
            if (this.settings.get('selectFirstChoice')) {
                var hasData = data && data.length > 0;
                if (hasData && !this._hasValue()) {
                    this.val(data[0].value);
                }
            }
            
            // Create the message area if one does not exist, and clear it.
            if (!this._$messageEl.length) {
                var $messageContainer = $("<div class='splunk-choice-input-message'></div>").appendTo(this.el);
                this._$messageEl = $("<span></span>").appendTo($messageContainer);
            }
            this._$messageEl.text('');
            
            // If there is no data, we disable the input, but if there is,
            // we may still need to disable it.
            if (!data || data.length === 0) {
                // We use the raw disable mechanism, because we don't want to
                // change our disabled state.
                this._disable(true);
            }
            else {
                this._onDisable();    
            }
        },
        
        _hasValue: function() { 
            var currentValue = this.val();
            return !_.isUndefined(currentValue) && !_.isNull(currentValue);
        },

        val: function(newValue) {
            if (arguments.length === 0) {
                return this.settings.get("value");
            }

            if (newValue !== this.settings.get("value")) {
                this.settings.set('value', newValue);
            }
            
            return this.settings.get('value');
        }
        
    });
    
    return BaseChoiceView;
});
