require.config({
    shim: {
        "splunkjs/contrib/jquery.deparam": {
            deps: ["jquery"],
            exports: "jQuery.fn.deparam"
        }
    }
});

define(function(require, exports, module) {    
    var _ = require("underscore");
    var Backbone = require("backbone");
    var Registry = require("./registry");
    var Resolver = require("JS_CACHE/urlresolver");
    var config = require("JS_CACHE/config");
    var deparam = require("../contrib/jquery.deparam");
    var protections = require("./protections");
    var TokenEscapeString = require("./tokenescapestring");
    var TokenSafeString = require("./tokensafestring");
    var utils = require("./utils");
    var sdk = require("splunkjs/splunk");
    var console = require("util/console");
    var splunkConfig = require('splunk.config');
    
    require('splunk.error');
    require('util/ajax_logging');

    var DEBUG = false;
    
    var MVCConstructor = function(options) {
        this.Components = new Registry();
        this.initialize.apply(this, arguments);
        this.STATIC_PREFIX = config.STATIC_URL;
    };
    _.extend(MVCConstructor.prototype, Backbone.Events, {
        started: false,
        
        initialize: function() {},
        
        drilldown: function(path, data) {
            var encoded = $.param(data || {});
            
            if (encoded) {
                path = path + "?" + encoded;
            }
            
            window.location = path;
        },

        loadDebugger: function(callback) {
            var that = this;
            require(['./debugger'], function(Debugger) {
                that.Debugger = new Debugger({'registry': that.Components});
                that.trigger("debugger:ready", that);
                if (callback) callback(that.Debugger);
            });
        },
        
        loadDrilldown: function(data) {
            var queryArgs = window.location.search.substr(1) || "";
            var args = _.extend({}, $.deparam(queryArgs), data);
            
            var that = this;
            _.each(args, function(datum, id) {
                if (that.Components.has(id)) {
                    var component = that.Components.getInstance(id);
                    
                    if (component.settings) {
                        component.settings.set(datum);
                    }
                    else if (component.set) {
                        component.set(datum);
                    }
                    else {
                        console.log("Could not find a setter for '" + id + "'", datum);
                    }
                }
            });
        },
        
        reverse: function(name, app, args) {
            return Resolver.reverse(name, app, args);
        },
        
        createService: function(options) {
            options = options || {};
            var http = options.http || new sdk.ProxyHttp(config.PROXY_PATH);
            options.version = options.version || "5.0";
            options.app = options.app || utils.getPageInfo().app || "-";
            options.owner = options.owner || encodeURIComponent(splunkConfig.USERNAME);
            
            return new sdk.Service(http, options);
        },
        
        _loadDOMComponents: function() {            
            var that = this;
            
            // Instantiate all views that we haven't instantiated yet
            var views = $(".splunk-view:not([data-initialized])");
            _.each(views, function(el, idx) {
                var site = $(el);
                var id = site.attr('id');
                var requiredFile = site.data("require");
                
                if (!requiredFile) {
                    return;
                }
                
                var options = that._decodeOptions(site);
                
                options.id = id;
                options.el = site;
                
                var Constructor = require(requiredFile);
                var view = new Constructor(options);
                view.render();
                
                site.attr("data-initialized", "");
            });
            
            // Instantiate all components that we haven't instantiated yet
            var managers = $(".splunk-manager:not([data-initialized])");
            _.each(managers, function(el, idx) {
                var site = $(el);
                var id = site.attr('id');
                var requiredFile = site.data("require");
                
                if (!requiredFile) {
                    return;
                }
                
                var options = that._decodeOptions(site);
                
                options.id = id;
                
                var Constructor = require(requiredFile);
                var manager = new Constructor(options);
                
                site.attr("data-initialized", "");
            });
        },
        
        _decodeOptions: function($el) {
            var that = this;
            
            var options = {};
            var optionsString = $el.attr('data-options');
            if (optionsString && optionsString.length > 0) {
                options = JSON.parse(optionsString);
                _.each(options, function(v, k) {
                    options[k] = that._decodeOptionValue(v);
                });
            }
            return options;
        },
        
        _decodeOptionValue: function(optionValue) {
            // Treat falsy values as literal values
            if (!optionValue) {
                return optionValue;
            }
            
            switch (optionValue.type) {
                case "token_safe":
                    return new TokenSafeString(optionValue.value);
                case undefined:
                    // Literal value
                    return optionValue;
                default:
                    throw Error("Unrecognize option value type: " + optionValue.type);
            }
        },
        
        /**
         * Marks the specified bare string to be interpreted as a
         * template with token values.
         * 
         * Since the default interpretation of a bare string is to be
         * treated as a literal string you must use mvc.tokenSafe() or
         * a {tokens: true} option to mark strings that you wish to be
         * interpreted as a template instead.
         * 
         * For example:
         *      new SearchManager({
         *          id: 'indexSize',
         *          search: mvc.tokenSafe('index=$indexName$ | stats count')
         *      });
         */
        // Historical Note: The name of this function is inspired by
        //                  Django's built-in 'safe' filter.
        tokenSafe: function(template) {
            return new TokenSafeString(template);
        },
        
        /**
         * Marks the specified bare string to be interpreted as a
         * literal string.
         * 
         * Bare strings are interpreted as literals by default so
         * this marking is only useful when a bare string is being
         * passed in a context when the default interpretation is
         * being overridden by a {tokens: true} option.
         * 
         * For example:
         *      new SingleView({
         *          id: 'costReport',
         *          beforeLabel: mvc.tokenEscape('USD $'),
         *          managerid: 'cost'
         *      }, {tokens: true});
         */
        // Historical Note: The name of this function is inspired by
        //                  Django's built-in 'escape' filter.
        tokenEscape: function(literal) {
            return new TokenEscapeString(literal);
        }
    });

    // Bootstrap onto the 'splunkjs' object if one exists, or create a new one.
    // TODO: What if one already exists with the '.mvc' namespace?
    //       Should we noConflict() it?
    if (!window.splunkjs) {
        window.splunkjs = {};
    }
    
    // Create splunkjs.mvc
    var mvc = new MVCConstructor();
    var ns = window.splunkjs.mvc = mvc;
    
    if (config.ENABLE_PROTECTIONS) {
        protections.enableCSRFProtection($);
        protections.enableUnauthorizationRedirection($, mvc.reverse(":login"), mvc.reverse(":logout"));
    }
    
    return mvc;
});
