define(function(require, exports, module) {
    var _ = require("underscore");
    var $ = require("jquery");
    var config = require('JS_CACHE/config');
    var mvc = require('./mvc');
    var Async = require('../splunk').Async;
    var BaseManager = require('./basemanager');
    var SplunkResultsModel = require('./splunkresultsmodel');
    var SearchModels = require('./searchmodel');
    var sdk = require('splunkjs/splunk');
    var TokenSafeString = require("./tokensafestring");
    var console = require('util/console');
    var JobTracker = require('./jobtracker');
    var Messages = require('./messages');

    var SearchManager = BaseManager.extend({ 
        moduleId: module.id,

        defaults: {
            autostart: true,
            cache: false,
            cancelOnUnload: true
        },

        constructor: function(attributes, options) {
            // This has to be in the constructor, otherwise
            // we will call Model.set before we have created these sub-models.
            this.query = attributes.queryModel || new SearchModels.SearchQuery({}, options);
            this.search = attributes.searchModel || new SearchModels.SearchJob({
                label: attributes.id
            }, options);

            // No need to set it on our model
            delete attributes.queryModel;
            delete attributes.searchModel;

            return BaseManager.prototype.constructor.apply(this, arguments);
        },

        initialize: function(options) {
            this.service = mvc.createService({app: options.app, owner: options.owner});
            if (!this.search.has("label")) {
                this.search.set("label", options.id);
            }

            this._debouncedStartSearch = _.debounce(this.startSearch);

            // We want to change behavior depending on whether autostart is set
            // or not. So we first bind a listener to it changing, and then
            // we invoke that listener manually
            this.on("change:autostart", this.handleAutostart, this);
            this.on("search:error", function(msg) { this.lastError = msg; return true; }, this);
            this.on('all', function(evt){
                if (evt.substring(0, 'search:'.length) === 'search:') {
                    this._lastSearchEvent = [ evt, Array.prototype.slice.call(arguments, 1) ];
                }
            }, this);
            this.handleAutostart();

            // Kick off the search. This will only do anything if autostart is enabled
            this._start();

            this._cancelOnUnload = _.bind(this._cancelOnUnload, this);
            this.on('change:cancelOnUnload change:cache', this.handleCancelOnUnload, this);
            this.handleCancelOnUnload();
        },

        handleAutostart: function() {
            if (this.get("autostart")) {
                this.query.on("change", this._debouncedStartSearch, this);
                this.search.on("change", this._debouncedStartSearch, this);
            }
            else {
                this.query.off("change", this._debouncedStartSearch, this);
                this.search.off("change", this._debouncedStartSearch, this);
            }
        },

        handleCancelOnUnload: function() {
            $(window).off('unload', this._cancelOnUnload);
            if (this.get('cancelOnUnload') && !this.get('cache')) {
                $(window).on('unload', this._cancelOnUnload);
            }
        },

        _cancelOnUnload: function() {
            // Cancel job is it isn't already done
            if (this.job && !this.job.properties().isDone) {
                $.ajax({
                    url: config.PROXY_PATH + this.job.qualifiedPath + '/control',
                    data: { action: 'cancel' },
                    type: 'POST',
                    // Force a synchronous XHR call since an asynchronous request might have no
                    // effect in the unload phase
                    async: false
                });
            }
        },

        // set(attributeName : String, attributeValue : mixed, options : Object|null)
        // set(attributes : Object, options : Object|null)
        set: function(key, value, options) {         
            var attrs;
            
            // Normalize the key-value into an object
            if ((_.isObject(key) && !_.isArray(key)) || key === null) {
                attrs = key;
                options = value;
            } else {
                attrs = {};
                attrs[key] = value;
            }
            
            var search = {};
            var query = {};
            
            // If the 'search' or 'query' input attributes are present
            // then use them as the base set of output attributes for
            // each respective model (and filter them out of the original
            // input attributes).
            if (_.has(attrs, "search") && this._isDictionary(attrs.search)) {
                search = attrs.search;
                delete attrs.search;
            }
            if (_.has(attrs, "query") && this._isDictionary(attrs.query)) {
                query = attrs.query;
                delete attrs.query;
            }
            
            // Partition all remaining input attributes based on whether they
            // are destined for the job model or the query model.
            // Ignore input attributes that are unrecognized.
            search = _.extend(search, _.pick(attrs, SearchModels.SearchJob.ALLOWED_ATTRIBUTES));
            query = _.extend(query, _.pick(attrs, SearchModels.SearchQuery.ALLOWED_ATTRIBUTES));
                
            // Finally, we set it on the child models, if there is anything to
            // set
            if (!_.isEmpty(search)) {
                this.search.set(search, options);
            }
            if (!_.isEmpty(query)) {
                this.query.set(query, options);
            }
            
            return BaseManager.prototype.set.call(this, attrs, options);
        },

        get: function(attr) {
            // These attributes are propagated to child objects.  We
            // should look to the child object for the "official"
            // version.
            if (_.contains(SearchModels.SearchJob.ALLOWED_ATTRIBUTES, attr)) {
                return this.search.get.apply(this.search, arguments);
            }

            if (_.contains(SearchModels.SearchQuery.ALLOWED_ATTRIBUTES, attr)) {
                return this.query.get.apply(this.query, arguments);
            }

            return BaseManager.prototype.get.call(this, attr);
        },
        
        _isDictionary: function(value) {
            // NOTE: This logic is fragile. Unfortunately there's no good
            //       way to detect a raw dictionary (via an object literal)
            //       as opposed to an object created through the 'new' keyword.
            return _.isObject(value) && !(value instanceof TokenSafeString);
        },
        
        _start: function() {         
            if (this.get("autostart")) {
                this.startSearch();
            }

            return this;
        },
        
        data: function(source, attrs) {
            if (!source) {
                throw new Error("Cannot get a results model without specifying the source.");
            }
            
            attrs = attrs || {};
            attrs.manager = this;
            attrs.source = source;
            return new SplunkResultsModel(attrs);
        },
        
        _startSearchFromSid: function(sid) {
            var that = this;
            var job = new sdk.Service.Job(this.service, sid);
            job.fetch(function(err) {
                if (err) {
                    that.trigger("search:error", _("Error in getting pre-existing search: ").t() + sid);
                    return;
                } 
                
                // Set our job and query model to the request parameters
                // for this search, in case we need to make any changes
                var request = job.properties().request;
                that.query.set(_.pick(request, SearchModels.SearchQuery.ALLOWED_ATTRIBUTES), {silent: true});
                that.search.set(_.pick(request, SearchModels.SearchJob.ALLOWED_ATTRIBUTES), {silent: true});
                
                that.createManager(job);
            });
        },
        
        startSearch: function() {
            this.lastError = '';
            this.cancel(); // Cancel any existing search

            // If we were provided with a SID, then we will
            // just try and use that
            if (this.has("sid")) {
                this._startSearchFromSid(this.get("sid"));
                return;
            }

            // Get the cache and convert it to a number:
            // false -> 0, never use cache
            // true -> -1, always use cache
            // positive integer -> positive integer, use cache if newer than
            var cache = this.get("cache") || false;
            if (_.isBoolean(cache)) {
                cache = cache ? -1 : 0;
            }
            
            var options = this._createOptionsForSearchJob();
            
            // Get the threshold for cache (which could be infinity)
            var threshold = cache < 0 ? 
                (new Date(0)).valueOf() :
                ((new Date()).valueOf() - cache * 1000);
            
            // Only start a search if we have a search string
            if (!options.search) {
                if (this.query.get("search", {tokens: true})) {
                    this.trigger("search:error", Messages.resolve("unresolved-search").message);
                }
                else {
                    this.trigger("search:error", _("No search query provided.").t());   
                }
                return;
            }

            // Skip fetching existing jobs if cache is disabled and directly create a new job
            if (cache === 0) {
                this.createSearch();
                return this;
            }
            
            var filter = {};
            if (options.label) {
                filter["search"] = "label="+options.label+"*";
            }

            var that = this;

            // First check and see if a matching job already exists, filtering
            // by the label
            this.service.jobs().fetch(filter, function(err, jobs) {
                if (err) {
                    console.log("Error fetching searches");
                    that.trigger("search:error", _("Error fetching searches").t());
                    return;
                }
                
                var list = jobs.list();
                
                // Splunk returns values as strings, so we have to make sure
                // our values are strings too, otherwise the comparison
                // will return false.
                _.each(options, function(val, key) {
                    options[key] = (_.isBoolean(val) || _.isNumber(val)) ? val.toString() : val;
                });
                
                // Find all jobs that match
                var matches = [];
                for (var i = 0; i < list.length; ++i) {
                    var potentialJob = list[i];
                    var request = potentialJob.properties().request;
                    var published = (new Date(potentialJob.published())).valueOf();
                    
                    if (_.isEqual(request, options) && published > threshold) {
                        matches.push(potentialJob);
                    }
                }
                
                // Sort to easily find the latest one
                matches.sort(function(a,b) { 
                    var aValue = (new Date(a.published())).valueOf();
                    var bValue = (new Date(b.published())).valueOf();
                    return bValue - aValue;
                });
                
                // If a job exists, then we will use it, otherwise,
                // create a new job.
                var job = matches[0];
                if (job) {
                    that.createManager(job);
                }                
                else {
                    that.createSearch();
                }
            });

            return this;
        },
        
        _createOptionsForSearchJob: function() {
            // Let options = this.search, ignoring undefined values
            var options = {};
            _.each(this.search.attributes, function(v, k) {
                if (v !== undefined) {
                    options[k] = v;
                }
            });
            
            // Perform token replacement on the search string using data
            // from the search query model.
            // 
            // TODO: Rewrite this functionality in terms of regular
            //       Token Binding. (DVPL-1636)
            options.search = this.query.resolve({ qualified: true });
            
            return options;
        },
        
        pause: function() {
            if (this.job) {
                this.job.pause();
            }
        },
        
        unpause: function() {
            if (this.job) {
                this.job.unpause();
            }
            return this;
        },
        
        finalize: function() {
            if (this.job) {
                this.job.finalize();
            }
            return this;
        },
        
        cancel: function() {
            // If this search is used with cache=true, then never cancel,
            // because somebody else might be using the same search.
            if (this.get("cache")) {
                return;
            }
            
            if (this.job) {
                JobTracker.untrack(this.job.sid);
                this.job.cancel();
                this.job = null;
                this.trigger("search:cancelled");
            }
            return this;
        },
        
        createSearchPinger: function(job) {
            var that = this;

            // As long as the job on the manager is the same
            // as the one we were told to iterate over, keep going.
            Async.whilst(
                function() { return that.job === job; },
                function(done) {
                    // Just ping the job so that it doesn't die. We will wait
                    // a third of the time it takes for the job to expire or to 
                    // be auto-cancelled before pinging it
                    var props = job.properties();
                    var expireDelay, autoCancel;

                    if (props.isDone || props.request === undefined || props.request.auto_cancel === undefined) {
                        // Use TTL if the job is done or there is not auto_cancel
                        expireDelay = props.ttl;
                    } else if (props.request !== undefined) {
                        autoCancel = props.request && parseInt(props.request.auto_cancel, 10) || Infinity;
                        expireDelay = Math.min(autoCancel, props.ttl);
                    }

                    // Do the final calculation, and make sure we have a sensible
                    // default of 10 seconds.
                    var delay = parseInt(((expireDelay || 30) * 1000) / 3, 10);

                    Async.sleep(delay, function() {
                        // Make sure it's still our job
                        if (that.job === job) {
                            job.fetch(done);
                        }
                    });
                },
                function() {
                    // Do nothing, this will fail if the job is cancelled, 
                    // and that's fine.
                }
            );
        },
        
        // Create a search job manager for the given job.
        createManager: function(job) {
            var that = this;
            var check = _.bind(this._checkJob, this); 
            var cacheEnabled = this.get('cache');
            
            this.job = job;
            this.createSearchPinger(job);
            JobTracker.track(job, {
                progress: function(job) {
                    // Check if the event is from current active job
                    if (!check(job, cacheEnabled)) {
                        return;
                    }
                    var properties = job.properties();
                    var state = job.state();
                    
                    that.trigger("search:progress", state, job);
                    
                    // Backbone does a deep equality check on sets, and since
                    // properties is an object, and the object contents may have
                    // not changed (e.g. due to a cached job), Backbone might not
                    // fire a "change" event. This ensures that a change
                    // event is fired every time by doing a silent unset followed
                    // by a set.
                    that.unset("data", {silent: true});
                    that.set("data", properties);
                },
                done: function(job) {
                    // Check if the event is from current active job
                    if (!check(job, cacheEnabled)) {
                        return;
                    }
                    var state = job.state();
                    that.trigger("search:done", state, job);
                },
                failed: function(job) {
                    // Check if the event is from current active job
                    if (!check(job, cacheEnabled)) {
                        return;
                    }
                    var state = job.state();
                    that.trigger("search:fail", state, job);
                }
            });

            this.trigger("search:start", job);
        },

        _checkJob: function(job, cache){
            if (this.job !== job) {
                if (!cache) {
                    job.cancel();
                }
                return false;
            }
            return true;
        },

        // Create a new search job from current options.
        createSearch: function() {
            var options = this._createOptionsForSearchJob();
            // Check if we currently have an inflight XHR for creating a job
            if (this._inflightCreateXHR) {
                // Abort the inflight XHR
                this._inflightCreateXHR.abort();
            }
            var that = this;
            var cache = this.get('cache');
            var req = this._inflightCreateXHR = this.service.jobs().create(options, function(err, job) {
                // Check if response is from an obsolete job creation XHR
                if (req !== that._inflightCreateXHR) {
                    if (!(err || cache)) {
                        job.cancel();
                    }
                    that._inflightCreateXHR = null;
                    return;
                }
                that._inflightCreateXHR = null;
                if (err) {
                    // We only want to trigger errors for non-abort cases. This is because
                    // we are the ones causing the abort (at the beginning of createSearch),
                    // and we should not cause an error for something we did. Furthermore,
                    // since we only abort in the case of creating a new search, we are guaranteed
                    // that the new search will cause some events to happen (even if those events
                    // are search:error ones).
                    if (err.status !== "abort") {
                        that.trigger("search:error", _("Could not create search.").t(), err, job);
                    }
                    return;
                }
                that.createManager(job);
            });
        },

        replayLastSearchEvent: function(listener) {
            var replayed = false;
            if (this._lastSearchEvent) {
                var evtName = this._lastSearchEvent[0];
                var listenerObj = _(this._events[evtName]).find(function(attachedListener){ return attachedListener.ctx === listener; });
                if (listenerObj) {
                    var args = this._lastSearchEvent[1];
                    listenerObj.callback.apply(listener, args);
                    replayed = true;
                }
            }
            return replayed;
        },
        
        getLastSearchEvent: function() {
            return this._lastSearchEvent;
        }
    });
    
    return SearchManager;
});
