define(function(require) {
    var _ = require('underscore');
    var $ = require('jquery');
    var BaseInput = require('./base');
    var Dashboard = require('../../simplexml/controller');
    var mvc = require('../../mvc');
    var TimeRangeView = require('../../timerangeview');
    
    var FACTORY_DEFAULT = {earliest_time: undefined, latest_time: undefined};
    
    var TimeRangeInput = BaseInput.extend({
        events: {
            // Destroy self when dashboard editor's remove button clicked
            'click .remove-timerange-picker a': function(e) {
                e.preventDefault();
                this.remove();
            }
        },
        
        /*
         * Private API:
         * 
         * @param options.isGlobal  Whether this time picker is considered
         *                          a global time picker by the dashboard
         *                          editor. If unspecified is determined
         *                          automatically based on what the
         *                          {earliest_time, latest_time} settings
         *                          are bound to.
         */
        initialize: function() {
            var inputOptions = _.defaults({
                el: $('<div class="splunk-view"></div>').appendTo(this.el),
                id: _.uniqueId(this.id + '-input')
            }, this.options);
            this.picker = new TimeRangeView(inputOptions);
            this.picker.on('change', this.handleChange, this);
            
            // Always use the inner view's settings, so we don't have two of them
            this.options.settings = this.picker.settings;
            BaseInput.prototype.initialize.call(this);
            
            // Attach to dashboard
            if (this._isGlobalTimeRangeInput()) {
                Dashboard.getStateModel().on('change:edit', this._onEditStateChange, this);
            }
        },
        
        hasValue: function() {
            var value = this.settings.get("value") || FACTORY_DEFAULT;
            var defaultValue = this.settings.get("default");
            var presetValue = this.settings.get("preset");
            
            // We have a value if one of these conditions is true:
            // 1. Value is not the factory default
            // 2. We have no default value
            return !_.isEqual(value, FACTORY_DEFAULT) || (!defaultValue && !presetValue);
        },
        
        /**
         * Update the time range input's value to be the default,
         * taking into account both the default value and the preset value.
         * 
         * @return The time range input object.
         */
        updateValueWithDefault: function() {   
            var defaultValue = this.settings.get("default");
            var presetValue = this.settings.get("preset");
            if (defaultValue) {
                this.val(defaultValue);
            }
            else if (presetValue) {
                // Synchronize the displayed preset and the actual set value.
                this.picker._onTimePresetUpdate();
            }
            
            return this;
        },
        
        handleChange: function(value, input) {
            value = value || FACTORY_DEFAULT;
                        
            // Update dashboard with new value
            if (this._isGlobalTimeRangeInput()) {
                // TODO: Try to convert dashboard to use the same time format
                //       as every other control.
                value = this.val();
                Dashboard.getStateModel().set('default_timerange', {
                    earliestTime: value.earliest_time,
                    latestTime: value.latest_time
                });
            }
            
            BaseInput.prototype.handleChange.apply(this, arguments);
        },
        
        _onEditStateChange: function(model) {
            if (this._isGlobalTimeRangeInput()) {
                if (model.get('edit')) {
                    // Render editing UI
                    $('<div class="remove-timerange-picker"><a href="#"><i class="icon-x-circle"></i></a></div>')
                        .appendTo(this.picker.$el);
                    $('<div class="timerange-edit-hint"></div>')
                        .html(_('Select default time range above. Time range only applies to <i class="icon-search-thin"></i> Inline Searches.').t())
                        .appendTo(this.picker.$el);
                    
                    // NOTE: The dashboard editor will directly update the
                    //       internal timerange picker
                    //       ('views/shared/timerangepicker/Master') and the
                    //       inner TimeRangeView will pick up the changes
                    //       automatically.
                } else {
                    // Unrender editing UI
                    this.$('.remove-timerange-picker').remove();
                    this.$('.timerange-edit-hint').remove();
                }
            }
        },
        
        _isGlobalTimeRangeInput: function() {
            if (this.settings.has('isGlobal')) {
                return (this.settings.get('isGlobal') === true);
            }
            
            return (
                (this.settings.get('earliest_time', {tokens: true}) === '$earliest$') &&
                (this.settings.get('latest_time', {tokens: true}) === '$latest$'));
        },
        
        remove: function() {
            // Dispose inner view
            this.picker.remove();

            // Dispose self
            mvc.Components.revokeInstance(this.id);
            BaseInput.prototype.remove.apply(this, arguments);

            // Detach from dashboard
            if (this._isGlobalTimeRangeInput()) {
                Dashboard.getStateModel().off(null, null, this);
                Dashboard.getStateModel().set('default_timerange', null);
            }
        },
        
        getValue: function() {
            return this.picker.val();
        },
        
        setValue: function(v) {
            return this.picker.val(v);
        },
        
        render: function() {
            this.renderLabel();
            this.picker.render();
            
            // Ensure dashboard editor chrome is displayed upon initial render
            this._onEditStateChange(Dashboard.getStateModel());
            this.handleChange(this.val(), this.picker);
            
            return this;
        },
        
        _getWrappedView: function() {
            return this.picker;
        }
    });
    
    return TimeRangeInput;
});
