define(function(require) {
    var BaseDashboardModel = require('models/Dashboard');
    var routerUtils = require('util/router_utils');
    var classicurl = require('models/classicurl');
    var Backbone = require('backbone');
    var $ = require('jquery');
    var _ = require('underscore');
    var mvc = require('../../mvc');
    var xmlUtils = require('util/xml_utils');
    var utils = require('../utils');
    var pdfUtils = require('util/pdf_utils');
    var console = require('util/console');
    var route = require('uri/route');
    var sharedModels = require('../sharedmodels');
    var splunkd_utils = require('util/splunkd_utils');
    var SplunkUtil = require('splunk.util');
    var protections = require('../protections');
    var Reports = require('collections/Reports'); 
    require('util/ajax_no_cache');

    protections.enableCSRFProtection($);
    protections.enableUnauthorizationRedirection($, SplunkUtil.make_url('account/login'), '/account/logout');

    var DashboardModel = BaseDashboardModel.extend({
        initialize: function(){
            BaseDashboardModel.prototype.initialize.apply(this, arguments);
            this.entry.content.on('change:eai:data', this.updateItemOrder, this);
        },
        /**
         *
         * @param options {
         *     tokens (boolean) -  whether the generated XML source should contain tokens or their values
         *     useLoadjob (boolean) - whether to use | loadjob <sid> instead of writing the actual search to the XML
         *     indent (boolean) - whether to generated pretty-printed XML or not
         * }
         * @returns {String} the serialized XML
         */
        getFlattenedXML: function(options) {
            if(!this._itemOrder) {
                throw new Error('Cannot create flattened XML without the item order being captured first');
            }
            var $n = xmlUtils.$node, that = this;
            options = options || {};

            var dashboard = $n('<dashboard/>');
            $n('<label/>').text(this.meta.get('label')).appendTo(dashboard);
            if(this.meta.has('description')) {
                $n('<description/>').text(this.meta.get('description')).appendTo(dashboard);
            }

            _(this._itemOrder).each(function(panels) {
                var row = $n('<row/>');

                if(_(panels).any(function(p) { return p.length > 1; })) { // Are there any grouped panels?
                    var groups = _(panels).map(function(p) { return p.length; }).join(',');
                    row.attr('grouping', groups);
                }

                _(panels).chain().flatten().each(function(id) {
                    var element = mvc.Components.get(id);
                    var settings = element.model.mapToXML(_.extend({ tokens: false, flatten: true }, options));
                    var newNode = xmlUtils.$node('<' + settings.type + '/>');

                    if(settings.attributes !== undefined) {
                        newNode.attr(settings.attributes);
                    }

                    if(settings.content) {
                        if(settings.cdata) {
                            newNode.append(xmlUtils.cdata(settings.content));
                        } else {
                            newNode.text(settings.content);
                        }
                    } else {
                        that._applyTitle(newNode, settings);
                        var manager = mvc.Components.get(element.model.entry.content.get('display.general.manager'));
                        if(options.useLoadjob !== false && manager && manager.job) {
                            that._applySearch(newNode, {
                                search: {
                                    type: 'inline',
                                    search: '| loadjob ' + manager.job.sid + ' ignore_running=t'
                                }
                            });
                        } else {
                            that._applySearch(newNode, settings);
                        }

                        that._applyOptions(newNode, settings);
                        that._applyTags(newNode, settings);
                    }

                    newNode.appendTo(row);
                });

                row.appendTo(dashboard);

            });

            if(options && options.indent) {
                xmlUtils.formatXMLDocument(dashboard);
            }

            return xmlUtils.serialize(dashboard);
        },
        /**
         * Update the dashboard element with the settings specified
         * @param id - the element id (string)
         * @param settings - a settings object containing the update information
         * @param options - persistence options
         *
         * settings contains: {
         *      type: (String) the element type (one of "table", "chart", "single", "map", "list", "html")
         *      search: (Object) {
         *          type: (String) one of 'inline', 'saved' or 'pivot'
         *          search: (String) the search string (for inline)
         *          earliest_time: (String) the earliest_time of the inline search
         *          latest_time: (String) the latest_time of the inline search
         *          name: (String) name of the saved search
         *          ... TODO pivot search params
         *      }
         *      options: (Object) options to added (or replaced) to the element (<option name="$name$">$value$</option>)
         *      removeOptions: (Array) options to be removed from the xml element
         *      tags: (Object) tags to be added (or replaced) to the element (<$tag$>$value$</$tag$>)
         * },
         * options: {
         *      clearOptions: (Boolean) true to remove all options nodes before updating the XML element
         * }
         */
        updateElement: function(id, settings, options) {
            console.log('About to update item=%o with settings=%o', id, settings);
            var $xml = this.get$XML();
            var cur = $xml.find('#' + id);
            if(!cur.length) {
                var itemIndex = _(this._itemOrder).chain().flatten().indexOf(id).value();
                console.log('Node with ID=%o not found. Search for item with index=%o', id, itemIndex);
                if(itemIndex === -1) {
                    throw new Error("Unable to find dashboard element with ID " + id);
                }
                cur = $xml.find('row').children()[ itemIndex ];
                if(!cur) {
                    throw new Error("Unable to find dashboard element with ID " + id);
                } else {
                    cur = $(cur);
                }
            }
            cur = cur[0];
            console.log("Updating XML node: ", cur);
            xmlUtils.stripEmptyTextNodes(cur);

            var newNode = xmlUtils.$node('<' + settings.type + '/>');

            if ($(cur).attr('id')) {
                newNode.attr('id', $(cur).attr('id'));
            }

            console.log('children', cur.childNodes);
            while(cur.childNodes.length) {
                newNode.append(cur.childNodes[0]);
            }

            if(options && options.clearOptions) {
                newNode.find('option').remove();
                delete settings.options; 
            }

            this._applyTitle(newNode, settings);
            if (settings.search && settings.search.type === 'global') {
                this._applyGlobalSearch($xml, settings);
            } else {
                this._applySearch(newNode, settings);
            }
            this._applyOptions(newNode, settings);
            this._applyTags(newNode, settings);

            $(cur).replaceWith(newNode);
            if(console.DEBUG_ENABLED) {
                xmlUtils.formatXMLDocument($xml);
                console.log(xmlUtils.serialize($xml));
            }
            this.set$XML($xml);
            return this.save();
        },
        deleteElement: function(id) {
            // Create new item order without the element which is to be deleted
            var newOrder = _(this._itemOrder).map(function(row) {
                return _(row).map(function(panel) {
                    var idx = _(panel).indexOf(id);
                    return idx > -1 ? panel.slice(0, idx).concat(panel.slice(idx + 1)) : _.clone(panel);
                });
            });

            return this.setItemOrder(newOrder);
        },
        addElement: function(id, settings) {
            var row = xmlUtils.$node('<row/>'), newNode = xmlUtils.$node('<' + settings.type + '/>');
            this._itemOrder.push([
                [settings.id]
            ]);
            row.append(newNode);
            this._applyTitle(newNode, settings);
            this._applySearch(newNode, settings);
            var $xml = this.get$XML();
            $xml.find(':eq(0)').append(row);
            if(console.DEBUG_ENABLED) {
                xmlUtils.formatXMLDocument($xml);
                console.log(xmlUtils.serialize($xml));
            }
            this.set$XML($xml);
            return this.save();
        },
        saveFormSettings: function(isForm, haveTimeRangePicker, timeRangePickerDefault) {
            var xml = this.get$XML(), root, fieldset;
            if(isForm) {
                xml = this._migrateViewType(xml, 'form');
                root = xml.find(':eq(0)');
                fieldset = xml.find(':eq(0)>fieldset');
                if(!fieldset.length) {
                    fieldset = xmlUtils.$node('<fieldset autoRun="true" submitButton="false"></fieldset>');
                    var el = xml.find(':eq(0)>description');
                    if(el.length) {
                        fieldset.insertAfter(el);
                    } else {
                        el = xml.find(':eq(0)>label');
                        if(el.length) {
                            fieldset.insertAfter(el);
                        } else {
                            fieldset.prependTo(root);
                        }
                    }
                }
                var timeInput = _(fieldset.find('input[type="time"]')).find(function(el){
                    return !$(el).attr('token');
                });
                if(haveTimeRangePicker) {
                    var defNode;
                    if(timeInput) {
                        defNode = $(timeInput).find('default');
                        if(!defNode.length) {
                            defNode = xmlUtils.$node('<default></default>').appendTo(timeInput);
                        }
                    } else {
                        timeInput = xmlUtils.$node('<input type="time" searchWhenChanged="true" />');
                        defNode = xmlUtils.$node('<default></default>').appendTo(timeInput);
                        fieldset.append(timeInput);
                    }

                    if(timeRangePickerDefault){
                        if(_.isString(timeRangePickerDefault)) {
                            defNode.text(timeRangePickerDefault);
                        } else {
                            defNode.empty();
                            if(timeRangePickerDefault.earliestTime) {
                                xmlUtils.$node('<earliestTime></earliestTime>').text(timeRangePickerDefault.earliestTime).appendTo(defNode);
                            }
                            if(timeRangePickerDefault.latestTime) {
                                xmlUtils.$node('<latestTime></latestTime>').text(timeRangePickerDefault.latestTime).appendTo(defNode);
                            }
                        }
                    } else {
                        defNode.remove();
                    }
                } else {
                    $(timeInput).remove();
                }
            } else {
                xml = this._migrateViewType(xml, 'dashboard');
                xml.find(':eq(0)>fieldset').remove();
            }
            if(console.DEBUG_ENABLED) {
                xmlUtils.formatXMLDocument(xml);
                console.log(xmlUtils.serialize(xml));
            }
            this.set$XML(xml);
            this.save();
        },
        _migrateViewType: function($xml, tagName) {
            var curRoot = $xml.find(':eq(0)');
            if(curRoot.prop('tagName') !== tagName) {
                $xml = $($.parseXML('<' + tagName + '/>'));
                var newRoot = $xml.find(':eq(0)'), cur = curRoot[0];
                _(cur.attributes).each(function(attr){
                    newRoot.attr(attr.nodeName, attr.nodeValue);
                });
                while(cur.childNodes.length) {
                    newRoot.append(cur.childNodes[0]);
                }
            }
            return $xml;
        },
        _applyOptions: function(newNode, settings) {
            if(settings.options) {
                _.each(settings.options, function(value, name) {
                    var curOption = newNode.find('option[name="' + name + '"]');
                    if(value === "" || value === null || value === void(0)) {
                        curOption.remove();
                    } else {
                        if(curOption.length) {
                            curOption.text(value);
                        } else {
                            xmlUtils.$node('<option/>').attr('name', name).text(value).appendTo(newNode);
                        }
                    }
                });
            }
            if(settings.removeOptions) {
                _(settings.removeOptions).each(function(name) {
                    newNode.find('option[name="' + name + '"]').remove();
                });
            }
        },
        _applyTags: function(newNode, settings) {
            if(settings.tags) {
                _.each(settings.tags, function(value, tag) {
                    newNode.find(tag).remove();
                    if((_.isArray(value) && value.length) || value) {
                        xmlUtils.$node('<' + tag + '/>').text(value).appendTo(newNode);
                    }
                });
            }
        },
        _applyTitle: function(newNode, settings) {
            var titleNode = newNode.find('title');
            if(settings.title) {
                if(!titleNode.length) {
                    titleNode = xmlUtils.$node('<title/>').prependTo(newNode);
                }
                titleNode.text(settings.title);
            }
        },
        _applySearch: function(newNode, settings) {
            if(settings.search) {
                // Clear current search info
                newNode.find('searchString,searchTemplate,searchName,searchPostProcess,pivotSearch,earliestTime,latestTime').remove();
                var titleNode = newNode.find('title');
                switch(settings.search.type) {
                    case 'inline':
                    case 'global': // Apply global search as an inline search for flattened XML
                        var searchNode = xmlUtils.$node('<searchString/>').text(settings.search.search);
                        if(titleNode.length) {
                            searchNode.insertAfter(titleNode);
                        } else {
                            searchNode.prependTo(newNode);
                        }
                        if(settings.search.latest_time) {
                            xmlUtils.$node('<latestTime/>').text(settings.search.latest_time).insertAfter(searchNode);
                        }
                        if(settings.search.earliest_time) {
                            xmlUtils.$node('<earliestTime/>').text(settings.search.earliest_time).insertAfter(searchNode);
                        }
                        break;
                    case 'postprocess':
                        var postSearchNode = xmlUtils.$node('<searchPostProcess/>').text(settings.search.search);
                        if(titleNode.length) {
                            postSearchNode.insertAfter(titleNode);
                        } else {
                            postSearchNode.prependTo(newNode);
                        }
                        break;
                    case 'saved':
                        var savedNode = xmlUtils.$node('<searchName/>').text(settings.search.name);
                        if(titleNode.length) {
                            savedNode.insertAfter(titleNode);
                        } else {
                            savedNode.prependTo(newNode);
                        }
                        break;
                    case 'pivot':
                        throw new Error("Pivot search not implemented!");
                    default:
                        throw new Error("Unknown search type: " + settings.search.type);
                }
            }
        },
        _applyGlobalSearch: function($xml, settings) {
            var root = $xml.find(':eq(0)');
            root.children('searchTemplate,earliestTime,latestTime').remove();
            var labelNode = root.children('label');
            var descNode = root.children('description');

            var searchNode = xmlUtils.$node('<searchTemplate/>').text(settings.search.search);
            if (descNode.length) {
                searchNode.insertAfter(descNode);
            } else if (labelNode.length) {
                searchNode.insertAfter(labelNode);
            } else {
                searchNode.prependTo(root);
            }
            if (settings.search.latest_time) {
                xmlUtils.$node('<latestTime/>').text(settings.search.latest_time).insertAfter(searchNode);
            }
            if (settings.search.earliest_time) {
                xmlUtils.$node('<earliestTime/>').text(settings.search.earliest_time).insertAfter(searchNode);
            }
        },
        _clearEmptyItems: function(itemOrder){
            return _(itemOrder).chain().map(function(row){
                return _(row).filter(function(panel){ return panel.length > 0; });
            }).filter(function(row){ return row.length > 0; }).value();
        },
        captureItemOrder: function(itemOrder) {
            itemOrder = this._clearEmptyItems(itemOrder);

            this._itemOrder = itemOrder;
            this._itemOrderMap = _.object(_.flatten(itemOrder), this.get$XML().find('row').children());

            this.validateItemOrder();

            if(console.DEBUG_ENABLED) {
                console.log('Captured dashboard item order: %o', JSON.stringify(this._itemOrder));
            }
        },
        updateItemOrder: function(){
            if (this._itemOrder && this.isXML()) {
                this.captureItemOrder(this._itemOrder);
            }
        },
        validateItemOrder: function() {
            _(this._itemOrderMap).each(function(node, id) {
                var nid = $(node).attr(id);
                if(!(nid === undefined || nid === id)) {
                    throw new Error('Invalid Item order. Expected node with ID ' + id + '. Instead saw ' + nid);
                }
            });
        },
        setItemOrder: function(itemOrder) {
            itemOrder = this._clearEmptyItems(itemOrder);
            var dfd;
            if(!this._itemOrder) {
                this.captureItemOrder(itemOrder);
            } else if(!_.isEqual(itemOrder, this._itemOrder)) {
                var $xml = this.get$XML(), itemMap = this._itemOrderMap;
                $xml.find('row').remove();

                _(itemOrder).each(function(r) {
                    var row = xmlUtils.$node('<row/>');
                    if(_(r).any(function(panel) { return panel.length > 1; })) {
                        var groups = _(r).map(function(p) { return p.length; }).join(',');
                        row.attr('grouping', groups);
                    }
                    _.chain(r).flatten().each(function(el) {
                        var item = itemMap[el];
                        $(item).find('script').remove();
                        row.append(item);
                    });
                    if(!row.children().length) {
                        if(console.DEBUG_ENABLED) {
                            console.log('Created empty row for items %o and map %o', r, itemMap);
                        }
                    }
                    $xml.find('dashboard,form').append(row);
                });
                this._itemOrder = itemOrder;
                this.set$XML($xml);
                console.log('Saving new item order', itemOrder, $xml[0]);
                dfd = this.save();
            } else {
                console.log('no changes');
                dfd = $.Deferred();
                dfd.resolve();
            }
            return dfd;
        },
        isEditable: function() {
            return this.isDashboard() || this.isForm();
        }
    });

    var DashboardRouter = Backbone.Router.extend({
        initialize: function(options) {
            this.model = options.model;
            this.app = options.app;
        },
        routes: {
            ':locale/app/:app/:page?*qs': 'view',
            ':locale/app/:app/:page': 'view',
            ':locale/app/:app/:page/?*qs': 'view',
            ':locale/app/:app/:page/': 'view',
            ':locale/app/:app/:page/edit?*qs': 'edit',
            ':locale/app/:app/:page/edit': 'edit',
            '*root/:locale/app/:app/:page?*qs': 'rootedView',
            '*root/:locale/app/:app/:page': 'rootedView',
            '*root/:locale/app/:app/:page/?*qs': 'rootedView',
            '*root/:locale/app/:app/:page/': 'rootedView',
            '*root/:locale/app/:app/:page/edit?*qs': 'rootedEdit',
            '*root/:locale/app/:app/:page/edit': 'rootedEdit',
            ':locale/manager/:app/:page?*qs': 'view',
            ':locale/manager/:app/:page': 'view',
            ':locale/manager/:app/:page/?*qs': 'view',
            ':locale/manager/:app/:page/': 'view',
            '*root/:locale/manager/:app/:page?*qs': 'rootedView',
            '*root/:locale/manager/:app/:page': 'rootedView',
            '*root/:locale/manager/:app/:page/?*qs': 'rootedView',
            '*root/:locale/manager/:app/:page/': 'rootedView',
            'dj/:app/:page/': 'splunkdj',
            'dj/:app/:page/?*qs': 'splunkdj',
            '*root/dj/:app/:page/': 'rootedSplunkdj',
            '*root/dj/:app/:page/?*qs': 'rootedSplunkdj'
        },
        view: function() {
            console.log('ROUTE: view');
            this.page.apply(this, arguments);
            this.model.set('edit', false);
        },
        edit: function() {
            console.log('ROUTE: edit');
            this.page.apply(this, arguments);
            this.model.set('edit', true);
        },
        rootedView: function(root) {
            this.app.set('root', root);
            this.view.apply(this, _.rest(arguments));
        },
        rootedEdit: function(root) {
            this.app.set('root', root);
            this.edit.apply(this, _.rest(arguments));
        },
        page: function(locale, app, page) {
            console.log('ROUTE: page(locale=%o, app=%o, page=%o)', locale, app, page);
            this.app.set({
                locale: locale,
                app: app,
                page: page
            });
            classicurl.fetch();
        },
        splunkdj: function(app, page) {
            this.page('en-US', app, page);
        },
        rootedSplunkdj: function(root) {
            this.app.set('root', root);
            this.splunkdj.apply(this, _.rest(arguments));
        },
        updateUrl: function() {
            var parts = [ this.app.get('root') || '', this.app.get('locale'), 'app', this.app.get('app'), this.app.get('page') ];
            if(this.model.get('edit')) {
                parts.push('edit');
            }
            var url = [ parts.join('/') ], params = classicurl.toURLEncodedQueryString();
            if(params.length) {
                url.push(params);
            }
            this.navigate(url.join('?'), { replace: false });
        }
    });

    // Singleton dashboard controller that sets up the router and holds a model representing the state of the dashboard
    var DashboardController = function() {
        this.readyDfd = $.Deferred();
        var model = this.model = new Backbone.Model();
        var collection = this.collection = {};
        
        // Set up the shared models/collections
        var app = this.model.app = sharedModels.get("app");
        var appLocal = this.model.appLocal = sharedModels.get("appLocal");
        var user = this.model.user = sharedModels.get("user");
        var times = this.collection.times = sharedModels.get("times");
        
        this.router = new DashboardRouter({
            model: this.model,
            app: app
        });
        routerUtils.start_backbone_history();
        var view = this.model.view = new DashboardModel();
        this.model.view.fetch({
            url: route.splunkdNS(app.get('root'), app.get('locale'), app.get('owner'), app.get('app'), [view.url, app.get('page')].join('/'))
        }).done(this._onViewModelLoad.bind(this));
        pdfUtils.isPdfServiceAvailable().always(function(available) {
            model.set('pdf_available', available);
        });
        this._onViewModelLoadDfd = $.Deferred();
        this.model.on('change:edit', this.router.updateUrl, this.router);
    };
    _.extend(DashboardController.prototype, Backbone.Events, {
        _onViewModelLoad: function() {
            var model = this.model;
            model.set('editable', model.view.isEditable());
            if(model.view.isXML()) {
                model.set('label', this.model.view.getLabel());
                model.set('description', this.model.view.getDescription());
                model.on('change:label change:description', function() {
                    model.view.setLabelAndDescription(model.get('label'), model.get('description'));
                    model.view.save();
                });
            }
            this._onViewModelLoadDfd.resolve(this.model.view);
        },
        onViewModelLoad: function(cb, scope) {
            this._onViewModelLoadDfd.done(cb.bind(scope || null));
        },
        getStateModel: function() {
            return this.model;
        },
        isEditMode: function() {
            return this.model.get('edit') === true;
        },
        onReady: function(callback) {
            $.when(this.readyDfd, this._onViewModelLoadDfd).then(callback);
        },
        ready: function() {
            this.readyDfd.resolve();
        },
        isReady: function(){
            return this.readyDfd.state() === "resolved";
        },
        fetchCollection: function() {
            this.reportsCollection = new Reports();
            this.reportsCollection.REPORTS_LIMIT = 150; 
            var appModel = this.model.app;
            var fetchParams = {
                data: {
                    count: this.reportsCollection.REPORTS_LIMIT, 
                    app: appModel.get('app'), 
                    owner: appModel.get('owner'), 
                    search: 'is_visible="1"'
                }
            };
            this.reportsCollection.initialFetchDfd = this.reportsCollection.fetch(fetchParams);
        }
    });

    var instance = new DashboardController();
    if(console.DEBUG_ENABLED) {
        window.Dashboard = instance;
    }
    return instance;
});
