define(function(require, exports, module) {
    var $ = require("jquery");
    var _ = require("underscore");
    var mvc = require("./mvc");
    var BaseModel = require("models/Base");
    var BaseSplunkView = require("./basesplunkview");
    var Messages = require("./messages");
    var SplunkUtil = require("splunk.util");
    var Map = require("views/shared/Map");
    var Backbone = require('backbone');
    var utils = require('./utils');
    var route = require("uri/route");
    var config = require('JS_CACHE/config');
    var console = require('util/console');
    var resizable = require('jquery.ui.resizable');
    var util = require('splunk.util');
    var TokenAwareModel = require('./tokenawaremodel');

    var TileSources = {
        openStreetMap: {
            url: 'http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
            maxZoom: 18
        }
        // TODO: dynamically add Splunk tile source info
    };

    var SplunkMapView = BaseSplunkView.extend({
        moduleId: module.id,

        className: "splunk-map",

        options: {
            data: "preview",
            height: '400px',
            tileSource: undefined,
            tileUrl: undefined,
            maxZoom: undefined,
            drilldown: true,
            drilldownRedirect: true,
            resizable: false
        },

        omitFromSettings: ['el', 'reportModel'],

        initialize: function() {
            this.configure();
            this.model = this.options.reportModel || TokenAwareModel._createReportModel();

            this.settings._sync = utils.syncModels(this.settings, this.model, {
                auto: true,
                prefix: 'display.visualizations.',
                alias: {
                    tileUrl: 'display.visualizations.mapping.tileLayer.url',
                    maxZoom: 'display.visualizations.mapping.tileLayer.maxZoom'
                },
                exclude: ['managerid','data','tileSource','drilldown','type', 'drilldownRedirect']
            });

            this.results = new Backbone.Model({
                columns: [],
                fields: []
            });

            this._currentHeight = parseInt(this.settings.get('height'), 10);

            // initialize containers as detached DOM
            this.$map = $('<div></div>');
            this.$msg = $('<div></div>');

            this.createMap = _.debounce(_.bind(this.createMap, this), 0);
            this.createMap();
            this.settings.on('change:tileSource', this._updateTileSource, this);
            this._updateTileSource();
            this.bindToComponentSetting('managerid', this._onManagerChange, this);
        },

        _updateTileSource: function() {
            if(this.settings.has('tileSource')) {
                var tileSource = TileSources[this.settings.get('tileSource')];
                if(tileSource) {
                    this.settings.set({
                        tileUrl: tileSource.url,
                        maxZoom: tileSource.maxZoom
                    });
                } else {
                    console.warn('Invalid tileSource parameter value=%o', tileSource);
                }
            }
        },

        destroyMap: function() {
            if(this.map) {
                this.map.off();
                this.map.remove();
                this.map = null;
            }
        },

        createMap: function() {
            this.destroyMap();
            this.map = new Map({
                el: this.$map.appendTo(this.el),
                model: {
                    searchResultsColumn: this.results,
                    state: this.model
                },
                height: this.settings.get('height')
            });
            this.map.on('boundsChanged', this._onBoundsChanged, this);
            this.map.on('mapClicked', this._onMapClicked, this);

            // force data to update according to new map bounds
            this._onBoundsChanged();

            // Setup resizing
            this._onResizeMouseup = _.bind(this._onResizeMouseup, this);
            this.settings.on('change:resizable', function(model, value, options) {
                value ? this._enableResize() : this._disableResize();
            }, this);
            if (this.settings.get('resizable')) {
                this._enableResize();
            }
        },

        _enableResize: function() {
            if (this._canEnableResize()) {
                this.map.$el.resizable({autoHide: true, handles: "s", stop: this._onResizeStop.bind(this)});
                // workaround until jquery ui is updated
                this.map.$el.on('mouseup', this._onResizeMouseup);
            }
        },
        
        _disableResize: function() {
            if (this._canEnableResize()) {
                this.map.$el.resizable('destroy');
                this.map.$el.off('mouseup', this._onResizeMouseup);
            }
        },
        
        // NOTE: Bound to 'this' by createMap
        _onResizeMouseup: function(event) {
            $(this).width("100%");
        },

        _onResizeStop: function(event, ui) {
            $(event.target).width("100%");
            this._currentHeight = parseInt(this.settings.get('height'), 10);
            this.$msg.height(this._currentHeight).css('overflow', 'hidden');
        },
        
        _canEnableResize: function() {
            // Disable resizing for safari 3 and below only
            return (!($.browser.safari && $.browser.version < "526"));
        },

        _onManagerChange: function(managers, manager) {
            if (this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }
            if (this.resultsModel) {
                this.resultsModel.off(null, null, this);
                this.resultsModel = null;
            }

            this._clearResults();

            if (!manager) {
                this.message("no-search");
                return;
            }

            this.manager = manager;
            this.manager.on("search:start", this._onSearchStart, this);
            this.manager.on("search:progress", this._onSearchProgress, this);
            this.manager.on("search:cancelled", this._onSearchCancelled, this);
            this.manager.on("search:fail", this._onSearchFail, this);
            this.manager.on("search:error", this._onSearchError, this);

            this.resultsModel = this.manager.data(this.settings.get("data"), {
                serialfetch: true,
                output_mode: "json_cols",
                offset: 0,
                count: this.map ? this.map.getMaxClusters() : 100,
                show_empty_fields: "True",
                search: this.map ? this.map.getPostProcessSearch() : ''
            });
            this.resultsModel.on("data", this._onDataUpdate, this);
            this.resultsModel.on("error", this._onSearchError, this);

            this.manager.replayLastSearchEvent(this);
        },

        _onSearchStart: function() {
            this._clearResults();
        },

        _onSearchProgress: function(properties) {
            // No action require on progress at the moment
        },

        _onSearchCancelled: function() {
            this.message('cancelled');
        },

        _onSearchFail: function(state) {
            var msg = Messages.getSearchFailureMessage(state);
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onSearchError: function(message, err) {
            var msg = Messages.getSearchErrorMessage(err) || message;
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onDataUpdate: function() {
            if (this.resultsModel.hasData()) {
                this.$msg.detach();
                this.$map.css("visibility", "");
                this.results.set(this.resultsModel.data());
            } else {
                this._clearResults();
            }
            this.render();
        },

        _onBoundsChanged: function(){
            if (this.resultsModel) {
                this.resultsModel.set('search', this.map.getPostProcessSearch());
            }
        },

        _onMapClicked: function(e) {
            if (!this.settings.get('drilldown')) {
                return;
            }

            e = e || {};

            // Set up the handlers for default drilldown behavior and preventing
            // the default behavior
            var preventRedirect = false;
            var defaultDrilldown = _.once(_.bind(this._doDrilldown, this, e, this.manager));
            var preventDefault = function() {
                preventRedirect = true;
            };

            e.preventDefault = preventDefault;
            e.drilldown = defaultDrilldown;

            this.trigger('drilldown click click:marker', e, this);

            // Now that the event is trigged, if there is a default action of
            // redirecting, we will execute it (depending on whether the user
            // executed preventDefault()).
            if (this.settings.get("drilldownRedirect") && !preventRedirect) {
                defaultDrilldown();
            }
        },

        _doDrilldown: function(event, manager) {
            if (!event || !manager)
                return;

            var data = event.data;
            if (!data || !data._geo_lat_field || !data._geo_long_field ||
                !data._geo_bounds_south || !data._geo_bounds_west || !data._geo_bounds_north || !data._geo_bounds_east)
                return;

            var options = {
                _geo_lat_field: data._geo_lat_field,
                _geo_lon_field: data._geo_long_field,
                _geo_bounds_south: data._geo_bounds_south,
                _geo_bounds_west: data._geo_bounds_west,
                _geo_bounds_north: data._geo_bounds_north,
                _geo_bounds_east: data._geo_bounds_east
            };

            if (event.altKey)
                options.negate = "true";

            var earliest, latest;
            if (manager.job) {
                earliest = manager.job.properties().searchEarliestTime;
                latest = manager.job.properties().searchLatestTime;
                if (!earliest && manager.job.properties().earliestTime){
                    earliest = util.getEpochTimeFromISO(manager.job.properties().earliestTime);
                }
                if (!latest && manager.job.properties().latestTime){
                    latest = util.getEpochTimeFromISO(manager.job.properties().latestTime);
                }
            }

            this._parseIntentions(manager.query.resolve(), "geoviz", options).done(function(searchString) {
                var params = { 
                    q: searchString,
                    earliest: earliest || '',
                    latest: latest || ''
                };

                var pageInfo = utils.getPageInfo();
                var url = route.search(pageInfo.root, pageInfo.locale, pageInfo.app, { data: params });
                utils.redirect(url, event.ctrlKey);
            });
        },

        _parseIntentions: function(searchString, action, options) {
            var dfd = $.Deferred();

            var params = {};
            for (var o in options) {
                if (options.hasOwnProperty(o))
                    params[o] = options[o];
            }
            params.q = SplunkUtil.addLeadingSearchCommand(searchString);
            params.action = action;
            params.output_mode = "json";

            $.ajax({
                url: SplunkUtil.make_url("/splunkd/__raw/services/search/intentionsparser"),
                data: params,
                type: "get",
                dataType: "json",
                async: true,
                success: function(data) {
                    dfd.resolve(SplunkUtil.stripLeadingSearchCommand(data.fullSearch));
                },
                error: function(xhr, status, error) {
                    dfd.reject(error);
                }
            });

            return dfd.promise();
        },

        _clearResults: function() {
            this.results.set({
                columns: [],
                fields: []
            });
        },

        message: function(info) {
            // Cannot use .hide() on map container because it screws up initial map rendering.
            // Map container dimensions must always evaluate to non-zero.
            // Using visibility hidden instead.
            this.$map.css("visibility", "hidden");
            this.$msg.detach();
            Messages.render(info, this.$msg);
            this.$msg.prependTo(this.$el);
        },

        render: function() {
            this.$el.css({ 'overflow': 'hidden', 'position': 'relative' });
            this.$msg.height(this._currentHeight).css({ 'overflow': 'hidden', 'position': 'absolute', 'width': '100%' });
            return this;
        },

        remove: function() {
            this.destroyMap();
            if(this.settings) {
                this.settings.off();
                if(this.settings._sync) {
                    this.settings._sync.destroy();
                }
            }
            BaseSplunkView.prototype.remove.call(this);
        }
    });

    return SplunkMapView;
});
