define(function (require, exports, module) {
    var $ = require("jquery");
    var _ = require("underscore");
    var mvc = require('./mvc');
    var utils = require('./utils');
    var SplunkUtil = require('splunk.util');
    var Backbone = require('backbone');
    var BaseSplunkView = require("./basesplunkview");
    var ResultsTableView = require('views/shared/results_table/ResultsTableMaster');
    var Messages = require("./messages");
    var Drilldown = require('./drilldown');
    var PaginatorView = require("./paginatorview");
    var BaseCellRenderer = require('views/shared/results_table/renderers/BaseCellRenderer');
    var console = require('util/console');
    var util = require('splunk.util');
    var TokenAwareModel = require('./tokenawaremodel');

    require("css!../css/results-table");

    // This regex will take a space or comma separated list of fields, with quotes
    // for escaping strings with spaces in them, and match each individual 
    // field.
    var fieldSplitterRegex = /(["'].*?["']|[^"',\s]+)(?=\s*|\s*,|\s*$)/g;
    
    // This regex will take a string that may or may not have leading quotes,
    // and strip them.
    var quoteStripperRegex = /^["']|["|']$/g;

    var TableView = BaseSplunkView.extend({
        moduleId: module.id,
        
        className: 'splunk-table',
        options: {
            data: 'preview',
            drilldown: 'row',
            drilldownRedirect: true,
            pagerPosition: 'bottom',
            fields: [],
            pageSize: 10
        },
        normalizeOptions: function(options) {
            if(options.hasOwnProperty('count')) {
                this.settings.set('pageSize', parseInt(this.options.count, 10));
            }
            if(options.hasOwnProperty('wrap')) {
                this.settings.set('wrap', SplunkUtil.normalizeBoolean(options.wrap) ? '1' : '0');
            }
            if(options.hasOwnProperty('rowNumbers')) {
                this.settings.set('displayRowNumbers', SplunkUtil.normalizeBoolean(options.rowNumbers) ? '1' : '0');
            } else if(options.hasOwnProperty('displayRowNumbers')) {
                this.settings.set('displayRowNumbers', SplunkUtil.normalizeBoolean(options.displayRowNumbers) ? '1' : '0');
            }
            if(options.hasOwnProperty('overlay')) {
                this.settings.set('dataOverlayMode', options['display.statistics.overlay']);
            }
            if(options.hasOwnProperty('drilldown')) {
                if (options.drilldown === 'all') {
                    this.settings.set('drilldown', 'cell');
                } else if (options.drilldown === 'off') {
                    this.settings.set('drilldown', 'none');
                } else {
                    this.settings.set('drilldown', options.drilldown);
                }
            }
            if(options.format) {
                var sparklineSettings = {};
                _.each(options.format, function(formats,field) {
                    _.each(formats, function(format){
                        if(format.type === 'sparkline') {
                            sparklineSettings[field] = format.options;
                        }
                    });
                });

                if(!_(sparklineSettings).isEmpty()) {
                    this.settings.set({
                        "sparkline.format": sparklineSettings
                    });
                }
            }
        },
        initialize: function (options) {
            this.configure();
            this.model = this.options.reportModel || TokenAwareModel._createReportModel();
            this.normalizeOptions(options);
            this.settings._sync = utils.syncModels(this.settings, this.model, {
                auto: true,
                alias: {
                    'displayRowNumbers': 'display.statistics.rowNumbers',
                    'dataOverlayMode': 'display.statistics.overlay',
                    'pageSize': 'display.prefs.statistics.count'
                },
                prefix: 'display.statistics.',
                include: ['displayRowNumbers','dataOverlayMode','pageSize','stripedRows','drilldown','sortKey',
                    'sortDirection','overlay','count','sparkline.format','wrap','fields'],
                exclude: ['drilldownRedirect']
            });

            var results = this.results = new Backbone.Model({
                rows: [],
                fields: []
            });
        
            var metadata = this.metadata = new Backbone.Model({
                pageSize: this.settings.get('pageSize') || 10,
                offset: 0
            });

            var table = this.table = new ResultsTableView({
                enableTableDock: false,
                el: $('<div></div>').appendTo(this.el),
                model: {
                    results: results,
                    metadata: metadata,
                    format: this.model
                }
            });

            this.listenTo(this.table, 'rendered', function(){ this.trigger('rendered', this); }, this);

            this.settings._syncMeta = utils.syncModels(this.settings, this.metadata, {
                include: ['sortKey','sortDirection', 'pageSize'],
                auto: 'push'
            });

            metadata.on('change', this.applyFetchOptions, this);

            if(this.options.showPager !== false) {
                var ct = $('<div></div>')[this.options.pagerPosition === 'bottom' ? 'appendTo':'prependTo' ](this.el);
                var p = this.paginator = new PaginatorView({
                    id: _.uniqueId(this.id+'-paginator'),
                    el: ct,
                    pageSize: metadata.get('pageSize')
                });
                p.settings.on('change:page', function(){
                    var count = metadata.get('pageSize'), page = p.settings.get('page');
                    this.metadata.set('offset', count*page);
                }, this);
            }

            this.transformFieldsToArray();
            this.listenTo(this.settings, 'change:fields', this.onFieldsChanged);

            table.on('cellClick', this.emitDrilldownEvent, this);
            this.bindToComponentSetting('managerid', this.onManagerChange, this);
            
            // If we don't have a manager by this point, then we're going to
            // kick the manager change machinery so that it does whatever is
            // necessary when no manager is present.
            if (!this.manager) {
                this.onManagerChange(mvc.Components, null);
            }
        },

        emitDrilldownEvent: function(e) {
            var data = {},
                manager = this.manager;
            _.extend(data,{
                "click.name": e.name,
                "click.value": e.value
            });
            if(e.hasOwnProperty('name2')) {
                _.extend(data, {
                    "click.name2": e.name2,
                    "click.value2": e.value2
                });
            } else {
                _.extend(data, {
                    "click.name2": e.name,
                    "click.value2": e.value
                });
            }
            _.extend(data, e.rowContext);
            
            // Get all original row data, as e.rowContext only has data
            // for the visible fields in the table.
            var originalRows = this.resultsModel ? this.resultsModel.data().rows : null;
            if (originalRows && (e.rowIndex < originalRows.length)) {
                var originalRow = originalRows[e.rowIndex];
                var originalFields = this.resultsModel.data().fields;
                _(originalFields).each(function(originalField, i) {
                    originalField = "row." + originalField;
                    if (!(originalField in data))
                        data[originalField] = originalRow[i];
                });
            }
            
            // Only get the job information if it exists (e.g. we are not
            // bound to a search with a job)
            if (manager.job) {
                var earliest = manager.job.properties().searchEarliestTime;
                var latest = manager.job.properties().searchLatestTime;
                if (!earliest && manager.job.properties().earliestTime){
                    earliest = util.getEpochTimeFromISO(manager.job.properties().earliestTime);
                }
                if (!latest && manager.job.properties().latestTime){
                    latest = util.getEpochTimeFromISO(manager.job.properties().latestTime);
                }
                _.extend(data, {
                    earliest: earliest || '',
                    latest: latest || ''
                });
            }
            else {
                // If we have no source of earliest/latest, default to ''.
                _.extend(data, {
                    earliest: '',
                    latest: ''
                });
            }

            // Set up the handlers for default drilldown behavior and preventing
            // the default behavior
            var preventRedirect = false;
            var defaultDrilldown = _.once(_.bind(this.drilldown, this, e));
            var preventDefault = function() {
                preventRedirect = true;  
            };
            
            this.trigger('drilldown click', 
                { 
                    field: e.name2 || e.name, 
                    data: data, 
                    event: e, 
                    preventDefault: preventDefault,
                    drilldown: defaultDrilldown
                },
                this
            );
            
            var rowIndex = e.rowIndex;
            var cellIndex = e.cellIndex;

            var model = this.collection.at(rowIndex);
            if (model) {
                var key = e.hasOwnProperty("name2") ? e.name2 : e.name;
                var value = e.hasOwnProperty("value2") ? e.value2 : e.value;
                var originalEvent = e.originalEvent;

                this.trigger("clicked:row click:row",
                    {
                        model: model,
                        index: rowIndex,
                        originalEvent: originalEvent,
                        event: originalEvent,
                        component: this,
                        preventDefault: preventDefault,
                        drilldown: defaultDrilldown
                    },
                    this
                );
                this.trigger("clicked:cell click:cell",
                    {
                        model: model,
                        index: rowIndex,
                        column: cellIndex,
                        key: key,
                        value: value,
                        originalEvent: originalEvent,
                        event: originalEvent,
                        component: this,
                        preventDefault: preventDefault,
                        drilldown: defaultDrilldown
                    },
                    this
                );
            }          
            
            // Now that the event is trigged, if there is a default action of
            // redirecting, we will execute it (depending on whether the user
            // executed preventDefault()).
            if (this.settings.get('drilldown') !== 'none' && this.settings.get("drilldownRedirect") && !preventRedirect) {
                defaultDrilldown();
            }
        },

        drilldown: function(clickInfo, target) {
            Drilldown.handleDrilldown(clickInfo, this.settings.get('drilldown'), this.manager, target);
        },

        onManagerChange: function (managers, manager) {
            if(this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }
            if(this.resultsModel) {
                this.resultsModel.off(null, null, this);
                this.resultsModel = null;
            }
            
            if (!manager) {
                this.renderMessage("no-search");
                return;
            }
            
            // Clear any messages, since we have a new manager.
            this.renderMessage("empty");
            
            this.manager = manager;
            this.resultsModel = this.manager.data(this.settings.get("data"), {
                autofetch: true,
                output_mode: "json_rows",
                show_empty_fields: "True",
                count: this.metadata.get('pageSize'),
                offset: this.metadata.get('offset')
            });
            this.manager.on("search:start", this.onSearchStart, this);
            this.manager.on("search:progress", this.onSearchProgress, this);
            this.manager.on("search:cancelled", this.onSearchCancelled, this);
            this.manager.on("search:fail", this.onSearchFailed, this);
            this.manager.on("search:error", this.onSearchError, this);
            this.manager.on("search:done", this.onSearchDone, this);
            this.resultsModel.on("data", this.renderResults, this);
            this.resultsModel.on("error", this.onSearchError, this);

            // Handle existing job
            var content = manager.get('data');
            if(content) {
                this.onSearchStart(content);
                this.onSearchProgress({ content: content });
                if(content.isDone) {
                    this.onSearchDone({ content: content });
                }
            }

            this.manager.replayLastSearchEvent(this);
        },
        _showMessages: function() {
            this.$('.msg').show();
            this.table.$el.hide();
            if (this.paginator) {
                this.paginator.$el.hide();
            }
        },
        _hideMessages: function() {
            this.$('.msg').hide();
            this.table.$el.show();
            if (this.paginator) {
                this.paginator.$el.show();
            }
        },
        renderMessage: function(msg) {
            var el = this.$('.msg');
            if(!el.length) {
                el = $('<div class="msg"></div>').appendTo(this.$el);
            }
            Messages.render(msg, el);
            this._showMessages();
            this.trigger('rendered', this);
        },
        onSearchStart: function() {
            this._searchRunning = true;
            this.clearResults();
            this.renderMessage('waiting');
            if (this.paginator) {
                this.paginator.settings.set('page', 0);
            }
        },
        onSearchDone: function(properties) {
            this._searchRunning = false;
            var previewCount = ((properties || {}).content || {}).resultPreviewCount || 0;
            this.updateCount(previewCount);
        },
        onSearchProgress: function(properties) {
            properties = properties || {};
            var job = properties.content || {};
            var previewCount = job.resultPreviewCount || 0;

            if(previewCount > 0) {
                this.updateCount(previewCount);
            }
        },
        onSearchCancelled: function() {
            this.renderMessage('cancelled');
        },

        onSearchError: function(message, err) {
            var msg = Messages.getSearchErrorMessage(err) || message;
            this.renderMessage({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },
        onSearchFailed: function(state, job) {
            var msg = Messages.getSearchFailureMessage(state);
            this.renderMessage({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },
        applyFetchOptions: function() {
            if(!this.resultsModel) {
                return;
            }
            var args = {
                count: this.metadata.get("pageSize"),
                offset: this.metadata.get("offset")
            };
            if(this.metadata.has('sortKey')) {
                args.search = [
                    '|','sort', this.metadata.get('sortDirection') === 'desc' ? '-':'',
                    JSON.stringify(this.metadata.get('sortKey'))
                ].join(' ');
            }
            this.resultsModel.set(args);
        },
        updateCount: function(count) {
            if (count === 0) {
                this.clearResults();
                if(this._searchRunning) {
                    this.renderMessage('waiting');
                } else {
                    this.renderMessage('no-results');
                }
            }
            if(this.paginator) {
                this.paginator.settings.set('itemCount', count);
            }
        },
        onFieldsChanged: function(){
            this.transformFieldsToArray();
            this.renderResults();
        },
        transformFieldsToArray: function(){
            var fields = this.settings.get('fields');
            if (_.isString(fields)) {
                fields = $.trim(fields);

                if(fields[0] === '[' && fields.slice(-1) === ']') {
                    fields = JSON.parse(fields);
                } else {
                    // Since this is a string, we're going to treat it as a
                    // space separated list of strings, with quoting. This is
                    // similar to what Splunk's 'fields' command takes.
                    fields = _.map(fields.match(fieldSplitterRegex), function(field) {
                        return field.replace(quoteStripperRegex, "");
                    });
                }
                this.settings.set('fields', fields, { silent: true });
            }
        },
        renderResults: function() {
            var rows, fields;
            if(!this.resultsModel || !this.resultsModel.hasData()){
                return;
            }
            this.collection = this.resultsModel.collection();
            if (this.settings.has('fields') && !_.isEmpty(this.settings.get('fields'))) {
                fields = this.settings.get('fields');
                var fieldsIdx = {}, resultsFields = this.resultsModel.data().fields;

                if (_(fields).contains('*')) {
                    fields = _(fields).chain().map(function(field) {
                        return field === '*' ? _(resultsFields).select(function(resField) {
                            return !_(fields).contains(resField);
                        }) : field;
                    }).flatten().unique().value();
                }

                _(resultsFields).each(function(f, i) { fieldsIdx[f] = i; });
                rows = _(this.resultsModel.data().rows).map(function(o){
                    return _(fields).map(function(f){ return fieldsIdx.hasOwnProperty(f) ? o[fieldsIdx[f]] : null; });
                });
            } else {
                rows = this.resultsModel.data().rows;
                fields = this.resultsModel.data().fields;
            }
            this.results.set({
                rows: rows,
                fields: fields
            });
            this._hideMessages();
        },
        clearResults: function() {
            if(this.paginator) {
                this.paginator.settings.set('itemCount', 0);
            }
            this.results.set({
                rows: [],
                fields: []
            });
        },
        render: function () {
            this.table.render();
            if(this.paginator) {
                this.paginator.render();
            }
            return this;
        },
        remove: function() {
            if(this.table) {
                this.table.remove();
            }
            if(this.paginator) {
                this.paginator.remove();
            }
            if(this.settings) {
                this.settings.off();
                if(this.settings._sync) {
                    this.settings._sync.destroy();
                }
                if(this.settings._syncMeta) {
                    this.settings._syncMeta.destroy();
                }
            }
            BaseSplunkView.prototype.remove.call(this);
        }
    });

    TableView.BaseCellRenderer = BaseCellRenderer;

    return TableView;
});
