define(function(require, exports, module) {
    var _ = require('underscore');
    var Registry = require('./registry');
    
    var ValueEscaper = {
        search: function(v) {
            return JSON.stringify(String(v));
        },
        url: function(v) {
            return encodeURIComponent(String(v));
        },
        html: function(v) {
            return _.escape(String(v));
        }
    };
    _.extend(ValueEscaper, {
        h: ValueEscaper.html,
        u: ValueEscaper.url,
        s: ValueEscaper.search
    });
    
    // NOTE: The regular expressions here for parsing tokens should attempt to
    //       be backward-compatible with the token parsing methods in
    //       exposed/js/utils.js such as discoverReplacementTokens() and
    //       replaceTokens().
    
    // ex: ns:
    var TOKEN_NAMESPACE_PREFIX_PATTERN = /(\w+:)?/.source;
    // ex: token.name|suh
    var TOKEN_NAME_CHARS_PATTERN = /([^$|:]+?)(\|[suh]+)?/.source;
    // ex: $ns:token.name|suh$
    var TOKEN_PATTERN = /\$/.source + TOKEN_NAMESPACE_PREFIX_PATTERN + TOKEN_NAME_CHARS_PATTERN + /\$/.source;
    
    var TOKEN_OR_DOLLAR_RE = new RegExp(TOKEN_PATTERN + '|' + /\$\$/.source, 'g');
    
    /**
     * Duplicates a RegExp literal so that it can be used safely for
     * operations like test() that mutate the RegExp object.
     */
    var r = function(regexp) {
        return new RegExp(regexp);
    };
    
    /**
     * Contains utilities for processing template strings that can contain tokens.
     * 
     * '$fullName$' is a simple example of a token that can appear in a
     * template string.
     * 
     * '$ns:foo.bar|suh$' is a more complex example.
     * Here the token has the name 'foo.bar' and is escaped using the
     * 's' (search), 'u' (url), and 'h' (HTML) escaping functions
     * (in that order). It also lives in the 'ns' namespace.
     * 
     * Some functions defined in this class will ignore namespaced tokens
     * to preserve backward compatibility:
     * 
     *      Namespace Aware   | Namespace Unaware
     *      ----------------- | -------------------
     *      getTokens         | getTokenNames
     *      replaceTokens     | replaceTokenNames
     *      hasToken          | hasTokenName
     *      isToken           | isTokenName
     *      -                 | getTokenName
     */
    var TokenUtils = {
        /**
         * Returns the tokens in the specified template string.
         * 
         * Example:
         *      TokenUtils.getTokenNames(
         *          '$firstName|u$ $ns:lastName|u$')
         *      ==
         *          [
         *              { namespace: 'default', name: 'firstName' },
         *              { namespace: 'ns',      name: 'lastName' }
         *          ]
         * 
         * @param searchTemplate    A template string whose tokens will be returned.
         * @param options.tokenNamespace
         *                          Token namespace to use when resolving
         *                          unqualified token references.
         *                          Defaults to 'default'.
         */
        /*
         * Private API:
         * 
         * @param options._isToken  If specified, alters return value to
         *                          indicate whether the specified template
         *                          string contains exactly one token
         *                          (and nothing else).
         */
        getTokens: function(searchTemplate, options) {
            options = options || {};
            var defaultTokenNamespace = options.tokenNamespace || 'default';
            
            /*
             * Looks for:
             *     (1a) $$ (literal $)
             *     (1b) $ns:token.name|suh$ (tokens)
             */
            var tokens = [];
            var rex = r(TOKEN_OR_DOLLAR_RE);
            while (true) {
                var match = rex.exec(searchTemplate);
                if (options._isToken) {
                    return match && 
                        (match.index === 0) &&
                        (match[0].length === searchTemplate.length);
                }
                if (!match) {
                    break;
                }
                var tokenNamespace = match[1];
                var tokenName = match[2];
                
                // NOTE: Some browsers return '' and others return undefined
                //       when there is no match.
                if (!tokenNamespace && !tokenName) {
                    // Matched '$$'
                    continue;
                }
                
                if (!tokenNamespace) {
                    tokenNamespace = defaultTokenNamespace;
                } else {
                    // Chop trailing ':'
                    tokenNamespace = tokenNamespace.substring(
                        0, tokenNamespace.length - 1);
                }
                
                tokens.push({ namespace: tokenNamespace, name: tokenName });
            }
            return tokens;
        },
        
        /**
         * Returns the token names in the specified template string,
         * ignoring namespaced tokens.
         * 
         * Example:
         *      TokenUtils.getTokenNames(
         *          '$firstName|u$ $lastName|u$ $ns:ignored$')
         *      ==
         *          ['firstName', 'lastName']
         */
        getTokenNames: function(searchTemplate) {
            var tokens = TokenUtils.getTokens(searchTemplate);
            
            // Do not recognize namespaced tokens for backward compatibility
            var tokenNames = [];
            _.each(tokens, function(token) {
                if (token.namespace === 'default') {
                    tokenNames.push(token.name);
                }
            });
            
            return tokenNames;
        },
        
        /**
         * Replaces all tokens in the specified template using the
         * values in the specified token registry.
         * 
         * An optional escaper function may be provided that will
         * be used to transform the values of all substituted tokens.
         *
         * Example:
         *      var tokenRegistry = mvc.Components;
         *      tokenRegistry.get('default').set('firstName', 'Bob');
         *      tokenRegistry.get('default').set('lastName', 'Waters');
         *      
         *      TokenUtils.replaceTokens(
         *          '$firstName$ $lastName$',
         *          tokenRegistry)
         *      ==
         *          'Bob Waters'
         * 
         * @param str               A template string whose tokens will be replaced.
         * @param tokenRegistry     Registry from which to read token values.
         * @param options.escaper   Transforms token values as they are substituted.
         * @param options.tokenNamespace
         *                          Token namespace to use when resolving
         *                          unqualified token references.
         *                          Defaults to 'default'.
         * @param escaper           Deprecated. Use options.escaper instead.
         */
        replaceTokens: function(str, tokenRegistry, optionsOrEscaper) {
            var options;
            if (_.isFunction(optionsOrEscaper)) {
                options = {escaper: optionsOrEscaper};
            } else {
                options = optionsOrEscaper || {};
            }
            
            var escaper = options.escaper;
            var defaultTokenNamespace = options.tokenNamespace || 'default';
            
            /*
             * Looks for:
             *     (1a) $$ (literal $)
             *     (1b) $ns:token.name|suh$ (tokens)
             *
             * Replaces with:
             *     (1a) $
             *     (1b) <token value, after optional escaping>
             */
            return str.replace(r(TOKEN_OR_DOLLAR_RE), function (
                match, tokenNamespace, tokenName, esc)
            {
                // NOTE: Some browsers return '' and others return undefined
                //       when there is no match.
                if (!tokenNamespace && !tokenName) {
                    // Matched '$$'
                    return '$';
                }
                
                if (!tokenNamespace) {
                    tokenNamespace = defaultTokenNamespace;
                } else {
                    // Chop trailing ':'
                    tokenNamespace = tokenNamespace.substring(
                        0, tokenNamespace.length - 1);
                }
                
                var tokenModel = tokenRegistry.get(tokenNamespace);
                if (tokenModel === undefined) {
                    return match;
                }
                var v = tokenModel.get(tokenName);
                if (v === undefined) {
                    return match;
                }
                if (esc) {
                    _(esc.substring(1).split('')).each(function(e){
                        var fn = ValueEscaper[e];
                        if(fn && fn !== escaper){
                            v = fn(v);
                        }
                    });
                }
                if (_.isFunction(escaper)) {
                    v = escaper(v);
                }
                return v;
            });
        },
        
        /**
         * Replaces all tokens in the specified template using the
         * specific map of token values, ignoring namespaced tokens.
         * 
         * An optional escaper function may be provided that will
         * be used to transform the values of all substituted tokens.
         *
         * Example:
         *      TokenUtils.replaceTokens('$firstName$ $lastName$', {
         *          'firstName': 'Bob',
         *          'lastName': 'Waters'
         *      }) == 'Bob Waters'
         */
        replaceTokenNames: function(str, valuesByTokenName, escaper) {
            // Only replace the specified tokens in the default namespace
            // for backward compatibility
            var tokenRegistry = new Registry();
            var defaultTokenModel = tokenRegistry.getInstance(
                'default', { create: true });
            defaultTokenModel.set(valuesByTokenName);
            
            return TokenUtils.replaceTokens(str, tokenRegistry, escaper);
        },
        
        /**
         * Returns whether the specified template string contains any tokens.
         * 
         * Example:
         *      TokenUtils.hasToken('Name: $fullName$') == true
         *      TokenUtils.hasToken('Name: $firstName$ $lastName$') == true
         *      TokenUtils.hasToken('foobar') == false
         *      TokenUtils.hasToken('$ns:fullName$') == true
         */
        hasToken: function(str) {
            return TokenUtils.getTokens(str).length > 0;
        },
        
        /**
         * Returns whether the specified template string contains any tokens,
         * ignoring namespaced tokens.
         * 
         * Example:
         *      TokenUtils.hasToken('Name: $fullName$') == true
         *      TokenUtils.hasToken('Name: $firstName$ $lastName$') == true
         *      TokenUtils.hasToken('foobar') == false
         *      TokenUtils.hasToken('$ns:fullName$') == false
         */
        hasTokenName: function(str) {
            return _.some(TokenUtils.getTokens(str), function(token) {
                return token.namespace === 'default';
            });
        },
        
        /**
         * Returns whether the specified template string consists of exactly
         * one token (and nothing else).
         * 
         * Example:
         *      TokenUtils.isToken('$fullName$') == true
         *      TokenUtils.isToken('Name: $fullName$') == false
         *      TokenUtils.isToken('Name') == false
         *      TokenUtils.isToken('$ns:fullName$') == true
         */
        isToken: function(p) {
            return TokenUtils.getTokens(p, {_isToken: true});
        },
        
        /**
         * Returns whether the specified template string consists of exactly
         * one non-namespaced token (and nothing else).
         * 
         * Example:
         *      TokenUtils.isToken('$fullName$') == true
         *      TokenUtils.isToken('Name: $fullName$') == false
         *      TokenUtils.isToken('Name') == false
         *      TokenUtils.isToken('$ns:fullName$') == false
         */
        isTokenName: function(p) {
            return TokenUtils.isToken(p) &&
                TokenUtils.getTokens(p)[0].namespace === 'default';
        },
        
        /**
         * Returns the token name of the solitary non-namespaced token in the 
         * specified template string.
         * 
         * If there is no such token a falsy value is returned.
         * 
         * Example:
         *      TokenUtils.getTokenName('$fullName$') == 'fullName'
         *      TokenUtils.getTokenName('Name') == false
         *      TokenUtils.getTokenName('$ns:fullName$') == false
         */
        getTokenName: function(tokenStr) {
            return TokenUtils.isTokenName(tokenStr) && 
                TokenUtils.getTokenNames(tokenStr)[0];
        },
        
        /**
         * Quotes a literal string as a template string.
         * 
         * Example:
         *      TokenUtils.quoteAsTokenString('$1.00') == '$$1.00'
         *      TokenUtils.quoteAsTokenString('C:\Windows') == 'C:\Windows'
         *      TokenUtils.quoteAsTokenString('identifier') == 'identifier'
         */
        quoteAsTokenString: function(literalStr) {
            if (!_.isString(literalStr)) {
                return literalStr;
            } else {
                // Replaces $ with $$
                return literalStr.replace(/\$/g, function(match) {
                    return '$$';
                });
            }
        },
        
        /**
         * Returns the escaping function corresponding to the specified
         * identifying character (i.e. one of 's', 'u', or 'h') or
         * long name (i.e. one of 'search', 'url', or 'html').
         */
        getEscaper: function(esc) {
            return ValueEscaper[esc];
        }
    };
    
    return TokenUtils;
});