define(['underscore', 'helpers/user_agent'], function(_, userAgent) {

    var VISIBLE_FIELD_REGEX = /^[^_]|^_time/,
        MAX_SERIES = 50,
        MAX_COLUMN_POINTS = (function() {
            if(!userAgent.isIE()) {
                return 1500;
            }
            if(!userAgent.isIELessThan(9)) {
                return 1200;
            }
            return 500;
        }()),
        MAX_LINE_POINTS = userAgent.isIELessThan(9) ? 2000 : 20000,
        MAX_CATEGORY_LABELS = 80;

    // sort of a "catch-all" method for adding display properties based on the data set and the web.conf config
    // this method is used by consumers of the JSCharting library to share custom logic that doesn't belong
    // in the library itself
    var getCustomDisplayProperties = function(chartData, webConfig) {
        webConfig = webConfig || {};
        var customProps = {};
        if(webConfig['JSCHART_TEST_MODE']) {
            customProps.testMode = true;
        }
        var isTimeData = chartData.length > 0 && chartData.fieldAt(0) === '_time'
                            && (chartData.hasField('_span') || chartData.seriesAt(0).length === 1);

        if(chartData.length > 0 && !isTimeData && chartData.seriesAt(0).length > MAX_CATEGORY_LABELS) {
            customProps['axisLabelsX.hideCategories'] = true;
        }
        if(chartData.hasField('_tc')) {
            customProps.fieldHideList = ['percent'];
        }
        return customProps;
    };

    var sliceResultsToSeriesLength = function(rawData, length) {
        var sliced = {
            fields: rawData.fields,
            columns: []
        };

        _(rawData.columns).each(function(column, i) {
            sliced.columns[i] = column.slice(0, length);
        });

        return sliced;
    };

    var fieldIsVisible = function(field) {
        var fieldName = _.isString(field) ? field : field.name;
        return VISIBLE_FIELD_REGEX.test(fieldName);
    };

    // pre-process chart data, truncating either the number of series or the number of points per series
    // default truncation constants are defined above, though a custom limit for total number of points can be
    // passed in as part of the display properties
    var preprocessChartData = function(rawData, displayProperties) {
        if(rawData.columns.length === 0 || rawData.columns[0].length === 0) {
            return rawData;
        }
        var chartType = displayProperties.chart || 'column';
        if(chartType in { pie: true, scatter: true, radialGauge: true, fillerGauge: true, markerGauge: true }) {
            return rawData;
        }

        if(rawData.fields.length >= MAX_SERIES) {
            var spanColumn,
                normalizedFields = _(rawData.fields).map(function(field) {
                    return _.isString(field) ? field : field.name;
                }),
                spanIndex = _(normalizedFields).indexOf('_span');

            if(spanIndex > -1 && spanIndex >= MAX_SERIES) {
                spanColumn = rawData.columns[spanIndex];
            }

            // slice the number of series
            rawData = {
                columns: rawData.columns.slice(0, MAX_SERIES),
                fields: rawData.fields.slice(0, MAX_SERIES)
            };

            // if our slicing removed _span, put it back
            if(spanColumn) {
                rawData.columns.push(spanColumn);
                rawData.fields.push('_span');
            }
        }

        var truncationLimit,
            userSpecifiedLimit = parseInt(displayProperties['chart.resultTruncationLimit'], 10)
                                    || parseInt(displayProperties['resultTruncationLimit'], 10);

        if(userSpecifiedLimit > 0) {
            truncationLimit = userSpecifiedLimit;
        }
        else {
            truncationLimit = (chartType in { line: true, area: true }) ? MAX_LINE_POINTS : MAX_COLUMN_POINTS;
        }
        var visibleFields = _(rawData.fields).filter(fieldIsVisible),
            numDataSeries = visibleFields.length - 1, // subtract one because the first field is the x-axis
            pointsPerSeries = rawData.columns[0].length,
            // numSeries is guaranteed not to be zero based on the first check in this method
            allowedPointsPerSeries =  Math.floor(truncationLimit / numDataSeries);

        if(pointsPerSeries > allowedPointsPerSeries) {
            return sliceResultsToSeriesLength(rawData, allowedPointsPerSeries);
        }
        return rawData;
    };

    return ({

        getCustomDisplayProperties: getCustomDisplayProperties,
        preprocessChartData: preprocessChartData

    });
});
