/**
 * @author jszeto
 * @date 11/9/12
 *
 * Main view for the Add/Edit views. It holds the header, the Save/Preview/Cancel buttons and the Preview view.
 * It creates a subview that shows the controls for adding or editing a particular Object or Attribute.
 * Each of these subviews subclasses views/data_model_editor/form_views/AddEditFormBaseView.js.
 *
 * Inputs:
 *
 * model: {
 *          application {models/Application}
 *          appLocal {models/services/AppLocal}
 *          dataModel {models/services/datamodel/DataModel} the data model currently being edited
 *          setting <Model> {
 *              Required:
 *              operation {String} - Whether to add or edit. Possible values = [add|edit]
 *              type {String} - The type of Object or Attribute to add or edit [object|child|transaction|Lookup|Rex|Eval|GeoIP]
 *              dataModel {DataModel} - clone of the DataModel being edited
 *
 *              Optional:
 *              objectName {String} - The name of the object to edit
 *              parentObjectName {String} - The name of the parent Object to either a child Object or a calculation (Lookup|Rex|Eval|GeoIP)
 *              calculationID {String} - The id of the calculation to edit. Only used for calculations
 *              lookupModel {TransformsLookups} - Model that contains the possible lookups. Used when type = lookup
 *         }
 *     }
 *     collection: {
 *         transformsLookups {collections/services/data/TransformsLookups} - Collection of lookups to use
 *              when add/edit a Lookup calculation
 *     }
 */
define([
    'jquery',
    'underscore',
    'backbone',
    'views/Base',
    'views/data_model_editor/form_views/AddEditObjectView',
    'views/data_model_editor/form_views/AddEditTransactionView',
    'views/data_model_editor/form_views/AddEditSearchObjectView',
    'views/data_model_editor/form_views/AddEditEvalView',
    'views/data_model_editor/form_views/AddEditLookupView',
    'views/data_model_editor/form_views/AddEditRexView',
    'views/data_model_editor/form_views/AddEditGeoIPView',
    'views/data_model_editor/form_components/SelectSampleSize',
    'views/shared/eventsviewer/Master',
    'views/shared/results_table/ResultsTableMaster',
    'views/shared/SearchResultsPaginator',
    'views/shared/jobstatus/Spinner',
    'views/shared/jobstatus/Count',
    'views/shared/controls/ControlGroup',
    'views/shared/controls/SyntheticSelectControl',
    'views/shared/FlashMessages',
    'models/Base',
    'models/SplunkDBase',
    'models/Job',
    'models/fetch_data/ResultsFetchData',
    'models/services/SavedSearch',
    'models/services/search/IntentionsParser',
    'models/services/search/jobs/Result',
    'models/services/search/jobs/ResultJsonRows',
    'models/services/search/jobs/Summary',
    'models/FlashMessage',
    'collections/Base',
    'collections/services/configs/EventRenderers',
    'collections/services/data/ui/WorkflowActions',
    'uri/route',
    'module'
],
    function(
        $,
        _,
        Backbone,
        BaseView,
        AddEditObjectView,
        AddEditTransactionView,
        AddEditSearchObjectView,
        AddEditEvalView,
        AddEditLookupView,
        AddEditRexView,
        AddEditGeoIPView,
        SelectSampleSize,
        EventsViewer,
        ResultsTableMaster,
        SearchResultsPaginator,
        Spinner,
        Count,
        ControlGroup,
        SyntheticSelectControl,
        FlashMessagesView,
        BaseModel,
        SplunkDBase,
        Job,
        ResultsFetchData,
        SavedSearch,
        IntentionsParser,
        ResultModel,
        ResultJsonRows,
        SummaryModel,
        FlashMessageModel,
        BaseCollection,
        EventRenderersCollection,
        WorkflowActionsCollection,
        route,
        module
        )
    {
        return BaseView.extend({
            className: "view-add-edit-container",
            moduleId: module.id,
            // Title to display at the top of the View
            ADHOC_SEARCH_LEVEL: "smart",
            headerTitle:"",
            isCalculation: false,
            showPreview: true,
            showPreviewButton: true,

            events: {
                'click .preview-button': function(e) {
                    e.preventDefault();
                    this.children.formView.preview();
                },
                'click .cancel-button': function(e) {
                    e.preventDefault();
                    this.children.formView.cleanup();
                    /**
                     * Cancels any edits and closes the Add / Edit view, returning the user back to the Data Model Editor
                     *
                     * @event AddEditContainerView#action:cancel
                     */
                    this.trigger('action:cancel');
                },
                'click .save-button': function(e) {
                    e.preventDefault();
                    // The child view understands how to apply its form controls to the model. It will send an
                    // action:save event if it passes validation.
                    this.children.formView.save();
                }
            },

            /**
             *  Passing along the action:save event from the form view to the application controller
             */
            formView_saveHandler: function(objectName) {
                /**
                 * Save the Data Model
                 *
                 * @event ObjectEditor#action:saveModel
                 * @param {string} Name of the object to save
                 */
                this.trigger('action:save', objectName);
            },

            /**
             * Called when user presses the Preview button. Posts the changes to the backend with the provisional flag.
             * Then retrieves the search string and kicks off a search job
             *
             * @param {string} objectName
             * @param {array} calculationFieldNames - array of strings of the fields to preview
             */
            formView_previewHandler: function(objectName, calculationFieldNames) {
                var value = this.selectSampleSize.getValue();

                // TODO [JCS] Move the provisional object save logic up here from the DataModelEditor router
                this.fetchPreviewSearchString(objectName, calculationFieldNames).then(function(searchString) {
                    this.startNewSearchJob(searchString, value.earliest, value.latest, value.head);
                }.bind(this));
            },

            /**
             * Used to preview Search Objects
             *
             * @param {string} searchString - the search string to use for the preview
             */
            formView_startSearchByStringHandler: function(searchString) {
                var value = this.selectSampleSize.getValue();

                this.startNewSearchJob(searchString, value.earliest, value.latest, value.head);
            },

            /**
             * If the user changes the the Sample Size, then start a new search job
             */
            selectSampleSizeChangeHandler: function() {
                this.children.formView.preview();
            },

            initialize: function(options) {
                BaseView.prototype.initialize.call(this, options);

                // Selecting the HEAD or time range
                this.selectSampleSize = new SelectSampleSize({});
                this.selectSampleSize.on("change", this.selectSampleSizeChangeHandler, this);
                this.children.selectSampleSizeGroup = new ControlGroup({label: _("Sample Size:").t(),
                    controls: [this.selectSampleSize]
                });

                // Search parser for checking the type of the search string (report/non-report)
                this.model.intentionsParser = new IntentionsParser();
                // A couple of collections to pass into the EventViewer.
                this.collection.workflowActions = new WorkflowActionsCollection();
                this.collection.eventRenderers = new EventRenderersCollection();

                // Keeps track of which buttons are visible
                this.model.buttonSettings = new BaseModel({save:"show",cancel:"show",preview:"show"});
                this.model.buttonSettings.on("change",this.updateButtons, this);

                // Setup the Flash Messages
                this.children.flashMessagesView = new FlashMessagesView({model:[this.model.dataModel, this.model.intentionsParser]});

                // Initialize the sub view based on the settings
                var type = this.model.setting.get("type");
                var objectName = this.model.setting.get("objectName");
                var operation = this.model.setting.get("operation");
                var calculationID = this.model.setting.get("calculationID");
                var subViewOptions = {
                    model:{dataModel:this.model.dataModel,
                           appLocal: this.model.appLocal,
                           application: this.model.application,
                           buttonSettings: this.model.buttonSettings},
                    objectName:objectName,
                    operation:operation,
                    flashMessagesHelper: this.children.flashMessagesView.flashMsgHelper
                };

                // Create the Subview based on the type and operation.
                if (type == "event")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Event Object").t();
                    else
                        this.headerTitle = _("Edit Constraints").t();
                    subViewOptions.type = type;
                    this.children.formView = new AddEditObjectView(subViewOptions);
                }
                else if (type == "child")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Child Object").t();
                    else
                        this.headerTitle = _("Edit Constraints").t();
                    subViewOptions.type = type;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");
                    this.children.formView = new AddEditObjectView(subViewOptions);
                }
                else if (type == "transaction")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Transaction Object").t();
                    else
                        this.headerTitle = _("Edit Transaction").t();

                    subViewOptions.type = type;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");
                    this.children.formView = new AddEditTransactionView(subViewOptions);
                }
                else if (type == "search")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Base Search").t();
                    else
                        this.headerTitle = _("Edit Base Search").t();

                    // Don't show the preview button
                    this.showPreviewButton = false;
                    subViewOptions.type = type;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");
                    this.children.formView = new AddEditSearchObjectView(subViewOptions);
                }
                else if (type == "Eval")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Attributes with an Eval Expression").t();
                    else
                        this.headerTitle = _("Edit Attributes with an Eval Expression").t();

                    this.isCalculation = true;
                    subViewOptions.calculationID = calculationID;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");
                    this.children.formView = new AddEditEvalView(subViewOptions);
                }
                else if (type == "Lookup")
                {
                    // Create the Lookup
                    //this.model.transformsLookups.fetch().then(function() {
                    if (operation == "add")
                        this.headerTitle = _("Add Attributes with a Lookup").t();
                    else
                        this.headerTitle = _("Edit Attributes with a Lookup").t();

                    this.isCalculation = true;
                    subViewOptions.calculationID = calculationID;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");
                    subViewOptions.collection = {};
                    subViewOptions.collection.transformsLookups = this.collection.transformsLookups;
                    this.children.formView = new AddEditLookupView(subViewOptions);
                }
                else if (type == "Rex")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Attributes with a Regular Expression").t();
                    else
                        this.headerTitle = _("Edit Attributes with a Regular Expression").t();

                    this.isCalculation = true;
                    subViewOptions.calculationID = calculationID;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");

                    this.children.formView = new AddEditRexView(subViewOptions);
                }
                else if (type == "GeoIP")
                {
                    if (operation == "add")
                        this.headerTitle = _("Add Geo Attributes with an IP Lookup").t();
                    else
                        this.headerTitle = _("Edit Geo Attributes with an IP Lookup").t();

                    this.isCalculation = true;
                    subViewOptions.calculationID = calculationID;
                    subViewOptions.parentObjectName = this.model.setting.get("parentObjectName");
                    subViewOptions.collection = {};
                    subViewOptions.collection.transformsLookups = this.collection.transformsLookups;

                    this.children.formView = new AddEditGeoIPView(subViewOptions);
                }
                else
                {
                    throw new Error("Unknown Calculation Type",type);
                }

                // Listen to events from the subview
                this.children.formView.on("action:save", this.formView_saveHandler, this);
                this.children.formView.on("action:preview", this.formView_previewHandler, this);
                this.children.formView.on("action:startSearchByString", this.formView_startSearchByStringHandler, this);
            },

            /**
             * This function will take any search string and run a search job. It will then display an events view or
             * table view depending upon the type of search.
             *
             * @param {string} searchString - string to use for the search
             * @param {string} earliest - earliest time for the search
             * @param {string} latest - latest time for the search
             * @param {number} head - number of results to return from head
             */
            startNewSearchJob: function(searchString, earliest, latest, head) {

                var searchParseDeferred = $.Deferred();
                var workflowActionsDeferred = $.Deferred();

                // Cleanup any existing search jobs
                this.destroySearchJob();

                // Remove any of the search job related UI elements
                if (this.children.spinner)
                    this.children.spinner.remove();

                if (this.children.count)
                    this.children.count.remove();

                if (this.children.eventsResultsPaginator)
                    this.children.eventsResultsPaginator.remove();

                if (this.children.resultsView)
                    this.children.resultsView.remove();

                if (this.children.pageCount)
                    this.children.pageCount.remove();

                // Create the Job
                this.model.searchJob = new Job({}, { delay: 1000, processKeepAlive: true });

                // Results paginator
                this.model.resultsFetchData = new ResultsFetchData({segmentation:"full", count:20, max_lines: 20});

                // Select the number of results per page
                this.children.pageCount = new SyntheticSelectControl({
                    menuWidth: "narrow",
                    className: "btn-group pull-left",
                    items: [
                        {value: 10, label: '10 per page'},
                        {value: 20, label: '20 per page'},
                        {value: 50, label: '50 per page'},
                        {value: 100, label: '100 per page'}
                    ],
                    model: this.model.resultsFetchData,
                    modelAttribute: 'count',
                    toggleClassName: 'btn-pill'
                });

                // Create the search job related UI elements
                this.children.spinner = new Spinner({model: this.model.searchJob});
                this.children.count = new Count({model: this.model.searchJob});

                // Fetch the workflowActions collection
                this.workflowActionsBootstrap(workflowActionsDeferred);

                $.when(workflowActionsDeferred).then(function() {
                    // First, we need to parse the search string to figure out if this is a non-transforming/transforming search
                    this.intentionsParserBootstrap(searchParseDeferred, searchString);
                    $.when(searchParseDeferred).then(function() {

                        var searchJobAttrs = {
                            search: searchString  || '',
                            earliest_time: earliest || null,
                            latest_time: latest || null,
                            auto_cancel: 100,
                            auto_finalize_ec: 0,
                            status_buckets: 0,
                            timeFormat: '%s.%Q',
                            ui_dispatch_app: this.model.application.get("app"),
                            rf: '*',
                            preview: true,
                            adhoc_search_level: this.ADHOC_SEARCH_LEVEL,
                            app: this.model.application.get("app"),
                            owner: this.model.application.get("owner")
                        };

                        if(head) {
                            searchJobAttrs.search += (' | head ' + head);
                        }

                        var reportType;

                        // We need two different result views depending upon the type of search. Each view takes a different
                        // type of result model
                        if (this.isCalculation || this.model.intentionsParser.isReportsSearch()) {
                            // Report searches and calculation previews return a table with columns for each field
                            this.model.resultsModel = new ResultJsonRows({}, {fetchData: this.model.resultsFetchData});
                            this.model.resultsFetchData.set({output_mode: 'json_rows'});

                            this.children.resultsView = new ResultsTableMaster({
                                model: {
                                    results: this.model.resultsModel,
                                    metadata: this.model.resultsFetchData,
                                    format: new BaseModel() // use an empty model for the formatting options
                                },
                                enableTableDock: true,
                                className: "scrolling-table-wrapper"
                            });

                            // Assign pointers to the results view
                            reportType = "results_preview";
                        } else {
                            // Event searches list the events
                            this.model.resultsModel = new ResultModel({}, {fetchData: this.model.resultsFetchData});
                            // TODO [JCS] Eventually we should do a fetch here to get the defaults SavedSearch from the server
                            this.model.reportModel = new SavedSearch();
                            this.model.reportModel.entry.content.set("display.events.type", "raw");

                            this.children.resultsView = new EventsViewer({
                                model: {
                                    result: this.model.resultsModel,
                                    summary: new SummaryModel(),
                                    searchJob: this.model.searchJob,
                                    report: this.model.reportModel,
                                    application: this.model.application
                                },
                                collection: {
                                    selectedFields: new BaseCollection(),
                                    workflowActions: this.collection.workflowActions,
                                    eventRenderers: this.collection.eventRenderers
                                },
                                selectableFields: false,
                                allowRowExpand: false,
                                headerMode: 'dock'
                            });
                            // Assign pointers to the results view
                            reportType = "events";
                        }

                        // Pass in a simple metadata model instead of the report model
                        // Also, pass in an empty collection since we don't have a timeline
                        this.children.eventsResultsPaginator = new SearchResultsPaginator({
                            model: {
                                state: this.model.resultsFetchData,
                                searchJob: this.model.searchJob
                            },
                            collection: new BaseCollection(),
                            mode: reportType
                        });

                        // render and append all of the UI Elements
                        this.children.spinner.render().appendTo(this.$(".job-status-placeholder"));
                        this.children.count.render().appendTo(this.$(".job-status-placeholder"));
                        this.children.pageCount.render().appendTo(this.$(".page-count-placeholder"));
                        this.children.eventsResultsPaginator.render().appendTo(this.$(".results-paginator-placeholder"));
                        this.children.resultsView.render().appendTo(this.$(".preview-table-wrapper"));

                        // Run the search job and start polling so we get progress updates
                        this.model.searchJob.save({}, { data: searchJobAttrs }).done(function() {
                            this.model.resultsModel.set("id", this.model.searchJob.entry.links.get(reportType));
                            this.model.searchJob.register(this.model.resultsModel);
                            this.model.searchJob.startPolling();
                        }.bind(this));
                    }.bind(this));
                }.bind(this));
            },

            /**
             * Helper function to run the search string through the search parser. We use this to determine the
             * search type (report/non-report)
             *
             * @param {Deferred} searchParseDeferred - deferred object that tracks the status of the request
             * @param {string} search - the search string to parse
             */
            intentionsParserBootstrap: function(searchParseDeferred, search){
                this.model.intentionsParser.fetch({
                    data: {
                        q: search,
                        adhoc_search_level: this.ADHOC_SEARCH_LEVEL,
                        app: this.model.application.get('app'),
                        owner: this.model.application.get('owner')
                    },
                    success: function(model, response) {
                        searchParseDeferred.resolve();
                    },
                    error: function(model, response) {
                        //the search parser blew up on our search and we won't be able to make a job from it
                        searchParseDeferred.reject();
                    }
                });
            },

            /**
             * Helper function to fetch the workflow Actions collection
             * @param {Deferred} workflowActionsDeferred
             */
            workflowActionsBootstrap: function(workflowActionsDeferred) {
                this.collection.workflowActions.fetch({
                    data: {
                        app: this.model.application.get('app'),
                        owner: this.model.application.get('owner'),
                        count: -1
                    },
                    success: function(model, response) {
                        workflowActionsDeferred.resolve();
                    },
                    error: function(model, response) {
                        workflowActionsDeferred.resolve();
                    }
                });
            },

            /**
             * Helper to clean up the search job
             */
            destroySearchJob: function() {
                if (this.model.searchJob) {
                    this.model.searchJob.destroy();
                    delete this.model.searchJob;
                }
                // TODO [JCS] Do we need to clear any event listeners on our models?
            },

            /**
             * Helper function to get the preview search string when adding or editing an Object.
             * We ask the backend to save a provisional copy of the dataModel and then ask it for the search string.
             *
             * @param {string} objectName - Name of the object that has been modified
             * @param {array} calculationFieldNames - Array of the fieldNames of the calculation's outputFields. If undefined,
             * then we just use previewSearch
             * @return {Deferred} - Returns a Deferred object that is done after the dataModel has been saved.
             */
            fetchPreviewSearchString: function(objectName, calculationFieldNames) {
                var dfd = $.Deferred();

                this.model.dataModel.save({}, {data: {provisional: true}}).then(function() {
                    var objectModel = this.model.dataModel.objectByName(objectName);
                    var searchString;

                    if (!_(calculationFieldNames).isUndefined()) {
                        searchString = objectModel.getCalculationPreviewSearch(calculationFieldNames);
                    } else {
                        searchString = objectModel.get("previewSearch");
                    }

                    dfd.resolve(searchString);
                }.bind(this));

                return dfd;
            },

            // Called when the view is being removed. Remove event listeners
            cleanup: function() {
                this.children.formView.off(null, null, this);
                this.destroySearchJob();
            },

            /**
             * Toggle button visibility based on the buttonSettings model
             */
            updateButtons: function() {
                this.updateButton(this.$(".save-button"), this.model.buttonSettings.get("save"));
                this.updateButton(this.$(".cancel-button"), this.model.buttonSettings.get("cancel"));
                this.updateButton(this.$(".preview-button"), this.model.buttonSettings.get("preview"));
                this.updateButton(this.$(".preview-placeholder"), this.model.buttonSettings.get("preview"));
            },

            /**
             * Helper function to update the visibility of a button
             *
             * @param {DOM element} buttonEl
             * @param {string} state - (show|hide)
             */
            updateButton: function(buttonEl, state) {
                if (state == "show") {
                    $(buttonEl).show();
                } else {
                    $(buttonEl).hide();
                }
            },

            render: function() {
                // Generate the breadcrumb
                var breadcrumb = "";

                if (this.isCalculation) {
                    var parentObjectName = this.model.setting.get("parentObjectName");
                    var parentObject = this.model.dataModel.objectByName(parentObjectName);

                    if (!_(parentObject).isUndefined()) {
                        var ancestors = this.model.dataModel.getAncestorsForObject(parentObjectName);

                        _(ancestors.reverse()).each(function(object) {
                            breadcrumb += object.get("displayName") + " > ";
                        }, this);

                        breadcrumb += parentObject.get("displayName");
                    }
                } else if (this.model.setting.get("operation") == "edit") {
                    var objectName = this.model.setting.get("objectName");
                    breadcrumb = this.model.dataModel.objectByName(objectName).get("displayName");
                }

                // Get the documentation URL
                var docUrl = route.docHelp(
                    this.model.application.get("root"),
                    this.model.application.get("locale"),
                    'learnmore.datamodel.addobject'
                );

                var html = _(this.template).template({
                    dataModel: this.model.dataModel,
                    docUrl: docUrl,
                    headerTitle: this.headerTitle,
                    objectBreadcrumb: breadcrumb,
                    showPreviewButton: this.showPreviewButton
                });
                this.$el.html(html);

                this.$(".flash-messages-placeholder").replaceWith(this.children.flashMessagesView.render().el);
                this.$(".form-view-placeholder").replaceWith(this.children.formView.render().el);
                this.$(".select-sample-size-placeholder").replaceWith(this.children.selectSampleSizeGroup.render().el);

                this.updateButtons();

                return this;
            },

            template: '\
                <div class="section-padded section-header">\
                    <div class="header-button-holder pull-right">\
                        <a href="<%- docUrl %>" target="_blank" class="btn external"><%- _("Documentation").t() %></a>\
                    </div>\
                    <h2 class="section-title"><%- headerTitle %></h2>\
                    <p class="section-description"><strong><%- _("Data Model:").t() %></strong> <span class="value"><%- dataModel.entry.content.get("displayName") %></span>   \
                    <% if (objectBreadcrumb != "") {%>\
                        <strong><%- _("Object:").t() %></strong> <span class="value"><%- objectBreadcrumb %></span>\
                     <% } %>\
                     </p>\
                </div>\
                <div class="divider"></div>\
                <div class="section-wrapper section-padded">\
                    <div class="flash-messages-placeholder"></div>\
                    <div class="form-view-placeholder"></div>\
                    <div class="divider"></div>\
                    <div class="job-status-placeholder"></div>\
                    <div class="buttons-wrapper pull-right">\
                        <a href="#" class="cancel-button btn"><%- _("Cancel").t() %></a>\
                        <% if (showPreviewButton) { %>\
                            <a href="#" class="preview-button btn"><%- _("Preview").t() %></a>\
                        <% } %>\
                        <a href="#" class="save-button btn btn-primary"><%- _("Save").t() %></a>\
                    </div>\
                </div>\
                <div class="gradient"></div>\
                <div class="preview-placeholder section-body section-padded">\
                    <div class="select-sample-size-placeholder"></div>\
                    <div class="page-count-placeholder pull-left"></div>\
                    <div class="results-paginator-placeholder pull-right"></div>\
                    <div class="preview-table-wrapper push-margins"></div>\
                </div>\
                '

        });

    });
