/**
 * @author sfishel, jszeto
 *
 * Master view for the "data_model_editor" page.  Used to perform edit operations on a single data model.
 *
 * Renders a selectable list of objects and an editor for the properties of the selected object.
 *
 * Child Views:
 *
 *     objectEditor - {views/data_model_editor/ObjectEditor} - Edit the selected Object
 *
 * Inputs:
 *
 *     model: {
 *         dataModel {models/services/datamodel/DataModel}
 *         setting {models/Base} - internal settings model
 *         {
 *             canCreateDataModel {boolean} - true if user can create a data model
 *             selectedObject {string} - name of the selected Object
 *         }
 *         application {models/Application}
 *         appLocal {models/services/AppLocal}
 *         user {models/services/authentication/User}
 *     }
 *     collection: {
 *         apps {collections/services/AppLocals}
 *     }
 *
 * Custom Events:
 *
 * @fires DataModelEditor#action:editDataModelTitle
 * @fires DataModelEditor#action:editPermissions
 * @fires DataModelEditor#action:editAcceleration
 * @fires DataModelEditor#action:cloneDataModel
 * @fires DataModelEditor#action:deleteDataModel
 * @fires DataModelEditor#action:fetchModel
 * @fires DataModelEditor#action:editCalculation
 * @fires DataModelEditor#action:addCalculation
 * @fires DataModelEditor#action:editObject
 * @fires DataModelEditor#action:saveModel
 * @fires DataModelEditor#action:removeObject
 * @fires DataModelEditor#action:addObject
 */

define([
            'jquery',
            'underscore',
            'views/Base',
            './ObjectEditor',
            'views/shared/DropDownMenu',
            'views/shared/FlashMessages',
            'views/data_model_manager/DataModelActionController',
            'models/services/datamodel/DataModel',
            'uri/route',
            'util/splunkd_utils',
            'module'
        ],
        function (
            $,
            _,
            BaseView,
            ObjectEditor,
            DropDownMenu,
            FlashMessagesView,
            DataModelActionController,
            DataModel,
            route,
            splunkDUtils,
            module
        ) {

    var CONSTS = {
        ALL_OBJECT_ITEMS : [{label:_("Root Event").t(), value:"event"},
            {label:_("Root Transaction").t(), value:"transaction"},
            {label:_("Root Search").t(), value:"search"},
            {label: _("Child").t(), value:"child"}],
        INITIAL_OBJECT_ITEMS : [{label:_("Root Event").t(), value:"event"},
            {label:_("Root Search").t(), value:"search"}],
        EDIT_ACCELERATED_MODEL_ID : "__EDIT_ACCELERATED_MODEL_ID__"
    };

    return BaseView.extend({

        className: 'data-model-editor',
        moduleId: module.id,


        events: {
            'click .object': function(e) {
                e.preventDefault();
                var objectName = $(e.currentTarget).attr('data-objid');
                this.model.setting.set({ selectedObject: objectName });
            }
        },

        dataModelChangeHandler: function() {
            this.updateSelection();
            this.debouncedRender();
        },

        dataModelObjectsChangeHandler: function() {
            // Reapply the Add Object items since changes to the objects might change the allowed items
            this.children.addObject.setItems(this.getEventItems());
            this.dataModelChangeHandler();
        },
        initialize: function(options) {
            BaseView.prototype.initialize.call(this, options);

            var pageRouter = route.getContextualPageRouter(this.model.application);

            //////////////////////////////////////////////////////////////////
            // SUBVIEWS
            //////////////////////////////////////////////////////////////////

            this.children.objectEditor = new ObjectEditor({collection:this.collection,
                                                           model: {application: this.model.application,
                                                                   dataModel: this.model.dataModel}
            });

            this.children.addObject = new DropDownMenu({label:_("Add Object").t(),
                items:this.getEventItems(),
                dropdownClassName: 'dropdown-menu-narrow'});

            this.children.flashMessagesView = new FlashMessagesView({model:this.model.dataModel});

            // Dynamically generate the Edit actions menu based on permissions and capabilities
            var subItems1 = [];
            var subItems2 = [];
            var items = [];

            if (!this.model.dataModel.isAccelerated()) {
                subItems1.push({label:_("Edit Title or Description").t(), value:"editDataModelTitle"});
            }

            if (this.model.dataModel.canChangePermissions()) {
                subItems1.push({label:_("Edit Permissions").t(), value:"editPermissions"});
            }
            if (this.model.user.canAccelerateDataModel()) {
                subItems1.push({label:_("Edit Acceleration").t(), value:"editAcceleration"});
            }
            if (this.model.setting.get("canCreateDataModel")) {
                subItems2.push({label:_("Clone").t(), value:"cloneDataModel"});
            }

            if (this.model.dataModel.canDelete()) {
                subItems2.push({label:_("Delete").t(), value:"deleteDataModel"});
            }

            if (subItems1.length > 0)
                items.push(subItems1);
            if (subItems2.length > 0)
                items.push(subItems2);

            if (items.length > 0) {
                // Dropdown menu that displays edit actions for this DataModel
                this.children.editObject = new DropDownMenu({label:_("Edit").t(),
                    items: items,
                           dropdownClassName: "dropdown-menu-narrow",
                           className: 'btn-combo create-drop-down',
                           popdownOptions: { attachDialogTo: 'body' }
                   });

                // Proxy events from editObjects to dataModelController
                this.children.editObject.on("all", function() {
                    var eventType = "action:" + arguments[1];
                    // If we are deleting a DataModel, we pass in the model we are deleting
                    if (eventType == "action:deleteDataModel") {
                        /**
                         * Delete a data model event
                         *
                         * @event DataModelEditor#action:deleteDataModel
                         * @param {models/services/datamodel/DataModel} - the data model to delete
                         */
                        this.children.dataModelController.trigger("action:deleteDataModel", this.model.dataModel);
                    } else {
                        // JSDoc for these are listed at bottom of file since these are generated dynamically
                        this.children.dataModelController.trigger(eventType, this.model.dataModel.id);
                    }
                }, this);
            }

            // Helper controller to handle edit actions from editObject
            this.children.dataModelController = new DataModelActionController({
                model: {
                    application: this.model.application,
                    user: this.model.user
                },
                collection: {apps: this.collection.apps}
            });
            this.children.dataModelController.on("action:fetchDataModel", function() {
                /**
                 * Fetch a Data Model model
                 *
                 * @event DataModelEditor#action:fetchModel
                 */
                this.trigger("action:fetchModel");
            }, this);
            this.children.dataModelController.on("action:updateDataModel", function(modelName, accelerationChanges, contentChanges) {
                    var dataModel = this.model.dataModel;

                    if (accelerationChanges) {
                        dataModel.entry.content.acceleration.set(accelerationChanges);
                    }
                    if (contentChanges) {
                        dataModel.entry.content.set(contentChanges);
                    }
            }, this);
            this.children.dataModelController.on("action:deletedDataModel", function() {
                window.location = pageRouter.data_model_manager();
            }, this);
            this.children.dataModelController.on("action:clonedDataModel", function(modelName) {
                window.location = pageRouter.data_model_editor({ data: { model: modelName } });
            }, this);

            //////////////////////////////////////////////////////////////////
            // EVENT LISTENERS
            //////////////////////////////////////////////////////////////////

            // TODO [JCS] Consider limiting the scope of what gets re-renderered when the model changes
            this.model.dataModel.entry.content.on('change associatedChange add remove reset', this.dataModelChangeHandler, this);
            this.model.dataModel.entry.content.objects.on('change associatedChange add remove reset', this.dataModelObjectsChangeHandler, this);
            this.model.setting.on('change:selectedObject', this.updateSelection, this);

            this.children.objectEditor.on('action:editCalculation', function(calculationID) {
                var objectName = this.model.setting.get('selectedObject');
                var selectedObject = this.model.dataModel.objectByName(objectName);
                var calculation = selectedObject.getCalculation(calculationID);
                var calculationType = calculation.get("calculationType");

                /**
                 * Edit an existing calculation
                 *
                 * @event DataModelEditor#action:editCalculation
                 * @param {string} - calculation type (Eval/GeoIP/Rex/Lookup)
                 * @param {string} - object name
                 * @param {string} - ID of the calculation
                 */
                this.trigger('action:editCalculation', calculationType, objectName, calculationID);
            }, this);

            this.children.objectEditor.on('action:addCalculation', function(calculationType) {
                var parentObjectName = this.model.setting.get('selectedObject');
                /**
                 * Add a new calculation
                 *
                 * @event DataModelEditor#action:addCalculation
                 * @param {string} - calculation type (Eval/GeoIP/Rex/Lookup)
                 * @param {string} - object name
                 */
                this.trigger('action:addCalculation', calculationType, parentObjectName);
            }, this);

            this.children.objectEditor.on('action:editObject', function(objectType, parentName, objectName) {
                /**
                 * Edit an existing object
                 *
                 * @event DataModelEditor#action:editObject
                 * @param {string} - object type (event/child/transaction/search)
                 * @param {string} - parent object's name
                 * @param {string} - object's name
                 */
                this.trigger("action:editObject", objectType, parentName, objectName);
            }, this);

            this.children.objectEditor.on('action:saveModel', function() {
                /**
                 * Save the Data Model
                 *
                 * @event DataModelEditor#action:saveModel
                 */
                this.trigger("action:saveModel");
            }, this);

            this.children.objectEditor.on('action:removeObject', function(objectModel) {
                /**
                 * Remove an Object
                 *
                 * @event DataModelEditor#action:removeObject
                 * @param {models/services/datamodel/private/Object} - object to delete
                 */
                this.trigger("action:removeObject", objectModel);
            }, this);

            /**
             * Listen for a itemClicked event from the AddObject dropdown menu
             *
             * @param itemData {string} value of the clicked item [event|transaction|child]
             */
            this.children.addObject.on("itemClicked", function(itemData) {
                /**
                 * Add a new Object
                 *
                 * @event DataModelEditor#action:addObject
                 * @param {string} - object type (event/child/transaction/search)
                 * @param {string} - selected object name
                 */
                this.trigger('action:addObject', itemData, this.model.setting.get('selectedObject'));
            }, this);
        },
        render: function() {
            // Get all of the sorted root trees
            var rootObjects = this.model.dataModel.entry.content.objects.getRootObjects();
            var eventsHtml = '<div class="content-group">';
            var transactionsHtml = '<div class="content-group">';
            var searchesHtml = '<div class="content-group">';
            var eventsCount = 0;
            var transactionsCount = 0;
            var searchesCount = 0;

            // Detach so jQuery doesn't remove all DOM event listeners when we set the html attribute
            this.children.objectEditor.detach();
            this.children.addObject.$el.detach();

            // Display error if the Data Model is accelerated
            if (this.model.dataModel.isAccelerated()) {
                this.children.flashMessagesView.flashMsgHelper.addGeneralMessage(CONSTS.EDIT_ACCELERATED_MODEL_ID,
                    { type: splunkDUtils.ERROR,
                        html: _("This Data Model cannot be edited because it is accelerated. " +
                                "Disable acceleration in order to edit the Data Model.").t()});

            } else {
                this.children.flashMessagesView.flashMsgHelper.removeGeneralMessage(CONSTS.EDIT_ACCELERATED_MODEL_ID);
            }

            // Generate the html for each section [Events/Transactions/Groups]
            // Each object is of type models/services/datamodel/private/Object
            _(rootObjects).each(function(object) {
                var parentName = object.get("parentName");
                if (parentName == DataModel.BASE_EVENT) {
                    eventsHtml += this.renderObjectTree(object);
                    eventsCount++;
                } else if (parentName == DataModel.BASE_TRANSACTION) {
                    transactionsHtml += this.renderObjectTree(object);
                    transactionsCount++;
                } else if (parentName == DataModel.BASE_SEARCH) {
                    searchesHtml += this.renderObjectTree(object);
                    searchesCount++;
                }
            }, this);

            eventsHtml += '</div>';
            transactionsHtml += '</div>';
            searchesHtml += '</div>';

            // Generate URL paths
            var pageRouter = route.getContextualPageRouter(this.model.application);
            var dataManagerUrl = pageRouter.data_model_manager();
            var pivotUrl = pageRouter.data_model_explorer({ data: { model: this.model.dataModel.id } });
            var docUrl = route.docHelp(
                this.model.application.get("root"),
                this.model.application.get("locale"),
                'learnmore.datamodel.about'
            );

            var html = _(this.template).template({
                displayName: this.model.dataModel.entry.content.get("displayName"),
                modelName: this.model.dataModel.entry.content.get("modelName"),
                docUrl: docUrl,
                pivotUrl: pivotUrl,
                eventsHtml: eventsHtml,
                transactionsHtml: transactionsHtml,
                searchesHtml: searchesHtml,
                showEvents: eventsCount,
                showSearches: searchesCount,
                showTransactions: transactionsCount,
                dataManagerUrl: dataManagerUrl,
                selected: this.model.setting.get('selectedObject')
            });

            this.$el.html(html);
            // TODO [JCS] Look into doing a replaceWith instead of an append
            if (!this.model.dataModel.isAccelerated()) {
                this.$('.add-button-holder').append(this.children.addObject.render().el);
            }

            if (this.children.editObject)
                this.$('.edit-button-holder').replaceWith(this.children.editObject.render().el);
            else
                this.$('.edit-button-holder').remove();
            this.$('.object-editor-wrapper').append(this.children.objectEditor.el);
            this.$('.flash-messages-placeholder').append(this.children.flashMessagesView.render().el);

            this.updateSelection();

            return this;
        },

        // Helper render function that takes a Data Model object root node, walks down the tree and generates
        // nested divs that represent the tree.
        renderObjectTree: function(objectNode) {
            // TODO [JCS] Investigate using _.memoize
            var linkHtml = _(this.objectLinkTemplate).template({objectName: objectNode.get("objectName"),
                                                                displayName : objectNode.get("displayName")});
            var html = linkHtml;

            if (objectNode.children) {
                var childrenLength = objectNode.children.length;
                if (childrenLength) {
                    html += '<div class="content-group">';
                    _(objectNode.children).each(function(object) {
                        html += this.renderObjectTree(object);
                    }, this);
                    html += '</div>';
                }
            }

            return html;
        },

        // Updates the object tree highlighting and internal model
        updateSelection: function() {
            this.$('.object').removeClass('selected');
            var selected = this.model.setting.get('selectedObject');
            if(selected) {
                this.$('.object[data-objid="' + selected + '"]').addClass('selected');
            }
            this.children.objectEditor.setModel(this.model.dataModel.objectByName(selected));
        },

        // Helper function to determine what items to display in the Add Object dropdown
        getEventItems: function() {
            if (this.model.dataModel.entry.content.objects &&
                this.model.dataModel.entry.content.objects.length == 0) {
                return CONSTS.INITIAL_OBJECT_ITEMS;
            } else {
                return CONSTS.ALL_OBJECT_ITEMS;
            }
        },

        // Template for an Object link
        objectLinkTemplate: '<a href="#" class="object" data-objid="<%- objectName %>"><%- displayName %></a>',

        template: '\
            <div class="section-padded section-header">\
                <div class="header-button-holder pull-right">\
                    <div class="btn-group">\
                        <span class="edit-button-holder btn"><%- _("Edit Placeholder").t() %></span>\
                        <a href="<%- pivotUrl %>" class="pivot-model btn"><%- _("Pivot").t() %></a>\
                        <a href="<%- docUrl %>" target="_blank" class="help-link btn"><%- _("Documentation").t() %> <i class="icon-external"></i></a>\
                    </div>\
                </div>\
                <h2 class="section-title"><%- displayName %></h2>\
                <p><%- modelName %></p>\
                <a href="<%- dataManagerUrl %>"><i class="icon-triangle-left-small icon-no-underline"></i><span><%- _("Back to Data Models").t() %></span></a>\
                <div class="flash-messages-placeholder"></div>\
            </div>\
            <div class="section-wrapper">\
                <div class="divider-gradient"></div>\
                <div class="object-list sidebar">\
                    <div class="object-list-header">\
                        <h3 class="title"><%- _("Objects").t() %></h3>\
                        <div class="add-button-holder">\
                        </div>\
                    </div>\
                    <% if(showEvents) { %>\
                        <h6 class="object-group-header"><%- _("Events").t() %></h6>\
                        <%= eventsHtml %>\
                    <% } %>\
                    <% if(showSearches) { %>\
                        <h6 class="object-group-header"><%- _("Searches").t() %></h6>\
                        <%= searchesHtml %>\
                    <% } %>\
                    <% if(showTransactions) { %>\
                        <h6 class="object-group-header"><%- _("Transactions").t() %></h6>\
                        <%= transactionsHtml %>\
                    <% } %>\
                </div>\
                <div class="object-editor-placeholder"></div>\
                <div class="object-editor-wrapper "></div>\
            </div>\
        '
    }, CONSTS);

});

/**
 * Edit a data model title
 *
 * @event DataModelEditor#action:editDataModelTitle
 * @param {string} - the data model name
 */
/**
 * Edit a data model permissions
 *
 * @event DataModelEditor#action:editPermissions
 * @param {string} - the data model name
 */
/**
 * Edit a data model acceleration
 *
 * @event DataModelEditor#action:editAcceleration
 * @param {string} - the data model name
 */
/**
 * Clone a data model
 *
 * @event DataModelEditor#action:cloneDataModel
 * @param {string} - the data model name
 */
