/**
 * @author sfishel
 * @author jszeto
 *
 * View for editing the attributes of a data model object.
 *
 * Display a list of constraints, fields (inherited/own), and calculations.  With calls to action for editing.
 *
 * The HTML template is stored in views/data_model_editor/ObjectEditor.html
 *
 * Child Views:
 *
 *  addAttributeButton - DropDownMenu control used to select an attribute to add to the object
 *
 *  deleteDialog - Shared dialog that deletes the object, an attribute or a calculation
 *
 *  renameDialog - Shared dialog that renames an object, attribute or calculation field
 *
 *  viewValuesDialog - Shared dialog that shows some search results for an attribute
 *
 *  addExtractedFieldsDialog - Dialog that displays the possible auto-extracted fields
 *
 *  editExtractedFieldDialog - Dialog for editing a single auto-extracted field
 *
 * Inputs:
 * options {Object} {
 *     model: {
 *         dataModel {models/services/datamodel/DataModel}
 *         application {models/Application}
 *         objectModel {models/services/datamodel/private/Object} - optional Object model. Can be set using this.setModel
 *     }
 *     collection: {
 *         apps {collections/services/AppLocals}
 *     }
 *  }
 *
 *  @fires ObjectEditor#action:editObject
 *  @fires ObjectEditor#action:editCalculation
 *  @fires ObjectEditor#action:addCalculation
 *  @fires ObjectEditor#action:removeObject
 *  @fires ObjectEditor#action:saveModel
 *
 */

define([
            'jquery',
            'underscore',
            'views/Base',
            'views/shared/DropDownMenu',
            'views/shared/dialogs/TextInputDialog',
            'views/shared/dialogs/TextDialog',
            'views/shared/datamodel/dialogs/ViewValuesDialog',
            'views/data_model_editor/form_views/AddExtractedFieldsDialog',
            'views/data_model_editor/form_views/EditExtractedFieldDialog',
            'util/datamodel/field_icon_utils',
            'models/services/datamodel/DataModel',
            'models/services/datamodel/private/Object',
            'jquery.ui.sortable',
            'contrib/text!views/data_model_editor/ObjectEditor.html',
            'uri/route',
            'module',
            'util/console',
            'splunk.util'
       ],
        function(
            $,
            _,
            BaseView,
            DropDownMenu,
            TextInputDialog,
            TextDialog,
            ViewValuesDialog,
            AddExtractedFieldsDialog,
            EditExtractedFieldDialog,
            fieldIconUtils,
            DataModel,
            ObjectModel,
            JQSortable,
            ObjectEditorTemplate,
            route,
            module,
            console,
            splunkUtils
        )
{
    return BaseView.extend({

        className: 'data-model-object-editor',
        moduleId: module.id,
        template: ObjectEditorTemplate,

        // Internal variables used to keep state for the rename and delete dialogs
        _deleteDialogType : "",
        _deleteDialogID : "",

        events: {
            ///////////////////////
            //  Object Actions
            ///////////////////////
            'click .rename-object-button': function(e) {
                e.preventDefault();
                this.showRenameDialog("object",this.model.objectModel.get("displayName"));
            },
            'click .delete-object-button': function(e) {
                e.preventDefault();
                this.showDeleteDialog("object",this.model.objectModel.get("displayName"), this.model.objectModel);
            },
            ///////////////////////
            //  Constraint Actions
            ///////////////////////
            'click .edit-constraint-button': function(e) {
                e.preventDefault();

                var parentName = this.model.objectModel.get("parentName");
                var type = "child";

                if (parentName == DataModel.BASE_TRANSACTION)
                    type = "transaction";
                else if (parentName == DataModel.BASE_SEARCH)
                    type = "search";
                else if (parentName == DataModel.BASE_EVENT)
                    type = "event";

                /**
                 * Edit an existing object
                 *
                 * @event ObjectEditor#action:editObject
                 * @param {string} - object type (event/child/transaction/search)
                 * @param {string} - parent object's name
                 * @param {string} - object's name
                 */
                this.trigger("action:editObject", type, parentName, this.model.objectModel.get("objectName"));
           },

            ///////////////////////
            //  Attribute Actions
            ///////////////////////
            'click .view-attribute-values-button': function(e) {
                e.preventDefault();
                var fieldModel = this.model.objectModel.getAnyField(this.resolveAttributeName(e.currentTarget),
                    this.resolveAttributeCalculationId(e.currentTarget));

                this.showViewValuesDialog(fieldModel);
            },
            'click .delete-attribute-button': function(e) {
                e.preventDefault();
                var fieldName = this.resolveAttributeName(e.currentTarget);
                this.showDeleteDialog("attribute", fieldName, this.model.objectModel.getField(fieldName));
            },
            'click .edit-attribute-button': function(e) {
                e.preventDefault();
                this.showEditExtractFieldsDialog(this.resolveAttributeName(e.currentTarget));
            },
            'click .show-attribute-button': function(e) {
                e.preventDefault();
                this.setFieldValue(e.currentTarget, "hidden", false);
            },
            'click .hide-attribute-button': function(e) {
                e.preventDefault();
                this.setFieldValue(e.currentTarget, "hidden", true);
            },
            'click .optional-attribute-button': function(e) {
                e.preventDefault();
                this.setFieldValue(e.currentTarget, "required", false);
            },
            'click .require-attribute-button': function(e) {
                e.preventDefault();
                this.setFieldValue(e.currentTarget, "required", true);
            },
            ///////////////////////
            //  Calculation Actions
            ///////////////////////
            'click .edit-calculation-button': function(e) {
                //console.log("Edit Calculation button clicked");
                e.preventDefault();
                /**
                 * Edit an existing calculation
                 *
                 * @event ObjectEditor#action:editCalculation
                 * @param {string} - ID of the calculation
                 */
                this.trigger('action:editCalculation', this.resolveCalculationId(e.currentTarget));
            },
            'click .delete-calculation-button': function(e) {
                e.preventDefault();
                var calcID = this.resolveCalculationId(e.currentTarget);
                this.showDeleteDialog("calculation", calcID, this.model.objectModel.getCalculation(calcID));
            },
            'click .view-calculation-values-button': function(e) {
                e.preventDefault();
                var fieldModel = this.model.objectModel.getAnyField(this.resolveAttributeName(e.currentTarget),
                                                                    this.resolveCalculationId(e.currentTarget));
                this.showViewValuesDialog(fieldModel);
            }
        },

        ////////////////////////////////////////
        // Event Handlers
        ////////////////////////////////////////

        /**
         * Listen for a itemClicked event from the Add Attribute dropdown menu
         *
         * @param itemData {Object} value of the clicked item
         */
        addAttributeButtonClickHandler: function(itemData) {
            if (itemData == "autoExtract") {
                var parentName = this.model.objectModel.get("parentName");
                if (parentName == DataModel.BASE_EVENT ||
                    parentName == DataModel.BASE_SEARCH) {
                    this.showAddExtractedFieldsDialog();
                }
            } else {
                /**
                 * Add a new calculation
                 *
                 * @event ObjectEditor#action:addCalculation
                 * @param {string} - calculation type (Eval/GeoIP/Rex/Lookup)
                 * @param {string} - display name TODO [JCS] Can we remove this?
                 */
                this.trigger('action:addCalculation', itemData, this.model.objectModel.get('displayName'));
            }
        },
        /**
         * Show a delete dialog given a type and the name of that type
         *
         * @param {string} type - [object|attribute|calculation]
         * @param {string} id - name of the object, attribute or calculation
         * @param {Model} model - model of the object, attribute or calculation
         */
        showDeleteDialog: function(type, id, model) {
            var attrCount = 1;
            var name = id;

            var text;
            var typeToDisplay = type == "object" ? "object" : "attribute";
            var title = type == "object" ? _("Delete Object").t() : _("Delete Attribute").t();

            if (type == "calculation") {
                attrCount = model.get("calculationType") == "GeoIP" ?
                                model.getVisibleOutputFieldsLength() :
                                model.outputFields.length;

                if (attrCount == 1) {
                    // Get the displayName of the only outputField
                    name = model.outputFields.at(0).get("displayName");
                }
            }

            if (attrCount > 1) {
                text = splunkUtils.sprintf(_("Are you sure you want to delete this %s and its %s attributes").t(),
                                             model.get("calculationType"), attrCount);
            } else if (typeToDisplay == "object") {
                text = splunkUtils.sprintf(_("Are you sure you want to delete the %s <em>%s</em> and all of its descendants").t(),
                                                typeToDisplay, _(name).escape());
            } else {
                text = splunkUtils.sprintf(_("Are you sure you want to delete the %s <em>%s</em>").t(),
                                                typeToDisplay, _(name).escape());
            }

            text += "?";

            // Store state for the delete dialog
            this._deleteDialogType = type;
            this._deleteDialogID = id;

            this.children.deleteDialog = new TextDialog({id: "modal_delete"});
            this.children.deleteDialog.settings.set("primaryButtonLabel",_("Delete").t());
            this.children.deleteDialog.settings.set("cancelButtonLabel",_("Cancel").t());
            this.children.deleteDialog.on('click:primaryButton', this.deleteDialogDeleteHandler, this);
            this.children.deleteDialog.on("click:closeButton", this.deleteDialogCancelHandler, this);
            this.children.deleteDialog.on("click:cancelButton", this.deleteDialogCancelHandler, this);

            this.children.deleteDialog.settings.set("titleLabel",_(title).escape());
            this.children.deleteDialog.setText(text);
            $("body").append(this.children.deleteDialog.render().el);
            this.children.deleteDialog.show();
        },
        /**
         * Handles when the user presses the Delete button in the delete dialog. Updates the model based on the
         * type of delete operation (object, attribute, of calculation).
         *
         * @param dialog
         */
        deleteDialogDeleteHandler: function(dialog) {
            var type = this._deleteDialogType;
            var name = this._deleteDialogID;

            if (type == "object") {
                /**
                 * Remove an Object
                 *
                 * @event ObjectEditor#action:removeObject
                 * @param {models/services/datamodel/private/Object} - object to delete
                 */
                this.trigger("action:removeObject", this.model.objectModel);
                this.saveModel();
            } else if (type == "attribute") {
                this.model.objectModel.removeField({ fieldName: name });
                this.saveModel();
            } else if (type == "calculation") {
                this.model.objectModel.removeCalculation(name);
                this.saveModel();
            }

            this.clearDeleteDialogState();
        },
        deleteDialogCancelHandler: function(dialog) {
            this.clearDeleteDialogState();
        },
        /**
         * Show a rename dialog given a type and the name of that type
         * TODO [JCS] Remove unused parameters
         */
        showRenameDialog: function(type, name, fieldName) {
            this.children.renameDialog = new TextInputDialog({id: "modal_rename",
                                                              parent: this,
                                                              model: this.model.objectModel,
                                                              modelAttribute: "displayName",
                                                              label:_("Name").t()});
            this.children.renameDialog.settings.set("titleLabel",_("Rename Object").t());
            this.children.renameDialog.on("click:cancelButton", this.renameDialogCancelHandler, this);
            this.children.renameDialog.on("click:closeButton", this.renameDialogCancelHandler, this);
            this.children.renameDialog.on("click:primaryButton", this.renameDialogChangeHandler, this);
            $("body").append(this.children.renameDialog.render().el);
            this.children.renameDialog.show();
        },
        /**
         * Called when the rename dialog cancel button is pressed
         * @param dialog
         */
        renameDialogCancelHandler: function(dialog) {
            this.children.renameDialog.remove();
        },
        /**
         * Called when the renameDialog sends us a change:textInputValue event
         *
         * @param {string} newValue - the new value from the text input dialog
         */
        renameDialogChangeHandler: function(newValue) {
            this.saveModel();
        },

        /**
         * Show the addExtractedFieldsDialog, set the model to the currently selected Object and kick off a new
         * search to get the field summary info
         */
        showAddExtractedFieldsDialog: function() {
            // Create a shared auto-extract dialog
            this.children.addExtractedFieldsDialog =
                new AddExtractedFieldsDialog({id: "modal_addExtractedFields",
                    parent: this,
                    model: {objectModel: this.model.objectModel,
                            application: this.model.application}});
            this.children.addExtractedFieldsDialog.on("action:saveModel", this.saveModel, this);
            this.children.addExtractedFieldsDialog.on("hidden", this.destroyAddExtractFieldsDialog, this);

            this.children.addExtractedFieldsDialog.show();
        },

        /**
         * Destroy the editExtractedFieldDialog whenever it has been hidden.
         */
        destroyAddExtractFieldsDialog: function() {
            this.children.addExtractedFieldsDialog.remove();
            delete this.children.addExtractedFieldsDialog;
        },

        /**
         * Show a new editExtractedFieldDialog when the user clicks on the edit button of an extracted field
         *
         * @param {string} fieldName - fieldName of the extracted field
         */
        showEditExtractFieldsDialog: function(fieldName) {
            // Make a copy of the extracted field model and pass this to the dialog
            var fieldModel = this.model.objectModel.getField(fieldName);

            this.children.editExtractedFieldDialog =
                new EditExtractedFieldDialog({id: "modal_editExtractedFields",
                    parent: this,
                    model: fieldModel.clone()});
            this.children.editExtractedFieldDialog.on("action:saveExtractedField", this.saveExtractedField, this);
            this.children.editExtractedFieldDialog.on("hidden", this.destroyEditExtractFieldsDialog, this);

            this.children.editExtractedFieldDialog.show();
        },

        /**
         * Destroy the editExtractedFieldDialog whenever it has been hidden.
         */
        destroyEditExtractFieldsDialog: function() {
            this.children.editExtractedFieldDialog.remove();
            delete this.children.editExtractedFieldDialog;
        },

        /**
         * Called when the editExtractedFieldDialog has validated and submitted its changes to the field model
         *
         * @param {models/services/datamodel/private/Field} workingField - working model of the extracted field being saved
         */
        saveExtractedField: function(workingField) {
            // Get the field model
            var field = this.model.objectModel.getField(workingField.get("fieldName"));

            // Copy over values
            field.set(workingField.attributes);

            // Save the DataModel
            this.saveModel();
        },

        /**
         * Show the View Values dialog for a given field or calculation output field
         * @param {models/services/datamodel/private/Field} fieldModel
         */
        showViewValuesDialog: function(fieldModel) {
            this.children.viewValuesDialog = new ViewValuesDialog({
                id: "modal_viewValues",
                model: {
                    objectModel: this.model.objectModel,
                    fieldModel: fieldModel,
                    application: this.model.application
                }
            });

            this.children.viewValuesDialog.show();
        },

        /**
         * Helper function to set an attribute on a field and save the model
         *
         * @param {DOM element} element - the DOM element used to extract the field
         * @param {string} attrName - the field's attribute name to set
         * @param {Object} attrValue - the value to set on the field's attribute
         */
        setFieldValue: function(element, attrName, attrValue) {
            var field = this.model.objectModel.getAnyField(this.resolveAttributeName(element),
                                                           this.resolveAttributeCalculationId(element));
            field.set(attrName, attrValue);
            this.saveModel();
        },

        ////////////////////////////////////////
        // Backbone.View Functions
        ////////////////////////////////////////

        initialize: function(options) {
            BaseView.prototype.initialize.apply(this, arguments);

            options = options || {};

            // Backbone Views automatically assign the options.model to this.model. However, we need to add bindings,
            // so we pass this through our setModel function.
            if (options.model.objectModel)
            {
                this.setModel(options.model.objectModel);
            }
        },

        render: function() {
            // If there is no objectModel, display a message
            if(!this.model.objectModel) {
                this.$el.html(_(this.emptyTemplate).template({}));

                return this;
            }

            if (this.children.addAttributeButton) {
                this.children.addAttributeButton.off();
                this.children.addAttributeButton.$el.detach();
            }

            var html = this.compiledTemplate({
                object: this.model.objectModel,
                getIcon: fieldIconUtils.fieldIconByType,
                getFieldFlags: this.getFieldFlags,
                getFieldHiddenIcon: fieldIconUtils.fieldHiddenIcon,
                getFieldRequiredIcon: fieldIconUtils.fieldRequiredIcon,
                getTimelineURL :_(this.getTimelineURL).bind(this),
                isEditable : !this.model.dataModel.isAccelerated()
            });
            this.$el.html(html);

            // Only non-accelerated Data Models can be edited.
            if (!this.model.dataModel.isAccelerated()) {
                // Create a subview to select an attribute to add
                var allowAutoExtracted = !(this.model.objectModel.isChildObject() ||
                    this.model.objectModel.get("parentName") == DataModel.BASE_TRANSACTION);

                this.children.addAttributeButton = new DropDownMenu({
                    className: '',
                    label:_("Add Attribute").t(),
                    items:[{label:_("Auto-Extracted").t(), value:"autoExtract", enabled:allowAutoExtracted},
                        {label:_("Eval Expression").t(), value:"Eval"},
                        {label:_("Lookup").t(), value:"Lookup"},
                        {label:_("Regular Expression").t(), value:"Rex"},
                        {label:_("Geo IP").t(), value:"GeoIP"}],
                    dropdownClassName: 'dropdown-menu-narrow'});

                this.children.addAttributeButton.on("itemClicked", this.addAttributeButtonClickHandler, this);
                this.$('.add-attribute-button-holder').append(this.children.addAttributeButton.render().el);
            }

            this.configureSortable();

            return this;
        },

        /**
         * Takes an Object model. Removes bindings from the old model and assigns them to the new one.
         *
         * @param {models/services/datamodel/private/Object} model - the Object model to edit
         */
        setModel: function(model) {

            // Remove binding from the previous model
            if (this.model.objectModel) {
                this.model.objectModel.off(null, null, this);
                this.model.objectModel.associatedOff(null, null, this);
            }

            if (model && !(model instanceof ObjectModel))
            {
                throw new Error("ObjectEditor model must be an Object model");
            }

            this.model.objectModel = model;

            if (this.model.objectModel)
            {
                this.model.objectModel.on('change', this.changeHandler, this);
                this.model.objectModel.on('associatedChange', this.associatedChangeHandler, this);
                this.model.objectModel.on('reset', this.resetHandler, this);
                //this.model.objectModel.fields.on('add remove update', this.debouncedRender, this);
                //this.model.objectModel.calculations.on('add remove update', this.debouncedRender, this);
            }
            this.debouncedRender();
        },

        // TODO [JCS] We can collapse the change/associatedChange/reset handlers into one. This was split for debugging
        changeHandler: function(model, options) {
            this.debouncedRender();
        },

        associatedChangeHandler: function(model, options) {
            this.debouncedRender();
        },

        resetHandler: function(model, options) {
            this.debouncedRender();
        },

        ////////////////////////////////////////
        // Helper Functions
        ////////////////////////////////////////

        saveModel: function() {
            /**
             * Save the Data Model
             *
             * @event ObjectEditor#action:saveModel
             */
            this.trigger("action:saveModel", this.model.objectModel);
        },

        /**
         * Sets up the configuration options for the jQuery UI sortable plugin used to manage the drag-and-drop resorting
         * of the objects calculations.
         *
         * For documentation, see:  http://jqueryui.com/sortable/ or http://api.jqueryui.com/sortable/
         */
        configureSortable: function() {
            var that = this;
            this.$('.calculations').sortable({
                items: '.calculation',
                placeholder: 'ui-sortabale-placeholder',
                forcePlaceholderSize: true,
                opacity: 0.7,
                tolerance: 'pointer',
                stop: function(event, ui) {
                    that.onSortStop();
                }
            });
        },

        /**
         * Called when the user has dropped the target calculation. Updates the Object model with the calculation order
         * in the UI. Then saves the model.
         */
        onSortStop: function() {
            var newOrder = _(this.$('.calculation')).map(function(elem) {
                return this.resolveCalculationId(elem);
            }, this);
            this.model.objectModel.reOrderCalculations(newOrder);
            this.saveModel();
        },

        /**
         * Helper function to get the Field name of a DOM element
         * @param {DOM element} elem
         * @return {string} the Field.fieldName associated with the DOM element
         */
        resolveAttributeName: function(elem) {
            return $(elem).closest('.attribute').attr('data-attribute-name');
        },
        /**
         * Helper function to get the Calculation ID of a DOM element. Used for inherited calculations.
         * @param {DOM element} elem
         * @return {string} the Calculation.calculationID associated with the DOM element
         */
        resolveAttributeCalculationId: function(elem) {
            return $(elem).closest('.attribute').attr('data-calc-id');
        },
        /**
         * Helper function to get the Calculation ID of a DOM element. Used for an Object's own calculations.
         * @param {DOM element} elem
         * @return {string} the Calculation.calculationID associated with the DOM element
         */
        resolveCalculationId: function(elem) {
            return $(elem).closest('.calculation').attr('data-calc-id');
        },
        clearDeleteDialogState: function() {
            this._deleteDialogID = "";
            this._deleteDialogType = "";
        },

        /**
         * Helper function called in the HTML template to generate the URL link for the View Events button
         * @param {string} lineage - The Object's objectName or lineage
         * @param {string} type - Either (search|transaction) which determines the type of search string to use
         * @return {string} URL for viewing events for the given lineage
         */
        getTimelineURL: function(lineage, type) {
            if (this.model) {
                var objectModel = this.model.dataModel.objectByLineage(lineage);

                if (objectModel) {
                    var searchString = type == "search" ? objectModel.get("baseSearch") : objectModel.get("previewSearch");

                    return route.search(this.model.application.get("root"),
                                          this.model.application.get("locale"),
                                          this.model.application.get("app"),
                                          { data: {q: searchString + " | head 1000"}});
                }
            }

            return "#";
        },

        /**
         * Helper function called in the HTML template to
         * @param field
         * @return {String}
         */
        getFieldFlags: function(field) {
            if (field.hidden && field.required)
                return "(required, hidden)";
            else if (field.required)
                return "(required)";
            else if (field.hidden)
                return "(hidden)";

            return "";
        },

        emptyTemplate: '<div class="alert alert-info">\
                            <i class="icon-alert"></i>\
                            <%- _("To get started, add an object using the menu to the left.").t() %>\
                        </div>'
    });

});
