/**
 * @author jszeto
 * @date 11/19/12
 *
 * A subview to add or edit a Lookup Calculation
 *
 * inputs:
 *      model
 *          calculation {models/services/datamodel/private/LookupCalculation} - The lookup calculation to edit
 *          objectModel {models/services/datamodel/private/Objects} - The object model that owns the calculation model
 *          dataModel {models/services/datamodel/DataModel} - The data model that owns the object model
 *      collection
 *          transformsLookups {collections/services/data/TransformsLookups} - The lookups available in the system
 *
 */

define([
    'jquery',
    'underscore',
    'views/data_model_editor/form_views/AddEditFormBaseView',
    'views/shared/controls/ControlGroup',
    'views/data_model_editor/form_components/LookupOutputFieldsView',
    'module',
    'util/console',
    'util/splunkd_utils'
],
    function(
        $,
        _,
        AddEditFormBaseView,
        ControlGroup,
        LookupOutputFieldsView,
        module,
        console,
        splunkDUtils
        )
    {
        return AddEditFormBaseView.extend({
            moduleId: module.id,
            className: 'add-edit-lookup-view',
            MISSING_OUTPUTFIELD_ID: "MISSING_OUTPUTFIELD_ID",

            /**
             * Handle a request from LookupOutputFieldsView to add a new field
             */
            // TODO [JCS] Handle case where you can't add a field if an existing field has an empty fieldName
            addOutputFieldHandler: function() {
                var newField = this.model.calculation.addField();
                if (newField)
                    this.flashMessagesHelper.register(newField);
            },

            /**
             * Handle a request from LookupOutputFieldsView to remove a specific field
             *
             * @param cid - cid of the field to remove
             */
            removeOutputFieldHandler: function(cid) {
                var existingField = this.model.calculation.getFieldByName(cid);

                if (existingField)
                    this.flashMessagesHelper.unregister(existingField);

                this.model.calculation.removeField(cid);
            },

            /**
             * @constructor
             * @param options {Object} {
                 }
             */
            initialize: function(options) {
                AddEditFormBaseView.prototype.initialize.call(this, options);

                this.model.objectModel = this.model.dataModel.objectByName(this.parentObjectName);

                if (_(this.model.objectModel).isUndefined()) {
                    this.addObjectError(this.parentObjectName);
                    return;
                }

                // Map the lookup tables into objects we can pass into a SyntheticSelct
                var lookupNames = _(this.collection.transformsLookups.getNames()).map(
                                        function(lookupName) {return {label:lookupName, value:lookupName};});

                // TODO [JCS] Figure out how we want the fields sorted
                var fieldNames = _(this.model.objectModel.getAllowedInputFields()).map(function(field) {
                                            return {label: field.displayName, value:field.fieldName};
                                        }
                );

                // TODO [JCS] If lookupNames or fieldNames is empty, don't allow any adding and show an error message

                if (this.operation == "add")
                {
                    var calcAttributes = {calculationType: "Lookup",
                                          inputField: fieldNames[0].value,
                                          lookupName: lookupNames[0].value,
                                          calculationID: ""};
                    // Create a new Calculation model
                    this.model.calculation = this.model.objectModel.createCalculation(calcAttributes);

                    // Create a new outputField model
                    this.model.calculation.addField();
                }
                else
                {
                    this.model.calculation = this.model.objectModel.getCalculation(this.calculationID);

                    if (_(this.model.calculation).isUndefined()) {
                        this.addCalculationError(this.calculationID);
                        return;
                    }
                }

                // TODO [JCS] Figure a way to divorce label from error message decorator
                this.children.selectLookupName = new ControlGroup({label: "",
                                                               controlType: "SyntheticSelect",
                                                               controlOptions:
                                                                    {model:this.model.calculation,
                                                                     modelAttribute:"lookupName",
                                                                     items:lookupNames,
                                                                     toggleClassName: 'btn'}});

                this.children.selectInputField = new ControlGroup({label: _("Attribute:").t(),
                                                              controlType: "SyntheticSelect",
                                                              controlOptions:
                                                                    {model:this.model.calculation,
                                                                     modelAttribute:"inputField",
                                                                     items:fieldNames,
                                                                     toggleClassName: 'btn'}});
                
                this.children.inputLookupField = new ControlGroup({label: _("Field in Lookup:").t(),
                                                                   controlType: "Text",
                                                                   controlOptions:
                                                                        {model:this.model.calculation,
                                                                         modelAttribute:"lookupField"}
                                                                   });

                this.children.outputFields = new LookupOutputFieldsView({model:this.model.calculation,
                                                                         flashMessagesHelper:options.flashMessagesHelper});
                this.children.outputFields.on("action:addOutputField", this.addOutputFieldHandler, this);
                this.children.outputFields.on("action:removeOutputField", this.removeOutputFieldHandler, this);

                this.flashMessagesHelper.register(this.model.calculation);
            },

            /**
             * Called by both save and preview to perform validation on the models
             * @return {boolean}
             */
            applyControlsToModel: function() {
                var isValid = this.model.calculation.set({}, {validate:true});

                if (this.model.calculation.outputFields.length == 0) {
                    this.flashMessagesHelper.addGeneralMessage(this.MISSING_OUTPUTFIELD_ID,
                                                               {type: splunkDUtils.ERROR,
                                                                html: _("Specify at least one Output Attribute.").t()});
                    isValid = false;
                } else {
                    this.flashMessagesHelper.removeGeneralMessage(this.MISSING_OUTPUTFIELD_ID);
                }

                this.model.calculation.withEachField(function(field) {
                    isValid = field.set({}, {validate:true}) && isValid;
                }, this);

                return isValid;
            },

            /**
             * Tells the view to save the inputs into the dataModel. resultCallback will be called after we get a response
             * from the splunkd
             *
             * @param resultCallback {Function}
             */
            save: function(resultCallback) {
                AddEditFormBaseView.prototype.save.call(this, resultCallback);

                if (this.applyControlsToModel()) {
                    var resultXHR = this.model.dataModel.save({});
                    if (resultXHR) {
                        resultXHR.done(_(function() {
                            this.trigger("action:save");
                        }).bind(this));
                    }
                }
            },

            /**
             * Tells the view to provisionally save the inputs into the dataModel. resultCallback will be called after we
             * get a response from splunkd.
             *
             * @param resultCallback {Function}
             */
            preview: function(resultCallback) {
                AddEditFormBaseView.prototype.preview.call(this, resultCallback);

                if (this.applyControlsToModel()) {
                    this.trigger("action:preview",
                        this.model.objectModel.get("objectName"),
                        this.model.calculation.getFieldNames());
                }
            },
            /**
             * Tells the view to perform unbinding and any other cleanup before it is destroyed.
             */
            cleanup: function() {
                AddEditFormBaseView.prototype.cleanup.call(this);
            },
            renderContent: function()
            {
                var html = _(this.template).template({
                    type: this.type,
                    operation: this.operation
                });
                this.$el.html(html);

                // lame hack to add classnames :(
                // $(this.children.selectLookupName.render().el).addClass('pull-left');

                this.$(".selectLookupName-placeholder").replaceWith(this.children.selectLookupName.render().el);
                this.$(".selectInputField-placeholder").replaceWith(this.children.selectInputField.render().el);
                this.$(".inputLookupField-placeholder").replaceWith(this.children.inputLookupField.render().el);
                this.$(".output-fields-placeholder").replaceWith(this.children.outputFields.render().el);

                return this;
            },
            template: '\
                <div class="col-1">\
                    <label class="control-label"><%- _("Lookup Table").t() %></label>\
                    <span class="selectLookupName-placeholder"/>\
                    <label class="control-label"><%- _("Input").t() %></label>\
                    <div class="lookup-controls">\
                        <span class="selectInputField-placeholder"/> = <span class="inputLookupField-placeholder"/>\
                    </div>\
                </div>\
                <div class="col-2">\
                    <label class="control-label"><%- _("Output").t() %></label>\
                    <span class="output-fields-placeholder"/>\
                </div>\
            '
        });
    });
