/**
 * @author jszeto
 * @date 12/18/12
 *
 * Inputs:
 *     options {
 *         parent {View} the view that created this dialog,
 *         model {
 *             objectModel {models/services/datamodel/private/Object} the Object being edited
 *             application {models/Application}
 *         }
 *     }
 *
 * @fires AddExtractedFieldsDialog#action:saveModel
 */
define(
[
    'jquery',
    'underscore',
    'backbone',
    'views/shared/dialogs/DialogBase',
    'views/shared/controls/ControlGroup',
    'views/data_model_editor/form_components/SelectSampleSize',
    'views/data_model_editor/form_components/FieldTable',
    'views/shared/jobstatus/Spinner',
    'views/shared/jobstatus/Count',
    'views/shared/FlashMessages',
    'models/custom/dataModel/ExtractedField',
    'models/Job',
    'models/services/search/jobs/Summary',
    'models/services/datamodel/DataModel',
    'util/splunkd_utils',
    'module',
    'util/console',
    'bootstrap.tooltip'
],
    function(
        $,
        _,
        Backbone,
        DialogBase,
        ControlGroup,
        SelectSampleSize,
        FieldTable,
        Spinner,
        Count,
        FlashMessagesView,
        ExtractedField,
        Job,
        Summary,
        DataModel,
        splunkdUtils,
        module,
        console
        /* bootstrap tooltip */
        )
{
    return DialogBase.extend({
        moduleId: module.id,
        className: "modal fade modal-wide",
        validSearch: true,

        initialize: function(options) {
            DialogBase.prototype.initialize.call(this, options);

            this.bodyClassName = "modal-body-scrolling";

            this.selectSampleSize = new SelectSampleSize({});
            this.selectSampleSize.on("change", this.selectSampleSizeChangeHandler, this);

            this.children.selectSampleSizeGroup = new ControlGroup({label: _("Sample Size:").t(),
                                                                    controls: [this.selectSampleSize]
                                                                   });

            this.children.flashMessagesView = new FlashMessagesView();

            this.startNewSearch();

            this.settings.set("primaryButtonLabel", _("Save").t());
            this.settings.set("cancelButtonLabel", _("Cancel").t());
            this.settings.set("titleLabel", _("Add Auto-Extracted Field").t());
        },

        selectSampleSizeChangeHandler: function(newValue, oldValue) {
            // Send a new search request
            // Put up the loading page
            this.startNewSearch();
            this.debouncedRender();
        },

        summaryChangeHandler: function() {
            this.parseSummary();
        },

        /**
         * Starts a new search job to retrieve the extracted fields for the Object
         */
        startNewSearch: function() {
            var value = this.selectSampleSize.getValue();
            var searchString;
            var rootParent = this.model.objectModel.get("rootParent");

            if (rootParent == DataModel.BASE_SEARCH)
                searchString = this.model.objectModel.get("baseSearch");
            else
                searchString = this.model.objectModel.get("autoextractSearch");

            this.validSearch = searchString != "";

            // Destroy old models and remove views
            this.destroySearchJob();
            if (this.children.fieldTable)
                this.children.fieldTable.remove();

            if (this.children.spinner)
                this.children.spinner.remove();

            if (this.children.count)
                this.children.count.remove();

            this.extractedFieldsCollection = new Backbone.Collection([], {model:ExtractedField});
            this.children.fieldTable = new FieldTable({collection: this.extractedFieldsCollection});

            // Create new Job and Summary
            this.model.searchJob = new Job({}, { delay: 1000, processKeepAlive: true });
            this.model.summary = new Summary();

            var fatalErrorWarning = [splunkdUtils.FATAL, splunkdUtils.ERROR, splunkdUtils.WARNING];
            this.children.flashMessagesView.flashMsgHelper.register(this.model.searchJob, fatalErrorWarning);
            this.children.flashMessagesView.flashMsgHelper.register(this.model.summary, fatalErrorWarning);

            // If the searchJob links change, then update the summary ID
            this.model.searchJob.entry.links.on('change', function() {
                this.model.summary.set("id", this.model.searchJob.entry.links.get('summary'));
            },this);

            // As the searchJob progresses, get the latest summary info
            this.model.searchJob.on("jobProgress", function() {
                this.model.summary.safeFetch();
            }, this);

            // need to debounce the listener and put it in two places in case summary the fetch gets an empty response
            var debouncedSummaryChangeHandler = _(this.summaryChangeHandler).debounce(0);
            this.model.summary.fields.on("add remove reset", debouncedSummaryChangeHandler, this);
            this.model.summary.on('sync', debouncedSummaryChangeHandler, this);

            var searchJobAttrs = {
                search: searchString || '',
                earliest_time: value.earliest || null,
                latest_time: value.latest || null,
                auto_cancel: 100,
                status_buckets: 0,
                timeFormat: '%s.%Q',
                ui_dispatch_app: this.model.application.get("app"), // TODO [JCS]
                rf: '*',
                preview: true,
                adhoc_search_level: "verbose",
                app: this.model.application.get("app"),
                owner: this.model.application.get("owner")
            };

            // if an event count limit was specified, chase the search with a 'head'
            // TODO this is pretty naive, are there edge cases where we need to be smarter?
            if(value.head) {
                searchJobAttrs.search += (' | head ' + value.head);
            }

            // Kick off the search job
            this.model.searchJob.save({}, { data: searchJobAttrs }).done(function() {
                this.model.searchJob.startPolling();
            }.bind(this));

            this.children.spinner = new Spinner({model: this.model.searchJob});
            this.children.count = new Count({model: this.model.searchJob});
        },

        /**
         * Cleanup the search job
         */
        destroySearchJob: function() {
            if (this.model.searchJob) {

                this.children.flashMessagesView.flashMsgHelper.unregister(this.model.searchJob);
                this.model.searchJob.destroy();
                delete this.model.searchJob;
            }
            if (this.model.summary) {
                this.children.flashMessagesView.flashMsgHelper.unregister(this.model.summary);
            }
        },

        /**
         * Called when the summary has changed
         */
        parseSummary: function() {
            //console.log("AddExtractedFieldsDialog.parseSummary summary",this.model.summary);

            if (!this.model.summary) {
                return;
            }

            if(this.model.searchJob.entry.content.get('isDone') && this.model.summary.fields.length === 0) {
                this.renderNoFieldsMessage();
                return;
            }

            this.model.summary.fields.each(function(field) {
                var fieldName = field.get("name");
                // Filter out any fields already in the Object model
                // Any duplication in the extractedFieldsCollection will be ignored because fieldName is its 'id' attribute
                if (!this.model.objectModel.getField(fieldName))
                {
                    // TODO [JCS] Use the Field model's new isNumeric function instead
                    var isNumeric = field.get('numeric_count') > field.get('count') / 2;

                    // Create a working model to represent the extracted field
                    var extractedField = new ExtractedField({fieldName: fieldName,
                        displayName: fieldName,
                        type: isNumeric ? "number" : "string",
                        isNumeric: isNumeric,
                        distinctValues: field.get("modes")
                    });
                    this.extractedFieldsCollection.add(extractedField);
                }
            }, this);
        },

        /**
         * Called when the save button is pressed
         */
        primaryButtonClicked: function() {
            DialogBase.prototype.primaryButtonClicked.call(this, arguments);

            // Save the results into our object. Iterate over each extracted field and add
            // the selected ones to the Object model
            this.extractedFieldsCollection.each(function(extractedField) {

                if (extractedField.get("selected")) {
                    this.model.objectModel.addField({fieldName: extractedField.get("fieldName"),
                                         displayName: extractedField.get("displayName"),
                                         type: extractedField.get("type"),
                                         hidden: extractedField.get("hidden"),
                                         required: extractedField.get("required")
                    });
                }
            }, this);

            /**
             * Save the Data Model
             *
             * @event AddExtractedFieldsDialog#action:saveModel
             */
            this.trigger("action:saveModel");
            // TODO [JCS] Need to handle validation and server errors
            this.hide();
        },

        renderBody : function($el) {
            //console.log("AddExtractedFieldsDialog.renderBody fieldTable",this.children.fieldTable);
            var html = _(this.bodyTemplate).template({validSearch: this.validSearch});
            $el.html(html);

            this.children.spinner.render().appendTo($el.find(".job-status-placeholder"));
            this.children.count.render().appendTo($el.find(".job-status-placeholder"));
            $el.find(".flash-messages-placeholder").append(this.children.flashMessagesView.render().el);
            $el.find(".select-sample-size-placeholder").append(this.children.selectSampleSizeGroup.render().el);
            $el.find(".field-table-placeholder").append(this.children.fieldTable.render().el);
        },

        renderFooter: function($el) {
            // TODO [JCS] Hide Save and Cancel when in loading state?
            DialogBase.prototype.renderFooter.call(this, $el);
        },

        renderNoFieldsMessage: function() {
            this.$(".field-table-placeholder").append(_(this.noFieldsMessageTemplate).template({}));
        },

        // Display the loading progress from the search job
/*            renderLoadingScreen: function($el) {
            var percentDone = this.searchJobModel.entry.content.get("doneProgress");
            var fieldCount = this.extractedFieldsCollection.length;

            if (_.isNaN(percentDone) || _.isUndefined(percentDone))
                percentDone = 0;
            else
                percentDone = Math.round(percentDone * 100);

            var html = _(this.loadingScreenTemplate).template({percentDone: percentDone, fieldCount: fieldCount});

            $el.html(html);
        },

        // Toggle the visibility of the loading screen and field table
        updateLoadingState: function() {

            var loadingScreen = this.$el.find(".loading-screen-placeholder");
            var fieldTable = this.$el.find(".field-table-placeholder");

            this.renderLoadingScreen(loadingScreen);

            if (this.isJobDone())
            {
                loadingScreen.hide();
                fieldTable.show();
            }
            else
            {
                loadingScreen.show();
                fieldTable.hide();
            }
        },

        isJobDone: function() {
            return this.searchJobModel.entry.content.get("isDone");
        },*/

        // Clean up the search job
        remove: function() {
            DialogBase.prototype.remove.call(this);

            this.destroySearchJob();
        },

        bodyTemplate: '\
            <div class="flash-messages-placeholder"></div>\
            <div class="select-sample-size-placeholder"></div>\
            <div class="job-status-placeholder"></div>\
            <% if (validSearch) { %>\
                <div class="field-table-placeholder"></div>\
            <% } else { %>\
                <div class="alert alert-error message-single"><i class="icon-alert"></i> \
                <%- _("You must specify a constraint on either this Object or its ancestor Objects").t() %></div>\
            <% } %>\
        ',

        noFieldsMessageTemplate: '\
            <div class="alert alert-error no-fields-message">\
                <i class="icon-alert"></i>\
                <%- _("No field summary information is available. This can mean that your search did not return any events, or that you are using a search command that does not provide summary information (i.e. inputcsv).").t() %>\
            </div>\
        '

    });

});
