/**
 * @author jszeto
 * @date 1/23/13
 *
 * Displays the filtering controls and the list of Data Models.
 *
 * Subviews:
 *
 *     views/data_model_manager/components/DataModelTable
 *
 * Inputs:
 *
 *     collection {
 *         apps {collections/services/AppLocals} collection of apps
 *         dataModels {collections/services/datamodel/DataModels}
 *         users {collections/services/authentication/Users}
 *     }
 *     model: {
 *         application {models/Application}
 *         appLocal {models/services/AppLocal}
 *         metadata {models/custom/EAIRequestParams} sorting, filtering and pagination attributes
 *         settings {Backbone.Model}
 *         user {models/services/authentication/User}
 *     }
 *
 * @fires DataModelManager#action:fetchDataModels
 */

define(
    [
        'underscore',
        'module',
        'views/Base',
        'views/data_model_manager/components/DataModelTable',
        'views/data_model_manager/DataModelActionController',
        'views/shared/CollectionPaginator',
        'views/shared/controls/SyntheticSelectControl',
        'views/shared/controls/TextControl',
        'splunk.util',
        'util/splunkd_utils',
        'uri/route'
    ],
    function(
        _,
        module,
        BaseView,
        DataModelTable,
        DataModelActionController,
        CollectionPaginator,
        SyntheticSelectControl,
        TextControl,
        splunkUtil,
        splunkDUtils,
        route
        )
    {


        return BaseView.extend({

            moduleId: module.id,
            dataModelToEdit: undefined,

            events: {
                ///////////////////////
                //  Object Actions
                ///////////////////////
                'click .new-data-model-button': function(e) {
                    e.preventDefault();
                    this.children.dataModelController.trigger("action:createDataModel");
                }
            },

            /**
             * Called when we get the action:editDataModel event. Redirects us to the Data Model Editor
             *
             * @param {models/services/datamodel/DataModel} dataModel - the data model to edit
             */
            editDataModelHandler: function(dataModel) {

                var app = dataModel.entry.acl.get("app");
                var sharing = dataModel.entry.acl.get("sharing");
                window.location = route.data_model_editor(this.model.application.get("root"),
                                                          this.model.application.get("locale"),
                                                          sharing == splunkDUtils.GLOBAL ?
                                                             this.model.application.get("app") :
                                                             app,
                                                          { data: { model: dataModel.get("id") } });
            },

            /**
             * Called when we get an action:fetchDataModel event
             *
             * @param {string} modelName - name of the model to fetch
             * @return {xhr} Returns the XHR object from the fetch operation.
             */
            fetchDataModelHandler: function(modelName) {
                var dataModel = this.collection.dataModels.get(modelName);

                if (dataModel)
                    return dataModel.fetch({data:{concise:true}});
            },

            /**
             * Applies changes directly to a DataModel instead of asking the DM to fetch itself. This allows for updates
             * to the DataModel to be rendered immediately instead of waiting for a fetch.
             *
             * @param {string} modelName - the data model name
             * @param {object} accelerationChanges - attributes to set on the Data Model's associated acceleration model
             * @param {object} contentChanges - attributes to set on the Data Model's associated content model
             */
            updateDataModelHandler: function(modelName, accelerationChanges, contentChanges) {
                var dataModel = this.collection.dataModels.get(modelName);

                if (accelerationChanges) {
                    dataModel.entry.content.acceleration.set(accelerationChanges);
                }
                if (contentChanges) {
                    dataModel.entry.content.set(contentChanges);
                }
            },

            /**
             * Handler for when the app filter control changes. If the value is set to "", which means "All Apps", then
             * we need to disable the "Created in the App" item in the visible filter control and set the metadata
             * visible property to true.
             * @param {string} value
             * @param {string} oldValue
             */
            appFilterChange: function(value, oldValue) {
                this.children.selectVisibleFilter.setItems(this.getVisibleItems(value != ""));

                if (value == "")
                    this.model.metadata.set("visible", "true");
            },

            /**
             * Helper to retrieve the items in the visible filter control.
             *
             * @param {boolean} enableCreated - if false, then set the "Created in the App" item to disabled
             * @return {array} of objects for a SyntheticSelectControl
             */
            getVisibleItems: function(enableCreated) {
                return [
                    {label: _("Created in the App").t(), value: "false", enabled: enableCreated},
                    {label: _("Visible in the App").t(), value: "true", enabled: true}
                ];
            },

            initialize: function(options) {
                BaseView.prototype.initialize.call(this, options);

                this.children.dataModelController = new DataModelActionController({
                    model: {
                        application: this.model.application,
                        user: this.model.user
                    },
                    collection: {apps: this.collection.apps}
                });

                this.children.dataModelController.on("action:clonedDataModel action:dataModelPermissionsChange",
                                                        function() {
                                                            /**
                                                             * Fetch all data models
                                                             *
                                                             * @event DataModelManager#action:fetchDataModels
                                                             */
                                                            this.trigger("action:fetchDataModels");
                                                        }, this);
                this.children.dataModelController.on("action:fetchDataModel", this.fetchDataModelHandler, this);
                this.children.dataModelController.on("action:updateDataModel", this.updateDataModelHandler, this);
                this.children.dataModelController.on("action:createdDataModel", this.editDataModelHandler, this);

                // Create the table that displays the Data Models
                this.children.dmList = new DataModelTable({
                    collection: this.collection.dataModels,
                    model: {
                        application: this.model.application,
                        metadata: this.model.metadata,
                        settings: this.model.settings,
                        user: this.model.user
                    }
                });

                // Proxy the table events so the controller can handle them
                this.children.dmList.on("all", function() {
                    var eventType = arguments[0];
                    var modelName = arguments[1];

                    // If we are deleting a DataModel, we pass in the model we are deleting
                    if (eventType == "action:editDataModel") {
                        this.editDataModelHandler(modelName);
                    } else if (eventType == "action:deleteDataModel") {
                        var dataModel = this.collection.dataModels.get(modelName);
                        this.children.dataModelController.trigger("action:deleteDataModel", dataModel);
                    } else {
                        this.children.dataModelController.trigger.apply(this.children.dataModelController, arguments);
                    }
                }, this);

                // App filter
                var staticAppItems = [{label:_('All Apps').t(), value: ''}];
                var dynamicAppItems = this.collection.apps.map(function(model, key, list){
                    return {
                        label: splunkUtil.sprintf(_('App: %s (%s)').t(), model.entry.content.get('label'), model.entry.get("name")),
                        value: model.entry.get('name')
                    };
                });
                var appItems = [{'items':staticAppItems},{'items':dynamicAppItems}];

                this.children.selectAppFilter = new SyntheticSelectControl({toggleClassName: 'btn-pill',
                                                                      menuWidth: 'wide',
                                                                      model: this.model.metadata,
                                                                      modelAttribute: 'app',
                                                                      groupedItems: appItems});

                this.children.selectAppFilter.on("change", this.appFilterChange, this);

                this.children.selectVisibleFilter = new SyntheticSelectControl({toggleClassName: 'btn-pill',
                                                                      menuWidth: 'narrow',
                                                                      model: this.model.metadata,
                                                                      modelAttribute: 'visible',
                                                                      items: this.getVisibleItems(this.model.metadata.get("app") != "")});

                // User filter
                var staticUserItems = [ {label:_('Owner: Any').t(), value:'*'},
                                        {label:_('Owner: None (nobody)').t(), value: 'nobody'}];
                var dynamicUserItems = this.collection.users.map(function(model, key, list){
                    return {
                        label: splunkUtil.sprintf(_('Owner: %s (%s)').t(),
                                                    model.entry.content.get('realname'),
                                                    model.entry.get('name')),
                        value: model.entry.get('name')
                    };
                });
                var userItems = [{'items':staticUserItems}, {'items':dynamicUserItems}];

                this.children.selectOwnerFilter = new SyntheticSelectControl({toggleClassName: 'btn-pill',
                    menuWidth: 'narrow',
                    model: this.model.metadata,
                    modelAttribute: 'owner',
                    groupedItems: userItems});

                this.children.textNameFilter = new TextControl({model: this.model.metadata,
                                                                modelAttribute: "nameFilter",
                                                                inputClassName: 'search-query',
                                                                placeholder: _("filter").t()});

                this.children.collectionPaginator = new CollectionPaginator({
                    collection: this.collection.dataModels,
                    model: this.model.metadata
                });
            },

            render: function() {

                // Detach so jQuery doesn't remove their event listeners when we reassign the DOM
                this.children.selectAppFilter.detach();
                this.children.selectVisibleFilter.detach();
                this.children.selectOwnerFilter.detach();
                this.children.textNameFilter.detach();
                this.children.collectionPaginator.detach();

                var docUrl = route.docHelp(
                    this.model.application.get("root"),
                    this.model.application.get("locale"),
                    'learnmore.datamodel.about'
                );

                var html = _(this.template).template({docUrl: docUrl, displayNewDataModel: this.model.settings.get("canCreateDataModel") });
                this.$el.html(html);
                this.$('.dm-list-placeholder').append(this.children.dmList.render().el);
                // TODO [JCS] Disable these for now until we've gotten the app and owner from the REST endpoint
                this.$('.select-app-filter-placeholder').append(this.children.selectAppFilter.render().el);
                this.$('.select-visible-filter-placeholder').append(this.children.selectVisibleFilter.render().el);
                this.$('.select-owner-filter-placeholder').append(this.children.selectOwnerFilter.render().el);
                this.$('.text-name-filter-placeholder').append(this.children.textNameFilter.render().el);
                this.$('.collection-paginator-placeholder').append(this.children.collectionPaginator.render().el);
                return this;
            },

            template: '\
                <div class="section-header section-padded">\
                    <% if (displayNewDataModel) { %>\
                        <div class="pull-right">\
                            <a href="#" class="btn btn-primary new-data-model-button"><%- _("New Data Model").t() %></a>\
                        </div>\
                    <% } %>\
                    <h2 class="section-title"><%- _("Data Models").t() %></h2>\
                    <%- _("Data models enable users to easily create reports in the Pivot tool.").t() %> \
                    <a href="<%- docUrl %>" target="_blank" class="help-link"><%- _("Learn More").t() %> <i class="icon-external"></i></a>\
                </div>\
                <div class="btn-toolbar pull-left">\
                    <span class="select-app-filter-placeholder"></span>\
                    <span class="select-visible-filter-placeholder"></span>\
                    <span class="select-owner-filter-placeholder"></span>\
                    <span class="text-name-filter-placeholder"></span>\
                </div>\
                <div class="pull-right collection-paginator-placeholder"></div>\
                <div class="dm-list-placeholder"></div>\
        '

        });

    });
