/**
 * @author sfishel
 *
 * A grid view that displays a list of data models.
 *
 * Adds expand/collapse behavior to show more details about a single data model.
 *
 * Custom Events:
 *
 * action:selectModel - triggered when a data model is selected from the list
 *     @param modelName {String} the name of the selected data model
 */

define([
            'jquery',
            'underscore',
            'module',
            'views/Base',
            'views/shared/delegates/RowExpandCollapse',
            'helpers/grid/RowIterator',
            'uri/route',
            'splunk.util',
            'splunk.i18n'
        ],
        function(
            $,
            _,
            module,
            Base,
            RowExpandCollapse,
            RowIterator,
            route,
            splunkUtils,
            i18n
        ) {

    return Base.extend({

        moduleId: module.id,

        /**
         * @constructor
         * @param options {Object} {
         *     collection: <collections/services/datamodel/DataModels> the collection of data models to list
         *     model: {
         *         application: <models/Application> the application state model
         *     }
         * }
         */

        initialize: function(options) {
            Base.prototype.initialize.call(this, options);
            this.children.rowExpandCollapse = new RowExpandCollapse({
                el: this.el,
                autoUpdate: false
            });
            this.expandedIds = [];
            this.collectionHasSynced = false;

            this.children.rowExpandCollapse.on(RowExpandCollapse.ROW_EXPAND + ' ' + RowExpandCollapse.ROW_COLLAPSE, function(id, expandedIds) {
                this.expandedIds = expandedIds;
                this.debouncedRender();
            }, this);
            this.collection.on('sync', function() {
                this.collectionHasSynced = true;
                this.debouncedRender();
            }, this);
        },

        render: function() {
            var headerMessage,
                filteredCollection = this.collection.filter(function(dataModel) { return !dataModel.isEmpty(); });

            if(this.collectionHasSynced) {
                headerMessage = splunkUtils.sprintf(
                    i18n.ungettext('1 Data Model', '%s Data Models', filteredCollection.length),
                    filteredCollection.length
                );
            }
            else {
                headerMessage = _('Loading Data Models...').t();
            }
            var that = this,
                rowIterator = new RowIterator({
                    expandedTest: function(model) { return _(that.expandedIds).indexOf(model.id) > -1; }
                }),
                html = this.compiledTemplate({
                    headerMessage: headerMessage,
                    dataModels: filteredCollection,
                    eachRow: rowIterator.eachRow,
                    headerCellClass: RowExpandCollapse.HEADER_CELL_CLASS,
                    toggleCellClass: RowExpandCollapse.TOGGLE_CELL_CLASS,
                    rowIdAttribute: RowExpandCollapse.ROW_ID_ATTR,
                    pageRouter: route.getContextualPageRouter(this.model.application),
                    collectionHasSynced: this.collectionHasSynced
                }),
                $html = $(html);

            this.children.rowExpandCollapse.update($html);
            this.$el.html($html);
        },

        template: '\
            <table class="table table-chrome table-striped table-bordered-lite table-row-expanding">\
                <thead>\
                    <tr>\
                        <th class="<%- headerCellClass %>"></th>\
                        <th>\
                            <%- headerMessage %>\
                        </th>\
                    </tr>\
                </thead>\
                <tbody>\
                    <% eachRow(dataModels, function(model, index, rowNumber, isExpanded) { %>\
                    \
                        <tr class="list-item <%- index%2 == 0 ? "even" : "odd" %>" <%- rowIdAttribute %>="<%- model.id %>">\
                            <td class="<%- toggleCellClass %>"></td>\
                            <td class="title col-description">\
                                <% var explorerHref = pageRouter.data_model_explorer({ data: { model: model.id } }); %>\
                                <a href="<%- explorerHref %>" class="select-model-button"><%- model.entry.content.get("displayName") %></a>\
                                <% if(isExpanded) { %>\
                                    <% var editorHref = pageRouter.data_model_editor({ data: { model: model.id } }); %>\
                                    <div class="expanded-info">\
                                        <p class="dm-description"><%- model.entry.content.get("description") || _("No description available").t() %></p>\
                                        <p>\
                                            <% if (model.canWrite()) { %>\
                                                <a href="<%- editorHref %>" class="edit-objects-button">\
                                                    <%- _("Edit Objects").t() %>\
                                                </a>\
                                            <% } %>\
                                        </p>\
                                    </div>\
                                <% } %>\
                            </td>\
                        </tr>\
                    \
                    <% }) %>\
                </tbody>\
            </table>\
            <% if(collectionHasSynced && dataModels.length === 0) { %>\
                <div class="no-models-message">\
                    <div class="alert alert-info">\
                        <i class="icon-alert"></i>\
                        <%- _("You have no Data Models available.").t() %>\
                        <a class="manage-models-link" href="<%- pageRouter.data_model_manager() %>" >\
                            <%- _("Manage Data Models").t() %>\
                        </a>\
                    </div>\
                </div>\
            <% } %>\
        '

    });

});