define([
            'jquery',
            'underscore',
            'module',
            'models/Job',
            'models/pivot/PivotJob',
            'models/services/summarization/TStatsSummarization',
            'views/Base',
            'views/pivot/DataSourceMenu',
            'views/shared/DropDownMenu',
            'views/shared/delegates/Popdown',
            'helpers/polling_manager'
        ],
        function(
            $,
            _,
            module,
            Job,
            PivotJob,
            TStatsSummarization,
            Base,
            DataSourceMenu,
            DropDownMenu,
            Popdown,
            pollingManager
        ) {

    var POLLING_OPTIONS = { delay: 2000, condition: function() { return true; } };

    return Base.extend({

        moduleId: module.id,

        events: {
            'click .clear-button': function(e) {
                e.preventDefault();
                this.model.report.trigger('clear');
            },
            'click .save-button': function(e) {
                e.preventDefault();
                this.model.report.trigger('save');
            }
        },

        /**
         * @constructor
         * @param options {
         *     report <models/pivot/PivotReport> the current report
         *     dataModel <models/services/datamodel/DataModel> the current data model
         *     application <models/Application> the current application state
         *     searchJob <models/pivot/PivotJob> the current pivot job
         * }
         */

        initialize: function() {
            Base.prototype.initialize.apply(this, arguments);

            this.children.createDropDown = new DropDownMenu({
                label: _('Save As...').t(),
                className: 'btn-combo create-drop-down',
                dropdownClassName: 'dropdown-menu-narrow',
                popdownOptions: { attachDialogTo: 'body' },
                items: [
                    { label: _('Report').t(), value: 'report' },
                    { label: _('Dashboard Panel').t(), value: 'panel' }
                ]
            });
            this.children.createDropDown.on('itemClicked', function(type) {
                if(type === 'report') {
                    this.model.report.trigger('saveAsReport');
                }
                else if(type === 'panel') {
                    this.model.report.trigger('saveAsDashboardPanel');
                }
            }, this);

            // determine what type of acceleration (if any) the job is using,
            // and create the corresponding model to be passed down to the data source menu
            var accelerationType = this.model.searchJob.getAccelerationType();
            if(accelerationType === PivotJob.ADHOC_ACCELERATION) {
                this.model.collectJob = new Job({ id: this.model.searchJob.getCollectId() }, POLLING_OPTIONS);
            }
            else if(accelerationType === PivotJob.MANAGED_ACCELERATION) {
                // TODO [sff] just a hack until we have a way to pass only the data model name
                var tstatsNamespace = "tstats:DM_" + this.model.dataModel.entry.acl.get('app')
                                    + "_" + this.model.dataModel.entry.content.get("modelName"),
                    tStatsId = this.model.dataModel.id.replace("datamodel/model", "admin/summarization")
                                    .replace(this.model.dataModel.entry.content.get("modelName"), tstatsNamespace);

                this.model.tstatsSummarization = new TStatsSummarization({ id: tStatsId });
                this.summarizationPoller = pollingManager.getPoller(
                    this.model.tstatsSummarization,
                    $.extend(POLLING_OPTIONS, {
                        condition: function(model) {
                            // no need to keep polling if the summary is complete
                           return model.entry.content.get('summary.complete') < 1;
                        }
                    })
                );
            }

            this.children.dataSourceMenu = new DataSourceMenu({
                model: {
                    report: this.model.report,
                    dataModel: this.model.dataModel,
                    application: this.model.application,
                    searchJob: this.model.searchJob,
                    collectJob: this.model.collectJob,
                    tstatsSummarization: this.model.tstatsSummarization
                }
            });

            this.children.popdown = new Popdown({
                el: this.el,
                toggle: '.data-source-button',
                dialog: '.data-source-menu',
                attachDialogTo: 'body'
            });

            this.children.popdown.on('show', this.onDataSourceShow, this);
            this.children.popdown.on('shown', this.onDataSourceShown, this);
            this.children.popdown.on('hidden', this.onDataSourceHidden, this);

            this.model.report.entry.on('change:name', this.updateReportName, this);
            this.model.report.on('change:id', function() {
                if(!this.model.report.isNew()) {
                    this.$('.save-button').show();
                }
                else {
                    this.$('.save-button').hide();
                }
            }, this);
        },

        render: function() {
            var object = this.model.dataModel.objectByName(this.model.report.getPivotObjectName());
            this.$el.html(this.compiledTemplate({
                objectName: object.get('displayName')
            }));
            if(this.model.report.isNew()) {
                this.$('.save-button').hide();
            }
            this.updateReportName();
            this.children.createDropDown.render().replaceAll(this.$('.save-as-dropdown-placeholder'));
            this.children.dataSourceMenu.render().replaceAll(this.$('.data-source-menu-placeholder'));
            return this;
        },

        remove: function() {
            this.onDataSourceHidden();
            return Base.prototype.remove.apply(this, arguments);
        },

        updateReportName: function() {
            var $nameHolder = this.$('.report-name'),
                isNew = this.model.report.isNew(),
                reportName = isNew ? _('New Pivot').t() : this.model.report.entry.get('name');

            $nameHolder.empty().text(reportName);
            if(isNew) {
                $nameHolder.prepend('<i class="icon-pivot"></i>');
            }
            else {
                $nameHolder.prepend('<i class="icon-report-pivot"></i>');
            }
        },

        onDataSourceShow: function() {
            if(this.model.collectJob) {
                this.model.collectJob.startPolling();
            }
            else if(this.model.tstatsSummarization) {
                this.summarizationPoller.start();
            }
        },

        onDataSourceShown: function() {
            // kind of hacky, but for 508 reach in and focus the first a tag
            this.children.dataSourceMenu.$('a').first().focus();
        },

        onDataSourceHidden: function() {
            if(this.model.collectJob) {
                this.model.collectJob.stopPolling();
            }
            else if(this.model.tstatsSummarization) {
                this.summarizationPoller.stop();
            }
        },

        template: '\
            <h2 class="report-name section-title"></h2>\
            <div class="action-bar pull-right">\
                <a href="#" class="save-button btn btn-primary">\
                    <%- _("Save").t() %>\
                </a>\
                <div class="report-action-buttons btn-group">\
                    <span class="save-as-dropdown-placeholder"></span>\
                    <a href="#" class="clear-button btn"><%- _("Clear").t() %></a>\
                </div>\
                <a href="#" class="data-source-button btn">\
                    <%- objectName %>\
                    <span class="caret"></span>\
                </a>\
                <div class="data-source-menu-placeholder"></div>\
            </div>\
        '

    });

});
