define([
            'jquery',
            'underscore',
            'module',
            'models/config',
            'js_charting/js_charting',
            'views/Base',
            'util/jscharting_utils',
            'splunk.util',
            'helpers/Printer'
        ],
        function(
            $,
            _,
            module,
            configModel,
            JSCharting,
            Base,
            jschartingUtils,
            splunkUtils,
            Printer
        ) {

    return Base.extend({
        VIZ_PROPERTY_PREFIX_REGEX: /^display\.visualizations\.charting\./,
        className: 'chart',
        moduleId: module.id,
        initialize: function(options) {
            Base.prototype.initialize.call(this, arguments);
            this.options = options || {};
            if(this.options.width) {
                this.$el.width(this.options.width);
            }
            if(this.options.height) {
                this.$el.height(this.options.height);
            }

            this.debouncedDrawChart = _.debounce(_(this.drawChart).bind(this), 25);
            this.model.state.on('change',function(){
                if(this.model.state.get('display.visualizations.type') === 'charting') {
                    var changedVizProps = this.model.state.filterChangedByWildcards(this.VIZ_PROPERTY_PREFIX_REGEX, {allowEmpty: true});
                    if(!_(changedVizProps).isEmpty()) {
                        this.debouncedDrawChart();
                    }
                }
            }, this);

            this.$chart = $('<div></div>');
            this.$inlineMessage = $('<div></div>').css('text-align', 'center');

            this.model.searchResultsColumn.on('change:fields change:columns', this.debouncedDrawChart, this);
            this.model.searchResultsColumn.on('destroy', this.empty, this);
            Printer.on(Printer.PRINT_START, this.reflow, this);
            Printer.on(Printer.PRINT_END, this.reflow, this);
        },
        empty:  function() {
            this.destroyChart();
            this.$chart.empty();
            this.$inlineMessage.empty();
            return this;
        },
        remove: function() {
            Printer.off(null, null, this);
            this.destroyChart();
            return Base.prototype.remove.apply(this, arguments);
        },
        // TODO [sff] this needs more thought, we want render to be snappy since it is often called in response to user action
        // but the actual draw can take a long time
        render: function() {
            this.$chart.appendTo(this.el);
            this.$inlineMessage.appendTo(this.el);
            this.debouncedDrawChart();
            return this;
        },
        setHeight: function(height) {
            this.$el.height(height);
        },
        drawChart: function() {
            if (!this.awake) {
                // it's possible we were put to sleep between when render was called and when drawChart was called
                this.touch = true;
                return;
            }
            this.$inlineMessage.empty();

            var rawResults = {
                fields: this.model.searchResultsColumn.get('fields') || [],
                columns: this.model.searchResultsColumn.get('columns') || []
            };
            var displayProperties = {};
            _.each(this.model.state.toJSON(), function(value, key){
                if(this.VIZ_PROPERTY_PREFIX_REGEX.test(key)) {
                    displayProperties[key.replace(this.VIZ_PROPERTY_PREFIX_REGEX, '')] = value;
                }
            }, this);

            var results = jschartingUtils.preprocessChartData(rawResults, displayProperties),
                maxResultCount = _.isFunction(this.model.searchResultsColumn.getFetchCount) ?
                                        this.model.searchResultsColumn.getFetchCount() : Infinity;

            // if the preprocessChartData method truncated the data, show a message to that effect
            if(results.columns.length > 0 &&
                    (results.columns.length < rawResults.columns.length ||
                        results.columns[0].length < rawResults.columns[0].length)) {
                this.renderResultsTruncatedMessage();
            }
            // otherwise if the number of results matches the max result count that was used to fetch,
            // show a message that we might not be displaying the full data set
            else if(results.columns.length > 0 && maxResultCount > 0 && results.columns[0].length >= maxResultCount) {
                this.renderMaxResultCountMessage(maxResultCount);
            }
            var chartReadyData = JSCharting.extractChartReadyData(results);
            $.extend(displayProperties, jschartingUtils.getCustomDisplayProperties(chartReadyData, configModel.toJSON()));

            if(!this.chart || !_.isEqual(displayProperties, this.chart.getCurrentDisplayProperties())) {
                this.destroyChart();
                this.chart = JSCharting.createChart(this.$chart, displayProperties);
            }
            this.updateChartContainerHeight();
            this.chart.prepareAndDraw(chartReadyData);
            var that = this;
            this.chart.on('pointClick', function(eventInfo) {
                that.trigger('chartClick', eventInfo);
            });
            this.chart.on('legendClick', function(eventInfo) {
                that.trigger('chartClick', eventInfo);
            });
        },
        destroyChart: function() {
            if(this.chart) {
                this.chart.off();
                this.chart.destroy();
                delete this.chart; // GC handler
            }
        },
        reflow: function() {
            if(this.chart && this.$el.height() > 0) {
                this.updateChartContainerHeight();
                this.chart.resize();
            }
            Base.prototype.reflow.call(this);
        },
        updateChartContainerHeight: function() {
            this.$chart.height(this.$el.height() - this.$inlineMessage.outerHeight());
        },
        renderResultsTruncatedMessage: function() {
            var message = _('These results may be truncated. Your search generated too much data for the current visualization configuration.').t();
            this.$inlineMessage.html(_(this.inlineMessageTemplate).template({ message: message, level: 'warning' }));
        },
        renderMaxResultCountMessage: function(resultCount) {
            var message = splunkUtils.sprintf(
                _('These results may be truncated. This visualization is configured to display a maximum of %s results per series, and that limit has been reached.').t(),
                resultCount
            );
            this.$inlineMessage.html(_(this.inlineMessageTemplate).template({ message: message, level: 'warning' }));
        },
        inlineMessageTemplate: '\
            <div class="alert alert-inline alert-<%= level %> alert-inline"> \
                <i class="icon-alert"></i> \
                <%= message %> \
            </div> \
        '
    });
 });
