/**
 * A delegate view to handle a statically positioned table header.
 */

define(['jquery', 'underscore', 'views/shared/delegates/DetachedTableHeader'], function($, _, DetachedTableHeader) {

    return DetachedTableHeader.extend({
        awake: true,
        touch: false,
        initialize: function(options) {
            var defaults = {
                    headerContainer: "> .header-table-static",
                    scrollContainer: "> .scroll-table-wrapper",
                    table: "> .scroll-table-wrapper > table",
                    offset: 0,
                    // index of the column to "flex" to fit remaining width, can be set to false for no flex
                    flexWidthColumn: -1,
                    defaultLayout:'auto'
                };
    
            _.defaults(this.options, defaults);
            
        
            DetachedTableHeader.prototype.initialize.call(this, options);
            
            this.awake && this.bindListeners();
        },
        bindListeners: function () {
            var debouncedResizeHandler = _.debounce(this.handleWindowResize, 50);
            $(window).on('resize.' + this.cid, _(debouncedResizeHandler).bind(this));
            $(this.$(this.options.scrollContainer)).on('scroll.' + this.cid, _(this.handleContainerScroll).bind(this));
        },
        unbindListeners: function () {
            $(window).off('resize.' + this.cid);
            $(this.$(this.options.scrollContainer)).off('scroll.' + this.cid);
        },
        destroy: function() {
            this.unbindListeners();
        },
        remove: function() {
            DetachedTableHeader.prototype.remove.apply(this);
            $(window).off('resize.' + this.cid);
            return this;
        },
        update: function() {
            if (!this.awake) {
                this.touch = true;
                return;            
            }
            
            this.$scrollContainer = this.$(this.options.scrollContainer);
            this.$table = this.$(this.options.table).first();
            this.$headerContainer = this.$(this.options.headerContainer);
            this.$headerContainer.empty();
            var $headerWrapper = $('<div class="header-table-wrapper"></div>').appendTo(this.$headerContainer);
            $headerWrapper.css({ 'margin-right': this.$scrollContainer[0].offsetWidth - this.$scrollContainer[0].clientWidth});
            this.$headerTable = $('<table>');
            this.$headerTable.attr('class', this.$table.attr('class')).width(this.$table.outerWidth()).css('table-layout', 'fixed');
            // SPL-77034, for in-place updates make sure we don't clone the visibility: hidden; that is added later
            this.$table.find('> thead').clone().css('visibility', 'visible').appendTo(this.$headerTable);
            $headerWrapper.prepend(this.$headerTable);

            // 508: make sure the browser doesn't tab to the original table header
            this.$table.find('thead').css('visibility', 'hidden');
            
            this.$headerTable.on('click', 'th', function (e) {
                e.preventDefault();
                e.stopPropagation();
                var colIndex = $(e.currentTarget).prevAll().length + 1;
                this.$table.find('> thead > tr > th:nth-child(' + colIndex + ')').click();
            }.bind(this));
            
            this.$headerTable.on('click', 'th a', function (e) {
                e.preventDefault();
            });

            // the newly rendered table head will be scrolled all the way to the left
            // reset this instance variable so that handleContainerScroll will do the right thing
            this.scrollLeft = 0;
            var updateDom = _(function() {
                this.syncColumnWidths();
                this.handleContainerScroll();
                this.trigger('updated', this.$headerTable);
            }).bind(this);

            if(this.$table.is(':visible')) {
                updateDom();
            } else {
                _.defer(updateDom);
            }
        },

        // ------ private methods ------ //
        handleContainerScroll: function() {
            // no-op if update has not been called yet
            if(!this.$table) {
                return;
            }
            var scrollLeft = this.$scrollContainer.scrollLeft();
            if(scrollLeft !== this.scrollLeft) {
                this.$headerTable.css('marginLeft', -scrollLeft + 'px');
                this.scrollLeft = scrollLeft;
            }
        },
        handleWindowResize: function() {
            // no-op if update has not been called yet
            if(!this.$table) {
                return;
            }
            this.syncColumnWidths();
        },
        wake: function() {
            this.awake = true;
            this.bindListeners();
            if (this.touch) {
                //do a full update
                _.defer(this.update.bind(this));
            } else {
                //just update the positions.
                _.defer(this.handleWindowResize.bind(this));
                _.defer(this.handleContainerScroll.bind(this));
            }
            return this;
        },
        sleep: function() {
            this.awake = false;
            this.unbindListeners();
            return this;
        }

    });

});