/**
 * @author sfishel
 *
 * Table renderer for search results
 *
 * Child Views:
 *
 * header <views/shared/results_table/ResultsTableHeader> the table header
 * rows <views/shared/results_table/ResultsTableRow> one for each row in the table
 *
 * Custom Events:
 *
 * cellClick - triggered when a cell is clicked for a drilldown action
 *     @param drilldownInfo {Object} {
 *         name {String} the name of the row-split field
 *         value {String} the value of the row-split field for the clicked row
 *         name2 {String} (only applicable if the user did not click on the first cell of the row)
 *                        the name of the column header for the clicked cell
 *         value2 {String} (only applicable if the user did not click on the first cell of the row)
 *                         the value contained in the clicked cell
 *         _span {Integer} (only applicable if the row-split field is time-based)
 *                         the number of seconds that the clicked row represents in time
 *         modifierKey {Boolean} whether the Ctrl/Cmd key was pressed when the cell was clicked
 *         rowContext {Object} a dictionary with key-value pairs for all of the cells in the clicked row
 *     }
 *
 *     In the case of a multivalue field, the value that was clicked will appear as 'value2',
 *     but the rowContext will include the entire array of values.
 */

define([
            'jquery',
            'underscore',
            'module',
            'views/Base',
            './ResultsTableHeader',
            './ResultsTableRow',
            'views/shared/delegates/ColumnSort',
            'views/shared/delegates/TableDock',
            'views/shared/delegates/TableHeadStatic',
            'helpers/grid/RowIterator',
            'helpers/Printer',
            'splunk.util',
            'util/general_utils',
            'util/time_utils',
            'util/math_utils',
            'jquery.sparkline',
            './renderers/BaseCellRenderer',
            './renderers/NullCellRenderer',
            './renderers/NumberCellRenderer',
            './renderers/SparklineCellRenderer',
            './renderers/StringCellRenderer',
            './renderers/TimeCellRenderer',
            'util/console'
        ],
        function(
            $,
            _,
            module,
            Base,
            ResultsTableHeader,
            ResultsTableRow,
            ColumnSort,
            TableDock,
            TableHeadStatic,
            RowIterator,
            Printer,
            splunkUtils,
            generalUtils,
            timeUtils,
            mathUtils,
            _sparkline,
            BaseCellRenderer,
            NullCellRenderer,
            NumberCellRenderer,
            SparklineCellRenderer,
            StringCellRenderer,
            TimeCellRenderer,
            console
        ) {

    var DATA_FIELD_REGEX = /^[^_]|^_time$|^_raw$/;

    var TIMESTAMP_FORMATS = {
        year: 'YYYY',
        month: 'YYYY-MM',
        day: 'YYYY-MM-dd',
        hour: 'YYYY-MM-dd HH:00',
        minute: 'YYYY-MM-dd HH:mm:00',
        second: 'YYYY-MM-dd HH:mm:ss'
    };

    var REPORT_FORMAT_ATTRIBUTES = [
        'display.statistics.wrap',
        'display.statistics.rowNumbers',
        'display.statistics.overlay',
        'display.statistics.sparkline.format'
    ];

    var MV_SUBCELL_SELECTOR = '.' + ResultsTableRow.MV_SUBCELL_CLASSNAME;

    var ENTER_KEY = 13;

    return Base.extend({

        moduleId: module.id,

        events: {
            'mouseover tbody td': function(e) {
                this.handleCellMouseover($(e.target));
            },
            'mouseout tbody td': function(e) {
                this.handleCellMouseout($(e.target));
            },
            'click tbody td': function(e) {
                e.preventDefault();
                this.handleCellClick($(e.target), e);
            },
            'keypress tbody td': function(e) {
                if(e.which === ENTER_KEY) {
                    e.preventDefault();
                    this.handleCellClick($(e.target), e);
                }
            },
            'keypress tbody tr': function(e) {
                if(e.which === ENTER_KEY) {
                    e.preventDefault();
                    // treat the target as the first cell in the row
                    this.handleCellClick($(e.target).find('td').first(), e);
                }
            }
        },

        /**
         * @constructor
         * @param options {Object} {
         *    model: {
         *        results <models/services/search/jobs/ResultsRow> the search results in row-major format
         *        metadata {Model} a model containing the fetch data for the results
         *        format <model> the table format settings
         *    },
         *    {String} enableTableDock: (Optional) whether or not to dock table header
         *    {Number} tableDockOffset: (Optional) the amount to offset the table header from the top when docked
         *    {String} sortKeyAttr: (Optional) attribute to set sort key on.  default is 'sortKey',
         *    {String} sortDirAttr: (Optional) attribute to set sort direction on. default is 'sortDirection',
         *    {String} countAttr: (Optional) attribute to set count on. default is 'count',
         *    {String} offsetAttr: (Optional) attribute to set offset on. default is 'offset'
         * }
         */

        initialize: function(options) {
            Base.prototype.initialize.call(this, options);

            if(this.options.enableStaticHeader && this.options.enableTableDock !== false) {
                throw new Error('Static header and docking header cannot be enabled at the same time');
            }

            var defaults = {
                sortKeyAttr: 'sortKey',
                sortDirAttr: 'sortDirection',
                countAttr: 'count',
                offsetAttr: 'offset',
                numRowSplits: 1
            };
            this.options = $.extend(true, defaults, this.options);

            this.rawFields = [];
            this.rawRows = [];
            this.dataFields = [];
            this.dataRows = [];

            this.columnTypes = {};
            this.fieldHeatDeltas = {};
            this.fieldExtremes = {};
            this.fieldTimestampFormats = {};
            this.sparklineFormatSettings = {};

            this._cellRenderers = [];

            // delegates
            this.children.columnSort = new ColumnSort({
                el: this.el,
                model: this.model.metadata,
                autoUpdate: true,
                sortKeyAttr: this.options.sortKeyAttr,
                sortDirAttr: this.options.sortDirAttr
            });
            if(this.options.enableTableDock !== false) {
                this.children.tableDock = new TableDock({ el: this.el, offset: this.options.tableDockOffset });
            }

            this.model.results.on('change', _.debounce(this.handleResultsChange, 0), this);
            this.model.format.on('change', this.handleFormatChange, this);
            this.updateFormatSettings();
            this.readResults();

            if (this.options.disable) {
                this.$el.addClass("table-disabled");
            }
            if(this.options.enableStaticHeader) {
                Printer.on(Printer.PRINT_END, function() {
                    this.children.staticHead.update();
                }, this);
            }

            this.addCellRenderer(new StringCellRenderer());
            this.addCellRenderer(new TimeCellRenderer());
            this.addCellRenderer(new NumberCellRenderer());
            this.addCellRenderer(new SparklineCellRenderer());
            this.addCellRenderer(new NullCellRenderer());
        },

        addCellRenderer: function(cellRenderer) {
            if ((typeof cellRenderer !== "function") && !(cellRenderer instanceof BaseCellRenderer)) {
                throw new Error("Parameter cellRenderer must be a function or an instance of views/shared/results_table/renderers/BaseCellRenderer.");
            }

            this._cellRenderers.unshift(cellRenderer);
        },

        removeCellRenderer: function(cellRenderer) {
            if ((typeof cellRenderer !== "function") && !(cellRenderer instanceof BaseCellRenderer)) {
                throw new Error("Parameter cellRenderer must be a function or an instance of views/shared/results_table/renderers/BaseCellRenderer.");
            }

            var cellRenderers = this._cellRenderers;
            for (var i = 0, l = cellRenderers.length; i < l; i++) {
                if (cellRenderers[i] === cellRenderer) {
                    cellRenderers.splice(i, 1);
                    break;
                }
            }
        },

        getCellRenderers: function() {
            return this._cellRenderers.concat();
        },

        remove: function() {
            Printer.off(null, null, this);
            return Base.prototype.remove.apply(this, arguments);
        },

        handleResultsChange: function() {
            this.readResults();
            // no need to debounce the render, this method is already debounced
            this.render();
        },

        readResults: function() {
            this.rawFields = this.getFieldList();
            this.rawRows = this.model.results.get('rows') || [];
            this.dataFields = _(this.rawFields).filter(this.isDataField, this);
            this.dataRows = _(this.rawRows).map(function(row) {
                return _(row).filter(function(data, i) { return this.isDataField(this.rawFields[i]); }, this);
            }, this);
        },

        handleFormatChange: function() {
            var changedFormatAttributes = this.model.format.changedAttributes();
            // only do work if an attribute changed that actually requires a re-render
            if(_(changedFormatAttributes).chain().pick(REPORT_FORMAT_ATTRIBUTES).isEmpty().value()) {
                // changes to drilldown do not require a re-render, but do have to be handled if a re-render is not happening
                if(changedFormatAttributes.hasOwnProperty('display.statistics.drilldown')) {
                    this.updateDrilldownClasses();
                }
                return;
            }
            this.updateFormatSettings();
            this.debouncedRender();
        },

        render: function() {
            if(this.options.enableStaticHeader && !this.el.innerHTML) {
                this.$el.html(this.staticHeaderTemplate);
                this.children.staticHead = new TableHeadStatic({
                    el: this.el,
                    flexWidthColumn: false
                });
            }
            var html = this.compiledTemplate({
                    wrapResults: splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.wrap')),
                    numDataColumns: this.dataFields.length
                }),
                $html = $(html);

            this.disposeOfChildren();
            if(this.dataFields.length > 0 && this.dataRows.length > 0) {
                console.debug('rendering results table with data:', { fields: this.dataFields, rows: this.dataRows });
                var columns = this.resultsAsColumns();
                this.updateColumnTypes(columns);
                this.insertHeader($html, this.dataFields);
                this.insertRows($html, this.dataFields, this.dataRows, columns);
                this.children.columnSort.update($html);
            }
            if(this.options.enableStaticHeader) {
                this.$('.scroll-table-wrapper').html($html);
                this.children.staticHead.update();
            }
            else {
                this.$el.html($html);
            }
            // render the sparklines that have been created with hidden elements or elements not attached to DOM
            $.sparkline_display_visible();
            if(this.children.tableDock) {
                this.children.tableDock.update();
            }
            
            if (this.options.disable) {
                this.$tableDisabled = $('<div class="table-disabled-screen"></div>');
                this.$el.append(this.$tableDisabled);
            }
            this.updateDrilldownClasses();

            // SPL-69820 trigger event when content has been updated
            this.trigger('rendered');

            this.reflow();
            return this;
        },
        reflow: function() {
            Base.prototype.reflow.apply(this, arguments);
            if(this.$tableDisabled) {
                this.$tableDisabled.width(this.$el[0].scrollWidth);
            }
        },
        handleCellMouseover: function($target) {
            var drilldown = this.model.format.get('display.statistics.drilldown'),
                rowNumbers = splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers'));

            if(drilldown === 'none') {
                return;
            }
            var $cell = $target.closest('td'),
                cellInfo = this.getCellInfo($cell),
                rowView = this.children['row_' + cellInfo.rowIndex];

            // If we didn't get a row view, then we should just stop
            if (!rowView) {
                return;
            }

            if(drilldown === 'cell') {
                // if the cell is multivalue but the mouse is not over one of the subcells, ignore the event
                if($cell.find(MV_SUBCELL_SELECTOR).length > 0 && !$target.is(MV_SUBCELL_SELECTOR)) {
                    return;
                }
                this.children.header.highlightColumn(cellInfo.cellIndex);
                // FIXME: this is pretty lame
                if(this.children.tableDock) {
                    this.children.tableDock.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + (rowNumbers ? 1 : 0)).addClass('highlighted');
                }
                if(this.children.staticHead) {
                    this.children.staticHead.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + (rowNumbers ? 1 : 0)).addClass('highlighted');
                }
                rowView.highlightCell($target, cellInfo.cellIndex);
            }
            else {
                // drilldown === 'row'
                rowView.highlightRow();
            }
        },

        handleCellMouseout: function($target) {
            var drilldown = this.model.format.get('display.statistics.drilldown'),
                rowNumbers = splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers'));

            if(drilldown === 'none') {
                return;
            }
            var $cell = $target.closest('td'),
                cellInfo = this.getCellInfo($cell),
                rowView = this.children['row_' + cellInfo.rowIndex];

            // If we didn't get a row view, then we should just stop
            if (!rowView) {
                return;
            }

            if(drilldown === 'cell') {
                // if the cell is multivalue but the mouse did not leave one of the subcells, ignore the event
                if($cell.find(MV_SUBCELL_SELECTOR).length > 0 && !$target.is(MV_SUBCELL_SELECTOR)) {
                    return;
                }
                this.children.header.unHighlightColumn(cellInfo.cellIndex);
                // FIXME: this is pretty lame
                if(this.children.tableDock) {
                    this.children.tableDock.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + (rowNumbers ? 1 : 0)).removeClass('highlighted');
                }
                if(this.children.staticHead) {
                    this.children.staticHead.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + (rowNumbers ? 1 : 0)).removeClass('highlighted');
                }
                rowView.unHighlightCell($target, cellInfo.cellIndex);
            }
            else {
                // drilldown === 'row'
                rowView.unHighlightRow();
            }
        },

        handleCellClick: function($target, clickEvent) {
            var drilldown = this.model.format.get('display.statistics.drilldown');
            if(drilldown === 'none') {
                return;
            }
            var $cell = $target.closest('td'),
                isMultiValue = $cell.find(MV_SUBCELL_SELECTOR).length > 0;

            // if the cell is multivalue but the click was not on one of the subcells, ignore the event
            if(drilldown === 'cell' && isMultiValue && !$target.is(MV_SUBCELL_SELECTOR)) {
                return;
            }

            var cellInfo = this.getCellInfo($cell),
                modifierKey = !!(clickEvent.ctrlKey || clickEvent.metaKey);

            if(isMultiValue) {
                cellInfo.mvIndex = this.getMultiValueIndex($target);
            }
            this.emitDrilldownEvent(cellInfo, modifierKey, clickEvent);
        },

        emitDrilldownEvent: function(cellInfo, modifierKey, clickEvent) {
            // If somebody clicked on an invalid row, we cannot create a drilldown
            // event for them.
            if (cellInfo.rowIndex === undefined || cellInfo.cellIndex === undefined) {
                return;
            }
            
            var rowIndex = cellInfo.rowIndex,
                cellIndex = cellInfo.cellIndex,
                rowName = this.dataFields[0],
                rowSplitIsTime = this.isTimeField(rowName),
                rowValue = this.dataRows[rowIndex][0],
                drilldownInfo = {
                    modifierKey: modifierKey,
                    name: rowName,
                    value: rowSplitIsTime ? splunkUtils.getEpochTimeFromISO(rowValue) : rowValue
                };

            // for a time-based row split, add the _span information for the row
            if(rowSplitIsTime) {
                var spanSeries = this.getSpanSeriesForTimeField(rowName);
                if(spanSeries) {
                    drilldownInfo._span = parseInt(spanSeries[cellIndex], 10);
                }
            }

            // can only add a name2/value2 if the click was not on the first cell of the row, unless it's a multi-value field
            var isMultiValue = _.isArray(this.dataRows[rowIndex][cellIndex]);
            if((cellIndex > this.options.numRowSplits - 1) || isMultiValue) {
                drilldownInfo.name2 = this.dataFields[cellIndex];
                // if value2 is an array (a multivalue field), use the mvIndex to pick the right value (or just take the first one)
                drilldownInfo.value2 = isMultiValue ? this.dataRows[rowIndex][cellIndex][cellInfo.mvIndex || 0] :
                                            this.dataRows[rowIndex][cellIndex];
            }

            // add in the row context information
            var rowContext = {};
            _(this.dataRows[rowIndex]).each(function(value, i) {
                rowContext['row.' + this.dataFields[i]] = value;
            }, this);
            drilldownInfo.rowContext = rowContext;
            drilldownInfo.rowIndex = parseInt(rowIndex, 10);
            drilldownInfo.cellIndex = parseInt(cellIndex, 10);
            drilldownInfo.originalEvent = clickEvent;

            this.trigger('cellClick', drilldownInfo);
        },

        insertHeader: function($rootEl, dataFields) {
            this.children.header = new ResultsTableHeader({
                rowNumbers: splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers')),
                fields: dataFields,
                columnTypes: this.columnTypes,
                sortKeyAttribute: ColumnSort.SORT_KEY_ATTR
            });
            $rootEl.find('thead').replaceWith(this.children.header.render().$el);
        },

        insertRows: function($rootEl, dataFields, dataRows, dataColumns) {
            var dataOverlay = this.model.format.get('display.statistics.overlay'),
                showRowNumbers = splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers'));

            if(dataOverlay === 'heatmap') {
                this.updateHeatDeltas(dataColumns);
            }
            else if(dataOverlay === 'highlow') {
                this.updateExtremes(dataColumns);
            }

            if(_(this.columnTypes).contains('timestamp')) {
                this.updateFieldTimestampFormats(dataFields, dataRows, dataColumns);
            }

            var rowIterator = new RowIterator({
                    offset: this.model.metadata.get(this.options.offsetAttr)
                }),
                $tbody = $rootEl.find('tbody');

            rowIterator.eachRow(dataRows, function(row, i, rowNumber) {
                var viewOptions = {
                    table: this,
                    cells: this.generateCellObjects(row, dataFields, dataOverlay),
                    dataOverlay: dataOverlay,
                    rowIndex: i,
                    numRowSplits: this.options.numRowSplits
                };

                if(showRowNumbers) {
                    viewOptions.rowNumber = rowNumber;
                }
                var rowView = this.children['row_' + i] = new ResultsTableRow(viewOptions);
                $tbody.append(rowView.render().$el);
            }, this);
        },

        generateCellObjects: function(row, fields, dataOverlay) {
            return _(row).map(function(value, i) {
                var field = fields[i];
                // TODO add configured format
                return ({
                    value: value,
                    field: field,
                    dataOverlay: dataOverlay,
                    columnType: this.columnTypes[field] || 'string',
                    heatDelta: this.fieldHeatDeltas[field],
                    extremes: this.fieldExtremes[field],
                    timestampFormat: this.fieldTimestampFormats[field],
                    sparklineFormat: this.sparklineFormatSettings[field],
                    index: i
                });
            }, this);
        },

        disposeOfChildren: function() {
            _(this.children).each(function(child, name) {
                if(name === 'header' || name.substr(0, 4) === 'row_') {
                    child.dispose();
                    delete this.children[name];
                }
            }, this);
        },

        resultsAsColumns: function() {
            var columns = {},
                rows = this.model.results.get('rows') || [];

            // make 'columns' into a dictionary of field name to array of values in the column
            _(this.getFieldList()).each(function(field, i) {
                columns[field] = _(rows).pluck(i);
            });
            return columns;
        },

        updateDrilldownClasses: function() {
            var drilldown = this.model.format.get('display.statistics.drilldown'),
                $table = this.$('.table'),
                $rows = $table.find('tbody > tr'),
                $cells = $table.find('tbody > tr > td');

            if(drilldown === 'none') {
                $table.removeClass('table-drilldown table-drilldown-row table-drilldown-cell');
                $cells.removeAttr('tabindex');
                $rows.removeAttr('tabindex');
            }
            else if(drilldown === 'cell') {
                $table.addClass('table-drilldown table-drilldown-cell').removeClass('table-drilldown-row');
                $cells.attr('tabindex', '0');
                $rows.removeAttr('tabindex');
            }
            else {
                // drilldown === 'row'
                $table.addClass('table-drilldown table-drilldown-row').removeClass('table-drilldown-cell');
                $rows.attr('tabindex', '0');
                $cells.removeAttr('tabindex');
            }
        },

        updateFormatSettings: function() {
            var settings = this.sparklineFormatSettings = {};
            if(this.model.format.has('display.statistics.sparkline.format')) {
                _.extend(settings, this.model.format.get('display.statistics.sparkline.format'));
            }
        },

        updateColumnTypes: function(columns) {
            this.columnTypes = {};
            _(columns).each(function(column, field) {
                // non-data fields (i.e. _span) do not get assigned a column type
                if(!this.isDataField(field)) {
                    return;
                }
                if(this.isTimeField(field)) {
                    this.columnTypes[field] = 'timestamp';
                }
                else {
                    this.columnTypes[field] = generalUtils.valuesAreNumeric(column) ? 'number' : 'string';
                }
            }, this);
        },

        updateHeatDeltas: function(columns) {
            // create a data structure that holds the adjusted range (5th percentile to 95th) for each numeric field
            // the current implementation applies the same global heat delta to all numeric series
            // but the interface is designed so that different deltas could be applied to different columns in future
            var allNumericValues = this.getAllNumericValues(columns),
                bounds = generalUtils.getPercentiles(allNumericValues.sort(function(a, b) { return a - b; }), 0.05, 0.95),
                heatDelta = Math.max(Math.abs(bounds.lower), Math.abs(bounds.upper));

            this.fieldHeatDeltas = {};
            _(columns).each(function(column, field) {
                if(this.columnTypes[field] === 'number') {
                    this.fieldHeatDeltas[field] = heatDelta;
                }
            }, this);
        },

        updateExtremes: function(columns) {
            // create a data structure that holds the extremes (min and max) for each numeric field
            // the current implementation applies the same global extremes to all numeric series
            // but the interface is designed so that different extremes could be applied to different columns in future
            var allNumericValues = this.getAllNumericValues(columns),
                extremes = {
                    min: _(allNumericValues).chain().without(null).min().value(),
                    max: _(allNumericValues).max()
                };

            this.fieldExtremes = {};
            _(columns).each(function(column, field) {
                if(this.columnTypes[field] === 'number') {
                    this.fieldExtremes[field] = extremes;
                }
            }, this);
        },

        // used for calculating heat deltas and extremes
        // finds all columns that contain numeric data, then returns all of their contents as a single flattened array
        getAllNumericValues: function(columns) {
            var normalizeToNumber = function(value) {
                // convert to null if the value can't be parsed to a number
                return mathUtils.strictParseFloat(value) || null;
            };

            // find all of the columns that have been deemed numeric, flatten, then normalize all values to numbers
            return _(columns).chain()
                .filter(function(column, field) { return this.columnTypes[field] === 'number'; }, this)
                // this is a deep flattening operation, so multi-value fields will also be flattened out
                .flatten()
                .map(normalizeToNumber)
                .value();
        },

        updateFieldTimestampFormats: function(fields, rows, columns) {
            this.fieldTimestampFormats = {};
            _(fields).each(function(field, i) {
                if(this.isTimeField(field)) {
                    if(this.isBucketedTimeField(field)) {
                        var labelGranularity = timeUtils.determineLabelGranularity(columns[field]);
                        this.fieldTimestampFormats[field] = TIMESTAMP_FORMATS[labelGranularity];
                    }
                    else {
                        this.fieldTimestampFormats[field] = TIMESTAMP_FORMATS['second'];
                    }
                }
            }, this);
        },

        // TODO [sff] a lot of these next methods should probably be extracted into a share-able helper

        // helper to normalize getting the list of field names
        // depending on whether 'show_metadata=true' was used to fetch the results
        // the "fields" could be a list of string names or a list of objects with a "name" property
        getFieldList: function() {
            return _(this.model.results.get('fields') || []).map(function(field) {
                return _(field).isString() ? field : field.name;
            });
        },

        isDataField: function(fieldName) {
            return DATA_FIELD_REGEX.test(fieldName);
        },

        isTimeField: function(fieldName) {
            return (fieldName in { _time: true, earliest_time: true, latest_time: true });
        },

        isBucketedTimeField: function(fieldName) {
            return (fieldName === '_time');
        },

        // hard-coded to use '_span' at the moment,
        // but we want an interface that could potentially support multiple time series with different spans
        getSpanSeriesForTimeField: function(fieldName) {
            var spanSeriesName = '_span';
            // find the column whose corresponding field matches the span series name
            return this.resultsAsColumns()[spanSeriesName];
        },

        getCellInfo: function($cell) {
            return ({
                cellIndex: $cell.attr(ResultsTableRow.CELL_INDEX_ATTR),
                rowIndex: $cell.closest('tr').attr(ResultsTableRow.ROW_INDEX_ATTR)
            });
        },

        getMultiValueIndex: function($el) {
            return $el.attr(ResultsTableRow.MV_INDEX_ATTR);
        },

        template: '\
            <table class="table table-chrome table-striped <%= wrapResults ? "wrapped-results" : "not-wrapped-results" %> <%= numDataColumns === 1 ? \'single-column-table\' : \'\' %>">\
                <thead></thead>\
                <tbody></tbody>\
            </table>\
        ',

        staticHeaderTemplate: '\
            <div class="header-table-static"></div>\
            <div class="scroll-table-wrapper"></div>\
        '

    },
    {
        TIMESTAMP_FORMATS: TIMESTAMP_FORMATS
    });

});
