define(
	[
        "jquery",
        "underscore",
        "backbone",
        "module",
        "models/services/search/TimeParser",
        "models/TimeRange",
        "views/Base",
        "views/shared/controls/SyntheticSelectControl",
        "views/shared/FlashMessages",
        "util/time_utils",
        "splunk.i18n"
    ],
    function($, _, Backbone, module, TimeParserModel, TimeRangeModel, BaseView, SyntheticSelectControl, FlashMessages, time_utils, i18n) {
        return BaseView.extend({
            moduleId: module.id,
            className: 'accordion-inner',
            initialize: function(options) {
                BaseView.prototype.initialize.call(this, arguments);
                this.label = _('Real-time').t();

                this.model = {
                    range: this.model,
                    workingRange: new TimeRangeModel($.extend(true, {}, this.model.toJSON())),
                    earliestTimeParser: new TimeParserModel({
                        key: time_utils.stripRTSafe(this.model.get("earliest"), false),
                        value: this.model.get("earliest_iso")
                    })
                };

                this.model.workingRange.set('realtime_range_type', 's');

                this.children.rangeTypeSelect = new SyntheticSelectControl({
                    model: this.model.workingRange,
                    items: [
                        {value: 's', label: _('Seconds Ago').t()},
                        {value: 'm', label: _('Minutes Ago').t()},
                        {value: 'h', label: _('Hours Ago').t()},
                        {value: 'd', label: _('Days Ago').t()},
                        {value: 'w', label: _('Weeks Ago').t()},
                        {value: 'mon', label: _('Months Ago').t()},
                        {value: 'q', label: _('Quarters Ago').t()},
                        {value: 'y', label: _('Years Ago').t()}
                    ],
                    modelAttribute: 'realtime_range_type',
                    className: 'btn-group timerangepicker-realtime_range_type',
                    toggleClassName: 'btn',
                    menuWidth: 'narrow',
                    popdownOptions: {attachDialogTo:'body'},
                    save: false
                });

                this.children.flashMessages = new FlashMessages({
                    model: {
                        workingRange: this.model.workingRange,
                        earliestTimeParser: this.model.earliestTimeParser
                    }
                });

                this.model.workingRange.on("change:realtime_range_type change:realtime_earliest_value", _.debounce(function() {
                    this.fetchEarliestTime();
                }.bind(this), 0), this);

                this.model.earliestTimeParser.on("change", this.updateEarliestHint, this);

                this.model.earliestTimeParser.on("error", function() {
                    this.$("#hint_" + this.cid).html(this.model.earliestTimeParser.error.get("messages")[0].message);
                    this.disableApply();
                }, this);

                this.model.range.on('change:earliest change:latest prepopulate', _.debounce(this.prepopulate, 0), this);

            },
            updateEarliestHint: function() {
                var date = time_utils.isoToDateObject(this.model.earliestTimeParser.get("value"));
                this.$("#hint_" + this.cid).html(i18n.format_datetime_microseconds(time_utils.jsDateToSplunkDateTimeWithMicroseconds(date), "short", "full"));
                this.enableApply();                  
            },
            prepopulate: function() {
               var earliestParse = this.model.range.getTimeParse('earliest'),
                   attrValues = {};

               if (!earliestParse) { return;}

               this.$('.earliest_input').val(earliestParse.amount);

               attrValues.realtime_earliest_value = earliestParse.amount;
               attrValues.realtime_range_type = earliestParse.unit;

               this.model.workingRange.set(attrValues);
            },
            supportsRange: function() {
                var earliestParse = this.model.range.getTimeParse('earliest'),
                    latestParse = this.model.range.getTimeParse('latest');

                if (!earliestParse) { return false; }

                return earliestParse.isRealTime && earliestParse.amount && !earliestParse.snapUnit
                        && latestParse.isRealTime && latestParse.isNow && !latestParse.snapUnit ;
            },
            getEarliestTime: function(withoutRT) {
                var amount = this.model.workingRange.get('realtime_earliest_value'),
                    type = this.model.workingRange.get('realtime_range_type') || "mon";

                if (!amount){
                    return null;
                }
                return (withoutRT ? "" : "rt") + "-" + amount + type;
            },
            fetchEarliestTime: function() {
                var time = this.getEarliestTime(true),
                    id = time_utils.stripRTSafe(this.model.earliestTimeParser.id, false);

                if (time && (time !== id)) {
                    this.model.earliestTimeParser.fetch({
                        data : {
                            time: time
                        }
                    });
                } else if (this.model.earliestTimeParser.get("value")) {
                    this.updateEarliestHint();
                } else {
                    this.$("#hint_" + this.cid).html("");
                    this.enableApply();
                }
            },
            events: {
                "keyup .earliest_input": function(event) {
                    this.model.workingRange.set('realtime_earliest_value', this.$("#earliest_" + this.cid).val());
                },
                "click .apply": function(event) {
                    if ($(event.currentTarget).hasClass("disabled")) {
                        event.preventDefault();
                        return;
                    }

                    var that = this,
                        earliest = this.getEarliestTime();
                    this.disableApply();

                    this.model.workingRange.save(
                            {
                                earliest: earliest,
                                latest: "rtnow"
                            },
                            {
                                success: function(model) {
                                    that.enableApply();
                                    that.model.range.set({
                                        'earliest': earliest,
                                        'latest': "rtnow",
                                        'earliest_epoch': model.get('earliest_epoch'),
                                        'latest_epoch': model.get('latest_epoch'),
                                        'earliest_iso': model.get('earliest_iso'),
                                        'latest_iso': model.get('latest_iso'),
                                        'earliest_date': new Date(model.get('earliest_date').getTime()),
                                        'latest_date': new Date(model.get('latest_date').getTime())
                                    });
                                    that.model.range.trigger("applied");
                                },
                                error: function(model, error) {
                                    that.enableApply();
                                }
                            }
                    );
                    event.preventDefault();
                }
            },
            enableApply: function() {
                this.$("#apply_" + this.cid).removeClass('disabled');
            },
            disableApply: function() {
                this.$("#apply_" + this.cid).addClass('disabled');
            },
            render: function() {
                var template = _.template(this.template, {
                    _: _,
                    cid: this.cid,
                    time: this.model.workingRange.get('realtime_earliest_value') || ""
                });
                this.$el.html(template);

                this.$el.prepend(this.children.flashMessages.render().el);
                this.children.rangeTypeSelect.render().$el.insertAfter(this.$(".earliest_input"));

                return this;
            },
            template:  '\
                <table class="form form-inline">\
                    <tbody>\
                        <tr>\
                            <td>\
                                <label class="control-label" for="earliest_<%- cid %>"><%- _("Earliest:").t() %></label>\
                            </td>\
                            <td>\
                                <div class="input-append"><input type="text" size="5" value="<%- time %>" class="earliest_input timerangepicker-real-time-earliest-time" id="earliest_<%- cid %>"/></div>\
                            </td>\
                            <td>\
                                <label class="control-label"><%- _("Latest:").t() %></label>\
                            </td>\
                            <td>\
                                <span class="form-value">now</span>\
                            </td>\
                            <td><button class="apply btn" id="apply_<%- cid %>"><%- _("Apply").t() %></button></td>\
                        </tr>\
                        <tr>\
                            <td></td>\
                            <td>\
                                <span id="hint_<%- cid %>" class="help-block help-block-timestamp"></span>\
                            </td>\
                        </tr>\
                    </tbody>\
                </table>\
            '
        });
    }
);
